/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of one of the parties involved in the TRANSORG project. For
 * 	  written permission, please contact diasz@ida.liu.se.
 *
 * 4. Products derived from this Software may not be called "TRANSORG"
 *    nor may "TRANSORG" appear in their names without prior written
 *    permission.
 *
 * 5. Due credit should be given to the TRANSORG Project
 *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
 *
 * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
 *
 * $Id: AFaultNotifier_impl.java,v 1.0  diasz Exp $
 *
 * Date         Author  Changes
 */

package org.omg.FT;
import org.omg.CosNotification.*;
import org.omg.CosNotifyComm.*;
import org.omg.CosNotifyFilter.*;
import org.omg.CosEventComm.*;
import java.util.*;
/** Class implementing the FaultNotifier interface
 * It contains methods for event notification receiving and sending
 * @author <a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
 */
public class AFaultNotifier_impl extends FaultNotifierPOA
{


    private class PushConsumerAId{
	public NotifyPublish spc;
	public long itsId;
	public Filter itsFilter;
	public PushConsumerAId(NotifyPublish s,long i,Filter f){
	    spc=s;
	    itsId=i;
	    itsFilter=f;
	}
    }



    Vector push_consumers;
    org.omg.CORBA.ORB orb;
    String[] grammars={"whatever"};

/** Constructor
 * @param o the ORB on top of which the fault notifier CORBA object runs
 */

    public AFaultNotifier_impl(org.omg.CORBA.ORB o){
	orb=o;
	push_consumers=new Vector();
    }

  /**
   * Method called to inform the notifier that an structured fault event occured; the notifier takes the event and sends it further to its
   * registered consumers
   * @param event the event to be sent to the consumers
   */

    public void push_structured_fault(StructuredEvent event){
	int i,l;
	PushConsumerAId crt;

	l=push_consumers.size();

	for(i=0;i<l;i++){
	    crt=(PushConsumerAId)push_consumers.elementAt(i);
	    if(crt.spc instanceof StructuredPushConsumer){

		if(crt.itsFilter==null||Utils.pass_filter(event,crt.itsFilter)){

		    try{
			((StructuredPushConsumer)(crt.spc)).push_structured_event(event);
		    }catch(Exception e){System.out.println("Some exception when push event");e.printStackTrace(new java.io.PrintStream(System.out));}

		}

	    }
	}
    }

	/**
	   * Method called to inform the notifier that a sequence of structured fault events occured; the notifier takes the events and sends it further to its
	   * registered consumers
	   * @param events the sequence of structured events to be sent to the consumers
   */

    public void push_sequence_fault(StructuredEvent[] events){
	int i,l,j;
	PushConsumerAId crt;

	l=push_consumers.size();
	for(i=0;i<l;i++){
	    crt=(PushConsumerAId)push_consumers.elementAt(i);
	    if(crt.spc instanceof SequencePushConsumer){
		for(j=0;j<events.length;j++)
		    if(!Utils.pass_filter(events[j],crt.itsFilter))
			return;
		try{
		    ((SequencePushConsumer)(crt.spc)).push_structured_events(events);
		}catch(org.omg.CosEventComm.Disconnected e){}
	    }
	}
    }

	/** Method used to create a filter according to what events the consumer is interested in
	 * @param constraint_grammar the grammar used by the filter
	 * @return A subscription filter
	 *
	 */

	public Filter create_subscription_filter(java.lang.String constraint_grammar)
	    throws InvalidGrammar{
	    /*  A simple verification is made about the grammar; actually, no factory is used to create the filter;
	       the notifier itself is the factory*/
	    int i;
	    boolean found=false;
	    for(i=0;i<grammars.length;i++)
		if(constraint_grammar.equals(grammars[i])){
		    found=true;
		    break;
		}
	    if(!found) throw new InvalidGrammar();
	    return (new FaultEventFilter(constraint_grammar))._this(orb);
	}

	/** Method used to register a consumer of a structured fault event
	 * @param	push_consumer the consumer of the structured event
	 *			filter the filter (previously created by create_subscription_filter) used by this consumer
	 * @return the index of this consumer in the consumer list (used later when disconnecting the consumer)
	 */

    public long connect_structured_fault_consumer(StructuredPushConsumer push_consumer, Filter filter){

	long l;
	l=(long)push_consumers.size();
	push_consumers.addElement(new PushConsumerAId(push_consumer,l,filter));
	return l;
    }

	/** Method used to register a consumer of a sequence of structured fault events
		 * @param	push_consumer the consumer of the event sequence
		 *			filter the filter (previously created by create_subscription_filter) used by this consumer
		 * @return the index of this consumer in the consumer list (used later when disconnecting the consumer)
	 */
    public long connect_sequence_fault_consumer(SequencePushConsumer push_consumer, Filter filter){
	long l;
	l=(long)push_consumers.size();
	push_consumers.addElement(new PushConsumerAId(push_consumer,l,filter));
	return l;
    }

	/** Method used to disconnect a consumer from the fault notifier (e.g., when the consumer fails
	 * @param connection the index of the consumer in the consumer list of the notifier
	 */
    public void disconnect_consumer(long connection)
	throws Disconnected{
	int i;
	long l;
	PushConsumerAId pci;
	boolean found=false;

	if(connection==-1) return;

	l=(long)push_consumers.size();

	for(i=0;i<l;i++){
	    pci=(PushConsumerAId)push_consumers.elementAt(i);
	    if(pci.itsId==connection){
		found=true;
		break;
	    }
	}
	if(found) push_consumers.remove(i);
	else
	    throw new Disconnected();
    }
}
