/**
  * Redistribution and use of this software and associated
  * documentation ("Software"), with or without modification, are
  * permitted provided that the following conditions are met:
  *
  * 1. Redistributions of source code must retain copyright statements
  *    and notices.  Redistributions must also contain a copy of this
  *    document.
  *
  * 2. Redistributions in binary form must reproduce the above
  *    copyright notice, this list of conditions and the following
  *    disclaimer in the documentation and/or other materials provided
  *    with the distribution.
  *
  * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
  *    products derived from this Software without prior written
  *    permission of one of the parties involved in the TRANSORG project. For
  * 	  written permission, please contact diasz@ida.liu.se.
  *
  * 4. Products derived from this Software may not be called "TRANSORG"
  *    nor may "TRANSORG" appear in their names without prior written
  *    permission.
  *
  * 5. Due credit should be given to the TRANSORG Project
  *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
  *
  * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  * PARTICULAR PURPOSE ARE DISCLAIMED.
  *
  * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
  * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
  * $Id: FaultMonitor_impl.java,v 1.0  diasz Exp $
  *
  * Date         Author  Changes
  */

package org.omg.FT;
import org.omg.CORBA.portable.*;
import java.util.*;
/** Class implementing the FaultMonitor interface
  * It is used to monitor application CORBA object for failures
  * @author	<a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
  * @version $Revision: 1.0 $ $Date:2003/06/23$
  */

public class FaultMonitor_impl extends FaultMonitorPOA{

    FaultNotifier fault_notifier;
    String ftDomId,location;
    Hashtable monitor_threads_objects;
    org.omg.CORBA.ORB orb;

    private class MonitorThread extends Thread{

	PullMonitorable obj_to_monitor;
	FaultMonitoringIntervalAndTimeoutValue f_mon_i_tv;
	String type_id;
	long objGrId;
	int monit_granularity;
	boolean is_sleep,stopped,iact_interrupted;
	java.lang.Object term_o;

	class IsAliveCallThread extends Thread{
	    MonitorThread parent;
	    PullMonitorable obj_to_mon;
	    boolean alive=true;

	    public IsAliveCallThread(MonitorThread p,PullMonitorable o_m){
		parent=p;
		obj_to_mon=o_m;
	    }

	    public void run(){
		try{
		    alive=obj_to_mon.is_alive();
		}catch(java.lang.Exception exc){alive=false;}
		if(!is_sleep){
		    parent.set_I_interrupted();
		    parent.interrupt();
		}

	    }

	    public boolean get_alive(){
		return alive;
	    }
	}

	/** Constructor
	 * @param	otm the CORBA object to monitor
	 *			mon_gran the monitoring granularity (can be MEMB=0,LOC =1, LOC_AND_TYPE =2; see the FTC IDL file for the details of the constants)
	 *			o_g_i the object group id of the monitored object
	 *			fmiv the Fault monitoring interval and timeout value
	 *			t_id type identifier of the monitored object
	 */

	public MonitorThread(PullMonitorable otm,int mon_gran,long o_g_i,FaultMonitoringIntervalAndTimeoutValue fmiv,String t_id){
	    obj_to_monitor=otm;
	    f_mon_i_tv=fmiv;
	    objGrId=o_g_i;
	    stopped=false;
	    type_id=t_id;
	    monit_granularity=mon_gran;
	    term_o=new java.lang.Object();
	    iact_interrupted=false;
	}

/** Method used to signal that the "is alive call thread" returned after the call to is_alive within the timeout interval
*/

	public void set_I_interrupted(){
	    iact_interrupted=true;
	}
/**
 * @return the monitoring granularity of the monitored object
 */
	public int get_mon_gran(){
	    return monit_granularity;
	}

/**
 * @return the object group identifier of the monitored object
 */

	public long get_g_i(){
	    return objGrId;
	}

/**
 * @return the fault monitoring and timeout interval for the monitored object
 */
	public FaultMonitoringIntervalAndTimeoutValue get_fmoni_tv(){
	    return f_mon_i_tv;
	}

	public void terminate(){
	    synchronized(term_o){
		stopped=true;
	    }
	}


	public boolean is_terminated(){
	    synchronized(term_o){
		return stopped;
	    }
	}

/** The monitoring thread calls periodically the is_alive method of the monitored CORBA application object; when calling the method it waits until a timeout
 * elapses
 * If a fault is detected the fault event is constructed and sent to the fault notifier
 */

	public void run(){

	    IsAliveCallThread iact;

	    org.omg.CosNotification.StructuredEvent fault_event=new org.omg.CosNotification.StructuredEvent();
	    org.omg.CORBA.Any any;
	    String str_a_v[],gr_i=Long.toString(objGrId),filt_data_v[]={ftDomId,location,type_id,gr_i},filt_data_n[]={"FTDomainId","Location","TypeId","ObjectGroupId"};

	    int i;
	    try{
		fault_event.header=new org.omg.CosNotification.EventHeader();
		fault_event.remainder_of_body=orb.create_any();
		fault_event.header.variable_header=new org.omg.CosNotification.Property[0];
		fault_event.header.fixed_header=new org.omg.CosNotification.FixedEventHeader();
		fault_event.header.fixed_header.event_name="fd_on_"+location;
		fault_event.header.fixed_header.event_type=new org.omg.CosNotification.EventType("FT_CORBA","ObjectCrashFault");
		fault_event.filterable_data=new org.omg.CosNotification.Property[4];
		for(i=0;i<4;i++){
		    fault_event.filterable_data[i]=new org.omg.CosNotification.Property();
		    fault_event.filterable_data[i].name = filt_data_n[i];
		}
		any=orb.create_any();
		str_a_v=new String[1];
		str_a_v[0]=ftDomId;
		org.omg.CORBA.StringSeqHelper.insert(any,str_a_v);
		fault_event.filterable_data[0].value = any;
		any=orb.create_any();
		str_a_v=new String[1];
		str_a_v[0]=location;
		org.omg.CORBA.StringSeqHelper.insert(any,str_a_v);
		fault_event.filterable_data[1].value = any;
		any=orb.create_any();
		str_a_v=new String[1];
		str_a_v[0]=type_id;
		org.omg.CORBA.StringSeqHelper.insert(any,str_a_v);
		fault_event.filterable_data[2].value = any;
		any=orb.create_any();
		str_a_v=new String[1];
		str_a_v[0]=gr_i;
		org.omg.CORBA.StringSeqHelper.insert(any,str_a_v);
		fault_event.filterable_data[3].value = any;

	    }catch(Exception e){System.out.println("Exception in monitor thread run line 62");e.printStackTrace(new java.io.PrintStream(System.out));}



	    boolean alive=true;
	    while(true){
		if(is_terminated()) break;
		is_sleep=false;
		iact=new IsAliveCallThread(this,obj_to_monitor);
		iact_interrupted=false;
		iact.start();

		try{
		    sleep(f_mon_i_tv.timeout);
		}catch(InterruptedException ie){ie.printStackTrace();}

		if(is_terminated()) break;

		if(!iact.isAlive()||iact_interrupted)
		    alive=iact.get_alive();
		else{
		    alive=false;
		}
		if(!alive) {
		    try{

			object_failed(obj_to_monitor,this);
			fault_notifier.push_structured_fault(fault_event);
			break;
		    }catch(Exception e){System.out.println("Exception if fault monitor when announcing fault");e.printStackTrace(new java.io.PrintStream(System.out));break;}
		}

		if(interrupted()) {System.out.println("Thread was interrupted");break;}
		try{
		    is_sleep=true;
		    sleep(f_mon_i_tv.monitoring_interval);
		}catch(InterruptedException e){System.out.println("Interrupted while sleeping");break;}
	    }
	}
    }

/** Constructor
 * @param	f_n the fault notifier to which the fault events are sent
 *			f_t_d the fault tolerance domain
 *			loc the location (host name) of this fault notifier
 *			o the ORB in which the fault monitor is running
 */


    public FaultMonitor_impl(FaultNotifier f_n,String f_t_d,String loc,org.omg.CORBA.ORB o){
	ftDomId=f_t_d;
	location=loc;
	fault_notifier=f_n;
	orb=o;
	monitor_threads_objects=new Hashtable();
    }

    private boolean monitors_obj(){
	return (monitor_threads_objects.size()>0);
    }

    private boolean monitors_obj_type(String type){
	int i,l;
	String t_i;
	Set entry_set;
	Iterator iter;
	Map.Entry entry;

	entry_set=monitor_threads_objects.entrySet();
	iter=entry_set.iterator();
	while(iter.hasNext()){
	    entry=(Map.Entry)iter.next();
	    t_i=Utils.type_id((org.omg.CORBA.Object)entry.getKey());
	    if(t_i.equals(type))
		return true;
	}
	return false;
    }


/** Method called to request the fault monitor to create a new thread for monitoring a new application object
  * @param	obj the application CORBA object to monitor
  *			o_g_id the object group identifier of the monitored object
  *			mon_gran the monitoring granularity for the monitored object
  *			fmiv fault monitoring interval and timeout value
  */

    public void do_monitor(PullMonitorable obj,long o_g_id,int mon_gran,FaultMonitoringIntervalAndTimeoutValue fmiv){


	String t_id=Utils.type_id(obj);

	switch(mon_gran){

	case MEMB.value:remove_threads_loc_or_type_loc(t_id);break;
	case LOC_AND_TYPE.value:
	    if(monitors_obj_type(t_id)) return;
	    remove_threads_loc();
	    break;
	case LOC.value:
	    if(monitors_obj()) return;
	    break;
	}

	MonitorThread mt=new MonitorThread(obj,mon_gran,o_g_id,fmiv,t_id);
	monitor_threads_objects.put(obj,mt);
	mt.start();
    }

/** Method called to stop the monitoring process of a CORBA object
 * @param the object reference of the application object that is monitored
 */


    public void stop_monitor(org.omg.CORBA.Object obj){
	MonitorThread m_thread;
	Set entry_set;
	Iterator iter;
	Map.Entry entry;
	PullMonitorable theObj=null;
	boolean found=false;

	entry_set=monitor_threads_objects.entrySet();
	iter=entry_set.iterator();
	while(iter.hasNext()){
	    entry=(Map.Entry)iter.next();
	    theObj=(PullMonitorable)entry.getKey();
	    if(Utils.refs_equal(orb,theObj,obj,-1)){
		m_thread=(MonitorThread)entry.getValue();
		if(m_thread!=null){
		    m_thread.terminate();
		    m_thread.interrupt();
		}
		found=true;
		break;
	    }
	}
	if(found)
	    monitor_threads_objects.remove(theObj);
    }


    private void remove_threads_loc_or_type_loc(String type_id){
	Set entrySet=monitor_threads_objects.entrySet();
	Iterator iter=entrySet.iterator();
	Map.Entry entry;
	MonitorThread value;
	java.util.Vector non_conc_buffer=new java.util.Vector();
	int i,l;

	while(iter.hasNext())
	    non_conc_buffer.addElement(iter.next());

	l=non_conc_buffer.size();
	for(i=0;i<l;i++){
	    entry=(Map.Entry)non_conc_buffer.elementAt(i);
	    value=(MonitorThread)entry.getValue();
	    if((value.monit_granularity==LOC_AND_TYPE.value&&value.type_id.equals(type_id))||value.monit_granularity==LOC.value){
		value.terminate();
		value.interrupt();
		monitor_threads_objects.remove(entry.getKey());
	    }
	}
    }


    private void remove_threads_type_loc(String type_id){
	Set entrySet=monitor_threads_objects.entrySet();
	Iterator iter=entrySet.iterator();
	Map.Entry entry;
	MonitorThread value;
	java.util.Vector non_conc_buffer=new java.util.Vector();
	int i,l;

	while(iter.hasNext())
	    non_conc_buffer.addElement(iter.next());

	l=non_conc_buffer.size();
	for(i=0;i<l;i++){
	    entry=(Map.Entry)non_conc_buffer.elementAt(i);
	    value=(MonitorThread)entry.getValue();
	    if(value.monit_granularity==LOC_AND_TYPE.value&&value.type_id.equals(type_id)){
		value.terminate();
		value.interrupt();
		monitor_threads_objects.remove(entry.getKey());
	    }
	}
    }

    private void remove_threads_loc(){
	Set entrySet=monitor_threads_objects.entrySet();
	Iterator iter=entrySet.iterator();
	Map.Entry entry;
	MonitorThread value;
	java.util.Vector non_conc_buffer=new java.util.Vector();
	int i,l;

	while(iter.hasNext())
	    non_conc_buffer.addElement(iter.next());

	l=non_conc_buffer.size();
	for(i=0;i<l;i++){
	    entry=(Map.Entry)non_conc_buffer.elementAt(i);
	    value=(MonitorThread)entry.getValue();
	    if(value.monit_granularity==LOC.value){
		value.terminate();
		value.interrupt();
		monitor_threads_objects.remove(entry.getKey());
	    }
	}
    }

    protected synchronized void object_failed(PullMonitorable obj,MonitorThread mt){
	monitor_threads_objects.remove(obj);
    }
}
