/**
  * Redistribution and use of this software and associated
  * documentation ("Software"), with or without modification, are
  * permitted provided that the following conditions are met:
  *
  * 1. Redistributions of source code must retain copyright statements
  *    and notices.  Redistributions must also contain a copy of this
  *    document.
  *
  * 2. Redistributions in binary form must reproduce the above
  *    copyright notice, this list of conditions and the following
  *    disclaimer in the documentation and/or other materials provided
  *    with the distribution.
  *
  * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
  *    products derived from this Software without prior written
  *    permission of one of the parties involved in the TRANSORG project. For
  * 	  written permission, please contact diasz@ida.liu.se.
  *
  * 4. Products derived from this Software may not be called "TRANSORG"
  *    nor may "TRANSORG" appear in their names without prior written
  *    permission.
  *
  * 5. Due credit should be given to the TRANSORG Project
  *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
  *
  * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  * PARTICULAR PURPOSE ARE DISCLAIMED.
  *
  * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
  * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
  *
  * $Id: GatewayForActiveRepl_impl.java,v 1.0  diasz Exp $
  *
  * Date         Author  Changes
  */

package org.omg.FT;

/** Class instantiated whenever an actively replicated group is created by the Replication Manager
*/

public class GatewayForActiveRepl_impl extends GatewayForActiveReplPOA{
	/* This is the group for whose failure of members this consumer is interested.
	   If a fault event arrives at this consumer it means that it passed the filter
	   of this consumer, meaning that a member of THIS GROUP (with group id= group_id)
	   failed, even if, in fact the event contains a different group id (for example
	   if this group has a monit. gran. of type LOC or LOC_AND_TYPE then the failure of
	   any member of agroup at the location or of a certain type at the location, will
	   trigger the sending of such an event to the present consumer
	 */
    class GatewayStructuredPushConsumer extends org.omg.CosNotifyComm.StructuredPushConsumerPOA{
	public void disconnect_structured_push_consumer(){}

	public void offer_change(org.omg.CosNotification.EventType[] added, org.omg.CosNotification.EventType[] removed){}

	public void push_structured_event(org.omg.CosNotification.StructuredEvent notification) throws org.omg.CosEventComm.Disconnected{
	    int l;
	    String location,value[];

	    try{
		l=notification.filterable_data.length;
		if(l<4) {return;}
		value=org.omg.CORBA.StringSeqHelper.extract(notification.filterable_data[1].value);
		location=value[0];
		update_memb_remove(location);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	}
    }

    private synchronized void update_memb_remove(String location){
	int i,l;
	ObjectLocStr crt;
	boolean found;


	group_members_loc.remove(location);
    }

    private class CountObj{
	int nr_waiting;

	public CountObj(){
	    nr_waiting=0;
	}
    }

    private class CallStr{
	org.omg.CORBA.Object target;
	int retention_id;
	CountObj synch_obj;

	public CallStr(org.omg.CORBA.Object target,int retention_id,CountObj synch_obj){
	    this.target=target;
	    this.retention_id=retention_id;
	    this.synch_obj=synch_obj;
	}
    }


    private class ReplyStr{
	org.omg.CORBA.Object target;
	int retention_id;
	org.omg.CORBA.Any reply;

	public ReplyStr(org.omg.CORBA.Object target,int retention_id,org.omg.CORBA.Any reply){
	    this.target=target;
	    this.retention_id=retention_id;
	    this.reply=reply;
	}
    }

    java.util.Hashtable group_members_loc;
    java.util.Vector synch_objects,replies,notified;
    org.omg.CORBA.ORB orb;
    org.omg.CORBA.Object[] old_group_members,new_group_members;
    byte[] obj_id;

/** Constructor
 * @param	o the ORB on which this CORBA object runs
 *			oid the object identifier of the active group
 */
    public GatewayForActiveRepl_impl(org.omg.CORBA.ORB o,byte[] oid){
	synch_objects=new java.util.Vector();
	notified=new java.util.Vector();
	replies=new java.util.Vector();
	orb=o;
	obj_id=oid;
	group_members_loc=new java.util.Hashtable();
    }

    private synchronized CountObj synch_obj(org.omg.CORBA.Object target,int retention_id){
	CallStr crt;
	int i,l;

	l=synch_objects.size();
	for(i=0;i<l;i++){
	    crt=(CallStr)synch_objects.elementAt(i);
	    if(crt.target==target&&crt.retention_id==retention_id)
		return crt.synch_obj;
	}
	return null;
    }

/** Method called when the group is created and its membership is communicated to the gateway
* @param	members the list of {location,replica object} coresponding to the initial group membership
*/
    public synchronized void init_membership(ObjectLocStr[] members){
	int i;
	for(i=0;i<members.length;i++)
	    group_members_loc.put(members[i].location,members[i].object);
	old_group_members=group_members();
	new_group_members=old_group_members;
	connect_to_fault_notifier();
    }

/** Method called when the membership of the group changes, usually due to failures
* @param	new_memb list of new members of the CORBA object group
*			location list of the locations where the new members reside
*/
    public synchronized void change_membership(org.omg.CORBA.Object[] new_memb,String location[]){
	int i;
	old_group_members=group_members();
	new_group_members=new org.omg.CORBA.Object[new_memb.length];
	for(i=0;i<location.length;i++){
	    group_members_loc.put(location[i],new_memb[i]);
	    new_group_members[i]=new_memb[i];
	}
    }


    public synchronized org.omg.CORBA.Object[] old_group_members(){
	return old_group_members;
    }

    public synchronized org.omg.CORBA.Object[] new_group_members(){
	return new_group_members;
    }

/** @return the current composition of te active CORBA object group*/

    public synchronized org.omg.CORBA.Object[] group_members(){
	java.util.Collection the_values=group_members_loc.values();
	org.omg.CORBA.Object[] toRet=new org.omg.CORBA.Object[the_values.size()];

	return (org.omg.CORBA.Object[])the_values.toArray(toRet);
    }

    private void connect_to_fault_notifier(){
	org.omg.CORBA.Object r_man_o;
	ReplicationManager r_man;
	FaultNotifier fault_notifier;

	int i,l;
	try{
	    r_man_o=orb.resolve_initial_references("ReplicationManager");
	    r_man=ReplicationManagerHelper.narrow(r_man_o);
	    fault_notifier=r_man.get_fault_notifier();
	}catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	if(fault_notifier!=null){
	    org.omg.CosNotifyFilter.Filter filter=null;
	    try{
		filter=fault_notifier.create_subscription_filter("whatever");
	    }catch(Exception ex){ex.printStackTrace(new java.io.PrintStream(System.out));filter=null;}
	    org.omg.CosNotifyFilter.ConstraintExp[] constraints=new org.omg.CosNotifyFilter.ConstraintExp[1];
	    org.omg.CosNotification.EventType[] event_types=new org.omg.CosNotification.EventType[1];
	    event_types[0]=new org.omg.CosNotification.EventType("FT_CORBA","ObjectCrashFault");
	    l=group_members_loc.size();
	    java.util.Set entrySet=group_members_loc.entrySet();
	    l=entrySet.size();
	    java.util.Iterator iter=entrySet.iterator();
	    java.util.Map.Entry entry;
	    String constraint_exp="$FTDomainId=="+"firstDomain";
	    constraint_exp+="&";
	    while(iter.hasNext()){
		entry=(java.util.Map.Entry)iter.next();
		constraint_exp+="$Location=="+(String)entry.getKey();
		if(iter.hasNext())
		    constraint_exp+="|";
	    }
	    constraint_exp+="&$ObjectGroupId=="+Utils.makeByteId(obj_id);

	    constraints[0]=new org.omg.CosNotifyFilter.ConstraintExp(event_types,constraint_exp);
	    if(filter!=null){
		try{

		    filter.add_constraints(constraints);
		}catch(org.omg.CosNotifyFilter.InvalidConstraint ic){ic.printStackTrace(new java.io.PrintStream(System.out));}

	    }

	    try{
		fault_notifier.connect_structured_fault_consumer(new GatewayStructuredPushConsumer()._this(orb),filter);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	}
    }

    public org.omg.CORBA.Any invoke(org.omg.CORBA.Object target,String type_id,int ret_id,String client_id,long req_start_time,String oper,org.omg.FT.ParametersHolder arguments,org.omg.CORBA.TypeCode[] exceptions,boolean resp_e,org.omg.CORBA.Any result){
	return null;
    }


}
