/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of one of the parties involved in the TRANSORG project. For
 * 	  written permission, please contact diasz@ida.liu.se.
 *
 * 4. Products derived from this Software may not be called "TRANSORG"
 *    nor may "TRANSORG" appear in their names without prior written
 *    permission.
 *
 * 5. Due credit should be given to the TRANSORG Project
 *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
 *
 * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
 *
 * $Id: LoggingRecoveryController_impl.java,v 1.0  diasz Exp $
 *
 * Date         Author  Changes
 */

package org.omg.FT;
import java.util.*;
/** Class instantiated on every host where replicas of server groups can be created; it has the role of logging and checkpointing for the application
* server replicas; to checkpoint the state of the object, it instantiates on demand a thread for each checkpointed object in which it periodically calls
* the state reading method on the CORBA object
* Also, it is used when logging of method calls is done in passive replication, and also reply logging when this is needed
* @author	<a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
*/

/* The readers-writers problem when recording a method call is solved for get_state and update
   operations, and for update-update operations (the latter one is because the order of replaying
   according to the log has to be identical to the actual execution order of the operations*/

/* In the log Hashatable for every element (object id) there is a Vector of Vectors
   Every Vector of Vectors contains 4 Vector elements:
   - Element 0: A vector of counters (Integers) and others,which contains e.g. the number of update method
    calls,the number of replies to update_methods,the index of the answer to get_state, the position up to
    where the outg req. queue has to be emptyed, a value which indicates if the get_state call is the first
    one or not, etc.
   - Element 1: A vector of CallLogRecords (get_state+read_opers+update_opers)
   - Element 2: A vector of ReplyLogRecords (reply+checkpoint+update)
   - Element 3: A vector of RequestLogRecords and RequestReplyLogRecords + get_state call record
*/


public class LoggingRecoveryController_impl extends  LoggingRecoveryControllerPOA{

    private class MyLong{
	long value;
	public MyLong(long v){
	    value=v;
	}

	public long get(){
	    return value;
	}

	public void set(long v){
	    value=v;
	}

	public void decr(){
	    value--;
	}

	public void incr(){
	    value++;
	}
    }

    private class NotifySObj{
	int nr_notify;

	public NotifySObj(){
	    nr_notify=0;
	}
    }


/** Class instantiated whenever a new replica is created and needs o be checkpointed (in passive replication) periodically, or not.
* The checkpoint thread reads the state of the object; if the replication style is warm passive the state is broadcast to the other
* replicas. In case cold passive replication is used, the state is stored locally at the logging recovery CORBA object. After the state
* is read, and operations are performed around it, the thread "goes to sleep" for the checkpointing interval given by the user (in case
* periodic checkpointing is used). When event triggered checkpointing is used, then whenever an update operation is logged in the logging
* object this triggers a state reading operation.
*/
    private class CheckpointThread extends Thread{

	Checkpointable obj_to_check;
	long checkpoint_interval;
	boolean transfer_state;
	Checkpointable[] transfer_to;
	java.lang.Object synch_obj;
	long object_id;
	boolean stopped;

	java.lang.Object wait_fall_asleep=new java.lang.Object();
	boolean working=false;
	NotifySObj event;

	public CheckpointThread(Checkpointable obj,long obj_id,long ch_int,Checkpointable trans_to[],boolean tr,NotifySObj ev){
	    obj_to_check=obj;
	    object_id=obj_id;
	    checkpoint_interval=ch_int;
	    transfer_state=tr;
	    transfer_to=trans_to;
	    event=ev;
	    stopped=false;
	    synch_obj=new java.lang.Object();
	}

	public void run(){
	    byte[] state=null;
	    int i;
	    Checkpointable[] t_t=null;
	    boolean t_s,error,first=true;
	    int flags=org.omg.CORBA.ARG_IN.value;
	    long timestamp;
	    org.openorb.CORBA.dii.Request dyn_request;
	    org.omg.CORBA.Any argument,any;
	    org.omg.CORBA.NVList dyn_arguments=orb.create_list(0);
	    org.omg.CORBA.ExceptionList dyn_exceptions=orb.create_exception_list();
	    org.omg.CORBA.NamedValue dyn_result;
	    dyn_result=orb.create_named_value("",orb.create_any(),-1);
	    org.openorb.CORBA.SystemExceptionHelper.id(new org.omg.CORBA.TIMEOUT());
	    long iter=0;
	    FTCurrORIServiceContext ctx_ori=null;
	    org.omg.IOP.ServiceContext serviceCtx_ori=new org.omg.IOP.ServiceContext();
	    org.omg.IOP.CodecFactory factory=null;
	    org.omg.IOP.Codec codec=null;

	    if(transfer_state)
		try{
		    factory =(org.omg.IOP.CodecFactory)orb.resolve_initial_references("CodecFactory");
		    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
		    serviceCtx_ori.context_id=org.omg.IOP.FT_CURRORI.value;
		}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}

	    while(true){
		error=false;
		state=null;
		if(terminated()) { break;}
		if(event!=null)
		   wait_update_method(event);
		try{

		    timestamp=System.currentTimeMillis();
		    record_method_call_par(object_id,"get_state",new org.omg.Dynamic.Parameter[0],"",-1,-1,org.omg.FT.Utils.READ,timestamp,transfer_state);
		    work();
		    state=obj_to_check.get_state();
		    fall_asleep();
		    record_rest_get_st_upd_reply("get_state",object_id,state,timestamp);
		}catch(java.lang.Exception ge){ge.printStackTrace(new java.io.PrintStream(System.out));error=true;}
		if(terminated()) break;
		if(!error&&state!=null){
		    synchronized(synch_obj){
			t_s=transfer_state;
			if(t_s)
			    t_t=transfer_to;
		    }
		    if(t_s){
			int ret_id=get_last_ret_id(object_id);
			argument=orb.create_any();
			org.omg.FT.StateHelper.insert(argument,state);
			dyn_arguments=orb.create_list(0);
			dyn_arguments.add_value("param0",argument,flags);
			/* for WARM_PASSIVE Replication*/
			for(i=0;i<t_t.length;i++){
			    try{
				work();
				ctx_ori=new FTCurrORIServiceContext(object_id,ret_id);
				any = orb.create_any();
				serviceCtx_ori=new org.omg.IOP.ServiceContext();
				serviceCtx_ori.context_id=org.omg.IOP.FT_CURRORI.value;
				org.omg.FT.FTCurrORIServiceContextHelper.insert(any,ctx_ori);
				serviceCtx_ori.context_data = codec.encode_value(any);
				dyn_request=new org.openorb.CORBA.dii.Request(t_t[i],"set_state",serviceCtx_ori,dyn_arguments,dyn_result,orb.create_environment(),dyn_exceptions,orb.create_context_list(),orb);
				dyn_request.invoke();
				fall_asleep();
				if(terminated()) break;
			    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
			    if(terminated())
				break;
			}
		    }
		}
		if(terminated()) {break;}
		if(event==null)
		    try{
			sleep(checkpoint_interval/10000);
		    }catch(InterruptedException e){break;}
		iter++;
	    }
	}

	public void set_transfer_to(Checkpointable[] t_to){
	    synchronized(synch_obj){
		transfer_to=t_to;
	    }
	}


	public void set_transfer_state(boolean v){
	    /* this method is called if (while) working, the replication style
	       changes from (in this case) WARM_PASSIVE, to other style
	       This is not probable to happen yet, in the current implementation*/
	    synchronized(synch_obj){
		transfer_state=v;
	    }
	}

	public synchronized boolean terminated(){
	    return stopped;
	}

	public synchronized void terminate(){
	    stopped=true;
	}

	public long get_obj_id()
	{
	    return object_id;
	}

	public void wait_working(){
	    synchronized(wait_fall_asleep){
		if(working)
		    try{
			wait_fall_asleep.wait();
		    }catch(InterruptedException e){}
		interrupt();
	    }
	}

	public void work(){
	    synchronized(wait_fall_asleep){
		working=true;
	    }
	}

	public void fall_asleep(){
	    synchronized(wait_fall_asleep){
		working=false;
		wait_fall_asleep.notify();
	    }
	}
    }

/** Class instantiated whenever a new replica is created and needs o be checkpointed (in passive replication) periodically, or not.
* The update thread reads the state change (update) of the object; if the replication style is warm passive the update is broadcast to the other
* replicas. In case cold passive replication is used, the update is stored locally at the logging recovery CORBA object. After the update
* is read, and operations are performed around it, the thread "goes to sleep" for the checkpointing interval given by the user (in case
* periodic checkpointing is used). When event triggered checkpointing is used, then whenever an update operation is logged in the logging
* object this triggers a state change reading operation.
*/
       private class UpdateThread extends Thread{


	   java.lang.Object wait_fall_asleep=new java.lang.Object();
	   boolean working=false;
	   Updateable obj_to_update;
	   long update_interval;
	   boolean transfer_state;
	   Updateable[] transfer_to;
	   java.lang.Object synch_obj;
	   long object_id;
	   boolean stopped;
	   NotifySObj event;

	public UpdateThread(Updateable obj,long obj_id,long up_int,Updateable trans_to[],boolean tr,NotifySObj ev){
	    obj_to_update=obj;
	    object_id=obj_id;
	    update_interval=up_int;
	    transfer_state=tr;
	    transfer_to=trans_to;
	    stopped=false;
	    synch_obj=new java.lang.Object();
	    event=ev;
	}

	public void run(){
	    byte[] state;
	    int i;
	    Updateable[] t_t=null;
	    boolean t_s,error,first=true;
	    int flags=org.omg.CORBA.ARG_IN.value;
	    long timestamp;

	    org.openorb.CORBA.dii.Request dyn_request;
	    org.omg.CORBA.Any argument,any;
	    org.omg.CORBA.NVList dyn_arguments=orb.create_list(0);
	    org.omg.CORBA.ExceptionList dyn_exceptions=orb.create_exception_list();
	    org.omg.CORBA.NamedValue dyn_result;
	    dyn_result=orb.create_named_value("",orb.create_any(),-1);
FTCurrORIServiceContext ctx_ori=null;
	    org.omg.IOP.ServiceContext serviceCtx_ori=new org.omg.IOP.ServiceContext();
	    org.omg.IOP.CodecFactory factory=null;
	    org.omg.IOP.Codec codec=null;

	    while(true){
		state=null;
		error=false;
		if(terminated()) break;
		if(event!=null)
		    wait_update_method(event);
		try{
		    timestamp=System.currentTimeMillis();
		    record_method_call_par(object_id,"get_update",new org.omg.Dynamic.Parameter[0],"",-1,-1,org.omg.FT.Utils.READ,timestamp,transfer_state);
		    work();
		    state=obj_to_update.get_update();
		    fall_asleep();
		    record_rest_get_st_upd_reply("get_update",object_id,state,timestamp);
		}catch(java.lang.Exception ge){error=true;}
		if(terminated()) break;
		if(!error&&state!=null){
		    synchronized(synch_obj){
			t_s=transfer_state;
			if(t_s)
			    t_t=transfer_to;
		    }
		    if(t_s){
			int ret_id=get_last_ret_id(object_id);
			argument=orb.create_any();
			argument.insert_long(ret_id);
			dyn_arguments=orb.create_list(0);
			dyn_arguments.add_value("param0",argument,flags);
			for(i=0;i<t_t.length;i++){
				try{
				 work();
				 ctx_ori=new FTCurrORIServiceContext(object_id,ret_id);
				 any = orb.create_any();
				 serviceCtx_ori=new org.omg.IOP.ServiceContext();
				 serviceCtx_ori.context_id=org.omg.IOP.FT_CURRORI.value;
				 org.omg.FT.FTCurrORIServiceContextHelper.insert(any,ctx_ori);
				 serviceCtx_ori.context_data = codec.encode_value(any);
				 dyn_request=new org.openorb.CORBA.dii.Request(t_t[i],"set_update",serviceCtx_ori,dyn_arguments,dyn_result,orb.create_environment(),dyn_exceptions,orb.create_context_list(),orb);
				 dyn_request.invoke();
				 fall_asleep();
				 if(terminated()) break;
				}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
				if(terminated())
					break;
			}
		    }
		}
		if(terminated()) break;
		if(event==null)
		    try{
			sleep(update_interval/10000);
		    }catch(InterruptedException e){break;}
	    }
	}

	public void set_transfer_to(Updateable[] t_to){
	    synchronized(synch_obj){
		transfer_to=t_to;
	    }
	}


	public void set_transfer_state(boolean v){
	    /* this method is called if (while) working, the replication style
	       changes from (in this case) WARM_PASSIVE, to other style
	       This is not probable to happen yet, in the current implementation*/
	    synchronized(synch_obj){
		transfer_state=v;
	    }
	}

	public synchronized boolean terminated(){
	    return stopped;
	}

	public synchronized void terminate(){
	    stopped=true;
	}

	public long get_obj_id()
	{
	    return object_id;
	}

	   public void wait_working(){
	    synchronized(wait_fall_asleep){
		if(working)
		    try{
			wait_fall_asleep.wait();
		    }catch(InterruptedException e){}
		interrupt();
	    }
	}

	public void work(){
	    synchronized(wait_fall_asleep){
		working=true;
	    }
	}

	public void fall_asleep(){
	    synchronized(wait_fall_asleep){
		working=false;
		wait_fall_asleep.notify();
	    }
	}

    }


    private class CheckpointThObjStr{
	CheckpointThread thread;
	Checkpointable chk_obj;
	long obj_id;

	public CheckpointThObjStr(CheckpointThread thread,Checkpointable chk_obj,long obj_id)
	{
	    this.thread=thread;
	    this.chk_obj=chk_obj;
	    this.obj_id=obj_id;
	}
    }

    private class UpdateThObjStr{
	UpdateThread thread;
	Updateable upd_obj;
	long obj_id;

	public UpdateThObjStr(UpdateThread thread,Updateable upd_obj,long obj_id)
	{
	    this.thread=thread;
	    this.upd_obj=upd_obj;
	    this.obj_id=obj_id;
	}
    }


    java.lang.Object wait_fall_asleep=new java.lang.Object();
    Hashtable logTable;
    LoggingRecoveryController[] neighbours;
    Vector checkpoint_threads,update_threads;
    org.omg.CORBA.ORB orb;


    public LoggingRecoveryController_impl(org.omg.CORBA.ORB o){
	logTable=new Hashtable();
	checkpoint_threads=new Vector();
	update_threads=new Vector();
	orb=o;
    }

/** Method called to obtain the retention identifier of the last outgoing CORBA call from the replicated group
* This identifier is needed in the new recovering member, in order for it to continue sending outgoing requests (if it has to send those)
* with retention identifiers starting not from 0, but from the number given by the last ret id in the failed replica
*/
    private int get_last_ret_id(long objI){
	java.util.Vector logVectorVector,replyLogVector;
	int i,l,toR=-1,j;
	LogRecord crt;
	CheckpointLogRecord clr;

	synchronized(logTable){
	    logVectorVector=logVectorVector(objI);
	    if(logVectorVector==null) {return -1;}

	}

	synchronized(logVectorVector){
	    replyLogVector=(Vector)logVectorVector.elementAt(2);
	    l=replyLogVector.size();


	    for(i=l-1;i>=0;i--){
		crt=(LogRecord)replyLogVector.elementAt(i);
		if(crt.the_log_record.discriminator().value()==RecordType._checkpoint){
		    toR=crt.the_log_record.checkp_record().current_retention_id;
		    break;
		}
		else
		    if(crt.the_log_record.discriminator().value()==RecordType._update){
			toR=crt.the_log_record.update_record().current_retention_id;
			break;
		    }
	    }
	}
	return toR;
    }

    private boolean equals(byte[] a,byte[] b){
	if(a==null) return (b==null);
	if(b==null) return (a==null);
	if(a.length!=b.length) return false;
	for(int i=0;i<a.length;i++)
	    if(a[i]!=b[i]) return false;
	return true;
    }


    private Vector logVectorVector(long id){
	return (Vector)logTable.get(new Long(id));
    }

  /** Method called to obtain the reply record for a call (normal method or get_state/update); if the call is of type checkpoint then
  * the reply is of type state record
  */

    private LogRecord get_its_reply(Vector logVector,String oper,String client_id,int ret_id,int s_pos){
	int i,l;
	LogRecord crt;
	ReplyLogRecord rlr;


	l=logVector.size();
	for(i=s_pos;i<l;i++)
	    if(logVector.elementAt(i).getClass().getName().equals("org.omg.FT.LogRecord")){
		crt=(LogRecord)logVector.elementAt(i);
		if(oper.equals("get_state")){
		    if(crt.the_log_record.discriminator().value()==RecordType._checkpoint)
			return crt;
		}
		else
		    if(oper.equals("get_update")){
			if(crt.the_log_record.discriminator().value()==RecordType._update)
			    return crt;
		    }
		    else
			if(crt.the_log_record.discriminator().value()==RecordType._reply){
			    rlr=crt.the_log_record.reply_record();
			    if(rlr.client_id.equals(client_id)&&rlr.retention_id==ret_id)
				return crt;
			}
	    }
	return null;
    }

/** Method used to record method call information
* @param	target the identifier of the group to whose member (primary) the call was addressed
*			operation the name of the called method
*			buff the byte array of raw data containing the parameters of the call (this buffer will be used later, at replay, if needed, to demarshal parameters)
*			big_e a boolean value indicating whether the ordering of octets is big endian
*			vers the version of the GIOP encoding of information
*			client_id the client identifier of the sender of the request
*			retention_id the retention identifier of the request
*			expiration_time the expiration time of the sent request
*			method_type the type of the called method (update or read-only)
*			broadcast if replication style is warm passive this parameter has the value true
* @return	a result record that is not null or empty in case the request was already executed and the result was stored
*/


    public ResultResultArgs record_method_call_as(long target,String operation,byte[] buff,boolean big_e,org.omg.GIOP.Version vers,String client_id,int retention_id,long expiration_time,char method_type,boolean broadcast){

	boolean call_logged=false,to_wait_r=false,to_wait_u=false;
	LogRecord crt;
	CallLogRecord clr;
	CallLogRecordAS clr_as;
	CallLogRecordPAR clr_par;
	ReplyLogRecord rlr;
	Vector logVectorVector,callLogVector,countersVector,outg_r_rLogVector,replyLogVector;
	int i;

	/* look first if there is an answer to the call or at least if the call is already logged*/
	ResultResultArgs _default=new ResultResultArgs(orb.create_any(),new org.omg.CORBA.Any[0]);

	synchronized(logTable){
	    logVectorVector=logVectorVector(target);
	    if(logVectorVector==null){
		logVectorVector=new Vector();
		for(i=0;i<4;i++)
		    logVectorVector.addElement(new Vector());
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new java.lang.Object()); /* synch object*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* expected id*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* current message_id*/
		Long hashcode=new Long(target);
		logTable.put(hashcode,logVectorVector);
	    }

	}

	if(logVectorVector==null) return _default;


	synchronized(logVectorVector){
	       countersVector=(Vector)logVectorVector.elementAt(0);
	    callLogVector=(Vector)logVectorVector.elementAt(1);
	    replyLogVector=(Vector)logVectorVector.elementAt(2);
	    outg_r_rLogVector=(Vector)logVectorVector.elementAt(3);
	    int l=callLogVector.size(),l1=replyLogVector.size(),disc,nr_to_wait=0;

	    MyLong nr_upd=(MyLong)countersVector.elementAt(1);
	    nr_to_wait=(int)nr_upd.get();

	    if(method_type!=Utils.READ||operation.equals("get_state")||operation.equals("get_update")){
		nr_upd.incr();
	    }

	    for(i=0;i<l1;i++){
		crt=(LogRecord)replyLogVector.elementAt(i);
		if(crt.the_log_record.discriminator().value()==RecordType._reply){
		    rlr=crt.the_log_record.reply_record();
		    if(rlr.client_id.equals(client_id)&&rlr.retention_id==retention_id){
			if(method_type!=Utils.READ)
			    nr_upd.decr();
			return new ResultResultArgs(rlr.reply,rlr.reply_arguments);
		    }
		}
	    }

	    /* In warm passive replication when requests are broadcasted to the neighbours is not necessary to
	       record the method calls in the primary's log*/

	    if(!broadcast){
		if(!operation.equals("get_state")&&!operation.equals("get_update")){
		    if(method_type==org.omg.FT.Utils.READ) return _default;
		    for(i=0;i<l;i++){
			crt=(LogRecord)callLogVector.elementAt(i);
			clr=org.omg.FT.Utils.get_clr(crt.the_log_record.discriminator().value(),crt.the_log_record);
			if(clr.retention_id==retention_id&&clr.client_id.equals(client_id)){
			    call_logged=true;
			    break;
			}
		    }
		}
		if(!call_logged){
		    SomeLogRecord slr=new SomeLogRecord();
		    slr.call_record_as(new CallLogRecordAS(new CallLogRecord(operation,client_id,retention_id,expiration_time,true,method_type),buff,big_e,vers));
		    callLogVector.addElement(new LogRecord(target,slr));
		    if(operation.equals("get_state")||operation.equals("get_update")){
			outg_r_rLogVector.addElement(new LogRecord(target,slr));
			((MyLong)countersVector.elementAt(3)).set(outg_r_rLogVector.size()-1);
		    }
		}
	    }
	    java.lang.Object last;
	    long replies_in_bw=0,replies_in_aw;
	    switch(method_type){
	    case org.omg.FT.Utils.UPDATE_NON_VOID:
	    case org.omg.FT.Utils.UPDATE_VOID:
	    case org.omg.FT.Utils.UPDATE_ONEWAY:
		while(nr_to_wait>0){
		    try{
			if(logVectorVector==null) return _default;
			replies_in_bw=((MyLong)countersVector.elementAt(2)).get();
			logVectorVector.wait();
		    }catch(InterruptedException ex){}
		    replies_in_aw=((MyLong)countersVector.elementAt(2)).get();
		    nr_to_wait-=replies_in_aw-replies_in_bw;
		    if(logVectorVector==null) return _default;
		    last=logVectorVector.lastElement();
		    if(last!=null&&last instanceof String) return _default;
		}
		trigger_one_thread(target);
		break;
	    case org.omg.FT.Utils.READ:
		if(!operation.equals("get_state")&&!operation.equals("get_update")) return _default;
		while(nr_to_wait>0){
		    try{
			if(logVectorVector==null) return _default;
			replies_in_bw=((MyLong)countersVector.elementAt(2)).get();
			logVectorVector.wait();
		    }catch(InterruptedException ex){}
		     replies_in_aw=((MyLong)countersVector.elementAt(2)).get();
		    nr_to_wait-=replies_in_aw-replies_in_bw;
		    if(logVectorVector==null) return _default;
		    last=logVectorVector.lastElement();
		    if(last!=null&&last instanceof String) return _default;
		}
		break;
	    }

	    if(broadcast&&(method_type!=Utils.READ||operation.equals("get_state")||operation.equals("get_update"))){
		long message_id=(long)((MyLong)countersVector.elementAt(7)).get();
		((MyLong)countersVector.elementAt(7)).incr();
		broadcast_method_call_as(target,operation,buff,big_e,vers,client_id,retention_id,expiration_time,method_type,message_id);
	    }
	    return _default;
	}
    }


/** Method used to record method call information (giving explicit list of parameters)
* @param	target the identifier of the group to whose member (primary) the call was addressed
*			operation the name of the called method
*			arguments the list of call parameters (value of parameters and their mode - IN, OT, INOUT)
*			client_id the client identifier of the sender of the request
*			retention_id the retention identifier of the request
*			expiration_time the expiration time of the sent request
*			method_type the type of the called method (update or read-only)
*			timestamp the moment when the call information is recorded
*			broadcast if replication style is warm passive this parameter has the value true
* @return	a result record that is not null or empty in case the request was already executed and the result was stored
*/

      public ResultResultArgs record_method_call_par(long target,String operation,org.omg.Dynamic.Parameter[] arguments,String client_id,int retention_id,long expiration_time,char method_type,long timestamp,boolean broadcast){

	boolean call_logged=false,to_wait_r=false,to_wait_u=false;
	LogRecord crt;
	CallLogRecord clr;
	ReplyLogRecord rlr;
	Vector logVectorVector,callLogVector,replyLogVector,countersVector,outg_r_rLogVector;
	int i,disc;

	/* look first if there is an answer to the call or at least if the call is already logged*/

	ResultResultArgs _default=new ResultResultArgs(orb.create_any(),new org.omg.CORBA.Any[0]);


	synchronized(logTable){
	    logVectorVector=logVectorVector(target);
	    if(logVectorVector==null){
		logVectorVector=new Vector();
		for(i=0;i<4;i++)
		    logVectorVector.addElement(new Vector());
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new java.lang.Object()); /* synch object*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* expected id*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* current message_id*/
		Long hashcode=new Long(target);
		logTable.put(hashcode,logVectorVector);
	    }

	}

	if(logVectorVector==null) return _default;

	synchronized(logVectorVector){
	     countersVector=(Vector)logVectorVector.elementAt(0);
	    callLogVector=(Vector)logVectorVector.elementAt(1);
	    replyLogVector=(Vector)logVectorVector.elementAt(2);
	    outg_r_rLogVector=(Vector)logVectorVector.elementAt(3);
	    int l=callLogVector.size(),l1=replyLogVector.size(),nr_to_wait=0;

	    MyLong nr_upd=(MyLong)countersVector.elementAt(1);
	    nr_to_wait=(int)nr_upd.get();

	    if(method_type!=Utils.READ||operation.equals("get_state")||operation.equals("get_update")){
		nr_upd.incr();
	    }


	    for(i=0;i<l1;i++){
		crt=(LogRecord)replyLogVector.elementAt(i);
		if(crt.the_log_record.discriminator().value()==RecordType._reply){
		    rlr=crt.the_log_record.reply_record();
		    if(rlr.client_id.equals(client_id)&&rlr.retention_id==retention_id)
			return new ResultResultArgs(rlr.reply,rlr.reply_arguments);
		}
	    }

	    if(!broadcast){
		if(!operation.equals("get_state")&&!operation.equals("get_update")){
		    if(method_type==org.omg.FT.Utils.READ) return _default;
		    for(i=0;i<l;i++){
			crt=(LogRecord)callLogVector.elementAt(i);
			clr=org.omg.FT.Utils.get_clr(crt.the_log_record.discriminator().value(),crt.the_log_record);
			if(clr.retention_id==retention_id&&clr.client_id.equals(client_id)){
			    call_logged=true;
			    break;
			}
		    }
		}
		if(!call_logged){
		    SomeLogRecord slr=new SomeLogRecord();
		    slr.call_record_par(new CallLogRecordPAR(new CallLogRecord(operation,client_id,retention_id,expiration_time,true,method_type),arguments,timestamp));
		    callLogVector.addElement(new LogRecord(target,slr));
		    if(operation.equals("get_state")||operation.equals("get_update")){
			outg_r_rLogVector.addElement(new LogRecord(target,slr));
			((MyLong)countersVector.elementAt(3)).set(outg_r_rLogVector.size()-1);
		    }
		}
	    }
	    java.lang.Object last;
	    long replies_in_bw=0,replies_in_aw;
	    switch(method_type){
	    case org.omg.FT.Utils.UPDATE_NON_VOID:
	    case org.omg.FT.Utils.UPDATE_VOID:
	    case org.omg.FT.Utils.UPDATE_ONEWAY:

		while(nr_to_wait>0){
		    try{
			if(logVectorVector==null) return _default;
			replies_in_bw=((MyLong)countersVector.elementAt(2)).get();
			logVectorVector.wait();
		    }catch(InterruptedException ex){}
		    replies_in_aw=((MyLong)countersVector.elementAt(2)).get();
		    nr_to_wait-=replies_in_aw-replies_in_bw;
		    if(logVectorVector==null) return _default;
		    last=logVectorVector.lastElement();
		    if(last!=null&&last instanceof String) return _default;

		}
		trigger_one_thread(target);
		break;
	    case org.omg.FT.Utils.READ:
		if(!operation.equals("get_state")&&!operation.equals("get_update"))
			return _default;
		while(nr_to_wait>0){
		    try{
			replies_in_bw=((MyLong)countersVector.elementAt(2)).get();
			logVectorVector.wait();
		    }catch(InterruptedException ex){}
		    replies_in_aw=((MyLong)countersVector.elementAt(2)).get();
		    nr_to_wait-=replies_in_aw-replies_in_bw;
		    if(logVectorVector==null) return _default;
		    last=logVectorVector.lastElement();
		    if(last!=null&&last instanceof String) return _default;
		}
		break;
	    }
	    if(broadcast&&(method_type!=Utils.READ||operation.equals("get_state")||operation.equals("get_update"))){
		long message_id=(long)((MyLong)countersVector.elementAt(7)).get();
		((MyLong)countersVector.elementAt(7)).incr();
		broadcast_method_call_par(target,operation,arguments,client_id,retention_id,expiration_time,method_type,message_id);
	    }
	    return _default;
	}
    }


    /* Method for recording the ORB based retention_id of the REPLICATED CLIENT*/


/** Method called to obtain the reply record (i present int he log) to a call with given identifier
* @param	target the identifier of the group to whose member the call identified by the client_id, and ret_id was sent
*			client_id the client identifier of the sender of the method call for which the reply is searched for
*			retention_id the retention identifier of the request for whose reply it is searched for
* @return 	a proper reply record for the specified call, if this reply exists in the log
*/

    public ResultResultArgs call_reply(long target,String client_id,int retention_id){
	ReplyLogRecord rlr;
	int l,i;
	LogRecord crt;
	Vector logVectorVector,replyLogVector;

	ResultResultArgs _default=new ResultResultArgs(orb.create_any(),new org.omg.CORBA.Any[0]);

	synchronized(logTable){
	    logVectorVector=logVectorVector(target);
	    if(logVectorVector==null) return _default;

	}

	synchronized(logVectorVector){
	    replyLogVector=(Vector)logVectorVector.elementAt(1);
	    l=replyLogVector.size();
	    for(i=0;i<l;i++){
		crt=(LogRecord)replyLogVector.elementAt(i);
		if(crt.the_log_record.discriminator().value()==RecordType._reply){
		    rlr=crt.the_log_record.reply_record();
		    if(rlr.client_id.equals(client_id)&&rlr.retention_id==retention_id)
			return new ResultResultArgs(rlr.reply,rlr.reply_arguments);
		}
	    }
	    return _default;
	}
    }

/** Method called to return a reply (if existent) to an outgoing request from this member of the replica group
* @param	cl_id the client identifier of the sender replica (of the group)
*			retention_id the retention identifier of the request sent by the replica to the third-tier server
*			to the CORBA object designating the destination third-tier server to which the request is sent
* @return	a proper reply record stored in a log, if the request was once sent out, executed, and a reply was stored in the log
*/

    public ResultResultArgs request_reply(String cl_id,int retention_id,org.omg.CORBA.Object to){
	int l,i;
	LogRecord crt;
	RequestReplyLogRecord rlr;
	boolean found=false;

	long from_where=Utils.extract_oid_l(cl_id);

	Vector logVectorVector,outg_reqLogVector;
	ResultResultArgs _default=new ResultResultArgs(orb.create_any(),new org.omg.CORBA.Any[0]);
	  try{
	synchronized(logTable){
	    logVectorVector=logVectorVector(from_where);

	    if(logVectorVector==null) {return _default;}
	    outg_reqLogVector=(Vector)logVectorVector.elementAt(3);

	}

	synchronized(outg_reqLogVector){

		l=outg_reqLogVector.size();
		for(i=0;i<l;i++){
		    crt=(LogRecord)outg_reqLogVector.elementAt(i);
		    if(crt.the_log_record.discriminator().value()==RecordType._request_reply){
			rlr=crt.the_log_record.request_r_record();
			if(rlr.retention_id==retention_id&&Utils.refs_equal(orb,rlr.server,to,-1)){
			    return new ResultResultArgs(rlr.reply,rlr.arguments);
			}
		    }
		}
	}
	  }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	  return _default;
    }

/** Method used to write in the log a set of call replies
* @param	replies the set of call replies to be written in the log
*/

    public void record_replies(LogRecord[] replies){
	int i;
	Vector logVectorVector,replyLogVector;
	boolean put=false;
	try{
	    if(replies.length<1) {return;}

	    logVectorVector=logVectorVector(replies[0].target);

	    if(logVectorVector==null){
		for(i=0;i<4;i++)
		    logVectorVector.addElement(new Vector());
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new java.lang.Object()); /* synch object*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* expected id*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* current message_id*/
		Long hashcode=new Long(replies[0].target);
		logTable.put(hashcode,logVectorVector);
	    }

	    replyLogVector=(Vector)logVectorVector.elementAt(2);

	    for(i=0;i<replies.length;i++)
		replyLogVector.addElement(replies[i]);
	}catch(java.lang.Exception e){e.printStackTrace();}
    }

/** Method used to log the reply of a call to a method on the server replica
* @param	source the identifier of the group whose member is the replica on which the call was previously made
*			client_id the client identifier of the caller of the method for which the reply is recorded now
*			retention_id the retention identifier of the call for which the reply is recorded now
*			reply the content of the reply as return value of the called method
*			reply_arguments the content of the reply as out/inout arguments that are written by the called method
*			method_type the type of the method (update or read-only) - used to know whether after recordong of the reply, the waiting methods
*			haveto be noitified or not
*			broadcast it is true in case of warm passive replication; in this case the reply has to be broadcast to the rest of replicas and logged there
*/

    public void record_reply(long source,String client_id,int retention_id,org.omg.CORBA.Any reply,org.omg.CORBA.Any[] reply_arguments,char method_type,boolean broadcast){
	/* search for the corresponding request message - if it was timed out, it is not
	   anymore in the list, and then the reply is not written either*/
       	Vector logVectorVector,replyLogVector,countersVector;
	long time=System.currentTimeMillis();

	synchronized(logTable){
	    logVectorVector=logVectorVector(source);
	    if(logVectorVector==null) return;

	}
	if(logVectorVector==null) return;
	synchronized(logVectorVector){
	    replyLogVector=(Vector)logVectorVector.elementAt(2);
	    countersVector=(Vector)logVectorVector.elementAt(0);

	    if(method_type!=Utils.READ){
		((MyLong)countersVector.elementAt(1)).decr();
		((MyLong)countersVector.elementAt(2)).incr();
	    }
	    SomeLogRecord slr=new SomeLogRecord();
	    slr.reply_record(new ReplyLogRecord(client_id,retention_id,0,reply,reply_arguments,method_type));
	    if(logVectorVector==null) return;
	    replyLogVector.addElement(new LogRecord(source,slr));
	    if(method_type!=Utils.READ)
		logVectorVector.notifyAll();
	    if(broadcast)
		broadcast_reply(source,client_id,retention_id,reply,reply_arguments,method_type);
	}
    }


/** This method is used when stateless or active replication is used, and the reply wants to be recorded; in these two cases there is no need for the method
* call for which information is recorded to wait for other methods executing before it to finish
* @param	objI the identifier of the group on whose member the method for which the reply is recorded was called
*			client_id the client identifier of the request corresponding to the method call
*			retention_id the retention identifier of the request correspondng to the method call
*			expiration_time the expiration time of the request
*			reply the content of the reply to the request as a return value
*			reply_arguments the content of the reply to the request as a set of OUT/INOUT arguments (if any) that were writen by the method call
*/
       public void record_reply_stateless(long objI,String client_id,int retention_id,long expiration_time,org.omg.CORBA.Any reply,org.omg.CORBA.Any[] reply_arguments){

	java.util.Vector logVectorVector,replyLogVector,toRemove=new java.util.Vector();
	int i,l;
	long now_time;
	LogRecord crt;
	ReplyLogRecord rlr;
	boolean put=false;

	synchronized(logTable){
	    logVectorVector=logVectorVector(objI);

	    if(logVectorVector==null){
		logVectorVector=new Vector();
		for(i=0;i<4;i++)
		    logVectorVector.addElement(new Vector());
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new java.lang.Object()); /* synch object*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* expected id*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* current message_id*/
		Long hashcode=new Long(objI);
		logTable.put(hashcode,logVectorVector);
	    }

	}

	synchronized(logVectorVector){
	    replyLogVector=(Vector)logVectorVector.elementAt(2);
	    now_time=System.currentTimeMillis();
	    l=replyLogVector.size();
	    for(i=0;i<l;i++){
		crt=(LogRecord)replyLogVector.elementAt(i);
		if(crt.the_log_record.discriminator().value()==RecordType._reply){
		    rlr=crt.the_log_record.reply_record();
		    if(now_time>rlr.expiration_time)
			toRemove.addElement(crt);
		}
	    }

	    replyLogVector.removeAll(toRemove);
	    SomeLogRecord slr=new SomeLogRecord();
	    slr.reply_record(new ReplyLogRecord(client_id,retention_id,expiration_time,reply,reply_arguments,'r'));
	    replyLogVector.addElement(new LogRecord(objI,slr));
	}
       }

/** Method used to record a reply to an outgoing request from the replica group (or even from a normal client that calls a server and has an interceptor that
* calls this method whnever a reply is received back)
* @param	cl_id the client identifier of the request for which the reply is logged now
*			to the target object of the request for which the reply is recorded now
*			arguments the set of OUT/INOUT parameters of the request that are written and returned by the method call
*			retention_id the retention identifier of the request sent by the client
*			reply the return value of the request
*/
    public  void record_request_repl(String cl_id,org.omg.CORBA.Object to,String operation,org.omg.CORBA.Any[] arguments,int retention_id,org.omg.CORBA.Any reply){
	int l,i;
	LogRecord crt;
	RequestLogRecord rlr;
	boolean put=false;
	long from_where=Utils.extract_oid_l(cl_id);
	Vector logVectorVector,outg_reqLogVector;


	synchronized(logTable){
	    logVectorVector=logVectorVector(from_where);
	    if(logVectorVector==null){
		logVectorVector=new Vector();
		for(i=0;i<4;i++)
		    logVectorVector.addElement(new Vector());
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new java.lang.Object()); /* synch object*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* expected id*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* current message_id*/
		Long hashcode=new Long(from_where);
		logTable.put(hashcode,logVectorVector);
	    }
	    outg_reqLogVector=(Vector)logVectorVector.elementAt(3);
	}

	synchronized(outg_reqLogVector){
	    SomeLogRecord slr=new SomeLogRecord();
	    slr.request_r_record(new RequestReplyLogRecord(operation,cl_id,arguments,retention_id,to,reply));
	    outg_reqLogVector.addElement(new LogRecord(-1,slr));
	}
    }


/** Method used to record a part of the reply to the get_state or get_update request that is sent by the chekpoint/update thread to the
* application replica - this operation is needed because the information recorded here is no longer available when the final state information
* is recorded
* @param	object the identifier of the group whose member is the checkpointed replica
*			current_retention_id the current retention identifier of outgoing CORBA calls from the replica whose state was checkpointed now
*			oper the name of the operation:get_state or get_update
*/

    public void record_part_get_st_upd_reply(long object,int current_retention_id,String oper){
	Vector logVector,logVectorVector,replyLogVector,countersVector;
	Integer index;

	synchronized(logTable){
	    logVectorVector=logVectorVector(object);
	    if(logVectorVector==null) {System.out.println("OOps!!Not found");return;}


	}

	synchronized(logVectorVector){
	    replyLogVector=(Vector)logVectorVector.elementAt(2);
	    countersVector=(Vector)logVectorVector.elementAt(0);
	    SomeLogRecord slr=new SomeLogRecord();
	    if(oper.equals("get_state"))
		slr.checkp_record(new CheckpointLogRecord(new byte[0],-1,current_retention_id));
	    else
		slr.update_record(new UpdateLogRecord(new byte[0],-1,current_retention_id));
	    replyLogVector.addElement(new LogRecord(object,slr));
	    try{
		((MyLong)countersVector.elementAt(0)).set(replyLogVector.size()-1);
	    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	}
    }

/**Method used to record the rest (state/update) of the reply to the get_state or get_update request that is sent by the chekpoint/update thread to the
* application replica
* @param	oper the name of the operation:get_state or get_update
*			object the identifier of the group whose member is the checkpointed replica
*			state the state/update that is recorded after the checkpointing method was called
*			timestamp a timestamp indicating to which get_state/get_update operation in the call log this reply record corresponds
*/

     protected void record_rest_get_st_upd_reply(String oper,long object,byte[] state,long timestamp){
	int i,l,index=-1,l1,k,disc,pos_g_s_orr=-1;
	long now_time;
	LogRecord crt,toR;
	CallLogRecord crt_c;
	ReplyLogRecord rlr;
	CheckpointLogRecord crt_cp;
	UpdateLogRecord crt_upd;
	Vector toRemoveC=new Vector(),toRemoveR=new Vector();
	boolean is_first_g_s_u=false,is_f_g_t;
	Vector logVectorVector,replyLogVector,countersVector,callLogVector,outg_r_rLogVector;

	synchronized(logTable){
	    logVectorVector=logVectorVector(object);
	    if(logVectorVector==null) {System.out.println("It is null");return;}

	}
	synchronized(logVectorVector){
	    replyLogVector=(Vector)logVectorVector.elementAt(2);
	    callLogVector=(Vector)logVectorVector.elementAt(1);
	    countersVector=(Vector)logVectorVector.elementAt(0);
	    outg_r_rLogVector=(Vector)logVectorVector.elementAt(3);
	    try{
		index=(int)((MyLong)countersVector.elementAt(0)).get();
		((MyLong)countersVector.elementAt(0)).set(-1);
		pos_g_s_orr=(int)((MyLong)countersVector.elementAt(3)).get();
		((MyLong)countersVector.elementAt(3)).set(-1);
		((MyLong)countersVector.elementAt(2)).incr();
		((MyLong)countersVector.elementAt(1)).decr();
		MyLong is_first=(MyLong)countersVector.elementAt(4);
		if(is_first.get()==-1){
		    is_first_g_s_u=true;
		    is_first.set(0);
		}
	    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}

	    now_time=System.currentTimeMillis();
	    crt=(LogRecord)replyLogVector.elementAt(index);
	    disc=crt.the_log_record.discriminator().value();
	    if(disc==RecordType._checkpoint&&oper.equals("get_state")){
		crt_cp=crt.the_log_record.checkp_record();
		crt_cp.state=new byte[state.length];
		crt.the_log_record.checkp_record().timestamp=timestamp;
		System.arraycopy(state,0,crt_cp.state,0,state.length);
	    }
	    else
		if(disc==RecordType._update&&oper.equals("get_update"))
		    {
			crt_upd=crt.the_log_record.update_record();
			crt_upd.state=new byte[state.length];
			crt.the_log_record.update_record().timestamp=timestamp;
			System.arraycopy(state,0,crt_upd.state,0,state.length);
		    }
	    /* Remove all outgoing calls and replies which were registered (logged) before the get_state
	       call was issued; pos_g_s_orr is the position were get_state call was recorded in the outg
	       req-repl table (NOTE: The get_state call is recorded there just for reference purposes, to know
	       from were to start the removal of the records)*/

	    if(pos_g_s_orr>=0)
		outg_r_rLogVector.subList(0,pos_g_s_orr+1).clear();

	    l=callLogVector.size();
	    l1=replyLogVector.size();

	    boolean exit_loop=false;

	    is_f_g_t=is_first_g_s_u;

	    if(oper.equals("get_update")){
		for(k=l1-1;k>=0;k--){
		    crt=(LogRecord)replyLogVector.elementAt(k);
		    if(crt.the_log_record.discriminator().value()==RecordType._update)
			break;
		}
		for(i=0;i<k;i++){
		    crt=(LogRecord)replyLogVector.elementAt(i);
		    if(crt.the_log_record.discriminator().value()==RecordType._reply){
			rlr=crt.the_log_record.reply_record();
			if(rlr.expiration_time<now_time)
			    toRemoveR.addElement(crt);
		    }
		}
	    }
	    else{
		for(k=0;k<l1;k++){
		    crt=(LogRecord)replyLogVector.elementAt(k);
		    disc=crt.the_log_record.discriminator().value();
		    if(disc==RecordType._reply){
			rlr=crt.the_log_record.reply_record();
			if(rlr.expiration_time<now_time)
			    toRemoveR.addElement(crt);
		    }
		    else
			if(disc==RecordType._checkpoint)
			    if(!is_first_g_s_u){
				toRemoveR.addElement(crt);
				is_first_g_s_u=true;
			    }
			    else
				break;
		}
	    }

	    is_first_g_s_u=is_f_g_t;

	    for(k=0;k<l;k++){
		crt=(LogRecord)callLogVector.elementAt(k);
		disc=crt.the_log_record.discriminator().value();
		if(disc==RecordType._call_par||disc==RecordType._call_as){
		    crt_c=org.omg.FT.Utils.get_clr(disc,crt.the_log_record);
		    if(crt_c.operation.equals(oper))
			if(!is_first_g_s_u){
			    toRemoveC.addElement(crt);
			    is_first_g_s_u=true;
			}
			else{
			    break;
			}
		    else
			if(crt_c.expiration_time<now_time)
			    toRemoveC.addElement(crt);
		}
	    }
	    callLogVector.removeAll(toRemoveC);
	    replyLogVector.removeAll(toRemoveR);
	    logVectorVector.notifyAll();
	}
     }

/** This method is used to signal when in the server interceptor and an exception is sent to the client, that the method was executed; thus, in case it has the type
* "update" the waiting method calls have to be notified
* @param	source the identifier of the group on whose replica the method generating the exception was called
*			client_id the client identifier of the request
*			retention_id the retention identifier of the request
*			method_type	the type of the method (update/read-only)
*/

    public void record_reply_exc(long source,String client_id,int retention_id,char method_type){
	/* search for the corresponding request message - if it was timed out, it is not
	   anymore in the list, and then the reply is not written either*/
	int i,l,disc;
	LogRecord crt;
	CallLogRecord clr;
	boolean found=false;
	Vector logVectorVector,countersVector;



	synchronized(logTable){
	    logVectorVector=logVectorVector(source);
	    if(logVectorVector==null) return;

	}

	synchronized(logVectorVector){
	    countersVector=(Vector)logVectorVector.elementAt(0);
	    if(method_type!=Utils.READ){
		((MyLong)countersVector.elementAt(1)).decr();
		((MyLong)countersVector.elementAt(2)).incr();
		logVectorVector.notifyAll();
	    }
	}
    }

/** Method used in a backup when warm passive replication is used, as soon as the set_state  method is called on that backup when the state is read from the
* primary; the "reaction" to that call of set_state is that the method calls logged since the last checkpoint are removed - this is what happens in the method
* react_set_....
* @param 	object the identifier of the group of which this backup is a member
*/

    public  void react_set_st_upd_reply(long object){
	int i,l,j,l1,disc;
	long now_time;
	LogRecord crt;
	CallLogRecord clr;
	ReplyLogRecord rlr;
	Vector toRemoveC=new Vector(),toRemoveR=new Vector();

	Vector logVectorVector,callLogVector,replyLogVector;

	synchronized(logTable){
	    logVectorVector=logVectorVector(object);
	    if(logVectorVector==null) {return;}

	}
	synchronized(logVectorVector){
	    callLogVector=(Vector)logVectorVector.elementAt(1);
	    replyLogVector=(Vector)logVectorVector.elementAt(2);
	    l=callLogVector.size();
	    l1=replyLogVector.size();
	    now_time=System.currentTimeMillis();
	    /* remove all expired replies from the log
	       NOTE: in case of warm passive replication, when transfer of requests to the
	       backups happens, only the calls are transfered for update methods, and only
	       replies for the read methods. Therefore, the replies in the reply log are
	       of read requests*/
	    try{
		for(i=0;i<l1;i++){
		    crt=(LogRecord)replyLogVector.elementAt(i);
		    if(crt.the_log_record.discriminator().value()==RecordType._reply){
			rlr=crt.the_log_record.reply_record();
			if(rlr.expiration_time<now_time)
			    toRemoveR.addElement(crt);
		    }
		}

		for(i=l-1;i>=0;i--){
		    crt=(LogRecord)callLogVector.elementAt(i);
		    clr=org.omg.FT.Utils.get_clr(crt.the_log_record.discriminator().value(),crt.the_log_record);
		    if(clr.operation.equals("get_state")||clr.operation.equals("get_update")) {
			/* for the warm backups, the get_state call record
			   is logged in the log, whenever a get_state happens
			   in the warm primary, in order to mark the point
			   where the state was taken*/
			for(j=0;j<i;j++){
			    crt=(LogRecord)callLogVector.elementAt(j);
			    clr=org.omg.FT.Utils.get_clr(crt.the_log_record.discriminator().value(),crt.the_log_record);
			    if(clr.operation.equals("get_state"))
				toRemoveC.addElement(crt);
			    else
				if(clr.expiration_time<now_time)
				    toRemoveC.addElement(crt);
			}
			break;
		    }
		}
		callLogVector.removeAll(toRemoveC);
		replyLogVector.removeAll(toRemoveR);
	    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	}
    }

/** Method called (mainly by the Replication Manager) when a new replica of a group is installed and needs to be checkpointed
* @param	checkpointed_obj the reference of the CORBA object replica that has to be checkpointed
*			obj_id the identifier of the group of which the checkpointed object is a member
*			checkpoint_interval the checkpointing interval (if checkpointing is periodic)
*			transfer_to in case of warm passive replication this contains the list of backups where the state has to be transfered
*			transfer a boolean variable indicating whether the state has to be transfered to the backups or not
*			periodic indicates whether checkpointing has to be done periodically
*/

    public void do_checkpointing(Checkpointable checkpointed_obj,long obj_id,long checkpoint_interval,Checkpointable[] transfer_to,boolean transfer,boolean periodic){
	if(checkpoint_interval<0&&periodic) {return;}
	NotifySObj event_object;
	if(periodic)
	    event_object=null;
	else
	    event_object=new NotifySObj();
	CheckpointThread ct=new CheckpointThread(checkpointed_obj,obj_id,checkpoint_interval,transfer_to,transfer,event_object);
	checkpoint_threads.addElement(new CheckpointThObjStr(ct,checkpointed_obj,obj_id));
	ct.start();
    }

/** Method called (mainly by the Replication Manager) when a new replica of a group is installed and needs to be checkpointed by reading state changes (updates)
* from time to time
* @param	updated_obj the reference of the CORBA object replica that has to be checkpointed by update readings
*			obj_id the identifier of the group of which the checkpointed object is a member
*			update_interval the checkpointing interval (if checkpointing is periodic)
*			transfer_to in case of warm passive replication this contains the list of backups where the state update has to be transfered
*			transfer a boolean variable indicating whether the state update has to be transfered to the backups or not
*			periodic indicates whether checkpointing has to be done periodically
*/

    public void do_updating(Updateable updated_obj,long obj_id,long update_interval,Updateable[] transfer_to,boolean transfer,boolean periodic){
	if(update_interval<0&&periodic) {return;}
	NotifySObj event_object;
	if(periodic)
	    event_object=null;
	else
	    event_object=new NotifySObj();
	UpdateThread ut=new UpdateThread(updated_obj,obj_id,update_interval,transfer_to,transfer,event_object);
	update_threads.addElement(new UpdateThObjStr(ut,updated_obj,obj_id));
	ut.start();
    }


    protected void wait_update_method(NotifySObj event){
	synchronized(event){
	    if(event.nr_notify<=0)
		try{
		    event.wait();
		}catch(java.lang.InterruptedException e){}
	    if(event.nr_notify>0)
		event.nr_notify--;
	}
    }

    private void trigger_one_thread(long obj_id){

	CheckpointThObjStr crt_c=null;
	UpdateThObjStr crt_u=null;
	boolean found=false;
	int i,l1,l2;

	l1=checkpoint_threads.size();
	l2=update_threads.size();

	for(i=0;i<l1;i++){
	    crt_c=(CheckpointThObjStr)checkpoint_threads.elementAt(i);
	    if(obj_id==crt_c.obj_id){
		if(crt_c.thread.event==null) break;
		synchronized(crt_c.thread.event){
		    crt_c.thread.event.nr_notify++;
		    crt_c.thread.event.notify();
		}
		break;
	    }
	}
	for(i=0;i<l2;i++){
	    crt_u=(UpdateThObjStr)update_threads.elementAt(i);
	    if(obj_id==crt_u.obj_id){
		if(crt_u.thread.event==null) break;
		synchronized(crt_u.thread.event){
		    crt_u.thread.event.nr_notify++;
		    crt_u.thread.event.notify();
		}
		break;
	    }
	}
    }

    private CheckpointThread get_ch_thread(Checkpointable checkpointed_obj){
	int i,l;
	CheckpointThObjStr crt=null;
	boolean found=false;

	l=checkpoint_threads.size();

	for(i=0;i<l;i++){
	    crt=(CheckpointThObjStr)checkpoint_threads.elementAt(i);
	    if(Utils.refs_equal(_orb(),checkpointed_obj,crt.chk_obj,-1)){
		found=true;
		break;
	    }
	}
	if(found){
	    checkpoint_threads.remove(i);
	    return crt.thread;
	}
	return null;
    }

       private UpdateThread get_up_thread(Updateable updated_obj){
	int i,l;
	UpdateThObjStr crt=null;
	boolean found=false;

	l=update_threads.size();

	for(i=0;i<l;i++){
	    crt=(UpdateThObjStr)update_threads.elementAt(i);
	    if(Utils.refs_equal(_orb(),updated_obj,crt.upd_obj,-1)){
		found=true;
		break;
	    }
	}
	if(found){
	    update_threads.remove(i);
	    return crt.thread;
	}
	return null;
    }

/** Method called (usually by the Replication Manager when a member of the group failed) to stop the checkpointing of the replica
*/

    public synchronized void stop_checkpointing(Checkpointable checkpointed_obj){
	CheckpointThread ct;


	ct=get_ch_thread(checkpointed_obj);
	if(ct==null) {return;}
	ct.terminate();
	ct.wait_working();
	ct.interrupt();
    }

/** Method called (usually by the Replication Manager when a member of the group failed) to stop the checkpointing(state change reading) of the replica
*/

       public synchronized void stop_updating(Updateable updated_obj){
	UpdateThread ut;


	ut=get_up_thread(updated_obj);
	if(ut==null) {return;}
	ut.terminate();
	ut.wait_working();
    }



    private void remove_wp_log(long id){
	Vector logVectorVector=logVectorVector(id);
	synchronized(logVectorVector){
	    logVectorVector.addElement(new String("over"));
	    logVectorVector.notifyAll();
	    logTable.remove(new Long(id));
	}
    }

    private LogRecord[] get_last_records_ch(long id,Vector logVectorVector,boolean clear){
	int l,i,l1,l2,disc;
	LogRecord crt,crt_r;
	Vector tempRetR,tempRetC,tempRet,callLogVector,replyLogVector;
	LogRecord[] toRet=new LogRecord[0];
	CallLogRecord clr;
	ReplyLogRecord rlr;
	CallLogRecordPAR g_s_u_rec;
	long timestamp=-1;

	if(logVectorVector==null) return new LogRecord[0];
	synchronized(logVectorVector){
	    try{
		callLogVector=(Vector)logVectorVector.elementAt(1);
		l1=callLogVector.size();
		replyLogVector=(Vector)logVectorVector.elementAt(2);
		l=replyLogVector.size();
	       	tempRetR=new Vector();
		tempRetC=new Vector();

		for(i=l-1;i>=0;i--){
		    crt=(LogRecord)replyLogVector.elementAt(i);
		    disc=crt.the_log_record.discriminator().value();
		    if(disc==RecordType._checkpoint){
			timestamp=crt.the_log_record.checkp_record().timestamp;
			tempRetR.insertElementAt(crt,0);
			break;
		    }
		    else
			if(disc==RecordType._reply){
			    rlr=crt.the_log_record.reply_record();
			    if(rlr.type==Utils.READ)
				tempRetR.insertElementAt(crt,0);
			}
		}
		for(i=l1-1;i>=0;i--){
		    crt=(LogRecord)callLogVector.elementAt(i);
		    disc=crt.the_log_record.discriminator().value();
		    clr=Utils.get_clr(disc,crt.the_log_record);
		    if(clr.operation.equals("get_state")){
			g_s_u_rec=crt.the_log_record.call_record_par();
			if(g_s_u_rec.timestamp==timestamp){
			    tempRetC.insertElementAt(crt,0);
			    break;
			}
		    }
		    else
			tempRetC.insertElementAt(crt,0);
		}

		tempRet=new Vector();
		if(tempRetR.size()>0)
		    tempRet.addElement(tempRetR.remove(0)); /* This should be the checkpoint element*/
		tempRet.addAll(tempRetC);
		tempRet.addAll(tempRetR);
		tempRet.addAll((Vector)logVectorVector.elementAt(3));
		toRet=new LogRecord[tempRet.size()];
		toRet=(LogRecord[])tempRet.toArray(toRet);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	    if(clear){
		logVectorVector.addElement(new String("over"));
		logVectorVector.notifyAll();
		logTable.remove(new Long(id));
	    }
	}
	return toRet;
    }




      private LogRecord[] get_last_records_upd(long id,Vector logVectorVector,boolean clear){
	int l,i,l1,disc,k;
	LogRecord crt,crt_r;
	Vector tempRetR,tempRetC,tempRet,callLogVector,replyLogVector;
	LogRecord[] toRet=new LogRecord[0];
	CallLogRecord clr;
	ReplyLogRecord rlr;
	CallLogRecordPAR g_s_u_rec;
	long timestamp=-1;

	if(logVectorVector==null) return new LogRecord[0];
	synchronized(logVectorVector){
	    try{
		callLogVector=(Vector)logVectorVector.elementAt(1);
		l1=callLogVector.size();
		replyLogVector=(Vector)logVectorVector.elementAt(2);
		l=replyLogVector.size();
	       	tempRetR=new Vector();
		tempRetC=new Vector();
		for(i=l-1;i>=0;i--){
		    crt=(LogRecord)replyLogVector.elementAt(i);
		    tempRetR.insertElementAt(crt,0);
		    disc=crt.the_log_record.discriminator().value();
		    if(disc==RecordType._update){
			timestamp=crt.the_log_record.update_record().timestamp;
			break;
		    }
		    else
			if(disc==RecordType._reply){
			    rlr=crt.the_log_record.reply_record();
			    if(rlr.type==Utils.READ)
				tempRetR.insertElementAt(crt,0);
			}
		}

		for(k=0;k<i;k++){
		    crt=(LogRecord)replyLogVector.elementAt(i);
		    if(crt.the_log_record.discriminator().value()==RecordType._update)
			tempRetR.insertElementAt(crt,0);
		}

		for(i=l1-1;i>=0;i--){
		    crt=(LogRecord)callLogVector.elementAt(i);
		    disc=crt.the_log_record.discriminator().value();
		    clr=Utils.get_clr(disc,crt.the_log_record);
		    if(clr.operation.equals("get_update")){
			g_s_u_rec=crt.the_log_record.call_record_par();
			if(g_s_u_rec.timestamp==timestamp){
			    tempRetC.insertElementAt(crt,0);
			    break;
			}
		    }
		    else
			tempRetC.insertElementAt(crt,0);
		}


		tempRet=new Vector();
		if(tempRetR.size()>0)
		    tempRet.addElement(tempRetR.remove(0)); /* This should be the checkpoint element*/
		tempRet.addAll(tempRetC);
		tempRet.addAll(tempRetR);
		tempRet.addAll((Vector)logVectorVector.elementAt(3));
		toRet=new LogRecord[tempRet.size()];
		toRet=(LogRecord[])tempRet.toArray(toRet);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	    if(clear){
		logVectorVector.addElement(new String("over"));
		logVectorVector.notifyAll();
		logTable.remove(new Long(id));
	    }
	}
	return toRet;
    }


/** Method called on the host where the failed primary resides, in order to install a backup as a primary, after replaying the last logged calls; this method is
* called in case the primary was checkpointed by update readings
* @param	o_id the identifier of the group of which this backup is a member
*			new_lrc the CORBA object reference of the Logging Recovery Controller residing on the machine where the backup that has to become primary resides;
*			this LRC will replay the last requests on the backup, after setting the updates (in case of cold replication)
*			to_recover the CORBA object reference of the backup that will become a primary
*			repl_s the replication style of the replicated group
*/

    public void recover_new_replica_upd(long o_id,LoggingRecoveryController new_lrc,Updateable to_recover,int repl_s){
	LogRecord[] last_records=new LogRecord[0],last_requests=new LogRecord[0];
	int i,current_retention_id=-1,nr_requests;
	byte[] update;
	Vector logVectorVector;

	try{
	    logVectorVector=logVectorVector(o_id);
	    if(logVectorVector==null) return;
	    synchronized(logVectorVector){
		switch(repl_s){
		case COLD_PASSIVE.value:
		    last_records=get_last_records_upd(o_id,logVectorVector,true);
		    if(last_records!=null&&last_records.length>0){
			i=0;
			while(i<last_records.length&&last_records[i].the_log_record.discriminator().value()==RecordType._update){
			    update=last_records[i].the_log_record.update_record().state;
			    try{
				to_recover.set_update(update);
			    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
			    i++;
			}
			if(i>0){
			    i--;
			    current_retention_id=last_records[i].the_log_record.update_record().current_retention_id;
			}
		    }
		    break;

		case WARM_PASSIVE.value:
		    break;
		case STATELESS.value:
		    current_retention_id=get_last_outgoing_ret_id_v(o_id,logVectorVector,true);
		}
	    }
	    new_lrc.replay_last_requests(to_recover,last_records,o_id,current_retention_id,repl_s);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
    }

/** Method called on the host where the failed primary resides, in order to install a backup as a primary, after replaying the last logged calls; this method is
* called in case the primary was checkpointed by state readings
* @param	o_id the identifier of the group of which this backup is a member
*			new_lrc the CORBA object reference of the Logging Recovery Controller residing on the machine where the backup that has to become primary resides;
*			this LRC will replay the last requests on the backup, after setting the state (in case of cold replication)
*			to_recover the CORBA object reference of the backup that will become a primary
*			repl_s the replication style of the replicated group
*/

      public void recover_new_replica_ch(long o_id,LoggingRecoveryController new_lrc,Checkpointable to_recover,int repl_s){
	  LogRecord[] last_records=new LogRecord[0],last_requests=new LogRecord[0];
	  int i,current_retention_id=-1,nr_rec_set;
	  byte[] new_state;
	  Vector logVectorVector;

	  try{
	      logVectorVector=logVectorVector(o_id);
	      if(logVectorVector==null) return;
	      synchronized(logVectorVector){
		  switch(repl_s){
		  case COLD_PASSIVE.value:
		      last_records=get_last_records_ch(o_id,logVectorVector,true);
		      if(last_records!=null&&last_records.length>0){
			  if(last_records[0].the_log_record.discriminator().value()==RecordType._checkpoint){
			      new_state=last_records[0].the_log_record.checkp_record().state;
			      current_retention_id=last_records[0].the_log_record.checkp_record().current_retention_id;
			      try{
				  to_recover.set_state(new_state);
			      }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}

			  }
		      }
		      break;
		  case WARM_PASSIVE.value:
		      remove_wp_log(o_id);
		      break;
		  case STATELESS.value:
		      current_retention_id=get_last_outgoing_ret_id_v(o_id,logVectorVector,true);
		      break;
		  }
	      }
	      new_lrc.replay_last_requests(to_recover,last_records,o_id,current_retention_id,repl_s);
	  }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
      }

/** Method called on the host where the failed primary resides, in order to install a backup as a primary, in case stateless replication is used, so that no replay of
* methods or state instalation is needed
* @param	o_id the identifier of the group of which this backup is a member
*			new_lrc the CORBA object reference of the Logging Recovery Controller residing on the machine where the backup that has to become primary resides
*			to_recover the CORBA object reference of the backup that will become a primary
*			failed_obj the CORBA reference of the failed primary
*/

      public void recover_new_replica_st_ch(long o_id,LoggingRecoveryController new_lrc,Checkpointable to_recover,Checkpointable failed_obj){
	  LogRecord[] last_records=new LogRecord[0],last_requests=new LogRecord[0];
	  int i,current_retention_id=-1,nr_rec_set;
	  byte[] new_state;
	  Vector logVectorVector;

	  try{
	      logVectorVector=logVectorVector(o_id);
	      if(logVectorVector==null) return;
	      stop_checkpointing(failed_obj);
	      remove_wp_log(o_id);
	      new_lrc.replay_last_requests(to_recover,last_records,o_id,current_retention_id,WARM_PASSIVE.value);
	  }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
      }


    private int get_last_outgoing_ret_id_v(long id,Vector v,boolean clear){
	int i,l;
	LogRecord crt;
	Vector outg_reqLogVector;


	outg_reqLogVector=(Vector)v.elementAt(3);

	synchronized(outg_reqLogVector){
	    l=outg_reqLogVector.size();
	    crt=(LogRecord)outg_reqLogVector.elementAt(l-1);
	    if(clear) {
		v.addElement(new String("over"));
		v.notifyAll();
		logTable.remove(new Long(id));
	    }
	    return crt.the_log_record.request_r_record().retention_id;
	}
    }

/** Method called to obtain the retention identifier of the last outgoing request sent from the replica residing on the machine of this LRC
*/

    public int get_last_outgoing_ret_id(long id,boolean clear){
	int i,l;
	LogRecord crt;
	Vector logVectorVector,outg_reqLogVector;

	synchronized(logTable){
	    logVectorVector=logVectorVector(id);
	    if(logVectorVector==null) return -1;
	    outg_reqLogVector=(Vector)logVectorVector.elementAt(3);
	}
	synchronized(outg_reqLogVector){
	    l=outg_reqLogVector.size();
	    crt=(LogRecord)outg_reqLogVector.elementAt(l-1);
	    if(clear) {
		logVectorVector.addElement(new String("over"));
		logVectorVector.notifyAll();
		logTable.remove(new Long(id));
	    }
	    return crt.the_log_record.request_r_record().retention_id;
	}
    }


    private Vector set_last_records(long id,LogRecord[] recs){
	int i;
	Vector logVectorVector,callLogVector,replyLogVector,outg_r_rLogVector;
	boolean put=false;

	/*This method transfers all the requests which arrived at the primary (in WARM_PASSIVE REPL STYLE)
	  between the "get_state" request and the "get_state" reply
	  It is called when the "get_state" gets its answer, at the interceptor level, thus before the
	  "set_state" on the neighbour replicas is called; so, a first empty checkpoint record will be created, which
	  will be filled only with the current outgoing retention_id, follwed by creation of normal call and reply
	  records depending on the transferred recs; the checkpoint record will be filled totally when the "set_state"
	  is called on the replica*/


	synchronized(logTable){
	    logVectorVector=logVectorVector(id);
	    if(logVectorVector==null){
		logVectorVector=new Vector();
		for(i=0;i<4;i++)
		    logVectorVector.addElement(new Vector());
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new java.lang.Object()); /* synch object*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* expected id*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* current message_id*/
		Long hashcode=new Long(id);
		logTable.put(hashcode,logVectorVector);
	    }
	    callLogVector=(Vector)logVectorVector.elementAt(1);
	    replyLogVector=(Vector)logVectorVector.elementAt(2);
	    outg_r_rLogVector=(Vector)logVectorVector.elementAt(3);
	}

	for(i=0;i<recs.length;i++)
	    switch(recs[i].the_log_record.discriminator().value()){
	    case RecordType._call_as:
	    case RecordType._call_par:callLogVector.addElement(recs[i]);break;
	    case RecordType._reply:
	    case RecordType._checkpoint:
	    case RecordType._update:
		replyLogVector.addElement(recs[i]);break;
	    case RecordType._request:
	    case RecordType._request_reply:outg_r_rLogVector.addElement(recs[i]);break;
	    }
	return logVectorVector;
    }

    private void broadcast_method_call_par(long objI,String oper,org.omg.Dynamic.Parameter[] args,String client_id,int retention_id,long expiration_time,char method_type,long mess_id){
	int i;
	for(i=0;i<neighbours.length;i++)
	    neighbours[i].record_method_no_wait_par(objI,oper,args,client_id,retention_id,expiration_time,method_type,mess_id);
    }

    private void broadcast_method_call_as(long objI,String oper,byte[] buffer,boolean big_e,org.omg.GIOP.Version vers,String client_id,int retention_id,long expiration_time,char meth_type,long mess_id){
	int i;
	for(i=0;i<neighbours.length;i++)
	    neighbours[i].record_method_no_wait_as(objI,oper,buffer,big_e,vers,client_id,retention_id,expiration_time,meth_type,mess_id);
    }

    private void broadcast_reply(long objI,String client_id,int retention_id,org.omg.CORBA.Any result,org.omg.CORBA.Any[] result_arguments,char meth_type){
	int i;
	for(i=0;i<neighbours.length;i++)
	    neighbours[i].record_reply_no_notify(objI,client_id,retention_id,result,result_arguments,meth_type);
    }

/** Method called in case a call information record has to be logged, but without waiting for other methods to finish execution first - this is the case when
* a method call record is broadcast to warm backups by the primary
* @param	objI the identifier of the group where the backup is a member
*			oper the operation (method) name for which the call record is logged
*			args the arguments of the call
*			client_id the client identifier associated with the request for which call info is logged
*			retention_id the retention identifier associated with the request for which call info is logged
*			expiration_time the expiration time associated with the request for which call info is logged
*			method_type the type of the method for which the information is logged
*			mess_id a unique message identifier so that the call info recording request is delivered in a total order
*/

    public void record_method_no_wait_par(long objI,String oper,org.omg.Dynamic.Parameter[] args,String client_id,int retention_id,long expiration_time,char method_type,long mess_id){
	Vector logVectorVector,callLogVector,countersVector;
	int i;
	long expected_id;
	MyLong exp_id;
	java.lang.Object wait_tot_ord;
	synchronized(logTable){
	    logVectorVector=logVectorVector(objI);
	    if(logVectorVector==null){
		logVectorVector=new Vector();
		for(i=0;i<4;i++)
		    logVectorVector.addElement(new Vector());
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new java.lang.Object()); /* synch object*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* expected id*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* current message_id*/
		Long hashcode=new Long(objI);
		logTable.put(hashcode,logVectorVector);
	    }

	}
	synchronized(logVectorVector){
	    countersVector=(Vector)logVectorVector.elementAt(0);
	    callLogVector=(Vector)logVectorVector.elementAt(1);
	    wait_tot_ord=countersVector.elementAt(5);
	}
	synchronized(wait_tot_ord){
	    exp_id=(MyLong)countersVector.elementAt(6);
	    if(exp_id.get()<mess_id){
		while(exp_id.get()<mess_id)
		    try{
			wait_tot_ord.wait();
		    }catch(InterruptedException e){}
	    }
	    exp_id.incr();
	    SomeLogRecord slr=new SomeLogRecord();
	    slr.call_record_par(new CallLogRecordPAR(new CallLogRecord(oper,client_id,retention_id,expiration_time,true,method_type),args,-1));
	    callLogVector.addElement(new LogRecord(objI,slr));
	    wait_tot_ord.notifyAll();
	}
    }

/** Method called in case a call information record has to be logged, but without waiting for other methods to finish execution first - this is the case when
* a method call record is broadcast to warm backups by the primary
* @param	objI the identifier of the group where the backup is a member
*			oper the operation (method) name for which the call record is logged
*			buffer the byte arraycontaining an unmarshalled format of the parameters of the call
*			big_e a boolean value indicating whether the ordering of octets is big endian
*			vers the version of the GIOP encoding of information
*			client_id the client identifier associated with the request for which call info is logged
*			retention_id the retention identifier associated with the request for which call info is logged
*			expiration_time the expiration time associated with the request for which call info is logged
*			method_type the type of the method for which the information is logged
*			mess_id a unique message identifier so that the call info recording request is delivered in a total order
*/

    public void record_method_no_wait_as(long objI,String oper,byte[] buffer,boolean big_e,org.omg.GIOP.Version vers,String client_id,int retention_id,long expiration_time,char method_type,long mess_id){
	Vector logVectorVector,callLogVector,countersVector;
	int i;
	long expected_id;
	MyLong exp_id;
	java.lang.Object wait_tot_ord;
	synchronized(logTable){
	    logVectorVector=logVectorVector(objI);
	    if(logVectorVector==null){
		logVectorVector=new Vector();
		for(i=0;i<4;i++)
		    logVectorVector.addElement(new Vector());
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(-1));
		((Vector)logVectorVector.elementAt(0)).addElement(new java.lang.Object()); /* synch object*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* expected id*/
		((Vector)logVectorVector.elementAt(0)).addElement(new MyLong(0)); /* current message_id*/
		Long hashcode=new Long(objI);
		logTable.put(hashcode,logVectorVector);
	    }

	}
	synchronized(logVectorVector){
	        countersVector=(Vector)logVectorVector.elementAt(0);
	    callLogVector=(Vector)logVectorVector.elementAt(1);
	    wait_tot_ord=countersVector.elementAt(5);

	}
	synchronized(wait_tot_ord){
	    exp_id=(MyLong)countersVector.elementAt(6);
	    if(exp_id.get()<mess_id){
		while(exp_id.get()<mess_id)
		    try{
			wait_tot_ord.wait();
		    }catch(InterruptedException e){}
	    }
	    exp_id.incr();
	    SomeLogRecord slr=new SomeLogRecord();
	    slr.call_record_as(new CallLogRecordAS(new CallLogRecord(oper,client_id,retention_id,expiration_time,true,method_type),buffer,big_e,vers));
	    callLogVector.addElement(new LogRecord(objI,slr));
	    wait_tot_ord.notifyAll();
	}
    }


/** Method called in case a reply information record has to be logged, but without notifying other waiting methods - this is the case when
* a method reply record is broadcast to warm backups by the primary
* @param	objI the identifier of the group where the backup is a member
*			client_id the client identifier associated with the request for which reply info is logged
*			retention_id the retention identifier associated with the request for which reply info is logged
*			result the result of the call in terms of a return value
*			result_arguments the result of the call in terms of OUT and INOUT call parameters that were written by the method call
*			meth_type the type of the method for which the information is logged
*/

    public void record_reply_no_notify(long objI,String client_id,int retention_id,org.omg.CORBA.Any result,org.omg.CORBA.Any[] result_arguments,char meth_type){
	Vector logVectorVector,replyLogVector;
	synchronized(logTable){
	    logVectorVector=logVectorVector(objI);
	    if(logVectorVector==null) return;
	}
	synchronized(logVectorVector){
	    replyLogVector=(Vector)logVectorVector.elementAt(2);
	    SomeLogRecord slr=new SomeLogRecord();
	    slr.reply_record(new ReplyLogRecord(client_id,retention_id,0,result,result_arguments,meth_type));
	    if(logVectorVector==null) return;
	    replyLogVector.addElement(new LogRecord(objI,slr));
	}
    }

/** Method called on the Logging Recovery Controller running on the host where the backup that becomes primary resides; the result will be that the last method calls
* logged since the last checkpoint are replayed on the backup
* @param	target the CORBA object reference of the object on which the requests are replayed
*			last_rec the last reords from the log; if replication style is cold passive, then the first element(s) in this array is the state (set of updates) to be
* 			set on the new primary before replay
*			objI the identifier of the group whose member the backup is
*			current_retention_id the value of the retention identifier for the last request sent out by the failed primary
*			repl_s the replication style
*/

    public void replay_last_requests(org.omg.CORBA.Object target,LogRecord[] last_rec,long objI,int current_retention_id,int repl_s){
	int l,i,lp,j,k,size;
	LogRecord crt_c,crt;
	CallLogRecord clr;
	ReplyLogRecord rlr;
	String oper,client_id=null,arg_cl_name;
	int retention_id=-1,pos=-1,disc;
	java.lang.reflect.Method method;
	Class[] par_cl;
	java.lang.Object[] params;
	org.omg.Dynamic.Parameter[] arguments;
	Vector last_r_r,last_r_nr,toRemoveC=new Vector(),toRemoveR=new Vector();
	LogRecord[] last_requests_r=new LogRecord[0],last_requests_nr=new LogRecord[0];
	boolean found,is_null=false;

	org.openorb.CORBA.dii.Request dyn_request;
	org.omg.CORBA.Any argument;
	org.omg.CORBA.NVList dyn_arguments;
	org.omg.CORBA.ExceptionList dyn_exceptions=orb.create_exception_list();
	org.omg.CORBA.NamedValue dyn_result=orb.create_named_value("",orb.create_any(),-1);
	int flags=org.omg.CORBA.ARG_IN.value;

	Vector logVectorVector,callLogVector=new Vector();

	logVectorVector=set_last_records(objI,last_rec);

	if(logVectorVector==null) {
	    is_null=true;
	}
	else{
	    callLogVector=(Vector)logVectorVector.elementAt(1);
	}

	if(!is_null){
	    synchronized(logVectorVector){
		size=callLogVector.size();
		if(size!=0&&repl_s!=STATELESS.value) {
		    try{
			i=0;
			last_r_r=new Vector();
			last_r_nr=new Vector();
			long now_time=System.currentTimeMillis();

			if(repl_s==WARM_PASSIVE.value){
			    for(i=size-1;i>=0;i--){
				crt_c=(LogRecord)callLogVector.elementAt(i);
				clr=org.omg.FT.Utils.get_clr(crt_c.the_log_record.discriminator().value(),crt_c.the_log_record);
				if(clr.operation.equals("get_state")||clr.operation.equals("get_update")){
				    i++;
				    break;
				}
			    }
			}

			if(i<0) i=0;
			while(i<=size-1){
			    crt_c=(LogRecord)callLogVector.elementAt(i);
			    clr=org.omg.FT.Utils.get_clr(crt_c.the_log_record.discriminator().value(),crt_c.the_log_record);
			    if(clr.type!=org.omg.FT.Utils.READ)
				/* this excludes the replay of get_state or get_update as well*/
				last_r_nr.addElement(crt_c);
			    i++;
			}
			last_requests_r=new LogRecord[last_r_r.size()];
			last_requests_nr=new LogRecord[last_r_nr.size()];

			last_requests_r=(LogRecord[])last_r_r.toArray(last_requests_r);
			last_requests_nr=(LogRecord[])last_r_nr.toArray(last_requests_nr);
		    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
		}
	    }
	}
	try{
	    dyn_arguments=orb.create_list(0);
	    argument=orb.create_any();
	    LogRecordsHelper.insert(argument,last_requests_r);
	    dyn_arguments.add_value("param0",argument,flags);
	    argument=orb.create_any();
	    LogRecordsHelper.insert(argument,last_requests_nr);
	    dyn_arguments.add_value("param1",argument,flags);
	    argument=orb.create_any();
	    LoggingRecoveryControllerHelper.insert(argument,this._this());
	    dyn_arguments.add_value("param2",argument,flags);
	    argument=orb.create_any();
	    argument.insert_long(current_retention_id);
	    dyn_arguments.add_value("param3",argument,flags);
	    dyn_request=new org.openorb.CORBA.dii.Request(target,"replay_last_requests",dyn_arguments,dyn_result,orb.create_environment(),dyn_exceptions,orb.create_context_list(),orb);
	    dyn_request.invoke();
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
    }


/** Method called to remove the log corresponding to a certain group
*/

    public void remove_log(long objId){
	logTable.remove(new Long(objId));
    }

/** Method called to set the other LRCs to which this LRC has to transfer method call info - in warm passive replication
*/

    public void set_neighbours(LoggingRecoveryController[] neigh){
	neighbours=neigh;
    }

}

