/**
  * Redistribution and use of this software and associated
  * documentation ("Software"), with or without modification, are
  * permitted provided that the following conditions are met:
  *
  * 1. Redistributions of source code must retain copyright statements
  *    and notices.  Redistributions must also contain a copy of this
  *    document.
  *
  * 2. Redistributions in binary form must reproduce the above
  *    copyright notice, this list of conditions and the following
  *    disclaimer in the documentation and/or other materials provided
  *    with the distribution.
  *
  * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
  *    products derived from this Software without prior written
  *    permission of one of the parties involved in the TRANSORG project. For
  * 	  written permission, please contact diasz@ida.liu.se.
  *
  * 4. Products derived from this Software may not be called "TRANSORG"
  *    nor may "TRANSORG" appear in their names without prior written
  *    permission.
  *
  * 5. Due credit should be given to the TRANSORG Project
  *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
  *
  * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  * PARTICULAR PURPOSE ARE DISCLAIMED.
  *
  * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
  * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
  *
  * $Id: LoggingRecoveryMechL.java,v 1.0  diasz Exp $
  *
  * Date         Author  Changes
  */

package org.omg.FT;
import org.omg.CosNaming.*;

/** Class used to create a Logging and REcovery Controller CORBA object
 * @author <a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
 */

public class LoggingRecoveryMechL{
    public static void main(String args[]){

	if(args.length<2){
	    System.out.println("Usage: java org.omg.FT.LoggingRecoveryMechL ftDomName logging_recovery_class_name");
	    return;
	}

	java.util.Properties props=System.getProperties();
	try{
	    props.put("org.omg.CORBA.ORBClass","org.openorb.CORBA.ORB");
	    props.put("org.omg.CORBA.ORBSingletonClass","org.openorb.CORBA.ORBSingleton");

	    System.setProperties(props);
	    org.omg.CORBA.Object poa_o;
	    org.omg.PortableServer.POA poa;
	    org.omg.CORBA.Object rootNC_o,r_man_o;
	    NamingContext rootNC,ftDomNC,loggRNC;
	    int i;
	    String location=null;

	    location=java.net.InetAddress.getLocalHost().getHostName();

	    String[] args_h=new String[args.length+1];

	    for(i=0;i<args.length;i++)
		args_h[i]=args[i];

	    args_h[i]="-ORBProfile=loggrip";

	    org.omg.CORBA.ORB orb=org.omg.CORBA.ORB.init(args_h,props);
	    poa_o=orb.resolve_initial_references("RootPOA");
	    poa=org.omg.PortableServer.POAHelper.narrow(poa_o);

	    rootNC_o=orb.resolve_initial_references("NameService");
	    rootNC=NamingContextHelper.narrow(rootNC_o);

	    org.omg.CosNaming.NameComponent fn=new NameComponent(args[0],"directory"),fn_nc[]={fn},lrn=new NameComponent(location,"directory"),lr_nc[]={lrn},lrin=new NameComponent("loggrecc","text"),lrin_nc[]={lrin};

	    ftDomNC=Utils.bind_context(rootNC,fn_nc);
	    if(ftDomNC==null)
		ftDomNC=Utils.resolve_n(rootNC,fn_nc);

	    loggRNC=Utils.bind_context(ftDomNC,lr_nc);
	    if(loggRNC==null)
		loggRNC=Utils.resolve_n(ftDomNC,lr_nc);

	    org.omg.FT.LoggingRecoveryControllerPOA inst;

	    java.lang.Class[] ctr_params_c=new java.lang.Class[1];
	    ctr_params_c[0]=java.lang.Class.forName("org.omg.CORBA.ORB");


	    java.lang.Object[] ctr_params=new java.lang.Object[1];
	    ctr_params[0]=orb;

	    java.lang.reflect.Constructor log_r_c_c=java.lang.Class.forName("org.omg.FT."+args[1]).getConstructor(ctr_params_c);
	    java.lang.Object inst_dyn=log_r_c_c.newInstance(ctr_params);
	    inst=(org.omg.FT.LoggingRecoveryControllerPOA)inst_dyn;

	    //LoggingRecoveryController_impl inst=new LoggingRecoveryController_impl(orb);

	    ((org.openorb.CORBA.ORB)orb).setFeature("ItsLoggRecC",inst);

	    LoggingRecoveryController the_interf;
	    the_interf=inst._this(orb);

	    loggRNC.rebind(lrin_nc,the_interf);

	    r_man_o=orb.resolve_initial_references("ReplicationManager");
	    ReplicationManager r_man;
	    r_man=ReplicationManagerHelper.narrow(r_man_o);


	    r_man.add_logging_recovery_c(new LoggingRecoveryControllerInfo(the_interf,lr_nc));

	    System.out.println("Instance of logging recovery controller created");
	    poa.the_POAManager().activate();
	    orb.run();
	}catch(Exception e){e.printStackTrace();}
    }
}
