/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of one of the parties involved in the TRANSORG project. For
 * 	  written permission, please contact diasz@ida.liu.se.
 *
 * 4. Products derived from this Software may not be called "TRANSORG"
 *    nor may "TRANSORG" appear in their names without prior written
 *    permission.
 *
 * 5. Due credit should be given to the TRANSORG Project
 *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
 *
 * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
 *
 * $Id: ReplicationManager_impl.java,v 1.0  diasz Exp $
 *
 * Date         Author  Changes
 */

package org.omg.FT;
/** This class implements the ReplicationManager interface that extends the PropertyManager, GenericFactory, and ObjectGroupManager interfaces.
* The main function of this class is to create replcia groups on demand from the application writer that calls the create_object method with certain parameters.
* A more dynamic function of the Replication Manager is to handle failover of the groups in which members fail. It receives fault notifications from the
* fault notifier and reacts to them by performing the appropriate reconfiguration operations on the server group.
* @author	<a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
 */
import java.util.*;
import org.omg.CORBA.*;
import org.omg.IOP.*;

/* When in an actively replicated group one member fails and one or more new members have to be joined to the group, the new membership of
the group is sent to the gateway before the state is transfered to the newcomers. For this case the gateway must also be changed, so that it does
not log calls received after the get_state request, in case the number of non-new replicas is bigger that one, but it sends all calls to all members.
What happens is that the replica which is the state giver will receive calls after the get_state, calls which have to synchronize with get_state.
The new members of the group will also receive those calls, but in their case the synchronization is done with set_state*/


public class ReplicationManager_impl extends ReplicationManagerPOA{

    static final int NR_NEW_REPL_W_F=1;	/* this represents the number of replicas which have to be created
					   to keep the minimum number replicas requirement.
					   It would be better if nr > 1 to avoid oscillation*/
    static final int INDEX_PRIMARY=0;
    java.lang.Object create_object_semaphor=new java.lang.Object(),delete_object_semaphor=new java.lang.Object(),recover_object_semaphor=new java.lang.Object();

    int flags=org.omg.CORBA.ARG_IN.value;

    ORB orb;
    Hashtable typeProperties,groupProperties;
    Vector defaultProperties;
    Vector currProperties;
    Hashtable groupEvid; /* a hashtable indexed by group_ref and values
			    are of type ObjGroupInfoStr*/
    long objGrId;
    java.lang.String ftDomName;
    org.omg.PortableServer.POA poa;
    FaultNotifier fault_notifier;
    Vector theFaultD_l,theLoggRecC_l;


    private class ObjGroupInfoStr{
	public long f_consumer_index;
	public int crash_after;
	public Any group_id;
	public Vector factlocEvid; /* a vector of ObjFactAndIdStr objects*/
	public Vector loggrecCtrls;
	public Vector faultMonits;
	public java.lang.Object theStub;

	public ObjGroupInfoStr(long f_cons_i,Any g_i,int c_a,Vector l_e,Vector l_c,Vector f_m,java.lang.Object st){
	    f_consumer_index=f_cons_i;
	    crash_after=c_a;
	    group_id=g_i;
	    factlocEvid=l_e;
	    loggrecCtrls=l_c;
	    faultMonits=f_m;
	    theStub=st;
	}
    }


    private class ObjFactAndIdStr{
	public org.omg.CORBA.Object obj,gateway;
	public boolean is_primary,is_active;
	public FactoryInfo cr_f_info;
	public Any fact_cr_id;

	public ObjFactAndIdStr(org.omg.CORBA.Object o,org.omg.CORBA.Object gw,boolean is_p,boolean is_a,FactoryInfo f_i,Any f_c_i){
	    obj=o;
	    gateway=gw;
	    cr_f_info=f_i;
	    fact_cr_id=f_c_i;
	    is_primary=is_p;
	    is_active=is_a;
	}
    }


/** the replication manager is represented by this consumer in the list of fault event consumers of the fault notifier
*/
    class StructuredFaultConsumer extends org.omg.CosNotifyComm.StructuredPushConsumerPOA{

	/* This is the group id of the group for whose failure of members this consumer is interested.
	   If a fault event arrives at this consumer it means that it passed the filter
	   of this consumer, meaning that a member of THIS GROUP (with group id= group_id)
	   failed, even if, in fact the event contains a different group id (for example
	   if this group has a monit. gran. of type LOC or LOC_AND_TYPE then the failure of
	   any member of agroup at the location or of a certain type at the location, will
	   trigger the sending of such an event to the present consumer
	 */


	long group_id;

	public StructuredFaultConsumer(long group_id){
	    this.group_id=group_id;
	}

	public void disconnect_structured_push_consumer(){}

	public void offer_change(org.omg.CosNotification.EventType[] added, org.omg.CosNotification.EventType[] removed){}

	public void push_structured_event(org.omg.CosNotification.StructuredEvent notification) throws org.omg.CosEventComm.Disconnected{

	    int l,i,lv,iv;
	    String location,type_id="",ft_d_name,value[];
	    Set elems;
	    Map.Entry entry;
	    Iterator setIter;
	    ObjGroupInfoStr crt;

	    org.omg.CORBA.Object group;
	    try{
		l=notification.filterable_data.length;
		if(l<4) {return;}
		value=org.omg.CORBA.StringSeqHelper.extract(notification.filterable_data[1].value);
		location=value[0];
		if(!notification.filterable_data[2].name.equals("TypeId")) {
		    return;
		}
		value=org.omg.CORBA.StringSeqHelper.extract(notification.filterable_data[2].value);
		type_id=value[0];
		if(!notification.filterable_data[3].name.equals("ObjectGroupId")) {
		    return;
		}
		value=org.omg.CORBA.StringSeqHelper.extract(notification.filterable_data[3].value);
		synchronized(groupEvid){
		    elems=groupEvid.entrySet();
		}
		setIter=elems.iterator();
		Vector non_conc_buffer=new Vector();
		ObjFactAndIdStr oals;

		while(setIter.hasNext()){
		    entry=(Map.Entry)setIter.next();
		    non_conc_buffer.addElement(entry);
		}

		lv=non_conc_buffer.size();

		/* in the following, the minimum number replicas requirement is checked to
		   be satisfied or not
		   If, after the failure, the number of replicas is big enough, the replication
		   manager does not have to create anything; otherwise, it has to create a new member
		   of the "damaged" object group(s)*/

		/*search for the obj. group with the given obj_gr_id*/
		/* the type_id is not really necessary, since group_ids are unique, not depending on
		   type_ids*/

		try{
		    for(iv=0;iv<lv;iv++){
			entry=(Map.Entry)non_conc_buffer.elementAt(iv);
			crt=(ObjGroupInfoStr)entry.getValue();
			if(crt.group_id.extract_longlong()==group_id) {
			    group=(org.omg.CORBA.Object)entry.getKey();
			    handle_obj_failure(group,location,crt,group_id,type_id);
			    break;
			}
		    }
		}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	}
    }

    public ReplicationManager_impl(ORB _o,org.omg.PortableServer.POA _p,java.lang.String ftDom){

	orb=_o;
	poa=_p;
	objGrId=0;
	ftDomName=ftDom;
	fault_notifier=null;
	groupEvid=new Hashtable();
	currProperties=new Vector();
	typeProperties=new Hashtable();
	groupProperties=new Hashtable();
	defaultProperties=new Vector();
	theFaultD_l=new Vector();
	theLoggRecC_l=new Vector();
    }

/** Method called when a member of a group failed
* @param	group the reference of the group from which the member failed
* 			location the location of the failed member
*			crt the internal Replication Manager held structure with info about the group
*			group_id the identifier of the group
*			type_id the type identifier of the CORBA objects members of the group
*/

    private void handle_obj_failure(org.omg.CORBA.Object group,String location,ObjGroupInfoStr crt,long group_id,String type_id){
	int l,i,s,min_nr_repl=-1,repl_s=-1,fault_m_g=-1;
	ObjFactAndIdStr oals;
	Vector gP,tP,propV;
	Property prop;
	long checkp_int=-1,mon_i=4000,mon_t_o=4000;
	boolean periodic=false;
	org.omg.CORBA.NVList dyn_arguments;
	org.omg.CORBA.ExceptionList dyn_exceptions=orb.create_exception_list();
	org.omg.CORBA.NamedValue dyn_result=orb.create_named_value("",orb.create_any(),-1);
	org.openorb.CORBA.dii.Request dyn_request;

	l=crt.factlocEvid.size();
	/* the size of the object group before the removal is l*/
	synchronized(recover_object_semaphor){
	    gP=(Vector)groupProperties.get(group);
	    tP=(Vector)typeProperties.get(type_id);

	    propV=Utils.prop_v_union_o(gP,tP,defaultProperties);
	    if(propV!=null){
		s=propV.size();
		/* the minimumnumberreplicas property is searched for*/
		for(i=0;i<s;i++){
		    prop=(Property)propV.elementAt(i);
		    if(prop.nam[0].id.equals("org.omg.ft.MinimumNumberReplicas")){
			min_nr_repl=prop.val.extract_long();
			if(min_nr_repl<2) min_nr_repl=2; /* Sorry, but if we use a fault tolerant ORB, then
							    make the server with at least 2 replicas*/
		    }
		    if(prop.nam[0].id.equals("org.omg.ft.ReplicationStyle"))
			repl_s=prop.val.extract_long();
		    if(prop.nam[0].id.equals("org.omg.ft.FaultMonitoringGranularity"))
			fault_m_g=prop.val.extract_long();
		    if(prop.nam[0].id.equals("org.omg.ft.CheckpointInterval"))
			checkp_int=prop.val.extract_longlong();
		    if(prop.nam[0].id.equals("org.omg.ft.CheckpointPeriodic"))
			periodic=true;
		    if(prop.nam[0].id.equals("org.omg.ft.FaultMonitoringInterval"))
			mon_i=prop.val.extract_longlong();
		    if(prop.nam[0].id.equals("org.omg.ft.FaultMonitoringTimeout"))
			mon_t_o=prop.val.extract_longlong();
		}
	    }

	    for(i=0;i<l;i++){
		oals=(ObjFactAndIdStr)crt.factlocEvid.elementAt(i);
		if(oals.cr_f_info.the_location[0].id.equals(location)){
		    if(!oals.is_active) return;
		    //if(repl_s!=ACTIVE.value&&repl_s!=ACTIVE_WITH_VOTING.value)
			//oals.cr_f_info.the_group_ref_man.remove_group_reference(oals.obj);
		    try{
			dyn_arguments=orb.create_list(0);
			dyn_request=new org.openorb.CORBA.dii.Request(oals.obj,"deactivate_object",dyn_arguments,dyn_result,orb.create_environment(),dyn_exceptions,orb.create_context_list(),orb);
			dyn_request.invoke();
			oals.cr_f_info.the_factory.delete_object(oals.fact_cr_id);
		    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
		    try{
			do_work_failure(group,oals,type_id,crt,group_id,repl_s,min_nr_repl,fault_m_g,checkp_int,periodic,mon_i,mon_t_o);
		    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
		    break; /* a group has only one member at a certain location*/
		}
	    }
	}
    }

/** Method used to return a list of fault monitors registered in the Replicatio Manager
*/

     private FaultMonitor[] find_monitors(String[] locations){
	int i;
	FaultMonitor toRet[];

	toRet=new FaultMonitor[locations.length];

	for(i=0;i<locations.length;i++)
	    toRet[i]=search_for_monitor(locations[i]);
	return toRet;
    }

    private void monitor_objects(java.lang.Object[] stubs,FaultMonitor[] monitors,int fault_m_g,long gr_id,long mon_i,long mon_t_o){
	int i;

	for(i=0;i<stubs.length;i++)
	    if(monitors[i]!=null)
		monitors[i].do_monitor((PullMonitorable)stubs[i],gr_id,fault_m_g,new FaultMonitoringIntervalAndTimeoutValue(mon_i,mon_t_o,(short)0));
    }

    private long setup_fault_consumer_connection(java.util.Vector obj_locs_fact_i,String type_id,int fault_mon_gran,long obj_gr_id,boolean delete,long old_f_c_index){

	long fault_consumer_index=-1;
	int l,i;
	if(fault_notifier==null) {return -1;}
	if(fault_notifier!=null){
	    org.omg.CosNotifyFilter.Filter filter=null;
	    try{
		filter=fault_notifier.create_subscription_filter("whatever");
	    }catch(java.lang.Exception ex){ex.printStackTrace(new java.io.PrintStream(System.out));filter=null;}
	    org.omg.CosNotifyFilter.ConstraintExp[] constraints=new org.omg.CosNotifyFilter.ConstraintExp[1];
	    org.omg.CosNotification.EventType[] event_types=new org.omg.CosNotification.EventType[1];
	    event_types[0]=new org.omg.CosNotification.EventType("FT_CORBA","ObjectCrashFault");
	    l=obj_locs_fact_i.size();
		String constraint_exp="$FTDomainId=="+ftDomName;
		constraint_exp+="&";
		for(i=0;i<l;i++){
		    constraint_exp+="$Location=="+((ObjFactAndIdStr)obj_locs_fact_i.elementAt(i)).cr_f_info.the_location[0].id;
		    if(i<l-1)
			constraint_exp+="|";
		}

		switch(fault_mon_gran){
		case LOC_AND_TYPE.value:constraint_exp+="&$TypeId=="+type_id;break;
		case MEMB.value:constraint_exp+="&$TypeId=="+type_id+"&$ObjectGroupId=="+Long.toString(obj_gr_id);break;
		}

		constraints[0]=new org.omg.CosNotifyFilter.ConstraintExp(event_types,constraint_exp);
		if(filter!=null){
		    try{
			filter.add_constraints(constraints);
		    }catch(org.omg.CosNotifyFilter.InvalidConstraint ic){ic.printStackTrace(new java.io.PrintStream(System.out));}
		}
		try{
		    if(delete)
			fault_notifier.disconnect_consumer(old_f_c_index);
		    fault_consumer_index=fault_notifier.connect_structured_fault_consumer(new StructuredFaultConsumer(obj_gr_id)._this(orb),filter);
		}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));fault_consumer_index=-1;}
	    }
	return fault_consumer_index;
    }


    private void checkpoint_object(long checkp_int,boolean periodic,java.lang.Object stub,long gr_id,ObjGroupInfoStr crt,org.omg.CORBA.Object primary,LoggingRecoveryController new_lg_r_c,int repl_s,java.lang.reflect.Method narrow_m,Class helper_class){
	int i,j,k,l;
	ObjFactAndIdStr oals;

	boolean transfer;
	Checkpointable transfer_st_to[]=new Checkpointable[0];
	j=0;
	k=0;
	l=crt.factlocEvid.size();
	LoggingRecoveryController l_r,log_rec_ctrl_n[]=new LoggingRecoveryController[l-1];
	java.lang.Object params[];
	try{
	    if(repl_s==WARM_PASSIVE.value){
		transfer=true;
		transfer_st_to=new Checkpointable[l-1];
		for(i=0;i<l;i++){
		    oals=(ObjFactAndIdStr)crt.factlocEvid.elementAt(i);
		    if(!Utils.refs_equal(orb,oals.obj,primary,-1)){
			params=new java.lang.Object[1];
			params[0]=oals.obj;
			transfer_st_to[j]=(Checkpointable)narrow_m.invoke(helper_class.newInstance(),params);
			l_r=search_for_loggrec_c(oals.cr_f_info.the_location[0].id);
			log_rec_ctrl_n[j]=l_r;
			if(l_r!=null) k++;
			j++;
		    }
		}
		if(k<j){
		    LoggingRecoveryController[] temp=new LoggingRecoveryController[k];
		    k=0;
		    for(i=0;i<j;i++)
			if(log_rec_ctrl_n[i]!=null){
			    temp[k]=log_rec_ctrl_n[i];
			    k++;
			}
		    log_rec_ctrl_n=temp;
		}
		new_lg_r_c.set_neighbours(log_rec_ctrl_n);
	    }

	    else
		transfer=false;
	    new_lg_r_c.do_checkpointing((Checkpointable)stub,gr_id,checkp_int,transfer_st_to,transfer,periodic);
	}catch(java.lang.Exception e){e.printStackTrace();}
    }

/** Method used to perform all the actions needed when a member of a group fails
 * Depending on the replication style it performs the proper failover procedures: depending on the numer of replicas remaining and the group size
 * constraints, it may create new replicas on hosts different from the ones where replicas failed
*/


    private void do_work_failure(org.omg.CORBA.Object group,ObjFactAndIdStr failed_info,String type_id,ObjGroupInfoStr crt,long group_id,int repl_s,int min_nr_repl,int fault_m_g,long checkp_int,boolean periodic,long mon_i,long mon_t_o) throws NoFactory{
	int s,p,m,i,j,k,l,r_nr_repl,nr_new_repl,b_r_nr_repl;
	boolean found_one=false,announced_start=false;
	Vector gP,tP,propV;
	byte[] gr_id_b;
	Property dyn_p[];
	ObjFactAndIdStr oals=null;
	LoggingRecoveryController new_lg_r_c=null;
	FactoryInfo[] factories=null;
	org.omg.CORBA.Object new_group=null,primary=null;
	org.omg.IOP.IOR p_ior;
	ObjectLocStr[] members_for_gw=new ObjectLocStr[0];
	GatewayForActiveRepl gateway=null;
	long new_consumer_index=-1;
	String new_locations[]=new String[0],new_location;
	java.lang.Object new_stubs[]=new java.lang.Object[0];
	java.lang.Object stub=new java.lang.Object();
	org.openorb.CORBA.dii.Request dyn_request;
	org.omg.CORBA.Any argument;
	org.omg.CORBA.NVList dyn_arguments;
	org.omg.CORBA.ExceptionList dyn_exceptions=orb.create_exception_list();
	org.omg.CORBA.NamedValue dyn_result=orb.create_named_value("",orb.create_any(),-1);

	try{
	    Class helper_class=Class.forName(Utils.get_class_helper(type_id));
	    Class[] par_cl=new Class[1];
	    par_cl[0]=Class.forName("org.omg.CORBA.Object");
	    java.lang.reflect.Method narrow_m=helper_class.getMethod("narrow",par_cl);
	    java.lang.Object[] params=new java.lang.Object[1];
	    org.omg.CORBA.Object[] objects=null;

	    gr_id_b=Utils.makeIdByte(group_id);

	    crt.factlocEvid.remove(failed_info);

	    l=crt.faultMonits.size();
	    FaultMonitor crt_fm;
	    boolean found=false;
	    for(i=0;i<l;i++){
		crt_fm=(FaultMonitor)crt.faultMonits.elementAt(i);
		if(failed_info.cr_f_info.the_location[0].id.equals(Utils.give_location(orb,crt_fm))){
		    found=true;
		    break;
		}
	    }
	    if(found)
		crt.faultMonits.remove(i);

	    r_nr_repl=crt.factlocEvid.size();

	    /*modify the group ior such that the failed member profile is
	      removed*/

	    IOR ge_ior=Utils.get_ior(group),f_ior;

	    if(repl_s!=ACTIVE.value&&repl_s!=ACTIVE_WITH_VOTING.value){
		f_ior=Utils.get_ior(failed_info.obj);
		ge_ior=Utils.update_group_ior_rem(orb,ge_ior,f_ior,true);
		synchronized(groupEvid){
		    groupEvid.remove(group);
		}
	    }

	    /* we suppose that l cannot be less than 1*/

	    /* if the remaining number of replicas (given as the dimension of
	       the factoriesEvid vector, or as the dimension of the locEvid
	       vector) is not big enough, then the repl. manager has to create
	       a new replica at one of the locations where there are factories*/

	    if(failed_info.is_primary){
		if(repl_s!=ACTIVE.value&&repl_s!=ACTIVE_WITH_VOTING.value){
		    /* The group is now one member less, still the minimum nr. replicas req.
		       is not violated; a different member has to be activated as primary (of
		       course, this should happen for a cold passive or stateless replicated
		       groups; in all other cases, all members are active, and just for the
		       warm passive, a new primary has to be set, not activated)*/
		    /*choose a group member to become the primary; for now, it is the first element
		      in the rest of the group elements vector
		    */
		    if(repl_s!=WARM_PASSIVE.value&&crt.loggrecCtrls.size()>0&&crt.theStub!=null)
			try{
			    ((LoggingRecoveryController)(crt.loggrecCtrls.elementAt(0))).stop_checkpointing((Checkpointable)crt.theStub);
			}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}

		    found_one=false;

		    l=crt.factlocEvid.size();
		    for(i=0;i<l;i++){
			oals=(ObjFactAndIdStr)crt.factlocEvid.elementAt(i);
			if(repl_s==COLD_PASSIVE.value||repl_s==STATELESS.value){
			    /* it has to be explicitly activated only in case of cold passive replication*/
			    try{
				dyn_arguments=orb.create_list(0);
				dyn_request=new org.openorb.CORBA.dii.Request(oals.obj,"activate_object",dyn_arguments,dyn_result,orb.create_environment(),dyn_exceptions,orb.create_context_list(),orb);
				dyn_request.invoke();
				//oals.cr_f_info.the_factory.activate_object(oals.fact_cr_id,group_id,false);
			    }catch(java.lang.Exception e){
				e.printStackTrace(new java.io.PrintStream(System.out));
				if(e instanceof org.omg.CORBA.BAD_CONTEXT){
				    continue;
				}
			    }
			    found_one=true;
			    break;
			}
			else{
			    found_one=true;
			    break;
			}
		    }

		    if(!found_one){
			factories=search_for_factories(1,failed_info.cr_f_info.the_location[0].id,crt.factlocEvid,(Vector)groupProperties.get(group),(Vector)typeProperties.get(type_id));
			if(factories!=null&&factories.length!=0){
			    create_active_object(ge_ior,factories[0],type_id,crt,repl_s);
			    oals=(ObjFactAndIdStr)crt.factlocEvid.lastElement();
			}
			else{/* This case should be very rare*/
			    /* it means that no backup could be set to primary, and no new primary
			       could be created either*/
			    return;
			}
		    }
		    primary=oals.obj;
		    p_ior=Utils.get_ior(primary);
		    ge_ior=Utils.change_primary(orb,ge_ior,p_ior,false);
		    params[0]=primary;
		    stub=narrow_m.invoke(helper_class.newInstance(),params);
		    oals.is_primary=true;
		    oals.is_active=true;

		    new_lg_r_c=search_for_loggrec_c(oals.cr_f_info.the_location[0].id);
		    if(new_lg_r_c!=null){
			if(crt.loggrecCtrls.size()>0){
			     if(repl_s!=WARM_PASSIVE.value)
				((LoggingRecoveryController)crt.loggrecCtrls.elementAt(0)).recover_new_replica_ch(group_id,new_lg_r_c,(Checkpointable)stub,repl_s);
			    else
				((LoggingRecoveryController)crt.loggrecCtrls.elementAt(0)).recover_new_replica_st_ch(group_id,new_lg_r_c,(Checkpointable)stub,(Checkpointable)crt.theStub);
			}
			crt.loggrecCtrls.clear();
			crt.loggrecCtrls.addElement(new_lg_r_c);
		    }
		    new_locations=new String[1];
		    new_locations[0]=oals.cr_f_info.the_location[0].id;
		    new_stubs=new java.lang.Object[1];
		    new_stubs[0]=stub;
		}
	    }
	    b_r_nr_repl=r_nr_repl;
	    if(r_nr_repl<min_nr_repl){
		/* we have to select a factory which can create an object
		   on a host different form the ones where there are members
		   of the object group*/
		factories=search_for_factories(NR_NEW_REPL_W_F,failed_info.cr_f_info.the_location[0].id,crt.factlocEvid,(Vector)groupProperties.get(group),(Vector)typeProperties.get(type_id));
		if(factories==null||factories.length==0){
		    /* The group remains one member less, since no factory is found to create
		       new replicas*/
		    if(repl_s!=ACTIVE.value&&repl_s!=ACTIVE_WITH_VOTING.value)
			new_group=new org.openorb.CORBA.ObjectStub(orb,ge_ior);
		}
		else{
		    /* create as many new replicas as the number of factories found,
		       under the maximum number NR_NEW_REPL_W_F condition*/
		    for(i=0;i<factories.length;i++){
			new_group=create_object(ge_ior,factories[i],type_id,crt,repl_s);
			if(new_group!=null)
			    ge_ior=Utils.get_ior(new_group);
		    }
		    r_nr_repl+=factories.length;
		}
	    }
	    else{
		/* the number of replicas remaining is not less than minimum number of replicas*/
		if(repl_s!=ACTIVE.value&&repl_s!=ACTIVE_WITH_VOTING.value)
		    new_group=new org.openorb.CORBA.ObjectStub(orb,ge_ior);
	    }

	    gP=(Vector)groupProperties.get(group);

	    if(new_group!=null&&gP!=null){
		dyn_p=new Property[gP.size()];
		dyn_p=(Property[])gP.toArray(dyn_p);
		set_properties_dynamically(new_group,dyn_p);
		remove_dyn_props(group);
	    }

	    if(repl_s==ACTIVE.value||repl_s==ACTIVE_WITH_VOTING.value&&b_r_nr_repl<r_nr_repl){
		org.omg.IOP.TaggedProfile[] profs=new org.omg.IOP.TaggedProfile[1];
		profs[0]=ge_ior.profiles[0];
		org.openorb.CORBA.ObjectStub gateway_feature=new org.openorb.CORBA.ObjectStub(orb,new IOR("IDL:omg.org/FT/GatewayForActiveRepl:1.0",profs));
		gateway=GatewayForActiveReplHelper.narrow(gateway_feature);
		ObjFactAndIdStr last_str;
		org.omg.CORBA.Object[] new_members=new org.omg.CORBA.Object[r_nr_repl-b_r_nr_repl];
		new_locations=new String[r_nr_repl-b_r_nr_repl];
		new_stubs=new java.lang.Object[r_nr_repl-b_r_nr_repl];
		for(i=0;i<r_nr_repl-b_r_nr_repl;i++){
		    last_str=(ObjFactAndIdStr)crt.factlocEvid.elementAt(b_r_nr_repl+i);
		    new_members[i]=last_str.obj;
		    new_locations[i]=last_str.cr_f_info.the_location[0].id;
		}
		org.omg.CORBA.Object intrm=null;
		java.lang.Object a_replica;
		if(new_members.length>0){
		    gateway.change_membership(new_members,new_locations);
		    intrm=((ObjFactAndIdStr)crt.factlocEvid.elementAt(0)).gateway;
		    Utils.type_id(type_id,intrm);
		    params[0]=intrm;
		    a_replica=narrow_m.invoke(helper_class.newInstance(),params);
		    byte[] state;
		    /* Special get_state invocation - the gateway will add to the result (in the first four bytes),
		       current_retention_id
		       the "state" is {byte0_ret_id,byte1_ret_id,...,real_state_bytes}
		       when "set_state" is invoked,the real state will be given as parameter, and the retention_id is set
		       with help of an extra method in the class ObjectImpl*/

		    state=((Checkpointable)a_replica).get_state();

		    dyn_arguments=orb.create_list(0);
		    argument=orb.create_any();
		    argument.insert_string(orb.object_to_string((org.omg.CORBA.Object)gateway_feature));
		    dyn_arguments.add_value("param0",argument,flags);
		    for(i=0;i<new_members.length;i++){
			last_str=(ObjFactAndIdStr)crt.factlocEvid.elementAt(b_r_nr_repl+i);
			params=new java.lang.Object[1];
			params[0]=new_members[i];
			new_stubs[i]=narrow_m.invoke(helper_class.newInstance(),params);
			((org.omg.CORBA.portable.ObjectImpl)new_stubs[i]).retention_id(Utils.readRetId(state));
			((Checkpointable)new_stubs[i]).set_state(Utils.readRealState(state));
			dyn_request=new org.openorb.CORBA.dii.Request(last_str.obj,"install_gateway",dyn_arguments,dyn_result,orb.create_environment(),dyn_exceptions,orb.create_context_list(),orb);
			dyn_request.send_oneway();
		    }
		}
	    }
	    else
		if(repl_s!=ACTIVE.value&&repl_s!=ACTIVE_WITH_VOTING.value){
		    if(checkp_int!=-1&&new_lg_r_c!=null&&repl_s!=STATELESS.value)
			checkpoint_object(checkp_int,periodic,stub,group_id,crt,primary,new_lg_r_c,repl_s,narrow_m,helper_class);
		    /* Announce all "server ORBs" about the change*/

		    dyn_arguments=orb.create_list(0);
		    argument=orb.create_any();
		    argument.insert_long(Utils.extract_group_version(orb,ge_ior));
		    dyn_arguments.add_value("param0",argument,flags);
		    argument=orb.create_any();
		    argument.insert_Object(new_group);
		    dyn_arguments.add_value("param1",argument,flags);
		    for(i=0;i<r_nr_repl;i++){
			oals=(ObjFactAndIdStr)crt.factlocEvid.elementAt(i);
			/* Here an UPDATE should happen - the old group reference for this member should
			   be replaced with the new reference, containing one member less (in case the
			   minimum nr. replicas requirement is not violated), or the same nr. of members,
			   but one of them new*/
			dyn_request=new org.openorb.CORBA.dii.Request(oals.obj,"set_latest_group_reference",dyn_arguments,dyn_result,orb.create_environment(),dyn_exceptions,orb.create_context_list(),orb);
			dyn_request.send_oneway();
		    }
		}
	    crt.f_consumer_index=setup_fault_consumer_connection(crt.factlocEvid,type_id,fault_m_g,group_id,true,crt.f_consumer_index);
	    if(repl_s!=ACTIVE.value&&repl_s!=ACTIVE_WITH_VOTING.value){
		crt.theStub=new_stubs[0];
		synchronized(groupEvid){
		    groupEvid.put(new_group,crt);
		}
	    }
	    /* Start the monitoring AFTER connecting to the fault_notifier*/

	    FaultMonitor[] monitors=find_monitors(new_locations);
	    for(i=0;i<monitors.length;i++)
		crt.faultMonits.addElement(monitors[i]);

	    monitor_objects(new_stubs,monitors,fault_m_g,group_id,mon_i,mon_t_o);

	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
    }


    /** this method is used only by the replication manager ("internally") to create an object
       instead of a failed one, if necessary*/



    private org.omg.CORBA.Object create_object(IOR ge_ior,FactoryInfo factory,String type_id,ObjGroupInfoStr the_i,int repl_s) throws NoFactory{
	org.omg.CORBA.Object gr_obj=null,ret;
	AnyHolder fci=new AnyHolder();
	int i,l;
	IOR group_ior,ior_ge=null,ior_grm=null;
	boolean active=false;
	long gr_id;
	byte[] gr_id_b;
	Property[] crit;

	gr_id=the_i.group_id.extract_longlong();
	gr_id_b=Utils.makeIdByte(gr_id);
	crit=set_criteria(repl_s,true,false,0,1,gr_id_b,the_i.crash_after);
	try{
	    gr_obj=factory.the_factory.create_object(type_id,crit,fci);
	}catch(NoFactory e) {return null;}
	catch(InvalidProperty e){System.out.println("Invalid Property in private create_object");return null;}
	catch(InvalidCriteria e){System.out.println("Invalid Criteria in private create_object");return null;}
	catch(ObjectNotCreated e){System.out.println("Object not created in private create_object"); return null;}
	catch(CannotMeetCriteria e){System.out.println("Cannot meet criteria in private create_object"); return null;}
	catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return null;}
	if(repl_s!=COLD_PASSIVE.value&&repl_s!=STATELESS.value)
	    active=true;
	ior_ge=Utils.get_ior(gr_obj);
	the_i.factlocEvid.addElement(new ObjFactAndIdStr(gr_obj,((ObjFactAndIdStr)the_i.factlocEvid.elementAt(0)).gateway,false,active,factory,fci.value));
	if(repl_s==ACTIVE.value||repl_s==ACTIVE_WITH_VOTING.value)
	    return null; /* "group reference" (= gateway_refs+repl_m_ref) known (or published) by the replication manager
			    remains unchanged - the new element is added to the fact_obj list*/
	ge_ior=Utils.update_group_ior_add(orb,ge_ior,ior_ge,ge_ior.profiles.length-1,false);
	ret=new org.openorb.CORBA.ObjectStub(orb,ge_ior);
	return ret;
    }

/** this method is used to create an object and install it as primary in a passive replicated group (usually this does not happen - one of the backups
* can take over the role of primary).The creation of a primary is needed when there are no backups to take over the role of primary
*/

    private org.omg.CORBA.Object[] create_active_object(IOR ge_ior,FactoryInfo factory,String type_id,ObjGroupInfoStr the_i,int repl_s) throws NoFactory{
	org.omg.CORBA.Object gr_obj=null,ret_g,ret_o,toRet[];
	AnyHolder fci=new AnyHolder();
	int i,l;
	IOR group_ior,ior_ge=null,ior_grm=null;
	long gr_id;
	byte[] gr_id_b;
	Property[] crit;

	gr_id=the_i.group_id.extract_longlong();
	gr_id_b=Utils.makeIdByte(gr_id);
	crit=set_criteria(repl_s,true,false,0,0,gr_id_b,the_i.crash_after);
	try{
	    gr_obj=factory.the_factory.create_object(type_id,crit,fci);
	}catch(NoFactory e) {System.out.println("No factory in private create_object");return null;}
	catch(InvalidProperty e){System.out.println("Invalid Property in private create_object");return null;}
	catch(InvalidCriteria e){System.out.println("Invalid Criteria in private create_object");return null;}
	catch(ObjectNotCreated e){System.out.println("Object not created in private create_object"); return null;}
	catch(CannotMeetCriteria e){System.out.println("Cannot meet criteria in private create_object"); return null;}
	catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return null;}
	ior_ge=Utils.get_ior(gr_obj);
	the_i.factlocEvid.addElement(new ObjFactAndIdStr(gr_obj,((ObjFactAndIdStr)the_i.factlocEvid.elementAt(0)).gateway,true,true,factory,fci.value));
	ge_ior=Utils.update_group_ior_add_pr(orb,ge_ior,ior_ge,INDEX_PRIMARY,false);
	toRet=new org.omg.CORBA.Object[2];
	ret_g=new org.openorb.CORBA.ObjectStub(orb,ge_ior);
	ret_o=new org.openorb.CORBA.ObjectStub(orb,ior_ge);
	toRet[0]=ret_g;
	toRet[1]=ret_o;
	return toRet;
    }


    private FaultMonitor search_for_monitor(String location){
	int l,j;
	l=theFaultD_l.size();
	FaultMonitorInfo crt_fmi;
	for(j=0;j<l;j++){
	    crt_fmi=(FaultMonitorInfo)theFaultD_l.elementAt(j);
	    if(crt_fmi.the_location[0].id.equals(location)) return crt_fmi.the_monitor;
	}
	return null;
    }

    private LoggingRecoveryController search_for_loggrec_c(String location){
	int l,j;

	l=theLoggRecC_l.size();

	LoggingRecoveryControllerInfo crt_lrci;

	for(j=0;j<l;j++){
	    crt_lrci=(LoggingRecoveryControllerInfo)theLoggRecC_l.elementAt(j);
    if(crt_lrci.the_location[0].id.equals(location)) return crt_lrci.the_loggrec_c;
	}
	return null;
    }


    /* this method should be called when factories (usually one factory) are needed to create
       new replica(s), in case a replica failed, and the minimum nr. replicas requirement is
       not any more satisfied; the criteria used when selecting the factories, is that the
       other members of the group are not at the same location as the chosen factory, and, if possible,
       the locationof the chosen factory is not a location where a member failed*/

    private FactoryInfo[] search_for_factories(int nr,String failure_location,Vector obj_a_loc,Vector grProp,Vector typeProp){
	Vector official=new Vector(),reserve=new Vector(),propV;
	int sp,sf,l,i,j,of_s,rs_s;
	FactoryInfo[] factories=null,toRet=null;
	Property prop;
	ObjFactAndIdStr oals;

	propV=Utils.prop_v_union_o(grProp,typeProp,defaultProperties);

	sp=propV.size();
	for(i=0;i<sp;i++){
	    prop=(Property)propV.elementAt(i);
	    if(prop.nam[0].id.equals("org.omg.ft.Factories")){
		factories=FactoryInfosHelper.extract(prop.val);
		break;
	    }
	}
	if(factories==null) {return null;}
	l=factories.length;
	if(l==0) return null;
	sf=obj_a_loc.size();
	for(j=0;j<l;j++){
	    for(i=0;i<sf;i++){
		oals=(ObjFactAndIdStr)obj_a_loc.elementAt(i);
		/* we test if the existing group members are not
		   in a location of one of the factories*/
		if(oals.cr_f_info.the_location[0].id.equals(factories[j].the_location[0].id))
		    break;
	    }
	    if(i>=sf){
		/*we test if the factory found is at the location where
		  the current object failed; if it is, we put it in a "reserve" list*/
		if(factories[j].the_location[0].id.equals(failure_location))
		    reserve.addElement(factories[j]);
		else
		    official.addElement(factories[j]);
		if(official.size()>=nr) break;
	    }
	}
	of_s=official.size();
	rs_s=reserve.size();
	if(of_s>=nr){
	    /* More than one member should be created, because next time when a failure
	       occurs, the same replacemant will happen, thus an oscillating process
	       goes on - not yet implemented this way*/

	    /* A new member is created to maintain the minimum nr. of replicas;*/
	    toRet=new FactoryInfo[nr];
	    for(i=0;i<nr;i++)
		toRet[i]=(FactoryInfo)official.elementAt(i);
	}

	else
	    if(of_s+rs_s>=nr){
		toRet=new FactoryInfo[nr];
		for(i=0;i<of_s;i++)
		    toRet[i]=(FactoryInfo)official.elementAt(i);
		j=i;
		for(;i<nr;i++)
		    toRet[i]=(FactoryInfo)reserve.elementAt(i-j);
	    }
	    else
		if(of_s+rs_s>0){
		    toRet=new FactoryInfo[of_s+rs_s];
		    for(i=0;i<of_s;i++)
			toRet[i]=(FactoryInfo)official.elementAt(i);
		    j=i;
		    for(;i<of_s+rs_s;i++)
			toRet[i]=(FactoryInfo)reserve.elementAt(i-j);
		}
		else {
		    /* if there could not be found a factory to create a new replica to maintain the min_nr_repl*/
		    toRet=null;
		}
	return toRet;
    }


    private void sort_by_load(FactoryInfo[] fis){
	FactoryInfo a;
	int i,j;
	boolean change;

	for(i=0;i<fis.length;i++){
	    change=false;
	    for(j=0;j<fis.length-i-1;j++){

		if(fis[j].the_factory.get_load()>fis[j+1].the_factory.get_load()){
		    a=fis[j];
		    fis[j]=fis[j+1];
		    fis[j+1]=a;
		    change=true;
		}
	    }
	    if(!change) break;
	}
    }

    private FactoryInfo[] search_for_factories(FactoryInfo[] choice,int nr,String[] pref_locs){
	int i,j,of_s;
	Vector official=new Vector(),reserve=new Vector();
	FactoryInfo[] toRet=new FactoryInfo[nr];
	if(choice==null) return null;
	//sort_by_load(choice);
	if(pref_locs==null){
	    for(i=0;i<nr;i++)
		toRet[i]=choice[i];
	    return toRet;
	}
	for(i=0;i<choice.length;i++){
	    for(j=0;j<pref_locs.length;j++)
		if(choice[i].the_location[0].id.equals(pref_locs[j])){
		    official.addElement(choice[i]);
		    break;
		}
	    if(official.size()>=nr)
		break;
	    else
		reserve.addElement(choice[i]);
	    }
	of_s=official.size();
	if(of_s>=nr)
	    for(i=0;i<nr;i++)
		toRet[i]=(FactoryInfo)official.elementAt(i);
	else{
	    for(i=0;i<of_s;i++)
		toRet[i]=(FactoryInfo)official.elementAt(i);
	    j=i;
	    for(;i<nr;i++)
		toRet[i]=(FactoryInfo)reserve.elementAt(i-j);
	}
	return toRet;
    }

/** Method used to prepare the property vector to be sent to the application object factories to create appropriate replicas for the group,
* according to what the application writer requested from the Replication Manager in terms of replication style, group size, checkpoinitng interval, etc.
*/
    private Property[] set_criteria(int replic_s,boolean join,boolean is_first,int index,int index_pr,byte[] objI,int crash_after){
	Property crit[]=new Property[1];
	switch(replic_s){
	case ACTIVE.value:
	case ACTIVE_WITH_VOTING.value:
	    if(join){
		crit=new Property[4];
		crit[3]=new Property();
		crit[3].nam=new org.omg.CosNaming.NameComponent[1];
		crit[3].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.Join","string");
		crit[3].val=orb.create_any();
	    }
	    else
		if(is_first){
		    crit=new Property[4];
		    crit[3]=new Property();
		    crit[3].nam=new org.omg.CosNaming.NameComponent[1];
		    crit[3].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.FirstFactory","string");
		    crit[3].val=orb.create_any();
		}
		else
		    crit=new Property[3];
	    crit[2]=new Property();
	    crit[2].nam=new org.omg.CosNaming.NameComponent[1];
	    crit[2].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.LifeLength","string");
	    crit[2].val=orb.create_any();
	    crit[2].val.insert_long(crash_after);
	    crit[1]=new Property();
	    crit[1].nam=new org.omg.CosNaming.NameComponent[1];
	    crit[1].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.Active","string");
	    crit[1].val=orb.create_any();
	    break;
	case WARM_PASSIVE.value:
	    if(index!=index_pr){
		crit=new Property[4];
		crit[3]=new Property();
		crit[3].nam=new org.omg.CosNaming.NameComponent[1];
		crit[3].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.Backup","string");
		crit[3].val=orb.create_any();
	    }
		else
		    crit=new Property[3];
	    crit[2]=new Property();
	    crit[2].nam=new org.omg.CosNaming.NameComponent[1];
	    crit[2].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.LifeLength","string");
	    crit[2].val=orb.create_any();
	    crit[2].val.insert_long(crash_after);
	    crit[1]=new Property();
	    crit[1].nam=new org.omg.CosNaming.NameComponent[1];
	    crit[1].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.WarmPassive","string");
	    crit[1].val=orb.create_any();
	    break;
	case STATELESS.value:

	    crit=new Property[4];
	    if(index==index_pr){
		crit[3]=new Property();
		crit[3].nam=new org.omg.CosNaming.NameComponent[1];
		crit[3].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.ActivateIt","string");
		crit[3].val=orb.create_any();
	    }
	    else{
		crit[3]=new Property();
		crit[3].nam=new org.omg.CosNaming.NameComponent[1];
		crit[3].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.Backup","string");
		crit[3].val=orb.create_any();
	    }
	    crit[2]=new Property();
	    crit[2].nam=new org.omg.CosNaming.NameComponent[1];
	    crit[2].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.LifeLength","string");
	    crit[2].val=orb.create_any();
	    crit[2].val.insert_long(crash_after);
	    crit[1]=new Property();
	    crit[1].nam=new org.omg.CosNaming.NameComponent[1];
	    crit[1].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.NoLogging","string");
	    crit[1].val=orb.create_any();
	    break;
	case COLD_PASSIVE.value:
		crit=new Property[3];
	    if(index==index_pr){
		crit[2]=new Property();
		crit[2].nam=new org.omg.CosNaming.NameComponent[1];
		crit[2].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.ActivateIt","string");
		crit[2].val=orb.create_any();
	    }
	    else{
		crit[2]=new Property();
		crit[2].nam=new org.omg.CosNaming.NameComponent[1];
		crit[2].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.Backup","string");
		crit[2].val=orb.create_any();
	    }
	    crit[1]=new Property();
	    crit[1].nam=new org.omg.CosNaming.NameComponent[1];
	    crit[1].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.LifeLength","string");
	    crit[1].val=orb.create_any();
	    crit[1].val.insert_long(crash_after);
	    break;
	}
	crit[0]=new Property();
	crit[0].nam=new org.omg.CosNaming.NameComponent[1];
	crit[0].nam[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.ObjectId","string");
	crit[0].val=orb.create_any();
	org.omg.PortableServer.ObjectIdHelper.insert(crit[0].val,objI);
	return crit;
    }

    /** Method called to add a new Fault Monitor to the ReplicationManager's list of FMs*/

    synchronized public void add_fault_monitor(FaultMonitorInfo fmi){
	if(!theFaultD_l.contains(fmi))
	    theFaultD_l.addElement(fmi);
    }
 /** Method called to add a Fault Notifier to the ReplicationManager's list of FNs*/
    synchronized public void add_logging_recovery_c(LoggingRecoveryControllerInfo lrci){
	if(!theLoggRecC_l.contains(lrci))
	    theLoggRecC_l.addElement(lrci);
    }


    /* BEGIN GenericFactory METHODS*/
/** Method specified in the GenericFactory interface
 * @param	type_id the type identifier of the CORBA objects members of the created group
 *			the_criteria set of properties sent by the application writer including replication style, group size, fault monitoring interval and timeout, etc.
 *			factory_creation_id an identifier returned to the caller of the method to use in case he/she wants to delete the created group
 * @return 	the collective object reference of the object group (the role of the Repl Manager is also to assemble such an object reference)
 */

    public org.omg.CORBA.Object create_object(java.lang.String type_id, Property[] the_criteria, AnyHolder factory_creation_id){

	long fault_consumer_index=-1;
	org.omg.CORBA.Object gr_obj,gateway=null,member;
	int i,i_nr_r=-1,min_n_repl=0,j,l,m_nr_r=-1,tr_to_s=0,tr_to_r=0;
	boolean chp_upd_per=false;
	AnyHolder fci_c=new AnyHolder();
	org.openorb.CORBA.ObjectStub ret;
	int replication_s=-1,membership_s=-1,fault_mon_gran=-1,index_pr,crash_after=-1;
	boolean primary=false,active=false;
	IOR ior_o,iors[]=new IOR[0],iors_grm[]=new IOR[0];
	TaggedProfile[] m_c_prof;
	FactoryInfo[] theFact_prop=null,theFact_l=null;
	Vector obj_locs_fact_i=new Vector(), log_r_v=new Vector(), fault_monits=new Vector();/*the vector for object id and location information about every member*/
	org.omg.IIOP.ProfileBody_1_1 pb11;
	Property crt,groupP[],crit[];
	long checkpoint_interval=-1,fault_mon_int=4000,fault_mon_t_o=4000;
	ObjectLocStr members_for_gw[]=new ObjectLocStr[0];

	org.openorb.CORBA.dii.Request dyn_request;
	org.omg.CORBA.Any argument;
	org.omg.CORBA.NVList dyn_arguments;
	org.omg.CORBA.ExceptionList dyn_exceptions=orb.create_exception_list();
	org.omg.CORBA.NamedValue dyn_result=orb.create_named_value("",orb.create_any(),-1);


	java.lang.Object stub=null,params[],stubs[]=new java.lang.Object[0];
	String fm_locations[]=new String[0];

	synchronized(create_object_semaphor){
	    factory_creation_id.value=orb.create_any();
	    try{
		Class helper_class=Class.forName(Utils.get_class_helper(type_id));
		Class[] par_cl=new Class[1];
		par_cl[0]=Class.forName("org.omg.CORBA.Object");
		java.lang.reflect.Method narrow_m=helper_class.getMethod("narrow",par_cl);
		byte[] objI=new byte[0];

		currProperties=Utils.prop_v_union((Vector)typeProperties.get(type_id),defaultProperties,0);
		l=currProperties.size();
		for(i=0;i<l;i++){
		    crt=(Property)currProperties.elementAt(i);
		    if(crt.nam[0].id.equals("org.omg.ft.MinimumNumberReplicas")){
			m_nr_r=crt.val.extract_long();
		    }

		    if(crt.nam[0].id.equals("org.omg.ft.LifeLength")){
			crash_after=crt.val.extract_long();
		    }
		    if(crt.nam[0].id.equals("org.omg.ft.InitialNumberReplicas"))
			{
			    i_nr_r=crt.val.extract_long();
			}
		    if(crt.nam[0].id.equals("org.omg.ft.ReplicationStyle")){
			replication_s=crt.val.extract_long();
		    }
		    if(crt.nam[0].id.equals("org.omg.ft.MembershipStyle")){
			membership_s=crt.val.extract_long();
		    }
		    if(crt.nam[0].id.equals("org.omg.ft.Factories")){
			theFact_prop=FactoryInfosHelper.extract(crt.val);
		    }
		    if(crt.nam[0].id.equals("org.omg.ft.FaultMonitoringGranularity")){
			fault_mon_gran=crt.val.extract_long();
		    }
		    if(crt.nam[0].id.equals("org.omg.ft.CheckpointInterval")){
			checkpoint_interval=crt.val.extract_longlong();
		    }
		    if(crt.nam[0].id.equals("org.omg.ft.CheckpointPeriodic")){
			chp_upd_per=true;
		    }
		    if(crt.nam[0].id.equals("org.omg.ft.FaultMonitoringInterval")){
			fault_mon_int=crt.val.extract_longlong();
		    }
		    if(crt.nam[0].id.equals("org.omg.ft.FaultMonitoringTimeout")){
			fault_mon_t_o=crt.val.extract_longlong();
		    }
		}

		TagFTGroupTaggedComponent all_prof_comp=new TagFTGroupTaggedComponent(new org.omg.GIOP.Version((byte)1,(byte)0),ftDomName,objGrId,0,replication_s);
		TagFTPrimaryTaggedComponent prof1_comp=new TagFTPrimaryTaggedComponent(true);

		if(membership_s==MEMB_APP_CTRL.value)
		    {
			/*create a group ref with only one profile with tag TAG_MULTIPLE_COMPONENTS*/

			m_c_prof=new TaggedProfile[1];
			pb11=new org.omg.IIOP.ProfileBody_1_1(new org.omg.IIOP.Version((byte)1,(byte)0),null,(short)-1,new byte[0],new TaggedComponent[1]);
			m_c_prof[0]=Utils.body_profile(orb,pb11);
			m_c_prof[0].tag=TAG_MULTIPLE_COMPONENTS.value;
			m_c_prof[0]=Utils.ext_profile_tag(orb,m_c_prof[0],all_prof_comp,TAG_FT_GROUP.value);
			factory_creation_id.value.insert_longlong(objGrId);
			ior_o=new IOR(type_id,m_c_prof);
		    }

		else{ /* MEMB_INF_CTRL*/
		    LoggingRecoveryController lg_r_c=null;
		    params=new java.lang.Object[1];
		    min_n_repl=Utils.min(i_nr_r,theFact_prop.length);
		    /* if the number of factories is less than the minimum
		       number of replicas then throw an exception*/
		    if(min_n_repl<m_nr_r) {throw new java.lang.Exception();}
		    /* there is 1 extra IOR - the IOR of the replication manager self*/
		    if(replication_s==ACTIVE.value||replication_s==ACTIVE_WITH_VOTING.value)
			iors=new IOR[2];
		    else
			iors=new IOR[min_n_repl+1];
		    Checkpointable[] transfer_state_to=new Checkpointable[0];
		    LoggingRecoveryController[] log_rec_ctrl_n=new LoggingRecoveryController[0];
		    org.omg.CORBA.Object the_warm_primary=null;
		    if(replication_s==ACTIVE.value||replication_s==ACTIVE_WITH_VOTING.value){
			index_pr=-1;
			iors_grm=new IOR[min_n_repl];
			stubs=new java.lang.Object[min_n_repl];
			fm_locations=new String[min_n_repl];
			members_for_gw=new ObjectLocStr[min_n_repl];
		    }
		    else {
			/* it is not necessarilly the case that index_pr is 0; by now, the primary profile
			   is the first profile from the composed IOR*/
			index_pr=INDEX_PRIMARY;
			if(replication_s==WARM_PASSIVE.value){
			    transfer_state_to=new Checkpointable[min_n_repl-1];
			    log_rec_ctrl_n=new LoggingRecoveryController[min_n_repl-1];
			}
		    }

		    objI=Utils.makeIdByte(objGrId);
		    theFact_l=search_for_factories(theFact_prop,min_n_repl,null); /*for now, the pref. locations is null*/
		    if(theFact_l==null) {return null;}
		    for(i=0;i<min_n_repl;i++){
			try{
			    crit=set_criteria(replication_s,false,(i==0),i,index_pr,objI,crash_after);
			    gr_obj=null;
			    /* Here, the factory from a wanted location should be selected.
			       It is not yet implemented that way!!!!!!*/
			    member=theFact_l[i].the_factory.create_object(type_id,crit,fci_c);
			}catch(NoFactory e) {System.out.println("No factory - After cr. obj in repl man.");return null;}
			catch(InvalidProperty e){System.out.println("Invalid Property - After cr. obj in repl man.");return null;}
			catch(InvalidCriteria e){System.out.println("Invalid Criteria - After cr. obj in repl man.");return null;}
			catch(ObjectNotCreated e){System.out.println("Object not created - After cr. obj in repl man."); return null;}
			catch(CannotMeetCriteria e){System.out.println("Cannot meet criteria - After cr. obj in repl man."); return null;}
			/* attach the object to a fault monitor, if necessary*/

			/* the fault monitoring of a member of a group happens if:
			   - the fault monitoring granularity is MEMB and this member HAS to be monitored,
			   i.e. the replication style is ACTIVE (meaning that all members of the group are
			   equal and monitorable), or the member is the primary of the group
			   - the fault monitoring granularity is LOC or LOC_AND_TYPE and the replication style is ACTIVE,
			   or the member is the primary of the group, and at its location and/or from its type there is
			   no other member monitored*/

			/* start the checkpointing for primary members*/

			primary=false;
			active=false;

			switch(replication_s){
			case STATELESS.value:
			case COLD_PASSIVE.value:
			    gr_obj=member;
			    if(index_pr==i){
				primary=true;
				active=true;
				params[0]=gr_obj;
				stub=narrow_m.invoke(helper_class.newInstance(),params);
				stubs=new java.lang.Object[1];
				fm_locations=new String[1];
				fm_locations[0]=theFact_l[i].the_location[0].id;
				stubs[0]=stub;
				if(replication_s==STATELESS.value||checkpoint_interval!=-1){
				    lg_r_c=search_for_loggrec_c(theFact_l[i].the_location[0].id);
				    if(lg_r_c!=null){
					log_r_v.addElement(lg_r_c);
					if(replication_s!=STATELESS.value)
					    lg_r_c.do_checkpointing((Checkpointable)stub,objGrId,checkpoint_interval,transfer_state_to,false,chp_upd_per);
				    }
				}
			    }
			    break;
			case WARM_PASSIVE.value:
			    gr_obj=member;
			    if(index_pr==i){
				primary=true;
				the_warm_primary=gr_obj;
			    }
			    else{
				lg_r_c=search_for_loggrec_c(theFact_l[i].the_location[0].id);
				if(lg_r_c!=null) tr_to_r++;
				log_rec_ctrl_n[tr_to_s]=lg_r_c;
				params[0]=gr_obj;
				transfer_state_to[tr_to_s]=(Checkpointable)narrow_m.invoke(helper_class.newInstance(),params);
				tr_to_s++;
			    }
			    active=true;
			    break;
			case ACTIVE.value:
			case ACTIVE_WITH_VOTING.value:
			    IOR total_ior=Utils.get_ior(member);
			    TaggedProfile[] ior_p;
			    ior_p=new TaggedProfile[1];
			    ior_p[0]=new TaggedProfile(total_ior.profiles[0].tag,total_ior.profiles[0].profile_data);
			    iors_grm[i]=new IOR(type_id,ior_p);
			    gr_obj=new org.openorb.CORBA.ObjectStub(orb,iors_grm[i]);
			    members_for_gw[i]=new ObjectLocStr(gr_obj,theFact_l[i].the_location[0].id);
			    if(i==0){
				ior_p=new TaggedProfile[1];
				ior_p[0]=new TaggedProfile(total_ior.profiles[1].tag,total_ior.profiles[1].profile_data);
				iors[i]=new IOR("IDL:omg.org/FT/GatewayForActiveRepl:1.0",ior_p);
				gateway=new org.openorb.CORBA.ObjectStub(orb,iors[i]);
			    }
			    params[0]=gr_obj;
			    stub=narrow_m.invoke(helper_class.newInstance(),params);
			    stubs[i]=stub;
			    fm_locations[i]=theFact_l[i].the_location[0].id;
			    primary=false;
			    active=true;
			    break;
			}

			/* add the information about the just created member
			   to the "location and object reference info" vector
			   associated to this group*/

			obj_locs_fact_i.addElement(new ObjFactAndIdStr(gr_obj,gateway,primary,active,theFact_l[i],fci_c.value));
			try {
			    /* remember the IOR of the created object (it will be used later
			       in merge_iors*/
			    if(replication_s!=ACTIVE.value&&replication_s!=ACTIVE_WITH_VOTING.value){
				iors[i] = Utils.get_ior(gr_obj);
				for(j=0;j<iors[i].profiles.length;j++)
				    iors[i].profiles[j]=Utils.ext_profile_tag(orb,iors[i].profiles[j],all_prof_comp,TAG_FT_GROUP.value);
			    }
			    else
				if(i==0)
				    for(j=0;j<iors[i].profiles.length;j++)
					iors[i].profiles[j]=Utils.ext_profile_tag(orb,iors[i].profiles[j],all_prof_comp,TAG_FT_GROUP.value);
				/* Add the TAG_FT_GROUP tagged component to the profiles of the IOR
				   (In general the number of profiles is 1)*/
			}catch(java.lang.Exception e) {
			    e.printStackTrace(new java.io.PrintStream(System.out));return null;
			}
		    } /*end for(i=0;i<min_n_repl;i++)*/

		    /* The checkpointing has to be started somewhere here*/

		    /* only for warm passive here, because I know to whom to transfer the state
		       only after all members were created*/

		    if(replication_s==WARM_PASSIVE.value){
			params[0]=the_warm_primary;
			stub=narrow_m.invoke(helper_class.newInstance(),params);
			stubs=new java.lang.Object[1];
			stubs[0]=stub;
			fm_locations=new String[1];
			fm_locations[0]=theFact_l[index_pr].the_location[0].id;
			if(checkpoint_interval!=-1){
			    lg_r_c=search_for_loggrec_c(theFact_l[index_pr].the_location[0].id);
			    if(lg_r_c!=null){
				log_r_v.addElement(lg_r_c);
				if(tr_to_r<log_rec_ctrl_n.length){
				    LoggingRecoveryController[] real_neighbour_s=new LoggingRecoveryController[tr_to_r];
				    int k=0;
				    for(i=0;i<log_rec_ctrl_n.length;i++)
					if(log_rec_ctrl_n[i]!=null){
					    real_neighbour_s[k]=log_rec_ctrl_n[i];
					    k++;
					}
				    log_rec_ctrl_n=real_neighbour_s;
				}
				lg_r_c.set_neighbours(log_rec_ctrl_n);
				lg_r_c.do_checkpointing((Checkpointable)stub,objGrId,checkpoint_interval,transfer_state_to,true,chp_upd_per);
			    }
			}
		    }

		    /* add tagprimarycomponent to the first profile of the first ior
		       - this happens only for non-ACTIVE replication style*/

		    if(replication_s!=ACTIVE.value&&replication_s!=ACTIVE_WITH_VOTING.value)
			iors[index_pr].profiles[0]=Utils.ext_profile_tag(orb,iors[index_pr].profiles[0],prof1_comp,TAG_FT_PRIMARY.value);
		    /* on the last position of the IOGR add the IOR of the replication manager self*/
		    if(replication_s==ACTIVE.value||replication_s==ACTIVE_WITH_VOTING.value)
			iors[1]=Utils.get_ior(this._this(orb));
		    else
			iors[min_n_repl]=Utils.get_ior(this._this(orb));
		    ior_o=Utils.merge_unique_iors(type_id,iors);
		    factory_creation_id.value.insert_longlong(objGrId);
		}

		ret=new org.openorb.CORBA.ObjectStub(orb,ior_o);


		if(membership_s==MEMB_INF_CTRL.value){
		    if(replication_s!=ACTIVE.value&&replication_s!=ACTIVE_WITH_VOTING.value){
			dyn_arguments=orb.create_list(0);
			argument=orb.create_any();
			argument.insert_long(0);
			dyn_arguments.add_value("param0",argument,flags);
			argument=orb.create_any();
			argument.insert_Object(ret);
			dyn_arguments.add_value("param1",argument,flags);
			for(i=0;i<min_n_repl;i++){
			    try{
				dyn_request=new org.openorb.CORBA.dii.Request(((ObjFactAndIdStr)obj_locs_fact_i.elementAt(i)).obj,"set_latest_group_reference",dyn_arguments,dyn_result,orb.create_environment(),dyn_exceptions,orb.create_context_list(),orb);
				dyn_request.send_oneway();
			    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
			}
		    }
		    else{
			Utils.get_ior(ret).type_id="IDL:omg.org/FT/GatewayForActiveRepl:1.0";
			dyn_arguments=orb.create_list(0);
			argument=orb.create_any();
			argument.insert_string(orb.object_to_string((org.omg.CORBA.Object)gateway));
			dyn_arguments.add_value("param0",argument,flags);
			for(i=0;i<min_n_repl;i++){
			    dyn_request=new org.openorb.CORBA.dii.Request(((ObjFactAndIdStr)obj_locs_fact_i.elementAt(i)).obj,"install_gateway",dyn_arguments,dyn_result,orb.create_environment(),dyn_exceptions,orb.create_context_list(),orb);
			    dyn_request.send_oneway();
			}
			//	theFact_l[i].the_factory.install_gateway(((ObjFactAndIdStr)obj_locs_fact_i.elementAt(i)).fact_cr_id,(org.omg.CORBA.Object)ret);
			GatewayForActiveReplHelper.narrow(gateway).init_membership(members_for_gw);
			Utils.get_ior(ret).type_id=type_id;
		    }
		}
		/* add the group reference to the hashtable which has the group evidence*/
	    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return null;}

	    /* set the group properties for this group (replication style, fault monitoring granularity,
	       minimum number of replicas*/
	    int nr;
	    nr=0;
	    Any any;
	    org.omg.CosNaming.NameComponent[] nc;
	    boolean r_s=false,f_g=false,m_n=false,c_i=false,m_s=false;
	    if(replication_s!=-1) {nr++;r_s=true;}
	    if(fault_mon_gran!=-1) {nr++;f_g=true;}
	    if(m_nr_r!=-1) {nr++;m_n=true;}
	    if(checkpoint_interval!=-1){nr++;c_i=true;}
	    if(membership_s!=-1){nr++;m_s=true;}
	    if(chp_upd_per){nr++;}
	    try{
		groupP=new Property[nr+2];
		nr=0;

		if(r_s) {
		    any=orb.create_any();
		    any.insert_long(replication_s);
		    nc=new org.omg.CosNaming.NameComponent[1];
		    nc[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.ReplicationStyle","string");
		    groupP[nr]=new Property(nc,any);
		    nr++;
		}
		if(f_g){
		    any=orb.create_any();
		    any.insert_long(fault_mon_gran);
		    nc=new org.omg.CosNaming.NameComponent[1];
		    nc[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.FaultMonitoringGranularity","string");
		    groupP[nr]=new Property(nc,any);
		    nr++;
		}
		if(m_n){
		    any=orb.create_any();
		    any.insert_long(m_nr_r);
		    nc=new org.omg.CosNaming.NameComponent[1];
		    nc[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.MinimumNumberReplicas","string");
		    groupP[nr]=new Property(nc,any);
		    nr++;
		}
		if(c_i){
		    any=orb.create_any();
		    any.insert_longlong(checkpoint_interval);
		    nc=new org.omg.CosNaming.NameComponent[1];
		    nc[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.CheckpointInterval","string");
		    groupP[nr]=new Property(nc,any);
		    nr++;
		}
		if(m_s){
		    any=orb.create_any();
		    any.insert_long(membership_s);
		    nc=new org.omg.CosNaming.NameComponent[1];
		    nc[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.MembershipStyle","string");
		    groupP[nr]=new Property(nc,any);
		    nr++;
		}
		if(chp_upd_per){
		    any=orb.create_any();
		    nc=new org.omg.CosNaming.NameComponent[1];
		    nc[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.CheckpointPeriodic","string");
		    groupP[nr]=new Property(nc,any);
		    nr++;
		}

		any=orb.create_any();
		any.insert_longlong(fault_mon_int);
		nc=new org.omg.CosNaming.NameComponent[1];
		nc[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.FaultMonitoringInterval","string");
		groupP[nr]=new Property(nc,any);

		nr++;
	    	any=orb.create_any();
		any.insert_longlong(fault_mon_t_o);
		nc=new org.omg.CosNaming.NameComponent[1];
		nc[0]=new org.omg.CosNaming.NameComponent("org.omg.ft.FaultMonitoringTimeout","string");
		groupP[nr]=new Property(nc,any);

		set_properties_dynamically(ret,groupP);
	    }catch(java.lang.Exception e){e.printStackTrace();}
	    /* register to the fault notifier, if the f.n. is set*/

	    fault_consumer_index=setup_fault_consumer_connection(obj_locs_fact_i,type_id,fault_mon_gran,objGrId,false,-1);
	    FaultMonitor[] monitors=find_monitors(fm_locations);
	    for(i=0;i<monitors.length;i++)
		fault_monits.addElement(monitors[i]);

	    synchronized(groupEvid){
		groupEvid.put(ret,new ObjGroupInfoStr(fault_consumer_index,factory_creation_id.value,crash_after,obj_locs_fact_i,log_r_v,fault_monits,(replication_s!=ACTIVE.value&&replication_s!=ACTIVE_WITH_VOTING.value)?stubs[0]:null));
	    }

	    monitor_objects(stubs,monitors,fault_mon_gran,objGrId,fault_mon_int,fault_mon_t_o);
	    objGrId++;
	    return ret;
	}
    }

/** Method specifid in the GenericFactory interface - deletes a certain group
* @param factory_creation_id the group identifier obtained when creating the group, returned by the Replication Manager
*/

    public void delete_object(Any factory_creation_id) throws ObjectNotFound{

	ObjFactAndIdStr value;
	GenericFactory crt_f;
	FaultMonitor crt_fm;
	Any crt_i;
	int i,l;
	long f_c_i;
	boolean found=false;
	ObjGroupInfoStr crt;
	Map.Entry entry=null;
	Set elems;
	Iterator setIter;

	synchronized(delete_object_semaphor){
	    elems=groupEvid.entrySet();
	    setIter=elems.iterator();
	    while(setIter.hasNext()){
		entry=(Map.Entry)setIter.next();
		crt=(ObjGroupInfoStr)entry.getValue();
		if(crt.group_id.extract_longlong()==factory_creation_id.extract_longlong()) {
		    l=crt.factlocEvid.size();
		    for(i=0;i<l;i++){
			value=(ObjFactAndIdStr)crt.factlocEvid.elementAt(i);
			value.cr_f_info.the_factory.delete_object(value.fact_cr_id);
			crt_fm=(FaultMonitor)crt.faultMonits.elementAt(i);
			if(crt_fm!=null)
			    crt_fm.stop_monitor(value.obj);
		    }
		    found=true;
		    break;
		}
	    }
	    if(found){
		groupEvid.remove(entry.getKey());
		if(fault_notifier!=null)
		    try{
			fault_notifier.disconnect_consumer(((ObjGroupInfoStr)entry.getValue()).f_consumer_index);
		    }catch(org.omg.CosEventComm.Disconnected e){}
	    }
	}
    }

    public int get_load(){return groupEvid.size();}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
    * It is not implemented in the Replication Manager
    */

    public void activate_object(org.omg.CORBA.Any f_c_i,long g_i,boolean f_t){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public void replay_last_requests(org.omg.CORBA.Any a, org.omg.FT.LogRecord[] lrr, org.omg.FT.LogRecord[] lrnr, org.omg.FT.LoggingRecoveryController lrc, int ret_id){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public void install_gateway(org.omg.CORBA.Any a,org.omg.CORBA.Object gw){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public void add_new_type_id(String type_id,String class_name,String[] new_arg_class_n,  java.lang.Object[] new_arg_values,  MethSignStr[] methods){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public void remove_type_id(String type_id){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public void modify_class_name_ti(String type_id,String new_class_name){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public void modify_class_name_ti_ctr(String type_id,String new_class_name,String[] class_names,  java.lang.Object[] arg_values){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public void add_new_method_type_id_s(String type_id,MethSignStr the_method){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public void add_new_method_type_id_d(String type_id,String method_n,  org.omg.CORBA.TypeCode[] par_types,  org.omg.CORBA.TypeCode[] exc_types,  int[] modes,char method_type){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public void remove_method_type_id(String type_id,String the_method){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public  void modify_meth_type_id_add_param(String type_id,String the_method,  org.omg.CORBA.TypeCode par_type,  int mode, int pos){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public void modify_meth_type_id_add_exc(String type_id,String the_method, org.omg.CORBA.TypeCode exc_type,int pos){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public void modify_meth_type_id_remove_param(String type_id,String the_method, int pos){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public   void modify_meth_type_id_remove_exc(String type_id,String the_method,int pos){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public  void modify_meth_type_id_modify_param(String type_id,String the_method,org.omg.CORBA.TypeCode par_type,int mode, int pos){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public  void modify_meth_type_id_modify_exc(String type_id,String the_method,org.omg.CORBA.TypeCode exc_type,int pos){}

    /** Extra method (not in the FT-CORBA specification) in the GenericFactory interface - used in the Object Factory
	    * It is not implemented in the Replication Manager
    */

    public  void modify_meth_type_id_modify_type(String type_id,String the_method,char meth_type){}

    /*END GenericFactory METHODS*/


    /*BEGIN ObjectGroupManager METHODS*/

   /** ObjectGroupManager interace method - creates a member of an already existing group according to the criteria given by the invoker (application writer)
    * Methods in ObjectGroupManager are used when the group membership is application controlled, and not infrastructure controlled
    * @param	group CORBA object identifying the group for which a new member is ceated and added
    *			loc the location where the new memebr has to be created
    *			type_id the type identifier of the member (of course, the same as that of all other members' of the group)
    *			crit the criteria given to the Object Factory from the user
    * @return	the CORBA object identiying the new group
    */

    public org.omg.CORBA.Object create_member(org.omg.CORBA.Object group, org.omg.CosNaming.NameComponent[] loc, java.lang.String type_id, Property[] crit) throws MemberAlreadyPresent, ObjectGroupNotFound{
	   FactoryInfo[] theFact_l=null;
	   ObjFactAndIdStr fact_loc_id_o;
	   int i,l,memb_s=-1;
	   Property crt;
	   ObjGroupInfoStr ogi=null;
	   org.omg.CORBA.Object obj=null,new_group=null;
	   AnyHolder fci=new AnyHolder(); /*factory creation id for the new member*/
	   Any fci_g=null; /* factory creation id of the object group*/
	   long g_i; /* the group id*/
	   Map.Entry entry;
	   Set elems; /* the elements of the factoriesEvid hashtable*/
	   /*it is used to search for the factories used in creating members of this group*/
	   Iterator setIter;
	   ogi=(ObjGroupInfoStr)groupEvid.get(group);

	   if(ogi==null) throw new ObjectGroupNotFound();
	   g_i=ogi.group_id.extract_longlong();
	   /*check whether there is already a member of the group at the given location*/
	   l=ogi.factlocEvid.size();
	   for(i=0;i<l;i++){
	     fact_loc_id_o=(ObjFactAndIdStr)ogi.factlocEvid.elementAt(i);
	     if(fact_loc_id_o.cr_f_info.the_location[0].id.equals(loc[0].id))
		  throw new MemberAlreadyPresent();
	   }
	   currProperties=Utils.prop_v_union_o((Vector)groupProperties.get(group),(Vector)typeProperties.get(type_id),defaultProperties);
	   l=currProperties.size();
	   for(i=0;i<l;i++){
	       crt=(Property)currProperties.elementAt(i);
	       if(crt.nam[0].id.equals("org.omg.ft.MembershipStyle")){
		   memb_s=crt.val.extract_long();
		   if(memb_s!=MEMB_APP_CTRL.value)
		       throw new ObjectGroupNotFound();
	       }
	       if(crt.nam[0].id.equals("org.omg.ft.Factories")){
		   theFact_l=FactoryInfosHelper.extract(crt.val);
		   break;
	       }
	   }
	   if(theFact_l==null) return group;
	   if(memb_s!=MEMB_APP_CTRL.value)
	       throw new ObjectGroupNotFound();

	   for(i=0;i<theFact_l.length;i++)
	       if(theFact_l[i].the_location[0].id.equals(loc[0].id)){
		   try{
		       obj=theFact_l[i].the_factory.create_object(type_id,crit,fci); }catch(NoFactory e) {System.out.println("No factory");return null;}
		   catch(InvalidProperty e){System.out.println("Invalid Property");return null;}
		   catch(InvalidCriteria e){System.out.println("Invalid Criteria");return null;}
		   catch(ObjectNotCreated e){System.out.println("Object not created"); return null;}
		   catch(CannotMeetCriteria e){System.out.println("Cannot meet criteria"); return null;}
		   break;
	       }
	   try{
	       new_group=add_member(group,loc,obj);
	   }catch(ObjectNotAdded e){System.out.println("Object Not added");return null;}
	   return new_group;
    }

/** ObjectGroupManager interace method - adds a member (from a certain location) to an already existing group
 * Methods in ObjectGroupManager are used when the group membership is application controlled, and not infrastructure controlled
 * @param	object_group the CORBA object identifying the replica group
 *			the_location the location where the new member of the group will run
 *			member the CORBA object designating the new member to be added to the group
 * @return  the CORBA object designatng the new extended replica group
 */

    public org.omg.CORBA.Object add_member(org.omg.CORBA.Object object_group, org.omg.CosNaming.NameComponent[] the_location, org.omg.CORBA.Object member)
	throws ObjectGroupNotFound, MemberAlreadyPresent, ObjectNotAdded{

	IOR ior,new_ior,ior_o;
	int i,j;
	long g_i_r;
	ObjGroupInfoStr ogi;
	Vector locs;
	org.omg.IIOP.ProfileBody_1_1 pb11;
	org.omg.CORBA.Object toRet;

	ogi=(ObjGroupInfoStr)groupEvid.get(object_group);
	if(ogi==null) throw new ObjectGroupNotFound();

	ior = Utils.get_ior(object_group);
	new_ior=Utils.get_ior(member);

	ior_o=Utils.update_group_ior_add(orb,ior,new_ior,ior.profiles.length-1,true);
	/* the member is just added; we do not care if there is a primary or not*/

	FactoryInfo fi=new FactoryInfo(null,the_location,null);

	/*the add_member operation of the GroupManager is called by the application
	  with the mentioned parameters. Thus, we do not know the factory which created this
	  object, the group reference manager which has to know the latest reference to this
	  group, or the factory creation id*/

	ogi.factlocEvid.addElement(new ObjFactAndIdStr(member,null,false /*Actually this is not known here*/,false,fi,null));
	groupEvid.remove(object_group);
	toRet=new org.openorb.CORBA.ObjectStub(orb,ior_o);
	groupEvid.put(toRet,ogi);

	return toRet;
    }

/** ObjectGroupManager interace method - removes a member (from a certain location) from an already existing group
 * Methods in ObjectGroupManager are used when the group membership is application controlled, and not infrastructure controlled
 * @param	object_group the CORBA object identifying the replica group
 *			the_location the location where the member to remove runs
 * @return  the CORBA object designatng the new reduced replica group
 */

    public org.omg.CORBA.Object remove_member(org.omg.CORBA.Object object_group, org.omg.CosNaming.NameComponent[] the_location)
	throws ObjectGroupNotFound, MemberNotFound{

	int i,l;
	long group_id;
	ObjFactAndIdStr crt=null;
	ObjGroupInfoStr crt_g;
	boolean found=false;
	org.omg.CORBA.Object to_rem=null,toRet;
	IOR ior,ior_m,new_ior;
	org.omg.IIOP.ProfileBody_1_1 pb111,pb112;
	TaggedProfile[] new_profiles;


	crt_g=(ObjGroupInfoStr)groupEvid.remove(object_group);
	if(crt_g==null) throw new ObjectGroupNotFound();
	l=crt_g.factlocEvid.size();
	for(i=0;i<l;i++){
	    crt=(ObjFactAndIdStr)crt_g.factlocEvid.elementAt(i);
	    if(crt.cr_f_info.the_location[0].id.equals(the_location[0].id)){
		found=true;
		crt_g.factlocEvid.remove(crt);
		to_rem=crt.obj;
		break;
	    }
	}
	if(!found) throw new MemberNotFound();

	ior = Utils.get_ior(object_group);
	ior_m=Utils.get_ior(to_rem);

	new_ior=Utils.update_group_ior_rem(orb,ior,ior_m,true);
	toRet=new org.openorb.CORBA.ObjectStub(orb,new_ior);

	groupEvid.put(toRet,crt_g);
	return toRet;
    }

/** ObjectGroupManager interace method - sets a member (from a certain location) as the primary of the group
 * Methods in ObjectGroupManager are used when the group membership is application controlled, and not infrastructure controlled
 * @param	object_group the CORBA object identifying the replica group
 *			the_location the location where the member that has to be set to primary runs
 * @return  the CORBA object designatng the new group
 */

    public org.omg.CORBA.Object set_primary_member(org.omg.CORBA.Object object_group, org.omg.CosNaming.NameComponent[] the_location) throws ObjectGroupNotFound, MemberNotFound, PrimaryNotSet, BadReplicationStyle{

	org.omg.CORBA.Object m_ref;
	IOR g_ior,m_ior,n_ior;
	int gr_ref_v;
	org.omg.CORBA.Object toRet;
	ObjFactAndIdStr value;

	if(!Utils.has_primary(orb,object_group)) throw new BadReplicationStyle();
	try{
	    m_ref=get_member_ref(object_group,the_location);
	}catch(MemberNotFound e){throw new MemberNotFound();}
	catch(ObjectGroupNotFound e){throw new ObjectGroupNotFound();}
	value=(ObjFactAndIdStr)groupEvid.get(object_group);
	m_ior = Utils.get_ior(m_ref);
	g_ior = Utils.get_ior(object_group);

	n_ior=Utils.change_primary(orb,g_ior,m_ior,true);
	toRet=new org.openorb.CORBA.ObjectStub(orb,n_ior);
	groupEvid.remove(object_group);
	groupEvid.put(toRet,value);
	return toRet;
    }

/** ObjectGroupManager interace method - returns the object group identifier of a group
 * @param	group the CORBA object identifying the replica group
 * @return  the group identifier (a number)
 */

    public long get_object_group_id(org.omg.CORBA.Object group) throws ObjectGroupNotFound{
       int i,l;
       ObjGroupInfoStr crt;
       crt=(ObjGroupInfoStr)groupEvid.get(group);
       if(crt==null) throw new ObjectGroupNotFound();
       return crt.group_id.extract_longlong();
    }

/** ObjectGroupManager interace method - returns the object group identifier of a group
 * @param	group_id the identifier of the group
 * @return  the group reference CORBA object
 */

    public org.omg.CORBA.Object get_group_ref(long group_id){
	int i,l;
	Set entries;
	Iterator setIter;
	Map.Entry entry;
	ObjGroupInfoStr crt;
	synchronized(recover_object_semaphor){
	    entries=groupEvid.entrySet();
	    setIter=entries.iterator();
	    while(setIter.hasNext()){
		entry=(Map.Entry)setIter.next();
		crt=(ObjGroupInfoStr)entry.getValue();
		if(crt.group_id.extract_longlong()==group_id)
		    return (org.omg.CORBA.Object)entry.getKey();
	    }
	return null;
	}
    }

/** ObjectGroupManager interace method - returns the list of locations of the members of a group
 * @param	group the CORBA object designating the group
 * @return  the list of locations of the group members
 */

    public org.omg.CosNaming.NameComponent[][] locations_of_members(org.omg.CORBA.Object group) throws ObjectGroupNotFound{

	int i,l;
	ObjGroupInfoStr ogis;
	Vector factloc_vector;
	org.omg.CosNaming.NameComponent[][] toRet;

	ogis=(ObjGroupInfoStr)groupEvid.get(group);

	factloc_vector=ogis.factlocEvid;
	l=factloc_vector.size();
	if(l==0) return null;
	toRet=new org.omg.CosNaming.NameComponent[l][];
	for(i=0;i<l;i++){
	    toRet[i]=((ObjFactAndIdStr)factloc_vector.elementAt(i)).cr_f_info.the_location;
	}
	return toRet;
    }

/** ObjectGroupManager interace method - returns the CORBA object reference for a member of a given group
 * @param	group the CORBA object identifying the replica group
 *			loc the location of the member whose reference is requested
 * @return  the CORBA object reference of the member
 */
    public org.omg.CORBA.Object get_member_ref(org.omg.CORBA.Object group, org.omg.CosNaming.NameComponent[] loc) throws ObjectGroupNotFound, MemberNotFound{
	ObjGroupInfoStr ogis;
	int i,l;
	ObjFactAndIdStr crt;


	ogis=(ObjGroupInfoStr)groupEvid.get(group);
	if(ogis==null) throw new ObjectGroupNotFound();

	l=ogis.factlocEvid.size();
	for(i=0;i<l;i++){
	    crt=(ObjFactAndIdStr)ogis.factlocEvid.elementAt(i);
	    if(crt.cr_f_info.the_location[0].id.equals(loc[0].id)){
		return crt.obj;
	    }
	}
	throw new MemberNotFound();
    }

/** ObjectGroupManager interace method - returns the object group reference of a group based on a group reference that could be outdated
 * @param	group the CORBA object identifying the replica group (as known by the caller)
 * @return  the CORBA object reference identifying the group (could be the same as group)
 */
    public org.omg.CORBA.Object get_object_group_ref(org.omg.CORBA.Object group) throws ObjectGroupNotFound{

	IOR ior;
	long group_id;
	int i,l;
	ObjGroupInfoStr crt;
	Set elems;
	Map.Entry entry;
	Iterator elemsIt;
	synchronized(create_object_semaphor){
	    ior = Utils.get_ior(group);
	    group_id=Utils.extract_group_id(orb,ior);
	    if(group_id==-1) return null;

	    elems=groupEvid.entrySet();
	    elemsIt=elems.iterator();
	    while(elemsIt.hasNext()){
		entry=(Map.Entry)elemsIt.next();
		if(((ObjGroupInfoStr)(entry.getValue())).group_id.extract_longlong()==group_id)
		    return (org.omg.CORBA.Object)entry.getKey();
	    }
	    throw new ObjectGroupNotFound();
	}
    }

    /*END ObjectGroupManager METHODS*/


    /*BEGIN PropertyManager METHODS*/

/** Method in the PropertyManager interface - sets the property list for a certain type
* @param	type the type identifier of groups for which the properties are set
*			props list of properties (e.g. replication style, checkpoinitg interval, fault monitoring granularity, etc.)
*/
    public void set_type_properties(java.lang.String type, Property[] props) {
	Vector typeP;
	Property crt;
	int i,j,l;
	boolean found;

	typeP=(Vector)typeProperties.get(type);

	if(typeP==null)
	    typeP=new Vector();
	l=typeP.size();
	for(i=0;i<props.length;i++){
	    found=false;
	    for(j=0;j<l;j++){
		crt=(Property)typeP.elementAt(j);
		if(props[i].nam[0].id.equals(crt.nam[0].id)){
		    found=true;
		    crt.val=props[i].val;
		    break;
		}
	    }
	    if(!found) typeP.addElement(props[i]);
	}
	typeProperties.put(type,typeP);
    }

/** Method in the PropertyManager interface - removes a set of properties
* @param	type the type identifier of groups for which the properties are removed
*			props the set of properties that have to be removed
*/

    public void remove_type_properties(java.lang.String type, Property[] props){
	Vector typeP;
	int i,j,l;
	Property crt;

	typeP=(Vector)typeProperties.get(type);
	l=typeP.size();
	if(l==0) return;
	for(j=0;j<props.length;j++){
	    for(i=0;i<l;i++){
		crt=(Property)typeP.elementAt(i);
		if(crt.nam[0].id.equals(props[j].nam[0].id)){
		    typeP.remove(i);
		    break;
		}
	    }
	    l=typeP.size();
	}
    }

/** Method in the PropertyManager interface
* @param	the type of the groups for which properties are requested at return
* @return	the propeties for the given type identifier
*/

    public Property[] get_type_properties(java.lang.String type){
	Property[] toRet;
	int l,i;
	Vector typeP;
	typeP=(Vector)typeProperties.get(type);
	if(typeP==null) return null;
	l=typeP.size();
	toRet=new Property[l];
	for(i=0;i<l;i++)
	    toRet[i]=(Property)typeP.elementAt(i);
	return toRet;
	}

/** Method in the PropertyManager interface
* @param	props the properties to be set as default for all groups that are created
*/

    public void set_default_properties(Property[] props) {

	int i,j,l;
	Property crt;
	boolean found;

	l=defaultProperties.size();

	for(j=0;j<props.length;j++){
	    found=false;
	    for(i=0;i<l;i++){
		crt=(Property)defaultProperties.elementAt(i);
		if(crt.nam[0].id.equals(props[j].nam[0].id)){
		    crt.val=props[j].val;
		    found=true;
		    break;
		}
	    }
	    if(!found) defaultProperties.addElement(props[j]);
	}
    }

/** Method in the PropertyManager interface
* @param	props the set of default properties that have to be removed
*/

    public void remove_default_properties(Property[] props){
	int i,j,l;
	Property crt;
	l=defaultProperties.size();
	if(l==0) return;
	for(j=0;j<props.length;j++){
	    for(i=0;i<l;i++){
		crt=(Property)defaultProperties.elementAt(i);
		if(crt.nam[0].id.equals(props[j].nam[0].id)){
		    defaultProperties.remove(i);
		    break;
		}
	    }
	    l=defaultProperties.size();
	}
    }

/** Method in the PropertyManager interface
* @return	the set of default properties
*/

    public Property[] get_default_properties(){
	    Property[] toRet;
	    int l,i;
	    l=defaultProperties.size();
	    toRet=new Property[l];
	    for(i=0;i<l;i++)
		toRet[i]=(Property)defaultProperties.elementAt(i);
	    return toRet;
	}

/** Method in the PropertyManager interface
* @param	group the CORBA object reference of the group for which
*			props the properties to be set for the specific group
*/

    public void set_properties_dynamically(org.omg.CORBA.Object group, Property[] props){
	Vector p_group;
	int l,j,i;
	boolean found;
	Property crt;

	p_group=(Vector) groupProperties.get(group);
	if(p_group==null)
	    p_group=new Vector();
	l=p_group.size();
	for(i=0;i<props.length;i++){
	    found=false;
	    for(j=0;j<l;j++){
		crt=(Property)p_group.elementAt(j);
		if(props[i].nam[0].id.equals(crt.nam[0].id)){
		    found=true;
		    crt.val=props[i].val;
		    break;
		}
	    }
	    if(!found) p_group.addElement(props[i]);
	}
	groupProperties.put(group,p_group);
    }

/** Method in the PropertyManager interface
* @param	group the group CORBA object reference for which the properties are requested
* @return	the properties for the specific group
*/

    public Property[] get_properties(org.omg.CORBA.Object group){
	Property[] toRet;
	int l,i;
	Vector p_group;
	p_group=(Vector)groupProperties.get(group);
	if(p_group==null) return null;
	l=p_group.size();
	if(l==0) return null;
	toRet=new Property[l];
	for(i=0;i<l;i++)
	    toRet[i]=(Property)p_group.elementAt(i);
	return toRet;
    }

/** Method in the PropertyManager interface
* @param	gr the group for which the properties will be removed
*/

    private void remove_dyn_props(org.omg.CORBA.Object gr){
	groupProperties.remove(gr);
    }

    /*END PropertyManager METHODS*/

    /*BEGIN ReplicationManager METHODS*/

/** Method specified in the ReplicationManager interface - used to register the fault notifier
*/
    public void register_fault_notifier(FaultNotifier fn){fault_notifier=fn;}

    /** Method specified in the ReplicationManager interface - used to return the registered fault notifier
	*/

    public  FaultNotifier get_fault_notifier(){return fault_notifier;}

    /*END ReplicationManager METHODS*/
}
