/**
  * Redistribution and use of this software and associated
  * documentation ("Software"), with or without modification, are
  * permitted provided that the following conditions are met:
  *
  * 1. Redistributions of source code must retain copyright statements
  *    and notices.  Redistributions must also contain a copy of this
  *    document.
  *
  * 2. Redistributions in binary form must reproduce the above
  *    copyright notice, this list of conditions and the following
  *    disclaimer in the documentation and/or other materials provided
  *    with the distribution.
  *
  * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
  *    products derived from this Software without prior written
  *    permission of one of the parties involved in the TRANSORG project. For
  * 	  written permission, please contact diasz@ida.liu.se.
  *
  * 4. Products derived from this Software may not be called "TRANSORG"
  *    nor may "TRANSORG" appear in their names without prior written
  *    permission.
  *
  * 5. Due credit should be given to the TRANSORG Project
  *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
  *
  * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  * PARTICULAR PURPOSE ARE DISCLAIMED.
  *
  * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
  * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
  *
  * $Id: Utils.java,v 1.0 2003/08/04 08:07:19 diasz Exp $
  *
  * Date         Author  Changes
  */

package org.omg.FT;
import org.omg.CORBA.*;
import org.omg.CORBA.portable.*;
import org.omg.IOP.*;
import java.util.*;

/**
 * This class mainly contains static methods and fields (constants) used as utilities (for example it contains methods to do conversions from byte arrays to long
 * and back, methods to read IOR profiles, etc.). Other utilities can be written if needed. Also, the author can be contacted to provide some methods that are
 * not in this file, but might needed for the well functioning of the infrastructure.
 * @author	<a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
 */

public class Utils{

    private static java.lang.Object semaphor;

    /** path of the file containing the host names from the fault tolerance domain
    *SHOULD BE CHANGED TO FIT NEEDS!! OR EVEN MAKE IT NOT HARD CODED!!
    */

    private static String ft_dom_name_fn="/home/diasz/Research/Experiments/trials/FT_DOMAINS";

/**
 * update-only method (return type is void)
 */
    public final static char UPDATE_VOID='v';
/**
 * update method - return type is not void
 */
    public final static char UPDATE_NON_VOID='u';
 /**
  *oneway update method
 */
    public final static char UPDATE_ONEWAY='o';
  /**
  * read-only method
  */
    public final static char READ='r';

/* NOTE: TIME TRACING WAS REMOVED FROM MOST OF THE INTERCEPTOR FILES; A SAMPLE TRACING IS GIVEN IN THE COLD PASSIVE SERVER INTERCEPTOR,
* TO ILLUSTRATE HOW THE TRACE METHODS ARE USED; THE METHOD trace IS DESCRIBED LATER IN THIS FILE IN TERMS OF PARAMETERS
* Also, time tracing calls can be placed anywhere it is suitable;still, the most advisable place to use them, is in the server and client
* interceptors.
*/


/**
* constant (used in time tracing) indicating the moment of start of the CORBA call in the client appl.
*/

    public static final int START_CALL_BY_CLIENT=0;
 /**
 * constant (used in time tracing) indicating the moment when the client interceptor is exited when sending the request
 */
    public static final int EXIT_SEND_REQUEST=1;

    /**
	 * constant (used in time tracing) indicating the moment when the server interceptor is entered when receiving the request
	 * at the server
    */
    public static final int ENTER_RECEIVE_REQUEST=2;

 /**
   * constant (used in time tracing) indicating the moment when in the server interceptor (at receiving of the request) it is found out that the
   * request is at the right replica (primary) - this constant is used in passive replication
 */
    public static final int YES_REQUEST_RIGHT_PLACE=3;
/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at receiving of the request) it is found out that the client
* knows a different group reference version than the server ORB; thus, a ForwardRequest exception is sent to the client containing the new version of the
* group reference
*/
    public static final int SEND_FORWARD_EXC=4;
/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at receiving of the request) it is found out that the request
* arrived to the wrong replica (not yet the primary); thus, a TRANSIENT exception is sent to the client
*/
    public static final int SEND_WR_PRIM_EXC=5;
/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at receiving of the request) it is found out that the request
* expiration time was exceeded
*/
    public static final int SEND_REQ_EXP_EXC=6;


/**
* constant (used in time tracing) indicating the moment when the request receiving method in the server interceptor is exited
*/

    public static final int EXIT_RECEIVE_REQUEST=9;

/**
* constant (used in time tracing) indicating the moment when the server interceptor is entered when the reply is sent to the client after the request
* was processed in the server
*/
    public static final int ENTER_SEND_REPLY=10;

/**
* constant (used in time tracing) indicating the moment when the server interceptor is exited when sending the reply to the client
*/
    public static final int EXIT_SEND_REPLY=11;

/**
* constant (used in time tracing) indicating the moment when the client interceptor is entered when the reply is received by the client
*/
    public static final int ENTER_RECEIVE_REPLY=12;

/**
* constant (used in time tracing) indicating the moment when the client interceptor is exited when the reply is received by the client
*/

    public static final int EXIT_RECEIVE_REPLY=13;

/**
* constant (used in time tracing) indicating that the client received the reply at the level of the application
*/

    public static final int END_CALL_BY_CLIENT=14;

/**
* constant (used in time tracing) indicating the moment when the gateway's server interceptor is entered when a request is received from the
* client to be broadcast to all replicas of an actively replicated server
*/

    public static final int ENTER_RECEIVE_REQUEST_G=15;

/**
* constant (used in time tracing) indicating the moment when in the gateway server interceptor, reading of the current membership of the
* active group starts to wait for a possible failover procedure that takes place at the same time and leads to a change in the group
*/

    public static final int BEG_SYNCH_FAILOVER=16;

/**
* constant (used in time tracing) indicating the moment when in the gateway server interceptor the group membership is read after waiting for a
* possible failover to take place
*/

    public static final int END_SYNCH_FAILOVER=17;

/**
* constant (used in time tracing) indicating the moment when the gateway server interceptor is exited when sending a reply to the client
* request
*/

    public static final int EXIT_SEND_REPLY_GW=18;

/**
* constant (used in time tracing) indicating the moment when in the gateway sends a TRANSIENT exception to the client calling a method on the
* actively replicated server
*/

    public static final int GW_SEND_TRANSIENT=20;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor at the gateway, the client request was sent to the replicas of the active
* group
*/

    public static final int GW_SEND_REQUESTS=21;


/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at the gateway) the first reply (from the fastest replica)
* is received
*/
    public static final int RECEIVE_FIRST_REPLY=22;

/**
* constant (used in time tracing) indicating the moment when the client interceptor is entered when the request is sent to the server
*/

    public static final int ENTER_SEND_REQUEST=23;

/**
* constant (used in time tracing) indicating the moment when in the client interceptor (at the active group member) the "invoke" method is called on the gateway
* when the active server group becomes a client for a third tier server
*/

    public static final int CALL_INVOKE=24;

/**
* constant (used in time tracing) indicating the moment when in the client interceptor (at the active group member) the method send_request is exited, when the
* invoke method is called
*/

    public static final int INV_EXITS_SEND_REQUEST=25;

/**
* constant (used in time tracing) indicating the moment when the gateway starts waiting for the answer to a request that it sent out on behalf of the
*actively replicated server group, to a third-tier server
*/

    public static final int START_WAIT_ANSWER_GW=26;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at the gateway) the answer to a request sent out on behalf of the
* actively replicated group is received
*/

    public static final int GET_ANSWER_GW=27;

/**
* constant (used in time tracing) indicating the moment when an exception is thrown in the client interceptor, saying that the answer to an outgoing
* request from the replicated server group is at the server
*/

    public static final int THROW_REPLY_HERE=28;

/**
* constant (used in time tracing) indicating the moment when the replication manager starts the recovery (failover) process of a group in which a replica failed
* (this time trace is put in the Replication Manager)
*/

    public static final int REPL_MAN_BEGIN_REC=29;

/**
* constant (used in time tracing) indicating the moment when the Replication Manager ends the recovery (failover) process for a group
*/
    public static final int REPL_MAN_END_REC=30;

/**
* constant (used in time tracing) indicating the moment when the replication manager received a fault notification from the failure notifier
*/

    public static final int REPL_MAN_REC_F_NOT=31;

/**
* constant (used in time tracing) indicating the moment when the Replication Manager starts creating a new replica when the number of members of the
* group that are up drops under the minimum number replicas limit
*/

    public static final int REPL_MAN_BEGIN_CR_OBJ=32;

/**
* constant (used in time tracing) indicating the moment when the Replication Manager finishes the creation of the new replica
*/

    public static final int REPL_MAN_END_CR_OBJ=33;

/**
* constant (used in time tracing) indicating the moment when the server interceptor (at receiving of the request) is exited because the reply to the
* request is already logged; thus, the request is not executed again and the answer is directly sent to the client
*/

    public static final int EXIT_REPLY_ALREADY_THERE=34;

/**
* constant (used in time tracing) indicating the moment when a client interceptor's receive_exception method is called showing that the server sent
* an exception as a result of the request execution
*/

    public static final int RECEIVE_EXCEPTION=35;

/**
* constant (used in time tracing) indicating the moment when in a client interceptor's receive_other method is called showing that the server sent
* something similar to an exception back to the client
*/

    public static final int RECEIVE_OTHER=36;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at the gateway) it is found out that the request
* expiration time was exceeded and a corresponding exception is sent to the client
*/

    public static final int SEND_REQ_EXP_EXC_GW=37;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at returning after a request) an exception is sent
* to the client (the method send_exception in the server interceptor is called)
*/

    public static final int SEND_EXCEPTION=38;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at returning after a request) something similar to an exception is sent
* to the client (the method send_other in the server interceptor is called)
*/

    public static final int SEND_OTHER=39;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor at the Replication Manager (at returning after a request) an exception is
* sent to the client (the method send_exception in the server interceptor is called)
*/
    public static final int SEND_EXCEPTION_RM=40;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor at the Replication Manager (at returning after a request) something similar
* to an exception is sent to the client (the method send_exception in the server interceptor is called)
*/

    public static final int SEND_OTHER_RM=41;


/**
* constant (used in time tracing) indicating the moment the call to get_update (reading of state change) on a server replica starts (this trace is usually
*placed in the object that does checkpointing)
*/
    public static final int START_GET_UPDATE=44;


/**
* constant (used in time tracing) indicating the moment the call to get_update on a server replica ends (this trace is usually placed in the object that
* does checkpointing)
*/
    public static final int STOP_GET_UPDATE=45;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at receiving of the request) for warm passive replication, the call
* information starts to be recorded by use of the Logging Recovery Mechanism (BR is short for broadcasting)
*/

    public static final int START_REC_BR_METHOD_CALL=46;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at receiving of the request) for warm passive replication, the call
* information recording stops
*/

    public static final int STOP_REC_BR_METHOD_CALL=47;


/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at replying to the request) for warm passive replication, the reply
* information starts to be recorded by use of the Logging Recovery Mechanism (BR is short for broadcasting)
*/

    public static final int START_REC_BR_REPLY=48;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at replying to the request) for warm passive replication, the reply
* information recording stops
*/

    public static final int STOP_REC_BR_REPLY=49;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at receiving of the request) for cold passive replication, the call
* information starts to be recorded by use of the Logging Recovery Mechanism
*/

    public static final int START_REC_METHOD_CALL=50;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at receiving of the request) for cold passive replication, the call
* information recording stops
*/

    public static final int STOP_REC_METHOD_CALL=51;


/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at replying to the request) for cold passive replication, the reply
* information starts to be recorded by use of the Logging Recovery Mechanism
*/
    public static final int START_REC_REPLY=52;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor (at replying to the request) for cold passive replication, the reply
* information recording stops
*/

    public static final int STOP_REC_REPLY=53;

/**
* constant (used in time tracing) indicating the moment when, while in the fail over process of a group, the Replication Manager reads the state of a replica
* (in case of active replication) to be transferred to a newly created replica. Note that this is the case only when the number of replicas in the group drops
* under the minimum number of replicas requested for the group.
*/

    public static final int REPL_MAN_BEG_GET_STATE=54;

/**
* constant (used in time tracing) indicating the moment when, while in the failover process of a group, the Replication Manager ends the state reading
*/

    public static final int REPL_MAN_END_GET_STATE=55;

/**
* constant (used in time tracing) indicating the moment when the Replication Manager starts setting the state on a new replica of an active group
*/

    public static final int REPL_MAN_BEG_SET_STATE=56;

/**
* constant (used in time tracing) indicating the moment when the Replication Manager ends setting the state on a new replica of an active group
*/

    public static final int REPL_MAN_END_SET_STATE=57;

/**
* constant (used in time tracing) indicating the moment when an update operation (when logged) waits for other update operations arrived before it to
* finish execution
*/

    public static final int START_WAIT_UPD_OPER=58;

/**
* constant (used in time tracing) indicating the moment when an update operation is ready to be executed on the server after all update methods arrived
* before it executed
*/

    public static final int STOP_WAIT_UPD_OPER=59;


/**
* constant (used in time tracing) indicating the moment when the call to get_state is initiated, but has to wait until all update method calls
* processing it will finish
*/

    public static final int START_GET_S_WAIT=60;


/**
* constant (used in time tracing) indicating the moment when the call to get_state can start, since all update methods executing in front of it
* finished
*/

    public static final int STOP_GET_S_WAIT=61;

/**
* constant (used in time tracing) indicating the moment when the call to get_update is initiated, but has to wait until all update method calls
* processing it will finish
*/

    public static final int START_GET_U_WAIT=62;

/**
* constant (used in time tracing) indicating the moment when the call to get_update can start, since all update methods executing in front of it
* finished
*/

    public static final int STOP_GET_U_WAIT=63;

/**
* constant (used in time tracing) indicating the moment the call to get_state on a server replica starts (this trace is usually placed in the object that
* does checkpointing)
*/

    public static final int START_GET_STATE_CALL=64;

/**
* constant (used in time tracing) indicating the moment the call to get_state on a server replica ends (this trace is usually placed in the object that
* does checkpointing)
*/

    public static final int STOP_GET_STATE_CALL=65;

/**
* constant (used in time tracing) indicating the moment when the recording of the state in the log starts (this trace should be placed in the object that
* does the checkpointing)
*/

    public static final int START_STATE_REC=66;

/**
* constant (used in time tracing) indicating the moment when the recording of the state in the log ends (this trace should be placed in the object that
* does the checkpointing)
*/

    public static final int STOP_STATE_REC=67;

/**
* constant (used in time tracing) indicating the moment when in the client interceptor (at the logging_recovery CORBA object) a get_state request is intercepted when
* it is sent to the server object to be checkpointed
*/

    public static final int SEND_GET_STATE_REQ=68;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor at the server replica a get_state request is received
*/

    public static final int REC_GET_STATE_REQ=69;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor at the server replica a reply to a get_state request is sent to the
* checkpointing object
*/


    public static final int SEND_GET_STATE_REPLY=70;

/**
* constant (used in time tracing) indicating the moment when in the client interceptor (at the logging_recovery CORBA object) a reply to get_state request is
* received from the server object that is checkpointed
*/

    public static final int REC_GET_STATE_REPLY=71;

/**
* constant (used in time tracing) indicating the moment the call to get_update on a server replica starts (this trace is usually placed in the object that
* does checkpointing)
*/
    public static final int START_GET_UPDATE_CALL=72;

/**
* constant (used in time tracing) indicating the moment the call to get_update on a server replica ends (this trace is usually placed in the object that
* does checkpointing)
*/
    public static final int STOP_GET_UPDATE_CALL=73;

/**
* constant (used in time tracing) indicating the moment when the recording of the state change (update) in the log starts (this trace should be placed in the object that
* does the checkpointing)
*/

    public static final int START_UPDATE_REC=74;

/**
* constant (used in time tracing) indicating the moment when the recording of the update in the log ends (this trace should be placed in the object that
* does the checkpointing)
*/
    public static final int STOP_UPDATE_REC=75;

/**
* constant (used in time tracing) indicating the moment when in the client interceptor (at the logging_recovery CORBA object) a get_update request is intercepted when
* it is sent to the server object to be checkpointed
*/

    public static final int SEND_GET_UPDATE_REQ=76;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor at the server replica a get_update request is received
*/

    public static final int REC_GET_UPDATE_REQ=77;

/**
* constant (used in time tracing) indicating the moment when in the server interceptor at the server replica a reply to a get_update request is sent to the
* checkpointing object
*/

    public static final int SEND_GET_UPDATE_REPLY=78;

/**
* constant (used in time tracing) indicating the moment when in the client interceptor (at the logging_recovery CORBA object) a reply to get_update request is
* received from the server object that is checkpointed
*/

    public static final int REC_GET_UPDATE_REPLY=79;

/** constant (used in time tracing) indicating when the setting of the state (by call to set_state) on a recovering replica (in cold passive replication)  is
* started
*/

    public static final int START_SET_STATE=80;

/** constant (used in time tracing) indicating when the setting of the state (by call to set_state) on a recovering replica (in cold passive replication)  is
* ended (the interval STOP_SET_STATE-START_SET_STATE is important to consider as a part of the failover time)
*/
    public static final int STOP_SET_STATE=81;


    public static java.lang.Object BACKUP_SYNCH=new java.lang.Object();

/**
*Class containing info about the fault tolerance domain - name and hosts
*/
    public static class FTDomInfo{
	public String ft_dom_name,hosts[];
	public FTDomInfo(String ft_dom_name,String[] hosts){
	    this.ft_dom_name=ft_dom_name;
	    this.hosts=hosts;
	}

    }


    /* BEGIN general utilities*/

    public static int max(int i,int j){
	if(i>j) return i;
	return j;
    }

    public static int min(int i,int j){
	if(i<j) return i;
	return j;
    }

    public static double max(double arr[]){
	double mx=arr[0];
	for(int i=1;i<arr.length;i++)
	    if(arr[i]>mx) mx=arr[i];
	return mx;
    }

    public static double min(double arr[]){
	double mn=arr[0];
	for(int i=1;i<arr.length;i++)
	    if(arr[i]<mn) mn=arr[i];
	return mn;
    }

    public static void set_true(boolean array[]){
	int i;
	for(i=0;i<array.length;i++)
	    array[i]=true;
    }


    public static void set_false(boolean array[],int where){
	array[where]=false;
    }

    public static boolean is_true(boolean array[],int how_many){
	int i,nr;
	nr=0;
	for(i=0;i<array.length;i++)
	    if(array[i]) {
		nr++;
		if(nr==how_many)
		    return true;
	    }
	return false;
    }

/**
* @param	id an identifier in a long format
* @return	a byte array (8 octets) representation of the long number (for example: id=890=3*256+122, the byte aray returned is:{122,3,0,0,0,0,0,0})
*/
    public static byte[] makeIdByte(long id)
    {
	byte[] toRet=new byte[8];
	int i;

	for(i=0;i<8;i++){
	    id=id >> (i*8);
	    toRet[i]=(byte)(id&0x00000000000000FF);
	}
	return toRet;
    }

    public static String[] makeByteArgs(byte[] id){
	String[] args=new String[id.length];
	for(int i=0;i<id.length;i++)
	    args[i]=Byte.toString(id[i]);
	return args;
    }
/**
* @param	id an identifier in a byte (octet) array format
* @return	a long representation of the array
*/

    public static long makeByteId(byte[] id)
    {
	long toRet;
	int i;

	if(id.length!=8) return -1;

	toRet=0;

	for(i=7;i>=0;i--){
	    toRet=toRet<< 8;
	    toRet+=id[i];
	    if(id[i]<0) toRet+=256;
	}
	return toRet;
    }

/** This method modifies the state array of bytes by inserting in the beginning of the array four octets that represent the integer retention identifier
* @param	state the state array to be extended
*			ret_id the retention identifier
* @return	the new extended byte array
*/

    public static byte[] add_ret_id_state(byte[] state,int ret_id){
	byte[] fake_state=new byte[state.length+4];
	byte[] ret_id_b=makeIntByte(ret_id);
	int i;
	for(i=0;i<4;i++)
	    fake_state[i]=ret_id_b[i];
	for(;i<4+state.length;i++)
	    fake_state[i]=state[i-4];
	return fake_state;
    }

/** This method extracts the retention identifier from a fake state array (the first four octets)
* @param	fake_state the array having in the first four positions the integer retention identifier's octets
* @return	the integer value of the retention identifier
*/

    public static int readRetId(byte[] fake_state){
	byte[] first_four=new byte[4];
	for(int i=0;i<4;i++)
	    first_four[i]=fake_state[i];
	return makeByteInt(first_four);
    }

/** This method extracts the real state a fake state array (the octets after position four)
* @param	fake_state the array having in the first four positions the integer retention identifier's octets
* @return	the byte array representing the state
*/


    public static byte[] readRealState(byte[] fake_state){
	if(fake_state.length<4) return new byte[0];

	byte[] real_state=new byte[fake_state.length-4];
	for(int i=4;i<fake_state.length;i++)
	    real_state[i-4]=fake_state[i];
	return real_state;
    }

/**
* @param	id an identifier in an int format
* @return	a byte array (4 octets) representation of the int number (for example: id=890=3*256+122, the byte aray returned is:{122,3,0,0})
*/
    public static byte[] makeIntByte(int nr){
	byte[] toRet=new byte[4];
	int i;

	for(i=0;i<4;i++){
	    nr=nr >> (i*8);
	    toRet[i]=(byte)(nr&0x000000FF);
	}
	return toRet;
    }
/**
* @param	id an identifier in a byte (octet) array format
* @return	an int representation of the array
*/
    public static int makeByteInt(byte[] b)
    {
	int toRet;
	int i;

	if(b.length!=4) return -1;

	toRet=0;

	for(i=3;i>=0;i--){
	    toRet=toRet<< 8;
	    toRet+=b[i];
	    if(b[i]<0)
		toRet+=256;
	}
	return toRet;
    }

/**
* @param	id an identifier in a short format
* @return	a byte array (2 octets) representation of the int number (for example: id=890=3*256+122, the byte aray returned is:{122,3})
*/
    public static byte[] makeShortByte(short v){
	byte[] toRet=new byte[2];

	toRet[0]=(byte)(v&0x00FF);
	toRet[1]=(byte)((v>>8)&0x00FF);
	return toRet;
    }

/**
* @param	id a String representing a client identifier (e.g. 0*ORB123214)
* @return	a byte array representation of the object group identifier (e.g. {0,0,0,0,0,0,0,0}) as extracted from the client id
*/
    public static byte[] extract_oid(String id){
	byte[] toRet=new byte[0];
	int i;

	i=id.indexOf('*');
	if(i==-1) return new byte[0];
	try{
	    toRet=makeIdByte(Long.parseLong(id.substring(0,i)));
	}catch(NumberFormatException e){toRet=new byte[0];}
	return toRet;
    }

/**
* @param	id a String representing a client identifier (e.g. 0*ORB123214)
* @return	a long representation of the object group identifier (e.g. 0) as extracted from the client id
*/

    public static long extract_oid_l(String id){
	int i;
	long toRet=-1;
	i=id.indexOf('*');
	if(i==-1) return -1;
	try{
	    toRet=Long.parseLong(id.substring(0,i));
	}catch(NumberFormatException e){toRet=-1;}
	return toRet;
    }

/** Tests whether two byte arrays are equal
* e.g. {1,2,2,4} is equal to {1,2,2,4}, but is not equal with {1,3,4,2} or {1,2,2}
*/

    public static boolean byte_a_equal(byte[] a1,byte[] a2){
	if(a1==null||a2==null) return false;
	if(a1.length!=a2.length) return false;
	int i;
	for(i=0;i<a1.length;i++)
	    if(a1[i]!=a2[i]) return false;
	return true;
    }
/**
* @param	repl_s an int representation of the replicaton style (constants definde in the FTC IDL file)
* @return	a String representation of the replation style
*/

    public static String repl_s_string(int repl_s){
	switch(repl_s){
	case STATELESS.value:return "STL";
	case COLD_PASSIVE.value:return "C_P";
	case WARM_PASSIVE.value:return "W_P";
	case ACTIVE.value:return "ACT";
	case ACTIVE_WITH_VOTING.value:return "ACT_W_V";
	}
	return "";
    }


    /**
    * @param	any an any object with a certain type
    * @return	a String representation of the type name of the object inside the any
    */

    public static String give_class_name(org.omg.CORBA.Any any){
	try{
	switch(any.type().kind().value()){

	case org.omg.CORBA.TCKind._tk_abstract_interface:return "java.lang.Object";

	case org.omg.CORBA.TCKind._tk_alias:return null;

	case     org.omg.CORBA.TCKind._tk_any:return "org.omg.CORBA.Any";

	case     org.omg.CORBA.TCKind._tk_array:return null;

	case     org.omg.CORBA.TCKind._tk_boolean:return "java.lang.Boolean";

	case     org.omg.CORBA.TCKind._tk_char:return "java.lang.Character";

	case     org.omg.CORBA.TCKind._tk_double:return "java.lang.Double";
	case      org.omg.CORBA.TCKind._tk_struct :
	case     org.omg.CORBA.TCKind._tk_enum:
	case    org.omg.CORBA.TCKind._tk_native:
	case org.omg.CORBA.TCKind._tk_value_box :
	case     org.omg.CORBA.TCKind._tk_union :return get_class_name(any.type().id());
	case     org.omg.CORBA.TCKind._tk_except:return null;
	case     org.omg.CORBA.TCKind._tk_fixed:return null;
	case     org.omg.CORBA.TCKind._tk_float:return "java.lang.Float";

	case     org.omg.CORBA.TCKind._tk_long:return "java.lang.Integer";

	case     org.omg.CORBA.TCKind._tk_longdouble:return null;

	case     org.omg.CORBA.TCKind._tk_longlong:return "java.lang.Long";
	case org.omg.CORBA.TCKind._tk_null:return null;

	case org.omg.CORBA.TCKind._tk_objref:return "org.omg.CORBA.Object";

	case     org.omg.CORBA.TCKind._tk_octet:return "java.lang.Byte";

	case     org.omg.CORBA.TCKind._tk_Principal :return null;

	case     org.omg.CORBA.TCKind._tk_sequence:return null;

	case     org.omg.CORBA.TCKind._tk_short :return "java.lang.Short";

	case     org.omg.CORBA.TCKind._tk_string:return "java.lang.String";

	case     org.omg.CORBA.TCKind._tk_TypeCode :return null;

	case     org.omg.CORBA.TCKind._tk_ulong :return "java.lang.Integer";

	case     org.omg.CORBA.TCKind._tk_ulonglong :return "java.lang.Long";

	case    org.omg.CORBA.TCKind._tk_ushort :return "java.lang.Short";

	case    org.omg.CORBA.TCKind._tk_value:return null;
	case org.omg.CORBA.TCKind._tk_void :return "java.lang.Void";
	case org.omg.CORBA.TCKind._tk_wchar:return null;
	case org.omg.CORBA.TCKind._tk_wstring:return null;
	}
	}catch(java.lang.Exception e){return null;}
	return null;

    }


/**
* @param	orb the ORB instance in which the application is running
*			result	result of a method invocation or an out/inout parameter writing when using the Java reflection (method.invoke returns an object)
*			holder	variable that indicates whether the parameter result is a simple parameter, or an inout or out variable that can be modified by the
*					execution of a method
* @return	an any representation of the object result
*/



    public static org.omg.CORBA.Any give_any(org.omg.CORBA.ORB orb,java.lang.Object result,boolean holder){
	org.omg.CORBA.Any any=orb.create_any();
	String type;
	java.lang.Object value;
	try{
	    if(result==null) {any.insert_TypeCode(orb.get_primitive_tc(org.omg.CORBA.TCKind.tk_void));return any;}
	    if(holder){
		value=result.getClass().getField("value").get(result);
		type=value.getClass().getName();
	    }
	    else{
		type=result.getClass().getName();
		value=result;
	    }
	    if(type.equals("java.lang.Integer")||type.equals("int"))
		any.insert_long(((java.lang.Integer)value).intValue());
	    if(type.equals("java.lang.Long")||type.equals("long"))
		any.insert_longlong(((java.lang.Long)value).longValue());
	    if(type.equals("java.lang.Boolean")||type.equals("boolean"))
		any.insert_boolean(((java.lang.Boolean)value).booleanValue());
	    if(type.equals("java.lang.String"))
		any.insert_string(((java.lang.String)value).toString());
	    if(type.equals("java.lang.Character")||type.equals("char"))
		any.insert_char(((java.lang.Character)value).charValue());
	    if(type.equals("java.lang.Short")||type.equals("short"))
		any.insert_short(((java.lang.Short)value).shortValue());
	    if(type.equals("java.lang.Float")||type.equals("float"))
		any.insert_float(((java.lang.Float)value).floatValue());
	    if(type.equals("java.lang.Double")||type.equals("double"))
		any.insert_double(((java.lang.Double)value).doubleValue());
	    if(type.equals("[B"))
		org.omg.CORBA.OctetSeqHelper.insert(any,(byte[])value);
	    if(type.equals("[I"))
		org.omg.CORBA.LongSeqHelper.insert(any,(int[])value);
	    if(type.equals("[J"))
		org.omg.CORBA.LongLongSeqHelper.insert(any,(long[])value);
	    if(type.equals("[Z"))
		org.omg.CORBA.BooleanSeqHelper.insert(any,(boolean[])value);
	    if(type.equals("[S"))
		org.omg.CORBA.ShortSeqHelper.insert(any,(short[])value);
	    if(type.equals("[C"))
		org.omg.CORBA.CharSeqHelper.insert(any,(char[])value);
	    if(type.equals("[F"))
		org.omg.CORBA.FloatSeqHelper.insert(any,(float[])value);
	    if(type.equals("[D"))
		org.omg.CORBA.DoubleSeqHelper.insert(any,(double[])value);
	    if(type.equals("[Ljava.lang.String;"))
		org.omg.CORBA.StringSeqHelper.insert(any,(String[])value);
	    if(type.equals("org.omg.CORBA.Any"))
		any.insert_any((org.omg.CORBA.Any)value);
	    if(type.equals("org.omg.CORBA.Object"))
		any.insert_Object((org.omg.CORBA.Object)value);
	}catch(java.lang.Exception e){e.printStackTrace();return null;}
	return any;
    }


/** This method is used when reexecuting methods that where logged when the primary was up, and their parameters where remembered in an Any format;
* at replay, the parameters are needed by the reflection operations in object formats
* @param	any an Any type object that has to be transformed in an object representation
* @return	an object that is contained in the Any structure
*/
      private static java.lang.Object give_object(org.omg.CORBA.Any any){

	switch(any.type().kind().value()){
	case org.omg.CORBA.TCKind._tk_abstract_interface:return null;
	case org.omg.CORBA.TCKind._tk_alias:
	    return extract_obj_from_any(any);
	case     org.omg.CORBA.TCKind._tk_any:return (java.lang.Object)(any.extract_any());
	case     org.omg.CORBA.TCKind._tk_boolean:return new java.lang.Boolean(any.extract_boolean());
	case     org.omg.CORBA.TCKind._tk_char:return new java.lang.Character(any.extract_char());
	case     org.omg.CORBA.TCKind._tk_double:return new java.lang.Double(any.extract_double());
	case     org.omg.CORBA.TCKind._tk_fixed:return null;
	case     org.omg.CORBA.TCKind._tk_float:return new java.lang.Float(any.extract_float());
	case     org.omg.CORBA.TCKind._tk_long:return new java.lang.Integer(any.extract_long());
	case     org.omg.CORBA.TCKind._tk_longdouble:return new java.lang.Double(any.extract_double());
	case     org.omg.CORBA.TCKind._tk_longlong:return new java.lang.Long(any.extract_longlong());
	case    org.omg.CORBA.TCKind._tk_native:return null;
	case org.omg.CORBA.TCKind._tk_null:return null;
	case org.omg.CORBA.TCKind._tk_objref:return null;
	case     org.omg.CORBA.TCKind._tk_octet:return new java.lang.Byte(any.extract_octet());
	case     org.omg.CORBA.TCKind._tk_Principal :return null;
	case     org.omg.CORBA.TCKind._tk_union :return extract_obj_from_any(any);
	case      org.omg.CORBA.TCKind._tk_struct :return extract_obj_from_any(any);
	case     org.omg.CORBA.TCKind._tk_sequence:return extract_obj_from_any(any);
	case     org.omg.CORBA.TCKind._tk_array:return extract_obj_from_any(any);
	case     org.omg.CORBA.TCKind._tk_except:return extract_obj_from_any(any);
	case     org.omg.CORBA.TCKind._tk_enum:return extract_obj_from_any(any);
	case     org.omg.CORBA.TCKind._tk_short :return new java.lang.Short(any.extract_short());
	case     org.omg.CORBA.TCKind._tk_string:return any.extract_string();
	case     org.omg.CORBA.TCKind._tk_TypeCode :return any.extract_TypeCode();
	case     org.omg.CORBA.TCKind._tk_ulong :return new java.lang.Integer(any.extract_ulong());
	case     org.omg.CORBA.TCKind._tk_ulonglong :return new java.lang.Long(any.extract_ulonglong());
	case    org.omg.CORBA.TCKind._tk_ushort :return new java.lang.Short(any.extract_ushort());
	case    org.omg.CORBA.TCKind._tk_value:return null;
	case org.omg.CORBA.TCKind._tk_value_box :return null;
	case org.omg.CORBA.TCKind._tk_void :return null;
	case org.omg.CORBA.TCKind._tk_wchar:return new java.lang.Character(any.extract_wchar());
	case org.omg.CORBA.TCKind._tk_wstring:return any.extract_wstring();
	}
	return null;
    }

/** Method used to return a holder object corresponding to an any object that is supposed to be modified by a method that receives it as out or inout
* parameter
* @param	any the Any out/inout parameter object that has to be provided as an Object type parameter to the reflectively called method
			typecode_only a boolean that indicates whether inside the any the object is only given only as a type (out param) or as a real object (inout param)
* @return	an object representation of the holder
*/
       private static java.lang.Object give_object_holder(org.omg.CORBA.Any any,boolean typecode_only){
	   int to_switch;
	   try{
	       if(typecode_only){
		   if(any.type().kind().value()!=org.omg.CORBA.TCKind._tk_TypeCode) {System.out.println("no read");return null;}
		   to_switch=any.extract_TypeCode().kind().value();
	       }
	       else
		   to_switch=any.type().kind().value();
	       switch(to_switch){

	       case org.omg.CORBA.TCKind._tk_abstract_interface:return null;

	       case org.omg.CORBA.TCKind._tk_alias:return null;

	       case     org.omg.CORBA.TCKind._tk_any:
		   if(typecode_only)
		       return new org.omg.CORBA.AnyHolder();
		   return new org.omg.CORBA.AnyHolder(any.extract_any());
	       case     org.omg.CORBA.TCKind._tk_boolean:
		   if(typecode_only)
		       return new org.omg.CORBA.BooleanHolder();
		   return new org.omg.CORBA.BooleanHolder(any.extract_boolean());

	       case     org.omg.CORBA.TCKind._tk_char:
		   if(typecode_only)
		       return new org.omg.CORBA.CharHolder();
		   return new org.omg.CORBA.CharHolder(any.extract_char());

	       case     org.omg.CORBA.TCKind._tk_double:
		   if(typecode_only)
		       return new org.omg.CORBA.DoubleHolder();
		   return new org.omg.CORBA.DoubleHolder(any.extract_double());
	       case     org.omg.CORBA.TCKind._tk_fixed:return null;

	       case     org.omg.CORBA.TCKind._tk_float:
		      if(typecode_only)
			  return new org.omg.CORBA.FloatHolder();
		      return new org.omg.CORBA.FloatHolder(any.extract_float());

	       case     org.omg.CORBA.TCKind._tk_long:
		   if(typecode_only)
		       return new org.omg.CORBA.IntHolder();
		   return new org.omg.CORBA.IntHolder(any.extract_long());

	       case     org.omg.CORBA.TCKind._tk_longdouble:return null;

	       case     org.omg.CORBA.TCKind._tk_longlong:
		   if(typecode_only)
		       return new org.omg.CORBA.LongHolder();
		   return new org.omg.CORBA.LongHolder(any.extract_longlong());

	       case    org.omg.CORBA.TCKind._tk_native:return null;

	       case org.omg.CORBA.TCKind._tk_null:return null;

	       case org.omg.CORBA.TCKind._tk_objref:return any.extract_Object();

	       case     org.omg.CORBA.TCKind._tk_octet:
		   if(typecode_only)
		       return new org.omg.CORBA.ByteHolder();
		   return new org.omg.CORBA.ByteHolder(any.extract_octet());
	       case     org.omg.CORBA.TCKind._tk_Principal :return null;
	       case     org.omg.CORBA.TCKind._tk_enum:
	       case     org.omg.CORBA.TCKind._tk_union :
	       case      org.omg.CORBA.TCKind._tk_struct :
	       case     org.omg.CORBA.TCKind._tk_sequence:
	       case     org.omg.CORBA.TCKind._tk_array:
	       case     org.omg.CORBA.TCKind._tk_except:
		   return holder_with_value(any,typecode_only);
	       case     org.omg.CORBA.TCKind._tk_short :
		   if(typecode_only)
		       return new org.omg.CORBA.ShortHolder();
		   return new org.omg.CORBA.ShortHolder(any.extract_short());
	       case     org.omg.CORBA.TCKind._tk_string:
		   if(typecode_only)
		       return new org.omg.CORBA.StringHolder();
		   return new org.omg.CORBA.StringHolder(any.extract_string());
	       case     org.omg.CORBA.TCKind._tk_TypeCode :
		   if(typecode_only)
		       return new org.omg.CORBA.TypeCodeHolder();
		   return new org.omg.CORBA.TypeCodeHolder(any.extract_TypeCode());
	       case     org.omg.CORBA.TCKind._tk_ulong :
		   if(typecode_only)
		       return new org.omg.CORBA.IntHolder();
		   return new org.omg.CORBA.IntHolder(any.extract_ulong());
	       case     org.omg.CORBA.TCKind._tk_ulonglong :
		   if(typecode_only)
		       return new org.omg.CORBA.LongHolder();
		   return new org.omg.CORBA.LongHolder(any.extract_ulonglong());
	       case    org.omg.CORBA.TCKind._tk_ushort :
		      if(typecode_only)
			  return new org.omg.CORBA.ShortHolder();
		   return new org.omg.CORBA.ShortHolder(any.extract_ushort());
	       case    org.omg.CORBA.TCKind._tk_value:return null;
	       case org.omg.CORBA.TCKind._tk_value_box :return null;
	       case org.omg.CORBA.TCKind._tk_void :return null;
	       case org.omg.CORBA.TCKind._tk_wchar:return null;
	       case org.omg.CORBA.TCKind._tk_wstring:
		   if(typecode_only)
		       return new org.omg.CORBA.StringHolder();
		   return new org.omg.CORBA.StringHolder(any.extract_wstring());
	       }
	   }catch(java.lang.Exception e){e.printStackTrace();}
	   return null;
     }

     /** Method used to insert the content of a Java object in an any structure, based on the type of the object
     * @param	orb the ORB on which th method caller object is running
     *			value the Java object whose value has to be inserted in th eAny structure
     * @return	the Any structure with the object inserted
     */

public static org.omg.CORBA.Any give_any_arg(org.omg.CORBA.ORB orb,java.lang.Object value){
	org.omg.CORBA.Any any=orb.create_any();
	String type;

	try{


		type=value.getClass().getName();

	    if(type.equals("java.lang.Integer")||type.equals("int"))
		{

		    any.insert_long(((java.lang.Integer)value).intValue());

		}
	    if(type.equals("java.lang.Long")||type.equals("long"))
		any.insert_longlong(((java.lang.Long)value).longValue());
	    if(type.equals("java.lang.Boolean")||type.equals("boolean"))
		any.insert_boolean(((java.lang.Boolean)value).booleanValue());
	    if(type.equals("java.lang.String"))
		any.insert_string(((java.lang.String)value).toString());
	    if(type.equals("java.lang.Character")||type.equals("char"))
		any.insert_char(((java.lang.Character)value).charValue());
	    if(type.equals("java.lang.Short")||type.equals("short"))
		any.insert_short(((java.lang.Short)value).shortValue());
	    if(type.equals("java.lang.Float")||type.equals("float"))
		any.insert_float(((java.lang.Float)value).floatValue());
	    if(type.equals("java.lang.Double")||type.equals("double"))
		any.insert_double(((java.lang.Double)value).doubleValue());
	    if(type.equals("[B"))
		org.omg.CORBA.OctetSeqHelper.insert(any,(byte[])value);
	    if(type.equals("[I"))
		org.omg.CORBA.LongSeqHelper.insert(any,(int[])value);
	    if(type.equals("[J"))
		org.omg.CORBA.LongLongSeqHelper.insert(any,(long[])value);
	    if(type.equals("[Z"))
		org.omg.CORBA.BooleanSeqHelper.insert(any,(boolean[])value);
	    if(type.equals("[S"))
		org.omg.CORBA.ShortSeqHelper.insert(any,(short[])value);
	    if(type.equals("[C"))
		org.omg.CORBA.CharSeqHelper.insert(any,(char[])value);
	    if(type.equals("[F"))
		org.omg.CORBA.FloatSeqHelper.insert(any,(float[])value);
	    if(type.equals("[D"))
		org.omg.CORBA.DoubleSeqHelper.insert(any,(double[])value);
	    if(type.equals("[Ljava.lang.String;"))
		org.omg.CORBA.StringSeqHelper.insert(any,(String[])value);
	    if(type.equals("org.omg.CORBA.Any"))
		any.insert_any((org.omg.CORBA.Any)value);
	    if(type.equals("org.omg.CORBA.Object"))
		any.insert_Object((org.omg.CORBA.Object)value);
	}catch(java.lang.Exception e){e.printStackTrace();return null;}
	return any;
    }


/** Method ued in the gateway server request interceptor when the gateway acts as an intermediate between the client and the replicated
* server group and broadcasts requests to the group members; the gateway has to read arguments to calls from the input stream and demarshal those
* @param	input the input stream from where the parameters are read
*			guide an Any object in which the result of the reading will be written; the guide contains at the time of the call, the type of the object
* 			to be read from the stream
*			typecode_only a boolean value indicating whether the guide contains a value (e.g. it is an INOUT parameter) or just the name of a type (n
* 			case of the result of a call or an OUT parameter)
*/

    public static void read_input_stream(org.omg.CORBA.portable.InputStream input,org.omg.CORBA.Any guide,boolean typecode_only){
	int to_switch;
	org.omg.CORBA.TypeCode its_typecode=null,alias_typecode=null;
	try{
	    if(typecode_only){
		if(guide.type().kind().value()!=org.omg.CORBA.TCKind._tk_TypeCode) {System.out.println("no read");return;}
		its_typecode=guide.extract_TypeCode();
	    }
	    else
		its_typecode=guide.type();
	    to_switch=its_typecode.kind().value();
	    switch(to_switch){
	    case org.omg.CORBA.TCKind._tk_abstract_interface:break;
	    case org.omg.CORBA.TCKind._tk_alias:
		alias_typecode=its_typecode.content_type();
		if(alias_typecode.kind().value()!=org.omg.CORBA.TCKind._tk_sequence){
		    guide.insert_TypeCode(alias_typecode);
		    read_input_stream(input,guide,true);
		}
		else{
		    if(typecode_only)
			guide.type(its_typecode);
		    insert_obj_read_seq(guide,input,alias_typecode);
		}
		break;
	    case org.omg.CORBA.TCKind._tk_any:org.omg.CORBA.Any any=input.read_any();guide.insert_any(any);break;
	    case org.omg.CORBA.TCKind._tk_array:break;
	    case org.omg.CORBA.TCKind._tk_boolean:boolean b=input.read_boolean();guide.insert_boolean(b);break;
	    case org.omg.CORBA.TCKind._tk_char:char c=input.read_char();guide.insert_char(c);break;
	    case org.omg.CORBA.TCKind._tk_double:double d=input.read_double();guide.insert_double(d);break;
	    case org.omg.CORBA.TCKind._tk_except:break;
	    case org.omg.CORBA.TCKind._tk_fixed:break;
	    case org.omg.CORBA.TCKind._tk_float:float f=input.read_float();guide.insert_float(f);break;
	    case org.omg.CORBA.TCKind._tk_long:int l=input.read_long();guide.insert_long(l);break;
	    case org.omg.CORBA.TCKind._tk_longdouble:break;
	    case org.omg.CORBA.TCKind._tk_longlong:
		long ll=input.read_longlong();guide.insert_longlong(ll);break;
	    case org.omg.CORBA.TCKind._tk_null:
		break;
	    case org.omg.CORBA.TCKind._tk_objref:
		org.omg.CORBA.Object obj=input.read_Object();
		guide.insert_Object(obj);
		break;
	    case org.omg.CORBA.TCKind._tk_octet:
		byte by=input.read_octet();guide.insert_octet(by);break;
	    case org.omg.CORBA.TCKind._tk_Principal :
		break;
	    case org.omg.CORBA.TCKind._tk_sequence:
		break;
	    case org.omg.CORBA.TCKind._tk_short :
		short s=input.read_short();guide.insert_short(s); break;
	    case    org.omg.CORBA.TCKind._tk_string:
		String str=input.read_string();guide.insert_string(str);break;
	    case      org.omg.CORBA.TCKind._tk_struct :
	    case org.omg.CORBA.TCKind._tk_enum:
	    case     org.omg.CORBA.TCKind._tk_union :
	    case    org.omg.CORBA.TCKind._tk_value:
	    case org.omg.CORBA.TCKind._tk_value_box :
	    case org.omg.CORBA.TCKind._tk_native:
		if(typecode_only==true)
		    guide.type(its_typecode);
		insert_obj_read(guide,input);
		break;
	    case     org.omg.CORBA.TCKind._tk_TypeCode :
		org.omg.CORBA.TypeCode tc=input.read_TypeCode();guide.insert_TypeCode(tc);break;
	    case     org.omg.CORBA.TCKind._tk_ulong :
		int ul=input.read_ulong();guide.insert_ulong(ul);break;
	    case     org.omg.CORBA.TCKind._tk_ulonglong :
		long ull=input.read_ulonglong();guide.insert_ulonglong(ull);break;
	    case    org.omg.CORBA.TCKind._tk_ushort :
		short us=input.read_ushort();guide.insert_ushort(us);break;
	    case org.omg.CORBA.TCKind._tk_void :
		break;
	    case org.omg.CORBA.TCKind._tk_wchar:
		char wc=input.read_wchar();guide.insert_wchar(wc);break;
	    case org.omg.CORBA.TCKind._tk_wstring:
		String wstr=input.read_wstring();guide.insert_wstring(wstr);break;
	    }
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
    }

/**
* @param	disc a discriminator to indicate whether the call record of a call logged together with explicit parameter is looked for, or the record for a
*			method call logged with parameter buffer information
*			slr	the log record
* @return 	the call log record given by disc
*/

   public static CallLogRecord get_clr(int disc,SomeLogRecord slr){
	if(disc==RecordType._call_par)
	    return slr._call_record_par.call_rec;
	else
	    return slr._call_record_as.call_rec;
    }

/** method used to read (demarshal) the arguments of a call from a stream, knowing the types of the arguments from the method signatures table
* This method is used when replaying methods on the recovering backup, by using nformation from the log of call information
* @param	orb the ORB  which the replica is running
*			method_s the method signatures table
*			is the stream from which the arguments are demarshalled
* @return	the list of method call arguments in the form of a Parameter array
*/

     private static org.omg.Dynamic.Parameter[] read_args(org.omg.CORBA.ORB orb,org.omg.FT.MethSignStr method_s,org.openorb.iiop.CDRInputStream is) throws org.omg.CORBA.BAD_PARAM{

	 int i;
	 org.omg.Dynamic.Parameter[] toRet;
	 org.omg.CORBA.Any guide;
	 org.omg.CORBA.TypeCode[] param_type_codes=method_s.param_types;
	 int[] modes=method_s.param_modes;
	 toRet=new org.omg.Dynamic.Parameter[modes.length];
	 for(i=0;i<modes.length;i++){
	     guide=orb.create_any();
	     guide.insert_TypeCode(param_type_codes[i]);
	     if(modes[i]!=org.omg.CORBA.ParameterMode._PARAM_OUT)
		 read_input_stream(is,guide,true);
	     toRet[i]=new org.omg.Dynamic.Parameter(guide,org.omg.CORBA.ParameterMode.from_int(modes[i]));
	 }
	 return toRet;
   }

/** Method used when a request does not need to be executed on the server and the reply is sent from the level of the server interceptor
*this method does marshaling of the arguments to be sent to the client as an answer
* @param	any	the parameter to be written on the output stream
*			os	the output stream
*/

    public static void write_output_stream(org.omg.CORBA.Any any, org.omg.CORBA.portable.OutputStream os){
	try{
	    switch(any.type().kind().value()){

	    case org.omg.CORBA.TCKind._tk_abstract_interface:break;
	    case org.omg.CORBA.TCKind._tk_any:os.write_any(any.extract_any());break;
	    case org.omg.CORBA.TCKind._tk_boolean:os.write_boolean(any.extract_boolean());break;
	    case org.omg.CORBA.TCKind._tk_char:os.write_char(any.extract_char());break;
	    case org.omg.CORBA.TCKind._tk_double:os.write_double(any.extract_double());break;
	    case     org.omg.CORBA.TCKind._tk_long:os.write_long(any.extract_long());break;
	    case org.omg.CORBA.TCKind._tk_float:os.write_float(any.extract_float());break;
	    case     org.omg.CORBA.TCKind._tk_longlong:os.write_longlong(any.extract_longlong());break;
	    case     org.omg.CORBA.TCKind._tk_octet:os.write_octet(any.extract_octet());break;
	    case     org.omg.CORBA.TCKind._tk_short :os.write_short(any.extract_short()); break;
	    case     org.omg.CORBA.TCKind._tk_string:String theStr=any.extract_string();os.write_string(theStr);break;
	    case     org.omg.CORBA.TCKind._tk_TypeCode :os.write_TypeCode(any.extract_TypeCode());break;
	    case     org.omg.CORBA.TCKind._tk_ulong :os.write_ulong(any.extract_ulong());break;
	    case     org.omg.CORBA.TCKind._tk_ulonglong :os.write_ulonglong(any.extract_ulonglong());break;
	    case    org.omg.CORBA.TCKind._tk_ushort :os.write_ushort(any.extract_ushort());break;
	    case org.omg.CORBA.TCKind._tk_wchar:os.write_wchar(any.extract_wchar());break;
	    case org.omg.CORBA.TCKind._tk_wstring:os.write_wstring(any.extract_wstring());break;
	    case     org.omg.CORBA.TCKind._tk_Principal :os.write_Principal(any.extract_Principal());break;
	    case     org.omg.CORBA.TCKind._tk_sequence:break;
	    case org.omg.CORBA.TCKind._tk_enum:
	    case      org.omg.CORBA.TCKind._tk_struct :
	    case     org.omg.CORBA.TCKind._tk_union :
	    case    org.omg.CORBA.TCKind._tk_value:
	    case org.omg.CORBA.TCKind._tk_array:
	    case org.omg.CORBA.TCKind._tk_value_box :
	    case org.omg.CORBA.TCKind._tk_except:
		write_obj_extract(any,os);
		break;
	    case org.omg.CORBA.TCKind._tk_alias:
		org.omg.CORBA.TypeCode alias_typecode=any.type().content_type();
		if(alias_typecode.kind().value()!=org.omg.CORBA.TCKind._tk_sequence){
		    any.type(any.type().content_type());
		    write_output_stream(any,os);
		    any.type(any.type());
		}
		else
		    write_obj_extract_seq(any,os,alias_typecode);
		break;
	    case org.omg.CORBA.TCKind._tk_fixed:break;
	    case     org.omg.CORBA.TCKind._tk_longdouble:break;
	    case    org.omg.CORBA.TCKind._tk_native:break;
	    case org.omg.CORBA.TCKind._tk_null:break;
	    case org.omg.CORBA.TCKind._tk_objref:os.write_Object(any.extract_Object());break;
	    case org.omg.CORBA.TCKind._tk_void :break;
	    }
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
    }

    /* END general utilities*/

     /* BEGIN NamingService utilities*/

/**
* @param	ncx the naming context in which the object we are looking for was bound
*			ncs the name sequence of the object we are looking for
* @return	the CORBA object bound in the given context
*/
    public static org.omg.CosNaming.NamingContext resolve_n(org.omg.CosNaming.NamingContext ncx,org.omg.CosNaming.NameComponent[] ncs){
	org.omg.CORBA.Object reto=null;
	org.omg.CosNaming.NamingContext ret=null;
	try{
	    reto=ncx.resolve(ncs);
	    ret=org.omg.CosNaming.NamingContextHelper.narrow(reto);
	}catch(org.omg.CosNaming.NamingContextPackage.InvalidName e){return null;}
	catch(org.omg.CosNaming.NamingContextPackage.NotFound e){return null;}
	catch(org.omg.CosNaming.NamingContextPackage.CannotProceed e){return null;}
	return ret;
    }

/**
* @param	ncx the naming context in which to bind the new name
*			ncs the name to be bound in the naming context
* @return	the naming context with the new name bound
*/

public static org.omg.CosNaming.NamingContext bind_context(org.omg.CosNaming.NamingContext ncx,org.omg.CosNaming.NameComponent[] ncs){
	org.omg.CosNaming.NamingContext ret=null;
	try{
	    ret=ncx.bind_new_context(ncs);
	}catch(org.omg.CosNaming.NamingContextPackage.AlreadyBound e){return null;}
	catch(org.omg.CosNaming.NamingContextPackage.InvalidName e){return null;}
	catch(org.omg.CosNaming.NamingContextPackage.NotFound e){return null;}
	catch(org.omg.CosNaming.NamingContextPackage.CannotProceed e){return null;}
	return ret;
    }

    /* END NamingService utilities*/


    /* BEGIN ReplicationManager utilities*/

/**
* @param	obj the CORBA object for which the IOR (interoperable object reference) is requested
* @return	the IOR for the object
*/
    public static IOR get_ior(org.omg.CORBA.Object obj){
	return ((org.openorb.CORBA.Delegate)((org.omg.CORBA.portable.ObjectImpl)obj)._get_delegate()).ior();
    }

/**
* @param	obj the CORBA object array for which the collective IOR is requested
* @return	the collective IOR for the object array
*/
    public static IOR get_ior(org.omg.CORBA.Object[] obj){

	IOR iors[]=new IOR[obj.length],toRet;
	int i;
	for(i=0;i<obj.length;i++)
	    iors[i]=((org.openorb.CORBA.Delegate)((org.omg.CORBA.portable.ObjectImpl)obj[i])._get_delegate()).ior();
	toRet=merge_unique_iors(iors[i-1].type_id,iors);
	return toRet;
    }

/** Method used to print an IOR in a readable format - giving host name, port number,etc.
* @param	orb the ORB on which the object is running
*			ior the IOR that has to be printed
*/

    public static void print_IOR(ORB orb,IOR ior){
	int i;
	org.omg.IIOP.ProfileBody_1_1 pb11;

	for(i=0;i<ior.profiles.length;i++){
	    pb11=profile_body(orb,ior.profiles[i]);
	    System.out.println("Profile "+i+" data : "+pb11.host+":"+(int)pb11.port);
	}
    }

/**
* @param	orb the ORB in which the object is running
*			pr the profile of the IOR to be printed
*/
    public static void print_profile(ORB orb,TaggedProfile pr){
	org.omg.IIOP.ProfileBody_1_1 pb11=profile_body(orb,pr);
	System.out.println(pb11.host+" "+(int)pb11.port);
    }

/**
* @param	orb the ORB in which the method calling object is running
*			obj the CORBA object whose location is asked for
* @return	the location of the CORBA object as a String
*/
    public static String give_location(ORB orb,org.omg.CORBA.Object obj){
	org.omg.IIOP.ProfileBody_1_1 pb11;
	IOR ior=get_ior(obj);
	pb11=profile_body(orb,ior.profiles[0]);
	return pb11.host;
    }

/**
* @param	orb the ORB in which the method calling object is running
*			ior the IOR of the CORBA object for which the location is requested
* @return	the location of the CORBA object as a String
*/
    public static String give_location(ORB orb,IOR ior){
	org.omg.IIOP.ProfileBody_1_1 pb11;
	pb11=profile_body(orb,ior.profiles[0]);
	return pb11.host;
    }

/**
* @param	orb the ORB in which the method calling object is running
*			ior the collective IOR of the CORBA object set for which the locations are requested
* @return	the locations of the CORBA objects as a String array
*/

    public static String[] give_locations(ORB orb,IOR ior){
	org.omg.IIOP.ProfileBody_1_1 pb11;
	String locations[]=new String[ior.profiles.length];
	for(int i=0;i<ior.profiles.length;i++){
	    pb11=profile_body(orb,ior.profiles[i]);
	    locations[i]=pb11.host;
	}
	return locations;
    }

/**
* @param	type_id the type identifier of the IOR that will be obtained after the merge (also the type identifier of the separate IORs)
*			iors a set of IORs that have to be assembled to a single IOR
* @return	the IOR after the merge
*/
    public static IOR merge_unique_iors(String type_id,IOR[] iors){
	TaggedProfile[] merged_profs;
	int nr_profs=0,i,j;

	for(i=0;i<iors.length;i++)
	    nr_profs+=iors[i].profiles.length;

	merged_profs=new TaggedProfile[nr_profs];

	nr_profs=0;

	for(i=0;i<iors.length;i++)
	    for(j=0;j<iors[i].profiles.length;j++){
		merged_profs[nr_profs]=iors[i].profiles[j];
		nr_profs++;
	    }
	return new IOR(type_id,merged_profs);
    }

    /* the value of override says which vector overrides the other:
       if override is 0 (or an even value) then the first vector overrides the second one*/
/** Method used to return the union of two sets represented as Vectors, one of them overriding the other;
* this means that the elements of one of the vectors (the non-overriding one) that are not present in the other one
* have to be inserted first in the union
*/
    public static Vector prop_v_union(Vector v1,Vector v2,int override){
	Vector toRet=new Vector();
	Property crt1,crt2,crt[]=new Property[2];
	Vector[] v12=new Vector[2];
	int l1,l2,i1,i2;
	boolean found;

	if(v1==null) {return v2;}
	if(v2==null) {return v1;}
	v12[override]=v1;
	v12[(override+1)%2]=v2;
	l1=v12[0].size();
	l2=v12[1].size();
	for(i2=0;i2<l2;i2++){
	    crt2=(Property)v12[1].elementAt(i2);
	    found=false;
	    for(i1=0;i1<l1;i1++){
		crt1=(Property)v12[0].elementAt(i1);
		if(crt2.nam[0].id.equals(crt1.nam[0].id)){
		    found=true;
		    break;
		}
	    }
	    if(!found) {
		toRet.addElement(crt2);
	    }
	}
	for(i1=0;i1<l1;i1++)
	    toRet.addElement(v12[0].elementAt(i1));
	return toRet;
    }

/** Method that uses prop_v_union twice to realize the union of three vectors
*/
    public static Vector prop_v_union_o(Vector v1,Vector v2,Vector v3){
	Vector ir,fin;
	ir=prop_v_union(v1,v2,0);
	fin=prop_v_union(ir,v3,0);
	return fin;
    }

    public static String type_id(org.omg.CORBA.Object o){

	IOR ior;

	ior = ((org.openorb.CORBA.Delegate)((ObjectImpl)o)._get_delegate()).ior();
	return ior.type_id;
    }


    public static void type_id(String type_id,org.omg.CORBA.Object o){
	IOR ior;

	ior = ((org.openorb.CORBA.Delegate)((ObjectImpl)o)._get_delegate()).ior();
	ior.type_id=type_id;
    }

/**
* @param	orb the ORB on which the method calling object is running
*			pr the tagged profile that has to be extended with a new tagged component
*			tftc the tagged component that has to be added to the profile
*			tag the code of the tag for the tagged component
* @return	the extended tagged profile
*/
    public static TaggedProfile ext_profile_tag(ORB orb,TaggedProfile pr,IDLEntity tftc,int tag){
	TaggedProfile toRet=new TaggedProfile();

	org.omg.IIOP.ProfileBody_1_1 pb11;

	Any any;

	byte[] to_decode;
	int i;

	CodecFactory factory;
	Codec codec;
	TaggedComponent[] ext_set;

	try{
	    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
	    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	    to_decode=pr.profile_data;
	    any=codec.decode_value(to_decode,org.omg.IIOP.ProfileBody_1_1Helper.type());
	    pb11=org.omg.IIOP.ProfileBody_1_1Helper.extract(any);
	    ext_set=new TaggedComponent[pb11.components.length+1];

	    for(i=0;i<pb11.components.length;i++)
		ext_set[i]=pb11.components[i];

	    ext_set[i]=new TaggedComponent();
	    ext_set[i].tag=tag;

	    if(tag==TAG_FT_GROUP.value)
		org.omg.FT.TagFTGroupTaggedComponentHelper.insert(any,(TagFTGroupTaggedComponent)tftc);
	    else
		org.omg.FT.TagFTPrimaryTaggedComponentHelper.insert(any,(TagFTPrimaryTaggedComponent)tftc);
	    ext_set[i].component_data=codec.encode_value(any);

	    pb11.components=ext_set;

	    org.omg.IIOP.ProfileBody_1_1Helper.insert(any,pb11);

	    toRet.profile_data=codec.encode_value(any);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out)); return null;}
	return toRet;
    }

/**
* @param	orb the ORB on which the method calling object is running
*			t_prof the tagged profile from which the primary tagged component is extracted
* @return	the tagged component with tag TAG_FT_PRIMARY
*/

    public static TaggedComponent extract_tagged_comp(ORB orb,TaggedProfile t_prof){

	Any any;

	org.omg.IIOP.ProfileBody_1_1 pb11;

	org.omg.FT.TagFTPrimaryTaggedComponent tpc;

	byte[] to_decode=t_prof.profile_data;

	boolean found_p=false,found_g=true;

	int j,i;

	CodecFactory factory;
	Codec codec;

	try{
	    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
	    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));

	    any=codec.decode_value(to_decode,org.omg.IIOP.ProfileBody_1_1Helper.type());
	    pb11=org.omg.IIOP.ProfileBody_1_1Helper.extract(any);

	    j=-1;

	    for(i=0;i<pb11.components.length;i++)
		{
		    if(pb11.components[i].tag==TAG_FT_PRIMARY.value){
			to_decode=pb11.components[i].component_data;
			any=codec.decode_value(to_decode,org.omg.FT.TagFTPrimaryTaggedComponentHelper.type());
			tpc=org.omg.FT.TagFTPrimaryTaggedComponentHelper.extract(any);
			if(tpc.primary) {
			    if(j!=-1)
				return pb11.components[j];
			    found_p=true;
			}
		    }
		    else
			if(pb11.components[i].tag==TAG_FT_GROUP.value){
			    if(found_p)
				return pb11.components[i];
			    j=i;
			}
		}
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return null;}
	return null;
    }
/**
* @param	orb the ORB o which the method calling object is running
*			group the CORBA object representing the replicated group with the multi-profile IOR
* @return	true if one of the profiles of the IOR contains a tagged component with tag=TAG_FT_PRIMARY
*/
    public static boolean has_primary(ORB orb,org.omg.CORBA.Object group){
	IOR ior;
	int i,j;
	org.omg.IIOP.ProfileBody_1_1 pb11;

	ior = ((org.openorb.CORBA.Delegate)((ObjectImpl)group)._get_delegate()).ior();
	for(i=0;i<ior.profiles.length;i++){
	    pb11=profile_body(orb,ior.profiles[i]);
	    for(j=0;j<pb11.components.length;j++)
		if(pb11.components[j].tag==TAG_FT_PRIMARY.value)
		    return true;
	}
	return false;
    }


/**
* @param	orb the ORB in which the method calling object is running
*			ior	the IOR to be extended with new profiles
*			extra_ior the IOR containing the extra profiles to be added to the ior
*			position the position where the new profiles are inserted
*			change_group_v if true it means that the group reference version has to be changed
* @return	the new extended IOR
*/

public static IOR update_group_ior_add(ORB orb,IOR ior,IOR extra_ior,int position,boolean change_group_v){

	org.omg.FT.TagFTGroupTaggedComponent tgtc=null;
	org.omg.IIOP.ProfileBody_1_1 pb11;
	Any any;

	int j,i;

	CodecFactory factory;
	Codec codec;


	if(ior==null || extra_ior==null) return null;

	try{
	    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
	    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(java.lang.Exception e) {e.printStackTrace(new java.io.PrintStream(System.out));return null;}

	TaggedProfile[] new_profs=new TaggedProfile[ior.profiles.length+extra_ior.profiles.length];

	int real_posit=min(position,ior.profiles.length-1);

	for(i=0;i<ior.profiles.length;i++){
	    pb11=profile_body(orb,ior.profiles[i]);
	    for(j=0;j<pb11.components.length;j++)
		if(pb11.components[j].tag==TAG_FT_GROUP.value)
		    try{
			any=codec.decode_value(pb11.components[j].component_data,org.omg.FT.TagFTGroupTaggedComponentHelper.type());
			tgtc=org.omg.FT.TagFTGroupTaggedComponentHelper.extract(any);
			if(tgtc!=null){
			    if(change_group_v)
				tgtc.object_group_ref_version++;
			    org.omg.FT.TagFTGroupTaggedComponentHelper.insert(any,tgtc);
			    pb11.components[j].component_data=codec.encode_value(any);
			}
			break;
		    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out)); return null;}
	}


	for(i=0;i<extra_ior.profiles.length;i++){
	    extra_ior.profiles[i]=ext_profile_tag(orb,extra_ior.profiles[i],new TagFTGroupTaggedComponent(new org.omg.GIOP.Version((byte)1,(byte)0),tgtc.ft_domain_id,tgtc.object_group_id,tgtc.object_group_ref_version,tgtc.repl_style),TAG_FT_GROUP.value);
	}

	for(i=0;i<real_posit;i++)
	    new_profs[i]=ior.profiles[i];

	j=i;

	for(i=0;i<extra_ior.profiles.length;i++)
	    new_profs[i+j]=extra_ior.profiles[i];

	for(;j<ior.profiles.length;j++)
	    new_profs[i+j]=ior.profiles[j];

	return new IOR(ior.type_id,new_profs);
    }

    /** The same as update_group_ior_add with the exception that the added group member becomes the primary
       (because there was no other backup to become the primary)*/

    public static IOR update_group_ior_add_pr(ORB orb,IOR ior,IOR extra_ior,int position,boolean change_group_v){

	org.omg.FT.TagFTGroupTaggedComponent tgtc=null;
	org.omg.FT.TagFTPrimaryTaggedComponent tptc=null;
	TaggedComponent[] new_comps;
	org.omg.IIOP.ProfileBody_1_1 pb11;
	Any any;

	int j,i,k;

	CodecFactory factory;
	Codec codec;


	if(ior==null || extra_ior==null) return null;

	try{
	    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
	    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(java.lang.Exception e) {e.printStackTrace(new java.io.PrintStream(System.out));return null;}

	TaggedProfile[] new_profs=new TaggedProfile[ior.profiles.length+extra_ior.profiles.length];

	int real_posit=min(position,ior.profiles.length-1);
	boolean found_pr_p=false,found_gr_p=false,changed=false;

	for(i=0;i<ior.profiles.length;i++){
	    pb11=profile_body(orb,ior.profiles[i]);
	    for(j=0;j<pb11.components.length;j++){
		if(pb11.components[j].tag==TAG_FT_PRIMARY.value){
		    found_pr_p=true;
		    if(found_pr_p){
			new_comps=new TaggedComponent[pb11.components.length-1];
			for(k=0;k<j;k++)
			    new_comps[k]=pb11.components[k];
			for(;k<pb11.components.length-1;k++)
			    new_comps[k]=pb11.components[k+1];
			pb11.components=new_comps;
			body_profile(orb,pb11);
			changed=true;
			break;
		    }
		}
		if(pb11.components[j].tag==TAG_FT_GROUP.value){
		    found_gr_p=true;
		    try{
			any=codec.decode_value(pb11.components[j].component_data,org.omg.FT.TagFTGroupTaggedComponentHelper.type());
			tgtc=org.omg.FT.TagFTGroupTaggedComponentHelper.extract(any);
			if(tgtc!=null){
			    if(change_group_v)
				tgtc.object_group_ref_version++;
			    org.omg.FT.TagFTGroupTaggedComponentHelper.insert(any,tgtc);
			    pb11.components[j].component_data=codec.encode_value(any);
			}
			if(found_pr_p&&!changed){
			    new_comps=new TaggedComponent[pb11.components.length-1];
			    for(k=0;k<j;k++)
				new_comps[k]=pb11.components[k];
			    for(;k<pb11.components.length-1;k++)
				new_comps[k]=pb11.components[k+1];
			    pb11.components=new_comps;
			    body_profile(orb,pb11);
			    changed=true;
			    break;
			}
		    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out)); return null;}
		}
	    }
	}

	extra_ior.profiles[0]=ext_profile_tag(orb,extra_ior.profiles[0],new TagFTGroupTaggedComponent(new org.omg.GIOP.Version((byte)1,(byte)0),tgtc.ft_domain_id,tgtc.object_group_id,tgtc.object_group_ref_version,tgtc.repl_style),TAG_FT_GROUP.value);
	extra_ior.profiles[0]=ext_profile_tag(orb,extra_ior.profiles[0],new TagFTPrimaryTaggedComponent(true),TAG_FT_PRIMARY.value);

	for(i=0;i<real_posit;i++)
	    new_profs[i]=ior.profiles[i];
	j=i;

	new_profs[j]=extra_ior.profiles[0];
	i=1;
	for(;j<ior.profiles.length;j++)
	    new_profs[i+j]=ior.profiles[j];

	return new IOR(ior.type_id,new_profs);
    }


/**
* @param	orb the ORB in which the method calling object is running
*			ior	the IOR that will be reduced with the profile of the failed group member
*			rem_ior the IOR corresponding to the member that ahs to be removed from the group
*			change_group_v if true it means that the group reference version has to be changed
* @return	the new IOR obtained after removing the failed member's IOR
*/
    public static IOR update_group_ior_rem(ORB orb,IOR ior,IOR rem_ior,boolean change_group_v){

	org.omg.FT.TagFTGroupTaggedComponent tgtc=null;
	org.omg.IIOP.ProfileBody_1_1 pb111,pb112;
	Any any;

	int j=-1,i;

	CodecFactory factory;
	Codec codec;
	TaggedProfile[] new_profiles;

	if(ior==null||rem_ior==null) return null;

	pb112=profile_body(orb,rem_ior.profiles[0]);
	try{
	    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
	    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}
	catch(java.lang.Exception e) {e.printStackTrace(new java.io.PrintStream(System.out));return null;}
	for(i=0;i<ior.profiles.length;i++)
	    if(ior.profiles[i].tag==TAG_INTERNET_IOP.value){
		pb111=profile_body(orb,ior.profiles[i]);
		if(profiles_eq(orb,pb111,pb112)){
		    j=i;
		    break;
		}
	    }
	if(j==-1) {System.out.println("IOR to remove not found");return null;}
	new_profiles=new TaggedProfile[ior.profiles.length-1];
	for(i=0;i<j;i++)
	    new_profiles[i]=ior.profiles[i];
	for(i=j;i<ior.profiles.length-1;i++)
	    new_profiles[i]=ior.profiles[i+1];
	for(i=0;i<new_profiles.length;i++){
	    pb111=profile_body(orb,new_profiles[i]);
	    for(j=0;j<pb111.components.length;j++)
		if(pb111.components[j].tag==TAG_FT_GROUP.value)
		    try{
			any=codec.decode_value(pb111.components[j].component_data,org.omg.FT.TagFTGroupTaggedComponentHelper.type());
			tgtc=org.omg.FT.TagFTGroupTaggedComponentHelper.extract(any);
			if(tgtc!=null){
			    if(change_group_v)
				tgtc.object_group_ref_version++;
			    org.omg.FT.TagFTGroupTaggedComponentHelper.insert(any,tgtc);
			    pb111.components[j].component_data=codec.encode_value(any);
			}
			break;
		    }
	    catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out)); return null;}
	    new_profiles[i]=body_profile(orb,pb111);
	}

	ior.profiles=new_profiles;
	return ior;
    }

/**
* @param	orb	the ORB on which the method calling object is running
*			tc the tagged component from which the group reference version is extracted
* @return	the group reference version
*/

    public static int  extract_group_version(ORB orb,TaggedComponent tc){

	Any any;

	org.omg.FT.TagFTGroupTaggedComponent tgtc;


	byte[] to_decode=tc.component_data;

	int j,i;

	CodecFactory factory;
	Codec codec;
	try{
	    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
	    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));

	    any=codec.decode_value(to_decode,org.omg.FT.TagFTGroupTaggedComponentHelper.type());
	    tgtc=org.omg.FT.TagFTGroupTaggedComponentHelper.extract(any);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return -1;}
	return tgtc.object_group_ref_version;
    }


/**
* @param	orb	the ORB on which the method calling object is running
*			profile the tagged profile in which the tagged component that contains the group reference version resides
* @return	the group reference version
*/

    public static int extract_group_version(ORB orb,TaggedProfile profile){

	Any any;
	byte[] to_decode;
	org.omg.IIOP.ProfileBody_1_1 pb11;
	org.omg.FT.TagFTGroupTaggedComponent tfgtc;
	int i;

	CodecFactory factory;
	Codec codec;
	TaggedComponent[] tcs;
	try{

	    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
	    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	    to_decode=profile.profile_data;
	    any=codec.decode_value(to_decode,org.omg.IIOP.ProfileBody_1_1Helper.type());
	    pb11=org.omg.IIOP.ProfileBody_1_1Helper.extract(any);

	    tcs=pb11.components;

	    for(i=0;i<tcs.length;i++)
		if(tcs[i].tag==TAG_FT_GROUP.value){
		    any=codec.decode_value(tcs[i].component_data,org.omg.FT.TagFTGroupTaggedComponentHelper.type());
		    tfgtc=org.omg.FT.TagFTGroupTaggedComponentHelper.extract(any);
		    return tfgtc.object_group_ref_version;
		}
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return -1;}
	return -1;
    }


/**
* @param	orb	the ORB on which the method calling object is running
*			ior the IOR in which one profile is examined for the tagged component that contains the group reference version
* @return	the group reference version
*/
    public static int extract_group_version(ORB orb,IOR ior){

	int i,grv;

	for(i=0;i<ior.profiles.length;i++){
	    grv=extract_group_version(orb,ior.profiles[i]);
	    if(grv!=-1) return grv;
	}
	return -1;
    }
/**
* @param	orb	the ORB on which the method calling object is running
*			profile the tagged profile in which the tagged component that contains the group reference version resides
* @return	the group reference version and the group identifier
*/

    public static long[] extract_group_vers_and_id(ORB orb,TaggedProfile profile){
	Any any;
	byte[] to_decode;
	org.omg.IIOP.ProfileBody_1_1 pb11;
	org.omg.FT.TagFTGroupTaggedComponent tfgtc;
	int i;

	long[] to_ret;

	CodecFactory factory;
	Codec codec;
	TaggedComponent[] tcs;
	try{

	    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
	    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	    to_decode=profile.profile_data;
	    any=codec.decode_value(to_decode,org.omg.IIOP.ProfileBody_1_1Helper.type());
	    pb11=org.omg.IIOP.ProfileBody_1_1Helper.extract(any);

	    tcs=pb11.components;

	    for(i=0;i<tcs.length;i++)
		if(tcs[i].tag==TAG_FT_GROUP.value){
		    any=codec.decode_value(tcs[i].component_data,org.omg.FT.TagFTGroupTaggedComponentHelper.type());
		    tfgtc=org.omg.FT.TagFTGroupTaggedComponentHelper.extract(any);
		    to_ret=new long[2];
		    to_ret[0]=(long)tfgtc.object_group_ref_version;
		    to_ret[1]=tfgtc.object_group_id;
		    return to_ret;
		}
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return new long[0];}
	return new long[0];
    }

/**
* @param	orb the ORB in which the method calling object is running
*			ior the IOR which is checked to discover whether its member residing on host location is the primary or not
*			location the location where the request from the client arrived
* @return   true if the replica to which the request arrived is really the primary
*/
    public static boolean is_primary(org.omg.CORBA.ORB orb,org.omg.IOP.IOR ior,String location){
		org.omg.IIOP.ProfileBody_1_1 pb11;
		int i,j;
		for(i=0;i<ior.profiles.length;i++){
			pb11=profile_body(orb,ior.profiles[i]);
			if(pb11.host.equals(location)){
			 for(j=0;j<pb11.components.length;j++)
				if(pb11.components[j].tag==TAG_FT_PRIMARY.value)
				  return true;
			 return false;
		 }
	 }
	return false;
    }


/**
* @param	orb the ORB in which the method calling object is running
*			obj the CORBA object (representing the group) which is checked to discover whether its member residing on host location is the primary or not
*			location the location where the request from the client arrived
* @return   true if the replica to which the request arrived is really the primary
*/
    public static boolean is_primary(org.omg.CORBA.ORB orb,org.omg.CORBA.Object obj,String location){
	org.omg.IOP.IOR ior=get_ior(obj);
	return is_primary(orb,ior,location);
    }

/**
* @param	orb the ORB in which the method calling object is running
*			component the tagged component that should have the TAG_FT_PRIMARY tag
* @return	false if either the tagged component does not have the expected tag, or if the primary field is set to false
*/
      public static boolean extract_primary(org.omg.CORBA.ORB orb,TaggedComponent component){
		Any any;
		byte[] to_decode;
		org.omg.FT.TagFTPrimaryTaggedComponent tfptc=null;
		int i;

		CodecFactory factory;
		Codec codec;

		if(component.tag!=TAG_FT_PRIMARY.value) return false;

		try{
		    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
		    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
		    to_decode=component.component_data;
		    any=codec.decode_value(to_decode,org.omg.FT.TagFTPrimaryTaggedComponentHelper.type());
		    tfptc=org.omg.FT.TagFTPrimaryTaggedComponentHelper.extract(any);
		}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return false;}

		if(tfptc==null) return false;

		return tfptc.primary;

	    }


	    /** change the primary member of the object group to be a different member of
	       the same group; thus, pr_ior has profiles which can be found in group_ior
	       * @param	orb the ORB in which the method calling object is running
	       *		group_ior the IOR of the group for which the primary has to be changed
	       *		pr_ior the IOR of the replica that will become the primary
	       *		change_gr_v if true then the group reference version has to be changed
	       * @return the new IOR with the changed primary
	       */

	    public static IOR change_primary(ORB orb,IOR group_ior, IOR pr_ior,boolean change_gr_v){

		int i,j,k;
		org.omg.IIOP.ProfileBody_1_1 pb11,pb11p;
		TaggedComponent[] new_comps;
		boolean found=false;
		Any any;
		CodecFactory factory;
		Codec codec;
		org.omg.FT.TagFTGroupTaggedComponent tgtc=null;

		try{
		    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
		    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
		}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out)); return null;}

		pb11p=profile_body(orb,pr_ior.profiles[0]);

		for(i=0;i<group_ior.profiles.length;i++){
		    pb11=profile_body(orb,group_ior.profiles[i]);
		    for(j=0;j<pb11.components.length;j++)
			if(pb11.components[j].tag==TAG_FT_PRIMARY.value){
			    new_comps=new TaggedComponent[pb11.components.length-1];
			    for(k=0;k<j;k++)
				new_comps[k]=pb11.components[k];
			    for(;k<pb11.components.length-1;k++)
				new_comps[k]=pb11.components[k+1];
			    pb11.components=new_comps;
			    group_ior.profiles[i]=body_profile(orb,pb11);
			    found=true;
			    break;
			}
		    if(found) break;
		}
		for(i=0;i<group_ior.profiles.length;i++){
		    pb11=profile_body(orb,group_ior.profiles[i]);
		    for(j=0;j<pb11.components.length;j++)
			if(pb11.components[j].tag==TAG_FT_GROUP.value){
			    try{
				any=codec.decode_value(pb11.components[j].component_data,org.omg.FT.TagFTGroupTaggedComponentHelper.type());
				tgtc=org.omg.FT.TagFTGroupTaggedComponentHelper.extract(any);
				if(change_gr_v)
				    tgtc.object_group_ref_version++;
				org.omg.FT.TagFTGroupTaggedComponentHelper.insert(any,tgtc);
				pb11.components[j].component_data=codec.encode_value(any);
				break;
			    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out)); return null;}
			}
		    group_ior.profiles[i]=body_profile(orb,pb11);
		    if(profiles_eq(orb,pb11,pb11p))
			group_ior.profiles[i]=ext_profile_tag(orb,group_ior.profiles[i],new TagFTPrimaryTaggedComponent(true),TAG_FT_PRIMARY.value);
		}
		if (tgtc==null) return null;
		return group_ior;
	    }

/**
* @param	orb the ORB in which the method calling object is running
*			ref the CORBA object reference corresponding to the group for which the fault tolerance domain name will be extracted from one
*			of the profiles of the IOR
* @return	the String representing the fault tolerance domain name
*/

public static String extract_ft_dom_n(ORB orb,org.omg.CORBA.Object ref){
	IOR ior=get_ior(ref);
	org.omg.IIOP.ProfileBody_1_1 pb11;
	int i;
	TagFTGroupTaggedComponent tfgtc;
	Any any;
	CodecFactory factory;
	Codec codec;
	byte[] to_decode;

	pb11=profile_body(orb,ior.profiles[0]);
	if(pb11.components[0].tag!=TAG_FT_GROUP.value) return "";
	try{
	    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
	    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	    to_decode=pb11.components[0].component_data;
	    any=codec.decode_value(to_decode,org.omg.FT.TagFTGroupTaggedComponentHelper.type());
	    tfgtc=org.omg.FT.TagFTGroupTaggedComponentHelper.extract(any);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return "";}
	if(tfgtc!=null)
	    return tfgtc.ft_domain_id;
	else
	    return "";
    }

/**
* @param	orb the ORB in which the method calling object is running
*			pb11 the profile body that will be converted to a tagged profile
* @return 	the tagged profile
*/
    public static TaggedProfile body_profile(ORB orb,org.omg.IIOP.ProfileBody_1_1 pb11){
	TaggedProfile tp=new TaggedProfile();
	Any any;
	CodecFactory factory;
	Codec codec;

	try{
	    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
	    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	    any=orb.create_any();
	    org.omg.IIOP.ProfileBody_1_1Helper.insert(any,pb11);
	    tp.profile_data=codec.encode_value(any);
	    tp.tag=TAG_INTERNET_IOP.value;
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return null;}
	return new TaggedProfile(tp.tag,tp.profile_data);
    }

/**
* @param	orb the ORB in which the method calling object is running
*			tp the tagged profile that will be converted to a profile body which has explicit fields showing host, port, etc.
* @return 	the profile body of this tagged profile
*/
    public static org.omg.IIOP.ProfileBody_1_1 profile_body(ORB orb,TaggedProfile tp){
	org.omg.IIOP.ProfileBody_1_1 pb11=new org.omg.IIOP.ProfileBody_1_1();
	Any any;

	CodecFactory factory;
	Codec codec;
	if(tp==null) return null;

	try{
	    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
	    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	    any=codec.decode_value(tp.profile_data,org.omg.IIOP.ProfileBody_1_1Helper.type());
	    pb11=org.omg.IIOP.ProfileBody_1_1Helper.extract(any);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return null;}
	return pb11;
    }

/**
* @param	orb the ORB in which the method calling object is running
*			ior the IOR which is checked whether it has a certain tagged component
* @return	true if the IOR contains a profile that contains a tagged component with tag TAG_FT_GROUP
*/
    public static boolean is_group_ref(ORB orb,IOR ior){
	/* Check whether the TaggedProfile has a TAG_FT_GROUP component*/
	org.omg.IIOP.ProfileBody_1_1 pb11=profile_body(orb,ior.profiles[0]);
	int i;

	for(i=0;i<pb11.components.length;i++)
	    if(pb11.components[i].tag==TAG_FT_GROUP.value)
		return true;
	return false;
    }

/**
* @param	orb	the ORB in which the method calling object is running
*			ior	the IOR which is used to extract the group identifier from
* @return	the group identifier of the group having ior as an IOR
*/
    public static long extract_group_id(ORB orb,IOR ior){
	TaggedProfile profile;
	org.omg.IIOP.ProfileBody_1_1 pb11;
	TagFTGroupTaggedComponent tfgtc;
	int i;
	Any any;
	CodecFactory factory;
	Codec codec;

	profile=ior.profiles[0];
	pb11=profile_body(orb,profile);
	try{
	    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
	    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));

	    for(i=0;i<pb11.components.length;i++)
		if(pb11.components[i].tag==TAG_FT_GROUP.value){
		any=codec.decode_value(pb11.components[i].component_data,TagFTGroupTaggedComponentHelper.type());
		tfgtc=TagFTGroupTaggedComponentHelper.extract(any);
		return tfgtc.object_group_id;
		}
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return -1;}
	return -1;
    }


     public static long extract_group_id(ORB orb,org.omg.CORBA.Object obj){
	 IOR ior=get_ior(obj);
	 return extract_group_id(orb,ior);
    }

/**
* @param	orb the ORB in which the method calling object is running
*			tp the tagged profile that is used to extract the replication style from
* @return	the relication style
*/
    public static int extract_repl_style(ORB orb,TaggedProfile tp){
	Any any;
	byte[] to_decode;
	org.omg.IIOP.ProfileBody_1_1 pb11;
	org.omg.FT.TagFTGroupTaggedComponent tfgtc;
	int i;

	CodecFactory factory;
	Codec codec;
	TaggedComponent[] tcs;
	try{

	    factory = (CodecFactory)orb.resolve_initial_references("CodecFactory");
	    codec = factory.create_codec(new Encoding(ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	    to_decode=tp.profile_data;
	    any=codec.decode_value(to_decode,org.omg.IIOP.ProfileBody_1_1Helper.type());
	    pb11=org.omg.IIOP.ProfileBody_1_1Helper.extract(any);

	    tcs=pb11.components;

	    for(i=0;i<tcs.length;i++)
		if(tcs[i].tag==TAG_FT_GROUP.value){
		    any=codec.decode_value(tcs[i].component_data,org.omg.FT.TagFTGroupTaggedComponentHelper.type());
		    tfgtc=org.omg.FT.TagFTGroupTaggedComponentHelper.extract(any);
		    return tfgtc.repl_style;
		}
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return -1;}
	return -1;
    }

/**
* @param	orb the ORB in which the method calling object is running
*			ior the IOR which is checked to obtain the index of the primary profile
* @return	the index of the profile containing the tagged component with tag TAG_FT_PRIMARY
*/

    public static int primary_prof_ind(ORB orb,IOR ior){
	int i,j;
	org.omg.IIOP.ProfileBody_1_1 pb11;
	for(i=0;i<ior.profiles.length;i++){
	    pb11=profile_body(orb,ior.profiles[i]);
	    for(j=0;j<pb11.components.length;j++)
		if(pb11.components[j].tag==TAG_FT_PRIMARY.value)
		    return i;
	}
	return -1;
    }

/** Method used to obtain the difference of two sets represented as Hashtables
*/
 public static Hashtable minus(Hashtable set_a, Hashtable set_b){
	if(set_b==null||set_b.size()==0) return set_a;
	if(set_a==null) return new java.util.Hashtable();
	Hashtable minus_set=new Hashtable();
	Set a_s=set_a.entrySet();
	Iterator set_iter_a=a_s.iterator();
	Map.Entry elem_a;
	while(set_iter_a.hasNext()){
	    elem_a=(Map.Entry)set_iter_a.next();
	    if(!set_b.containsKey(elem_a.getKey()))
		minus_set.put(elem_a.getKey(),elem_a.getValue());
	}
	return minus_set;
    }

/** Method used to obtain the difference of two sets represented as arrays of long
*/
    public static long[] minus(long[] a,long[] b){
	int i,j;
	boolean found;
	if(a==null) return new long[0];
	if(b==null||b.length==0) return a;
	Vector temp=new Vector();
	for(i=0;i<a.length;i++){
	    found=false;
	    for(j=0;j<b.length;j++){
		if(a[i]==b[j]){
		    found=true;
		    break;
		}
	    }
	    if(!found)
		temp.addElement(new Long(a[i]));
	}

	long[] ret=new long[temp.size()];
	for(i=0;i<ret.length;i++)
	ret[i]=((Long)temp.elementAt(i)).longValue();
	return ret;
    }

/** Method used to obtain the differences A-B and B-A and the intersection of two sets represented as Hashtables
*/
    public static Hashtable[] a_minus_b_a_inters_b_b_minus_a(Hashtable a,Hashtable b){
	Hashtable three_sets[]=new Hashtable[3], b_copy=new Hashtable(),a_minus_b=new Hashtable(),a_inters_b=new Hashtable(),b_minus_a=new Hashtable();
	Set a_s;
	Iterator set_iter_a;
	Map.Entry elem_a;
	Vector elem_to_rem_b=new Vector();

	int i,l;

	a_s=a.entrySet();
	set_iter_a=a_s.iterator();
	b_copy.putAll(b);
	while(set_iter_a.hasNext()){
	    elem_a=(Map.Entry)set_iter_a.next();
	    if(!b_copy.containsKey(elem_a.getKey()))
		a_minus_b.put(elem_a.getKey(),elem_a.getValue());
	    else{
		a_inters_b.put(elem_a.getKey(),elem_a.getValue());
		elem_to_rem_b.addElement(elem_a.getKey());
	    }
	}
	l=elem_to_rem_b.size();
	for(i=0;i<l;i++)
	b_copy.remove(elem_to_rem_b.elementAt(i));
	b_minus_a.putAll(b_copy);
	three_sets[0]=a_minus_b;
	three_sets[1]=a_inters_b;
	three_sets[2]=b_minus_a;
	return three_sets;
    }


/** Method used to obtain the differences A-B and B-A of two sets represented as arrays of long
*/
    public static long[][] a_minus_b_b_minus_a(long[] a,long[] b){
	long[] two_sets[]=new long[2][], a_minus_b,b_minus_a,b_copy;
	Vector a_minus_b_v=new Vector();
	int i,j,size_b_m_a;
	boolean found;
	if(a==null) {
	    two_sets[0]=new long[0];
	    if(b==null)
		two_sets[1]=new long[0];
	    else
		two_sets[1]=b;
	    return two_sets;
	}
	if(b==null) {
	    two_sets[0]=a;
	    two_sets[1]=new long[0];
	    return two_sets;
	}

	size_b_m_a=b.length;
	b_copy=new long[size_b_m_a];
	System.arraycopy(b,0,b_copy,0,size_b_m_a);
	for(i=0;i<a.length;i++){
	    found=false;
	    for(j=0;j<b_copy.length;j++)
		if(a[i]==b_copy[j]){
		    found=true;
		    break;
		}
	    if(!found)
		a_minus_b_v.addElement(new Long(a[i]));
	    else{
		b_copy[j]=Long.MIN_VALUE;
		size_b_m_a--;
	    }
	}
	b_minus_a=new long[size_b_m_a];
	j=0;
	for(i=0;i<b_copy.length;i++)
	if(b_copy[i]!=Long.MIN_VALUE){
	    b_minus_a[j]=b_copy[i];
	    j++;
	}
	a_minus_b=new long[a_minus_b_v.size()];
	for(i=0;i<a_minus_b.length;i++)
	a_minus_b[i]=((Long)a_minus_b_v.elementAt(i)).longValue();
	two_sets[0]=a_minus_b;
	two_sets[1]=b_minus_a;
	return two_sets;
    }

/** Method used to check whether two tagged profiles are equal (identical) or not (uses the comparison of profile bodies)
*/

    public static boolean profiles_eq(ORB orb,TaggedProfile pr1,TaggedProfile pr2){
	return profiles_eq(orb,profile_body(orb,pr1),profile_body(orb,pr2));
    }

/** Method used to check whether two tagged profiles are equal (identical) or not by comparing the profile bodies
*/
    public static boolean profiles_eq(ORB orb,org.omg.IIOP.ProfileBody_1_1 pr1,org.omg.IIOP.ProfileBody_1_1 pr2){

	int i;
	if(pr1==null)
	    if(pr2==null)
		return true;
	    else
		return false;

	if(!pr1.host.equals(pr2.host)||pr1.port!=pr2.port||pr1.object_key.length!=pr2.object_key.length)
	    return false;
	if(pr1.iiop_version.minor!=pr2.iiop_version.minor||pr1.iiop_version.major!=pr2.iiop_version.major)
	    return false;
	for(i=0;i<pr1.object_key.length;i++)
	    if(pr1.object_key[i]!=pr2.object_key[i]) return false;
	return true;
    }

/** Method used to check if two CORBA object references are identical by checking the equality of their profiles
*/

    public static boolean refs_equal(org.omg.CORBA.ORB orb,org.omg.CORBA.Object ref1,org.omg.CORBA.Object ref2,int how_many_profiles){
	IOR ior1,ior2;
	int i,nr=how_many_profiles;


	ior1=get_ior(ref1);
	ior2=get_ior(ref2);
	if(ior1.profiles.length!=ior2.profiles.length) {
	    return false;
	}
	if(how_many_profiles==-1)
	    nr=ior1.profiles.length;

	for(i=0;i<nr;i++)
	    if(!profiles_eq(orb,profile_body(orb,ior1.profiles[i]),profile_body(orb,ior2.profiles[i]))){
		return false;
	    }
	return true;
    }

/** Method used to return the stub class corresponding to a type identifier
  Example: the type identifier IDL:a/b/c/D:1.0 has the corresponding stub class name a.b.c._DStub (a.b.c is the name of the package)
*/
    public static java.lang.Class get_class(String type_id){
	java.lang.Class toRet=null;
	String class_n=null;
	StringBuffer class_n_b=null;
	try{
	    class_n=get_class_name(type_id);
	    int i=class_n.lastIndexOf('.');
	    class_n_b=new StringBuffer(class_n);
	    class_n_b.insert(i+1,'_');
	    class_n_b.append("Stub");
	    toRet=java.lang.Class.forName(class_n_b.toString());
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));toRet=null;}
	return toRet;
    }

 /**
 * @return helper class name corresponding to the type identifier type_id
 */

    public static String get_class_helper(String type_id){
	if(type_id==null) return "";
	return get_class_name(type_id)+"Helper";
    }

    public static String get_class_helper_s(String class_n){
	return class_n+"Helper";
    }

/** Method used to return the class name corresponding to a type identifier
  Example: the type identifier IDL:a/b/c/D:1.0 has the corresponding stub class name a.b.c.D (a.b.c is the name of the package)
*/

    public static String get_class_name(String type_id){
	int i,j;
	StringTokenizer path,part_path;
	String part_token;
	StringBuffer toRet=new StringBuffer();

	if(type_id==null) return "";

	i=type_id.indexOf(":");
	j=type_id.indexOf(":",i+1);
	path=new StringTokenizer(type_id.substring(i+1,j),"/");
	while(path.hasMoreTokens()){
	    part_token=path.nextToken();
	    if(part_token.indexOf('.')!=-1){
		part_path=new StringTokenizer(part_token,".");
		while(part_path.hasMoreTokens()){
		    toRet.insert(0,part_path.nextToken());
		    if(part_path.hasMoreTokens()) toRet.insert(0,'.');
		    else if(path.hasMoreTokens()) toRet.append('.');
		}
	    }
	    else{
		toRet.append(part_token);
		if(path.hasMoreTokens()) toRet.append('.');
	    }
	}
	return toRet.toString();
    }


/**
* @param	any the Any structure in which the read object will be inserted
*			input the input stream from where the object to be put in the Any is read
*/
    public static void insert_obj_read(org.omg.CORBA.Any any,org.omg.CORBA.portable.InputStream input){
	String class_n;
	String helper;
	Class helper_class,param_cl[];
	java.lang.reflect.Method is_read_m,insert_method;
	java.lang.Object params[],result,helper_class_inst;
	try{
	    class_n=get_class_name(any.type().id());
	    helper=get_class_helper_s(class_n);
	    helper_class=Class.forName(helper);
	    helper_class_inst=helper_class.newInstance();
	    param_cl=new Class[1];
	    param_cl[0]=Class.forName("org.omg.CORBA.portable.InputStream");
	    is_read_m=helper_class.getMethod("read",param_cl);
	    param_cl=new Class[2];
	    param_cl[0]=Class.forName("org.omg.CORBA.Any");
	    param_cl[1]=Class.forName(class_n);
	    insert_method=helper_class.getMethod("insert",param_cl);
	    params=new java.lang.Object[1];
	    params[0]=input;
	    result=is_read_m.invoke(helper_class_inst,params);
	    params=new java.lang.Object[2];
	    params[0]=any;
	    params[1]=result;
	    insert_method.invoke(helper_class_inst,params);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
    }

/** similar to insert_obj_read, just that the object read from the input stream is a sequence (array) of elements whose typecode is given as parameter
*/
    public static void insert_obj_read_seq(org.omg.CORBA.Any any,org.omg.CORBA.portable.InputStream input,org.omg.CORBA.TypeCode seq_tc){
	String class_n,seq_class_n;
	String helper;
	Class helper_class,param_cl[];
	java.lang.reflect.Method is_read_m,insert_method;
	java.lang.Object params[],result,helper_class_inst;
	org.omg.CORBA.TypeCode any_tc,seq_e_typecode;
	String the_seq_id;
	try{
	    any_tc=any.type();
	    seq_e_typecode=seq_tc.content_type();
	    the_seq_id=any_tc.id();
	    class_n=get_class_name(the_seq_id);
	    helper=get_class_helper_s(class_n);
	    helper_class=Class.forName(helper);
	    helper_class_inst=helper_class.newInstance();
	    param_cl=new Class[1];
	    param_cl[0]=Class.forName("org.omg.CORBA.portable.InputStream");
	    is_read_m=helper_class.getMethod("read",param_cl);
	    param_cl=new Class[2];
	    param_cl[0]=Class.forName("org.omg.CORBA.Any");
	    seq_class_n="[L"+get_class_name(seq_e_typecode.id())+";";
	    param_cl[1]=Class.forName(seq_class_n);
	    insert_method=helper_class.getMethod("insert",param_cl);
	    params=new java.lang.Object[1];
	    params[0]=input;
	    result=is_read_m.invoke(helper_class_inst,params);
	    params=new java.lang.Object[2];
	    params[0]=any;
	    params[1]=result;
	    insert_method.invoke(helper_class_inst,params);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
    }

/**
* @param	any the Any object that contains the object that has to be written to the output stream
*			os	the output stream where the content of the Any is written
*/
    public static void write_obj_extract(org.omg.CORBA.Any any,org.omg.CORBA.portable.OutputStream os){
	String class_n;
	String helper;
	Class helper_class,param_cl[];
	java.lang.reflect.Method os_write_m,extract_method;
	java.lang.Object params[],result,helper_class_inst;
	try{
	    class_n=get_class_name(any.type().id());
	    helper=get_class_helper_s(class_n);
	    helper_class=Class.forName(helper);
	    helper_class_inst=helper_class.newInstance();
	    param_cl=new Class[1];
	    param_cl[0]=Class.forName("org.omg.CORBA.Any");
	    extract_method=helper_class.getMethod("extract",param_cl);
	    params=new java.lang.Object[1];
	    params[0]=any;
	    result=extract_method.invoke(helper_class_inst,params);
	    param_cl=new Class[2];
	    param_cl[0]=Class.forName("org.omg.CORBA.portable.OutputStream");
	    param_cl[1]=Class.forName(class_n);
	    os_write_m=helper_class.getMethod("write",param_cl);
	    params=new java.lang.Object[2];
	    params[0]=(java.lang.Object)os;
	    params[1]=result;
	    os_write_m.invoke(helper_class_inst,params);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
    }


/** Method similar to write_obj_extract, but the content of the Any object is a sequence of elementas whose typecode is given as a parameter (seq_tc)
*/
    public static void write_obj_extract_seq(org.omg.CORBA.Any any,org.omg.CORBA.portable.OutputStream os,org.omg.CORBA.TypeCode seq_tc){
	String class_n,seq_cl_n;
	String seq_el_id;
	String helper;
	Class helper_class,param_cl[];
	java.lang.reflect.Method os_write_m,extract_method;
	java.lang.Object params[],result,helper_class_inst;
	try{
	    class_n=get_class_name(any.type().id());
	    seq_cl_n="[L"+get_class_name(seq_tc.content_type().id())+";";
	    helper=get_class_helper_s(class_n);
	    helper_class=Class.forName(helper);
	    helper_class_inst=helper_class.newInstance();
	    param_cl=new Class[1];
	    param_cl[0]=Class.forName("org.omg.CORBA.Any");
	    extract_method=helper_class.getMethod("extract",param_cl);
	    params=new java.lang.Object[1];
	    params[0]=any;
	    result=extract_method.invoke(helper_class_inst,params);
	    param_cl=new Class[2];
	    param_cl[0]=Class.forName("org.omg.CORBA.portable.OutputStream");
	    param_cl[1]=Class.forName(seq_cl_n);
	    os_write_m=helper_class.getMethod("write",param_cl);
	    params=new java.lang.Object[2];
	    params[0]=(java.lang.Object)os;
	    params[1]=result;
	    os_write_m.invoke(helper_class_inst,params);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
    }

/**
* @param	any the Any from where the complex structure object is extracted
* @return	the Object corresponding to the content of the Any
*/
    public static java.lang.Object extract_obj_from_any(org.omg.CORBA.Any any){
	String helper;
	Class helper_class,param_cl[];
	java.lang.reflect.Method os_write_m,extract_method;
	java.lang.Object params[],result,helper_class_inst;
	try{
	    helper=get_class_helper(any.type().id());
	    helper_class=Class.forName(helper);
	    helper_class_inst=helper_class.newInstance();
	    param_cl=new Class[1];
	    param_cl[0]=Class.forName("org.omg.CORBA.Any");
	    extract_method=helper_class.getMethod("extract",param_cl);
	    params=new java.lang.Object[1];
	    params[0]=any;
	    result=extract_method.invoke(helper_class_inst,params);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return null;}
	return result;
    }

/** Method used to return an object corresponding to an instance of the Holder class of the object contained by the Any;
* the instance of the holder class will contain as value the content of the Any
*/
    public static java.lang.Object holder_with_value(org.omg.CORBA.Any any,boolean typecode_only){

	String class_n;
	String helper;
	Class helper_class,param_cl[];
	java.lang.reflect.Field value_field;
	java.lang.reflect.Method os_write_m,extract_method;
	java.lang.Object params[],result=null,helper_class_inst,holder_class_inst;
	try{
	    class_n=get_class_name(any.type().id());
	    helper=get_class_helper_s(class_n);
	    helper_class=Class.forName(helper);
	    helper_class_inst=helper_class.newInstance();
	    param_cl=new Class[1];
	    param_cl[0]=Class.forName("org.omg.CORBA.Any");
	    extract_method=helper_class.getMethod("extract",param_cl);
	    params=new java.lang.Object[1];
	    params[0]=any;
	    result=extract_method.invoke(helper_class_inst,params);
	    holder_class_inst=Class.forName(class_n+"Holder").newInstance();
	    if(!typecode_only){
		value_field=Class.forName(class_n+"Holder").getField("value");
		value_field.set(holder_class_inst,result);
	    }
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return null;}

	return holder_class_inst;
    }

/** Method used to invoke the operation oper from the helper class of the class with name class_n, with parameters params
*/
    public static void invoke_helper_method(String oper,String class_n,java.lang.Object[] params){
	Class helper_class,par_cl[];
	java.lang.reflect.Method method;
	int i;
	try{
	    helper_class=Class.forName(get_class_helper_s(class_n));
	    par_cl=new Class[params.length];
	    for(i=0;i<params.length;i++)
		par_cl[i]=params[i].getClass();
	    method=helper_class.getMethod(oper,par_cl);
	    method.invoke(helper_class.newInstance(),params);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
    }

/** Method used to call the narrow method of the helper class of the type identifier type_id, with the obj as parameter
*/

    public static java.lang.Object narrow(org.omg.CORBA.Object obj,String type_id){
	Class helper_class,par_cl[];
	java.lang.reflect.Method narrow_m;
	java.lang.Object helper_class_inst,params[],toRet=null;
	try{
	    helper_class=Class.forName(Utils.get_class_helper(type_id));
	    par_cl=new Class[1];
	    par_cl[0]=Class.forName("org.omg.CORBA.Object");
	    narrow_m=helper_class.getMethod("narrow",par_cl);
	    helper_class_inst=helper_class.newInstance();
	    params=new java.lang.Object[1];
	    params[0]=obj;
	    toRet=narrow_m.invoke(helper_class_inst,params);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	return toRet;
    }

/** Method used to check whether a certain method of the server is read-only or update (using the method signature table)
*/

     public static boolean is_read_method(String mn,MethStr[] method_table){
	int i;
	if(method_table==null) return false;

	for(i=0;i<method_table.length;i++){
	    if(method_table[i].name.equals(mn))
		if(method_table[i].type=='r')
		    return true;
		else
		    return false;
	}
	return false;
    }

/** Method used to return the type of a server method
*/

    public static char method_type(String mn,MethSignStr[] method_table){
	int i;
	if(method_table==null) return 'u';

	for(i=0;i<method_table.length;i++)
	    if(method_table[i].name.equals(mn))
		return method_table[i].type;
	return 'u';
    }

/** Method used to return the method signature for the method with name mn
*/

    public static MethSignStr method_structure(String mn,MethSignStr[] method_table){
	int i;
	if(method_table==null) return null;

	for(i=0;i<method_table.length;i++)
	    if(method_table[i].name.equals(mn))
		return method_table[i];
	return null;
    }

    /* END ReplicationManager utilities*/

    /*BEGIN Notification utilities*/

/** Method used to extract properties used in filtering events, from aconsraint expression given as parameter
*/

    public static org.omg.CosNotification.Property[] parse_constraint_expr(String constraint_expr){
	StringTokenizer props_tok=new StringTokenizer(constraint_expr,"&"),mult_value_prop,name_value;
	Vector props=new Vector(),mult_val=new Vector();
	org.omg.CosNotification.Property[] properties;
	org.omg.CORBA.ORB orb=org.omg.CORBA.ORB.init();
	org.omg.CORBA.Any any;
	int i,j,lmp,lp;
	String mult_values[],mv_p;

	while(props_tok.hasMoreTokens())
	    props.addElement(props_tok.nextToken());

	lp=props.size();
	properties=new org.omg.CosNotification.Property[lp];

	for(i=0;i<lp;i++){
	    properties[i]=new org.omg.CosNotification.Property();
	    mv_p=(String)props.elementAt(i);
	    mult_value_prop=new StringTokenizer(mv_p,"|");
	    mult_val.clear();
	    while(mult_value_prop.hasMoreTokens())
		mult_val.addElement(mult_value_prop.nextToken());

	    lmp=mult_val.size();
	    mult_values=new String[lmp];
	    for(j=0;j<lmp;j++){
		name_value=new StringTokenizer((String)mult_val.elementAt(j),"==");
		if(j==0)
		    properties[i].name=name_value.nextToken().substring(1); /* the name is in the form
									       $name*/
		else
		    name_value.nextToken();
		mult_values[j]=name_value.nextToken();
	    }
	    any=orb.create_any();
	    org.omg.CORBA.StringSeqHelper.insert(any,mult_values);
	    properties[i].value=any;
	}
	return properties;
    }


/** Method called to decide whether an event (StructuredEvent) has to be sent or not to a certain consumer, based on the filter
* information (what events it is interested in) provided about that consumer
*/

    public static boolean pass_filter(org.omg.CosNotification.StructuredEvent e,org.omg.CosNotifyFilter.Filter f){
	org.omg.CosNotifyFilter.ConstraintInfo[] constraints;
	String t_n_e,d_n_e,t_n_c,d_n_c,vals[];
	org.omg.CosNotification.Property[] props,filt_data;
	int k,j,i,l,m;

	try{
	    if(e==null) {System.out.println("ev is null");return true;}
	    if(f==null) {System.out.println("f is null");return true;}
	    d_n_e=e.header.fixed_header.event_type.domain_name;
	    t_n_e=e.header.fixed_header.event_type.type_name;
	    filt_data=e.filterable_data;
	    constraints=f.get_all_constraints();
	    if(constraints==null) {System.out.println("cons are null");return true;}
	    for(j=0;j<constraints.length;j++){

		for(k=0;k<constraints[j].constraint_expression.event_types.length;k++){
		    d_n_c=constraints[j].constraint_expression.event_types[k].domain_name;
		    t_n_c=constraints[j].constraint_expression.event_types[k].type_name;
		    if(d_n_c.equals(d_n_e)&&t_n_c.equals(t_n_e)){
			props=Utils.parse_constraint_expr(constraints[j].constraint_expression.constraint_expr);
			if(include(filt_data,props)) return true;
			else return false;
		    }
		}
	    }
	}catch(java.lang.Exception ex) {ex.printStackTrace(new java.io.PrintStream(System.out));}
	return false;
    }


/** Method used to determine whether one set is a subset of another (sets are expressed as CosNotification.Property arrays)
*/

    private static boolean include(org.omg.CosNotification.Property[] a,org.omg.CosNotification.Property[] b){
	int i,j,k,l;
	String[] values_a,values_b;
	boolean intersect,found;

	for(i=0;i<b.length;i++){
	    found=false;
	    for(j=0;j<a.length;j++)
		if(b[i].name.equals(a[j].name)){
		    found=true;
		    values_a=org.omg.CORBA.StringSeqHelper.extract(a[j].value);
		    values_b=org.omg.CORBA.StringSeqHelper.extract(b[i].value);
		    for(k=0;k<values_a.length;k++){
			intersect=false;
			for(l=0;l<values_b.length;l++){
			    if(values_a[k].equals(values_b[l])) {
				intersect=true;
				break;
			    }
			}
			if(intersect)
			    break;
			else
			    return false;
		    }
		    break;
		}
	    if(!found)
		return false;
	}
	return true;
    }


    /* END Notification Utilities*/

   /** Method used to read an array of long from an Object Input Stream
*/

  public static long[] readlongObject(java.io.ObjectInputStream ois){
	int l,i;
	long[] obj=new long[0];
	try{
	l=ois.readInt();
	obj=new long[l];
	for(i=0;i<l;i++)
	obj[i]=ois.readLong();
	}catch(java.lang.Exception e){}
	return obj;
    }

/** Method used to write an array of long to an Object Output Stream
*/
    public static void writelongObject(java.io.ObjectOutputStream oos,long[] obj){
	int i;
	try{
	    oos.writeInt(obj.length);
	    for(i=0;i<obj.length;i++)
		oos.writeLong(obj[i]);
	}catch(java.lang.Exception e){}
    }

/** Method used to read the host names form the FT domain from the file with name file_name
 */

 public static FTDomInfo[] parse_ft_dom_file_i(String file_name){
	Vector ft_d_infos=new Vector(),hosts=new Vector();
	String crt_dom_n="";
	String line;
	String crt_hosts[];
	FTDomInfo[] toRet;
	long size;
	int l;
	try{
	    java.io.RandomAccessFile raf=new java.io.RandomAccessFile(file_name,"rw");
	    size=raf.length();
	    while(raf.getFilePointer()<size){
		line=raf.readLine();
		if(line.length()>2)
		    if(line.substring(0,2).toUpperCase().equals("DB"))
			crt_dom_n=line.substring(2);
		    else
			if(line.substring(0,2).toUpperCase().equals("DE")){
			    crt_hosts=new String[hosts.size()];
			    crt_hosts=(String[])hosts.toArray(crt_hosts);
			    ft_d_infos.addElement(new FTDomInfo(crt_dom_n,crt_hosts));
			    hosts.clear();
			}
			else
			    hosts.addElement(line.substring(1));
	    }
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return null;}

	l=ft_d_infos.size();
	if(l==0) return null;
	toRet=new FTDomInfo[l];
	toRet=(FTDomInfo[])ft_d_infos.toArray(toRet);
	return toRet;
    }

/** Method used to parse the file where the host names from the FT domain are and return their list where the filename is hard coded
*/

    public static String[] parse_ft_dom_file(String ft_dom_name){
	Vector h_names=null;
	String host_names[],line;
	int l,i;
	long size;


	try{
	    java.io.RandomAccessFile raf=new java.io.RandomAccessFile(ft_dom_name_fn,"rw");
	    size=raf.length();
	    while(raf.getFilePointer()<size){
		line=raf.readLine();
		if(line.length()>2&&line.substring(0,2).toUpperCase().equals("DB"))
		    if(line.substring(2).equals(ft_dom_name)){
			h_names=new Vector();
			while(true){
			    line=raf.readLine();
			    if(line.length()>2&&line.substring(0,2).toUpperCase().equals("DE"))
				break;
			    if(line.length()>0&&line.charAt(0)=='H')
				h_names.addElement(line.substring(1));
			}
			break;
		    }
	    }
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return null;}

	if(h_names==null) return null;
	l=h_names.size();
	host_names=new String[l];
	for(i=0;i<l;i++)
	    host_names[i]=(String)h_names.elementAt(i);
	return host_names;
    }



/** Method used to write a message in a file
*/

    public static void trace(String message,java.io.RandomAccessFile raf){
	try{
	    raf.writeChars(message);
	}catch(java.lang.Exception e){}
    }

/** Method used to record time traces for measurement purposes
* @param	moment_code identifies the moment when the trace was taken (e.g. when the replication manager finished some recovery action)
*			retention_id the retention identifier of the request for which the trace was taken
*			client_id the client identifier of the request for which the trace was taken
*			time	time in milliseconds when the trace was taken
*			oper_name the operation (method) name for which the trace was done
*			raf the file where to write the trace info
*/
    public static void trace(int moment_code,int retention_id,String client_id,long time,String oper_name,java.io.RandomAccessFile raf){
	if(raf==null) return;
	synchronized(raf){
	    try{
		raf.writeInt(0);
		raf.writeInt(moment_code);
		raf.writeInt(retention_id);
		raf.writeLong(time);
		raf.writeBytes(client_id+"\n");
		raf.writeBytes(oper_name+"\n");
	    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	}
    }


/** Method used to record time traces for measurement purposes in active replication
* @param	which_replica the index of the replica of the active group that traced the information
*			moment_code identifies the moment when the trace was taken (e.g. when the replication manager finished some recovery action)
*			retention_id the retention identifier of the request for which the trace was taken
*			client_id the client identifier of the request for which the trace was taken
*			time time in milliseconds when the trace was taken
*			oper_name the operation (method) name for which the trace was done
*			raf the file where to write the trace info
*/

        public static void trace(int which_replica,int moment_code,int retention_id,String client_id,long time,String oper_name,java.io.RandomAccessFile raf){
	if(raf==null) return;
	synchronized(raf){
	    try{
		raf.writeInt(1);
		raf.writeInt(which_replica);
		raf.writeInt(moment_code);
		raf.writeInt(retention_id);
		raf.writeLong(time);
		raf.writeBytes(client_id+"\n");
		raf.writeBytes(oper_name+"\n");
	    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	}
    }

/** Method used to record time traces for measurement purposes
* @param	moment_code identifies the moment when the trace was taken (e.g. when the replication manager finished reading some state)
*			size	size of the state
*			time	time in milliseconds when the trace was taken
*			obj_id	the object identifier of the group for which the trace was taken
*			raf the file where to write the trace info
*/

    public static void trace(int moment_code,int size,long time,long obj_id,java.io.RandomAccessFile raf){
	if(raf==null) return;
	synchronized(raf){
	    try{
		raf.writeInt(moment_code);
		raf.writeInt(size);
		raf.writeLong(obj_id);
		raf.writeLong(time);
	    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	}
      }

/** Method used to record time traces for measurement purposes
* @param	moment_code identifies the moment when the trace was taken (e.g. when the replication manager finished some recovery action)
*			group_id	the object identifier of the group for which the trace was taken
*			nr_of_repl	number of replicas in the group when the trace was taken
*			replication_s the replication_style
*			time	time in milliseconds when the trace was taken
*			raf the file where to write the trace info
*/
    public static void trace(int moment_code,long group_id,int nr_of_repl,int replication_s,long time,java.io.RandomAccessFile raf){
	if(raf==null) return;
	synchronized(raf){
	    try{
		raf.writeInt(moment_code);
		raf.writeLong(group_id);
		raf.writeInt(nr_of_repl);
		raf.writeInt(replication_s);
		raf.writeLong(time);
	    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	}
    }

/** Method used in replay_last_requests to replay the logged methods on the recoverin bakup
*/
    private static void exec_meth(java.lang.Object obj,long o_i,org.omg.FT.LogRecord[] last_req, org.omg.FT.LoggingRecoveryController cb,org.omg.FT.MethSignStr[] method_sign,org.omg.CORBA.ORB orb){
	int i,k,lp,j,disc;
	org.omg.FT.CallLogRecord clr;
	org.omg.FT.CallLogRecordPAR clr_par;
	org.omg.FT.CallLogRecordAS clr_as;
	String oper,arg_cl_name;
	Class[] par_cl;
	org.omg.Dynamic.Parameter[] arguments;
	org.omg.FT.MethSignStr one_meth_sign;
	char meth_type='r';
	java.lang.reflect.Method method=null,methods[];
	java.lang.Object[] params;
	java.lang.Object ret;
	org.omg.FT.LogRecord[] replies;
	java.util.Vector log_records=new java.util.Vector();
	org.omg.CORBA.Any ret_args[],ret_res;
	int nr_of_out_inout;

	methods=obj.getClass().getMethods();


	for(i=0;i<last_req.length;i++){
	    arguments=null;
	    disc=last_req[i].the_log_record.discriminator().value();
	    if(disc!=org.omg.FT.RecordType._call_as&&disc!=org.omg.FT.RecordType._call_par) {continue;}
	    clr=org.omg.FT.Utils.get_clr(disc,last_req[i].the_log_record);
	    oper=clr.operation;
	    if(disc==org.omg.FT.RecordType._call_par){
		clr_par=last_req[i].the_log_record.call_record_par();
		arguments=clr_par.arguments;
	    }
	    else{
		clr_as=last_req[i].the_log_record.call_record_as();
		for(j=0;j<method_sign.length;j++)
		    if(method_sign[j].name.equals(oper)){
			try{
			    org.openorb.iiop.CDRInputStream fake=new org.openorb.iiop.CDRInputStream(orb,clr_as.bigEndian,clr_as.version,new org.openorb.io.StorageBuffer(clr_as.buffer,0,clr_as.buffer.length));
			    arguments=org.omg.FT.Utils.read_args(orb,method_sign[j],fake);
			    meth_type=method_sign[j].type;
			    break;
			}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
		    }
	    }

	    nr_of_out_inout=0;
	    ret_args=new org.omg.CORBA.Any[0];
	    method=null;
	    for(j=0;j<methods.length;j++)
		if(methods[j].getName().equals(oper)){
		    method=methods[j];
		    break;
		}
	    if(method!=null){
		params=new java.lang.Object[0];
		try{
		    if(arguments!=null){
			lp=arguments.length;
			params=new java.lang.Object[lp];
			for(k=0;k<lp;k++){
			    if(arguments[k].mode.value()!=org.omg.CORBA.ParameterMode._PARAM_IN){
				params[k]=org.omg.FT.Utils.give_object_holder(arguments[k].argument,arguments[k].mode.value()==org.omg.CORBA.ParameterMode._PARAM_OUT);
				nr_of_out_inout++;
			    }
			    else
				params[k]=org.omg.FT.Utils.give_object(arguments[k].argument);
			}
		    }
		    ret=method.invoke(obj,params);
		    if(cb!=null){
			ret_args=new org.omg.CORBA.Any[nr_of_out_inout];
			ret_res=org.omg.FT.Utils.give_any(orb,ret,false);
			j=0;
			boolean one_null=false;
			for(k=0;k<params.length;k++){
			    int i_o=params[k].getClass().getName().indexOf("Holder");
			    if(i_o!=-1&&params[k].getClass().getName().substring(i_o).equals("Holder")){
				ret_args[j]=org.omg.FT.Utils.give_any(orb,params[k],true);
				if(ret_args[j]==null) {one_null=true;break;}
				j++;
			    }
			}
			if(!one_null&&ret_res!=null){
			    org.omg.FT.SomeLogRecord slr=new org.omg.FT.SomeLogRecord();
			    slr.reply_record(new org.omg.FT.ReplyLogRecord(clr.client_id,clr.retention_id,clr.expiration_time,ret_res,ret_args,meth_type));
			    log_records.addElement(new org.omg.FT.LogRecord(o_i,slr));
			}
		    }
		    /*temporary; it should be something with ret in it*/
		}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	    }
	}
	if(cb!=null) {
	    try{
		replies=new org.omg.FT.LogRecord[log_records.size()];
		replies=(org.omg.FT.LogRecord[])log_records.toArray(replies);
		cb.record_replies(replies);
	    }catch(java.lang.Exception e){e.printStackTrace();}
	}
    }

/** Method called when at failver requests have to be replayed on the recovering backup (in passive replication)
* @param	the_obj Object on which the methods are replayed
*			obj_id the identifier of the object group whose member the_obj is
*			last_requests the list ofrequests that have to be replayed on the backup
*			last_requests_nr the list of requests to be replayed for which no reply was recorded in the primary
*			callback the logging and recovery CORBA object
*			retention_id the retention identifier of the last outgoing CORBA request (if any) from the primary
*			orb the ORB in which the method calling object is running
*			methods the table of method signatures of the CORBA object on which the requests will be replayed
*/
    public static void replay_last_requests(java.lang.Object the_obj,long obj_id,org.omg.FT.LogRecord[] last_requests,org.omg.FT.LogRecord[] last_requests_nr,org.omg.FT.LoggingRecoveryController callback,int retention_id,org.openorb.CORBA.ORB orb,org.omg.FT.MethSignStr[] methods){


	if(last_requests.length>0){
	    exec_meth(the_obj,obj_id,last_requests,null,methods,orb);
	}
	if(last_requests_nr.length>0){
	    exec_meth(the_obj,obj_id,last_requests_nr,callback,methods,orb);
	}
	orb.setFeature("backup",null);
    }

/** Method used in the FA-CORBA infrastructure when the outcome register of a witness has to be reduced, by removing the
* update records since the last state record, and replace them by one state record
* @param	reg the vector representing the register of the witness
*			k the position in the total order of requests of the request that was processed before sending the state
*/
     public static void prune_outcome_vector(java.util.Vector reg,int k){
  	int i,l;
  	universal.Outcome crt;
  	java.util.Vector toRemove=new java.util.Vector();
  	l=reg.size();
  	for(i=0;i<l;i++){
  	    crt=(universal.Outcome)reg.elementAt(i);
  	    if(crt.after_n_req<k)
  		toRemove.addElement(crt);
  	}
  	reg.removeAll(toRemove);
      }

/** Method that returns the element in the Outcome vector (used in FA-CORBA infrastructure) that corresponds to an update
* that happened after k number of requests processed
* Since the number of records in the vector is not the sane as the requests processed, in order to return the element "k", it
* is not possible to return the element on position k, but the element for which the field after_n_req is k
*/

      public static universal.Outcome elementAt(java.util.Vector reg,int k){
  	int i,l;
  	universal.Outcome crt;
  	l=reg.size();
  	for(i=0;i<l;i++){
  	    crt=(universal.Outcome)reg.elementAt(i);
  	    if(crt.after_n_req==k)
  		return crt;
  	}
  	return null;
    }
}
