/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of one of the parties involved in the TRANSORG project. For
 * 	  written permission, please contact diasz@ida.liu.se.
 *
 * 4. Products derived from this Software may not be called "TRANSORG"
 *    nor may "TRANSORG" appear in their names without prior written
 *    permission.
 *
 * 5. Due credit should be given to the TRANSORG Project
 *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
 *
 * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
 *
 * $Id: ExceptionalClRMInterceptor.java,v 1.0  diasz Exp $
 *
 * Date         Author  Changes
 */


package org.omg.FT;
/** Client interceptor class attached to the Replication Manager CORBA object (interceptor gets created when the profile excrmpr is chosen when
 * creating the Replication Manager CORBA object
 * @author	<a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
 */

public class ExceptionalClRMInterceptor extends org.omg.CORBA.LocalObject implements org.omg.PortableInterceptor.ClientRequestInterceptor{

    org.omg.PortableInterceptor.ORBInitInfo info;
    int ft_slot;
    org.omg.CORBA.ORB orb;
    String location;
    String ftDomName;
    LoggingRecoveryController log_rec_ctrl=null;
    org.omg.CosNaming.NamingContext rootC=null,ftDomC=null,hostC=null;
    org.omg.CosNaming.NameComponent loggrcNC=new org.omg.CosNaming.NameComponent("loggrecc","text"),loggrcNCs[]={loggrcNC},fn=new org.omg.CosNaming.NameComponent(),fn_nc[],hostNCs[]=new org.omg.CosNaming.NameComponent[1];

 	/** Constructor
 	 * @param	inf ORB information used to obtain e.g. the instance of the ORB on which the Replication Manager is running
 	 *			ft_s the slot of this interceptor
 	 */

    public ExceptionalClRMInterceptor(org.omg.PortableInterceptor.ORBInitInfo inf, int ft_s)
	{
	    info = inf;
	    ft_slot = ft_s;

	    java.util.Properties pr=System.getProperties();
	    orb=((org.openorb.PI.OpenORBInitInfo)info).orb();

	    ftDomName=(String)pr.get("fault.tolerance.domain.name");

	    try{
		location=java.net.InetAddress.getLocalHost().getHostName();
	    }catch(Exception e){location="localhost";}

	    String ft_dom_name;

	    ft_dom_name=(String)pr.get("fault.tolerance.domain.name");
	    if(ft_dom_name==null) ft_dom_name="";

	    fn.id=ft_dom_name;
	    fn.kind="directory";

	    fn_nc=new org.omg.CosNaming.NameComponent[1];
	    fn_nc[0]=fn;

	    hostNCs[0]=new org.omg.CosNaming.NameComponent(location,"directory");
	}

  /** @return the name of the interceptor
	*/
    public java.lang.String name()
    {
	return "ExceptionalClRMInterceptor";
    }

/** Method called from a specific interception point, when the Replication Manager receives an exception as the result of a request sent by it to another
  * CORBA object
  * @param ri the client request information
  */

    public void receive_exception(org.omg.PortableInterceptor.ClientRequestInfo ri) {

    }

/** Method called from a specific interception point, when the Replication Manager receives a response similar to an exception, as a result of a request sent by it to another CORBA object
  * @param ri the client request information
  */

    public void receive_other(org.omg.PortableInterceptor.ClientRequestInfo ri) {

    }

/** Method called from a specific interception point, when the Replication Manager receives a normal reply to a request sent by it to another CORBA object
  * @param ri the client request information
  */

    public void receive_reply(org.omg.PortableInterceptor.ClientRequestInfo ri) {

    }

    public void send_poll(org.omg.PortableInterceptor.ClientRequestInfo ri) {}

/** Method called from a specific interception point, when the Replication Manager sends a request to another CORBA object
  * @param ri the client request information
  */

    public void send_request(org.omg.PortableInterceptor.ClientRequestInfo ri){

	org.omg.IOP.IOR ior;
	int i,gr_version;
	long gr_id;
	int ret_id;
	String client_id,oper;
	org.omg.IOP.CodecFactory codecFactory;
	org.omg.IOP.Encoding encoding;
	org.omg.IOP.Codec codec;
	org.omg.CORBA.Any any;

	org.omg.CORBA.Object theTarget=ri.target();

	oper=ri.operation();

	org.omg.IOP.ServiceContext serviceCtx_ar = null,serviceCtx_rg=null;


	if(oper.equals("get_state")||oper.equals("set_state")){
	    try{
		codecFactory = info.codec_factory();
		encoding = new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value,(byte)1,(byte)2);
		codec = codecFactory.create_codec(encoding);
		any = orb.create_any();
		FTActiveReplAccessServiceContextHelper.insert(any,new FTActiveReplAccessServiceContext(false));
		serviceCtx_ar=new org.omg.IOP.ServiceContext();
		serviceCtx_ar.context_id=org.omg.IOP.FT_GATEWAY_SOURCE.value;
		serviceCtx_ar.context_data = codec.encode_value(any);
		any = orb.create_any();
		FTReplManToGatewayServiceContextHelper.insert(any,new FTReplManToGatewayServiceContext());
		serviceCtx_rg=new org.omg.IOP.ServiceContext();
		serviceCtx_rg.context_id=org.omg.IOP.FT_RM_TO_GATEWAY.value;
		serviceCtx_rg.context_data = codec.encode_value(any);
	    }
	    catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));serviceCtx_ar=null;serviceCtx_rg=null;}
	    if(serviceCtx_ar!=null)
		ri.add_request_service_context(serviceCtx_ar,true);
	    if(serviceCtx_rg!=null)
		ri.add_request_service_context(serviceCtx_rg,true);
	    System.out.println("Context added");
	}
    }
}
