/**
  * Redistribution and use of this software and associated
  * documentation ("Software"), with or without modification, are
  * permitted provided that the following conditions are met:
  *
  * 1. Redistributions of source code must retain copyright statements
  *    and notices.  Redistributions must also contain a copy of this
  *    document.
  *
  * 2. Redistributions in binary form must reproduce the above
  *    copyright notice, this list of conditions and the following
  *    disclaimer in the documentation and/or other materials provided
  *    with the distribution.
  *
  * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
  *    products derived from this Software without prior written
  *    permission of one of the parties involved in the TRANSORG project. For
  * 	  written permission, please contact diasz@ida.liu.se.
  *
  * 4. Products derived from this Software may not be called "TRANSORG"
  *    nor may "TRANSORG" appear in their names without prior written
  *    permission.
  *
  * 5. Due credit should be given to the TRANSORG Project
  *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
  *
  * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  * PARTICULAR PURPOSE ARE DISCLAIMED.
  *
  * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
  * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
  *
  * $Id: ExceptionalReqRMInterceptor.java,v 1.0  diasz Exp $
  *
  * Date         Author  Changes
  */

package org.omg.FT;
/** Server interceptor class attached to the Replication Manager CORBA object (interceptor gets created when the profile excrmpr is chosen when
 * creating the Replication Manager CORBA object
 * @author	<a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
 */

public class ExceptionalReqRMInterceptor extends org.omg.CORBA.LocalObject implements org.omg.PortableInterceptor.ServerRequestInterceptor{

    org.omg.PortableInterceptor.ORBInitInfo info;
    int ft_slot;
    org.omg.CORBA.ORB orb;
    ReplicationManagerOperations repl_m=null;
    /* the logging and recovery mechanism; it should "know" other such objects on
       other machines and communicate them the information to log
       This is not yet implemented*/

/** Constructor
 	 * @param	inf ORB information used to obtain e.g. the instance of the ORB on which the Replication Manager is running
 	 *			ft_s the slot of this interceptor
 	 */

    public ExceptionalReqRMInterceptor(org.omg.PortableInterceptor.ORBInitInfo inf, int ft_s)
    {
	info = inf;
	ft_slot = ft_s;
	orb=((org.openorb.PI.OpenORBInitInfo)info).orb();
    }

	/** Method called from a specific interception point, when the Replication Manager receives a request; the service contexts attached to the request
	 * can be read here
	 * @param ri the server request information
	 */
	public void receive_request_service_contexts(org.omg.PortableInterceptor.ServerRequestInfo ri) throws org.omg.PortableInterceptor.ForwardRequest
	{


	    org.omg.IOP.CodecFactory factory;
	    org.omg.IOP.Codec codec;
	    org.omg.CORBA.Any any=null;
	    String client_id="";
	    long expiration_time;
	    int retention_id=-1;
	    String oper;
	    boolean ex_req=true;

	    if(ri.operation().equals("_is_a")) return;
	    oper=ri.operation();
	    org.omg.IOP.ServiceContext serviceCtx_gi = null, serviceCtx_req=null;

	    try{
		factory = info.codec_factory();
		codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	    }catch(Exception e){System.out.println("Error in req interc. cont.");e.printStackTrace(new java.io.PrintStream(System.out));return;}

	    try
		{
		    serviceCtx_gi = ri.get_request_service_context(org.omg.IOP.FT_GROUP_ID.value);
		}
	    catch (org.omg.CORBA.BAD_PARAM bp){System.out.println("Operation "+ri.operation()+" There is no group id service context");/*bp.printStackTrace();*/return;}

	    if(serviceCtx_gi==null) {System.out.println("Operation "+ri.operation()+" Group id service context is null");return;}

	    try{
		any = codec.decode_value(serviceCtx_gi.context_data, org.omg.FT.FTGroupIdServiceContextHelper.type());
	    }catch(Exception e){System.out.println("Error in req interc. cont.");e.printStackTrace(new java.io.PrintStream(System.out));return;}

	    org.omg.FT.FTGroupIdServiceContext ftctx = org.omg.FT.FTGroupIdServiceContextHelper.extract(any);

	    /* get the information about the group with group_id for which no initial member is alive*/

	    org.omg.CORBA.Object group_ref;

	    try{
		repl_m=(ReplicationManagerOperations)((org.openorb.CORBA.ORB)orb).getFeature("ItsReplicationManager");
	    }catch(java.lang.Exception e){System.out.println("Error when getFeature");e.printStackTrace(new java.io.PrintStream(System.out));}
	    if(repl_m==null) {System.out.println("Operation "+ri.operation()+" Replication manager is null");throw new org.omg.CORBA.BAD_PARAM("No replication manager");}


	    group_ref=repl_m.get_group_ref(ftctx.object_group_id);

	    if(group_ref==null) {System.out.println("Operation "+ri.operation()+" No group reference for id given "+ftctx.object_group_id);throw new org.omg.CORBA.TRANSIENT(org.omg.CORBA.OMGVMCID.value | 3, org.omg.CORBA.CompletionStatus.COMPLETED_NO);}


	    org.omg.IOP.IOR ior_g= Utils.get_ior(group_ref);

	    try{
		serviceCtx_req=ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	    }catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}

	    if(serviceCtx_req!=null&&ex_req){
		try{
		    any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		    ri.set_slot(ft_slot,any);
		}catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
		org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
		client_id=frctx.client_id;
		retention_id=frctx.retention_id;
	    }

	    throw new org.omg.PortableInterceptor.ForwardRequest(group_ref,true);
	}

    /**
     * receive request operation
     */

    public void receive_request(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {

    }

    /**
     * send reply operation
     */
    public void send_reply(org.omg.PortableInterceptor.ServerRequestInfo ri)
    {}

    public void send_exception(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {

    }

    /**
     * send other operation
     */
    public void send_other(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {

    }


    public java.lang.String name()
    {
	return "FTServerRequestInterceptor";
    }

}
