/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of one of the parties involved in the TRANSORG project. For
 * 	  written permission, please contact diasz@ida.liu.se.
 *
 * 4. Products derived from this Software may not be called "TRANSORG"
 *    nor may "TRANSORG" appear in their names without prior written
 *    permission.
 *
 * 5. Due credit should be given to the TRANSORG Project
 *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
 *
 * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
 *
 * $Id: FTARClientRequestInterceptor.java,v 1.0  diasz Exp $
 *
 * Date         Author  Changes
 */

package org.omg.FT;
/** Class instantiated when active replication is used and the corresponding profile is given
 *  It intercepts outgoing requests from, as well as replies to requests of, an actively replicated group member
 * @author	<a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
 */
public class FTARClientRequestInterceptor extends org.omg.CORBA.LocalObject implements org.omg.PortableInterceptor.ClientRequestInterceptor{

    org.omg.PortableInterceptor.ORBInitInfo info;
    int ft_slot;
    org.omg.CORBA.ORB orb;
    String location;
    LoggingRecoveryController log_rec_ctrl;
    int last_retention_id;
    String last_client_id;


    private void set_logg_rec_c(){
	log_rec_ctrl=(LoggingRecoveryController)((org.openorb.CORBA.ORB)orb).getFeature("logging_recovery_ctrl");
    }

	/** Constructor
	  * @param	inf the ORB info used e.g. to obtain the instance of the ORB on which the CORBA object is running
	  *			ft_s the slot of this interceptor
	  */

    public FTARClientRequestInterceptor(org.omg.PortableInterceptor.ORBInitInfo inf, int ft_s)
    {
	info = inf;
	ft_slot = ft_s;

	orb=((org.openorb.PI.OpenORBInitInfo)info).orb();
	try{
	    location=java.net.InetAddress.getLocalHost().getHostName();
	}catch(java.lang.Exception e){location="localhost";}
	log_rec_ctrl=null;
    }

    public java.lang.String name()
    {
	return "FTARClientRequestInterceptor";
    }


    public void receive_exception(org.omg.PortableInterceptor.ClientRequestInfo ri) {
   }

    public void receive_other(org.omg.PortableInterceptor.ClientRequestInfo ri) {
  }

/** Method called from a specific interception point, when the application object replica receives a normal reply to a request sent by it to another CORBA object
  * @param ri the client request information
  */
    public void receive_reply(org.omg.PortableInterceptor.ClientRequestInfo ri) {
	int ret_id;
	String oper;
	boolean ex_repl=true;

	oper=ri.operation();

	System.out.println("Receive AR reply for "+oper+" "+ri.request_id()+" "+((org.openorb.net.ClientRequest)ri).retention_id());
	if(oper.equals("_is_a"))
	    return;
	if(oper.equals("_get_domain_managers"))
	    return;
	if(oper.equals("_interface"))
	    return;
	if(oper.equals("_non_existent") || oper.equals("_not_existent"))
	    return;
	if(oper.equals("_component"))
	    return;

	if(oper.equals("invoke"))
	    return;
	ret_id=((org.openorb.net.ClientRequest)ri).retention_id();
	if(log_rec_ctrl==null)
	    set_logg_rec_c();

	org.omg.IOP.ServiceContext serviceCtx_repl=null;
	org.omg.IOP.CodecFactory factory;
	org.omg.IOP.Codec codec;
	org.omg.CORBA.Any any=null;

	if(log_rec_ctrl!=null){
	    try{
		serviceCtx_repl=ri.get_reply_service_context(org.omg.IOP.FT_REPLY.value);
	    }catch(java.lang.Exception bio){System.out.println("No reply serv. context");ex_repl=false;}

	    if(serviceCtx_repl!=null&&ex_repl){
		try{
		    factory = info.codec_factory();
		    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
		    any = codec.decode_value(serviceCtx_repl.context_data, org.omg.FT.FTReplyServiceContextHelper.type());
		    org.omg.FT.FTReplyServiceContext frctx = org.omg.FT.FTReplyServiceContextHelper.extract(any);
		    log_rec_ctrl.record_request_repl(((org.openorb.net.ClientRequest)ri).client_id(),ri.target(),ri.operation(),frctx.result_arguments,ret_id,frctx.result);
		}catch(java.lang.Exception e){e.printStackTrace();return;}
	    }
	}
    }

    public void send_poll(org.omg.PortableInterceptor.ClientRequestInfo ri) {}


/** Method called from a specific interception point, when the application object replica sends a request to another CORBA object; in active replication, the request
  * is forwarded to the gateway that does the further sending to the destination
  * @param ri the client request information
  */

    public void send_request(org.omg.PortableInterceptor.ClientRequestInfo ri){

	org.omg.IOP.IOR ior;
	int i,gr_version;
	long gr_id,req_start_time;
	int ret_id;
	String client_id,oper,type_id;
	org.omg.IOP.CodecFactory codecFactory;
	org.omg.IOP.Encoding encoding;
	org.omg.IOP.Codec codec=null;
	org.omg.CORBA.Any any;

	org.omg.CORBA.Object theTarget=ri.target();
	oper=ri.operation();

	if(oper.equals("_is_a"))
	    return;
	if(oper.equals("_get_domain_managers"))
	    return;
	if(oper.equals("_interface"))
	    return;
	if(oper.equals("_non_existent") || oper.equals("_not_existent"))
	    return;
	if(oper.equals("_component"))
	    return;

	ret_id=((org.openorb.net.ClientRequest)ri).retention_id();
	client_id=((org.openorb.net.ClientRequest)ri).client_id();
	req_start_time=((org.openorb.net.ClientRequest)ri).request_start_time();

	long req_dur=DEFAULT_REQUEST_DURATION.value;
	/* Calculate here the expiration time*/
	Integer the_dur=(Integer)((org.openorb.CORBA.ORB)orb).getFeature("request_duration");

	if(the_dur!=null)
	    req_dur=the_dur.intValue();

	FTRequestServiceContext ctx_req=new FTRequestServiceContext(client_id,ret_id,System.currentTimeMillis()+req_dur,location);
	org.omg.IOP.ServiceContext serviceCtx_req = new org.omg.IOP.ServiceContext();
	serviceCtx_req.context_id=org.omg.IOP.FT_REQUEST.value;

	try{
	    codecFactory = info.codec_factory();
	    encoding = new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value,(byte)1,(byte)2);
	    codec = codecFactory.create_codec(encoding);
	    any = orb.create_any();
	    org.omg.FT.FTRequestServiceContextHelper.insert(any,ctx_req);
	    serviceCtx_req.context_data = codec.encode_value(any);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}

	try{
	    ri.get_request_service_context(org.omg.IOP.FT_RTOBR.value);
	    ri.add_request_service_context(serviceCtx_req,true);
	    return;
	}catch(org.omg.CORBA.BAD_PARAM bp){System.out.println("No resend of request");}

	String ids[]=((org.omg.CORBA.portable.ObjectImpl)theTarget)._ids();
	for(i=0;i<ids.length;i++)
	    if(ids[i].equals("IDL:omg.org/FT/LoggingRecoveryController:1.0"))
		return;

	GatewayForActiveRepl gateway;
	java.lang.String gateway_feature=null;

	gateway_feature=(java.lang.String)((org.openorb.CORBA.ORB)orb).getFeature("gateway");
	if(gateway_feature==null) {gateway=null;}
	else
	    gateway=GatewayForActiveReplHelper.narrow(orb.string_to_object(gateway_feature));

	if(log_rec_ctrl==null)
	    set_logg_rec_c();

	gr_id=Utils.extract_group_id(orb,Utils.get_ior(theTarget));
	FTGroupIdServiceContext ctx_gi=new FTGroupIdServiceContext(gr_id);
	org.omg.IOP.ServiceContext serviceCtx_gi=new org.omg.IOP.ServiceContext();
	serviceCtx_gi.context_id = org.omg.IOP.FT_GROUP_ID.value;
	try{
	    any = orb.create_any();
	    org.omg.FT.FTGroupIdServiceContextHelper.insert(any,ctx_gi);
	    serviceCtx_gi.context_data = codec.encode_value(any);
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}

	if(gateway!=null){
	    for(i=0;i<ids.length;i++){
		System.out.println(ids[i]);
		if(ids[i].equals("IDL:omg.org/FT/GatewayForActiveRepl:1.0")){
		    if(oper.equals("invoke")){
			ri.add_request_service_context(serviceCtx_gi,true);

		    }
		    return;
		}
	    }

	    /* Check for reply*/
	    ResultResultArgs result_result_args;
	    org.omg.CORBA.Any reply=null,reply_args[]=new org.omg.CORBA.Any[0];
	    boolean error=false;
	    int tkvr=-1,tkvrc=Integer.MIN_VALUE;

	    if(log_rec_ctrl!=null){
		result_result_args=log_rec_ctrl.request_reply(client_id,ret_id,theTarget);
		try{
		    result_result_args.result.type().id();
		}catch(org.omg.CORBA.TypeCodePackage.BadKind bk){error=true;}
		if(!error){
		    if(ri.response_expected()){
			reply=result_result_args.result;
			reply_args=result_result_args.result_arguments;
			org.omg.CORBA_2_3.portable.OutputStream interm=(org.omg.CORBA_2_3.portable.OutputStream)orb.create_output_stream();
			if(reply!=null){
			    tkvr=reply.type().kind().value();
			    if(tkvr==org.omg.CORBA.TCKind._tk_TypeCode)
				tkvrc=reply.extract_TypeCode().kind().value();
			    if(tkvr!=org.omg.CORBA.TCKind._tk_TypeCode||tkvrc!=org.omg.CORBA.TCKind._tk_void)
				Utils.write_output_stream(reply,interm);
			    for(i=0;i<reply_args.length;i++)
				Utils.write_output_stream(reply_args[i],interm);
			    throw new org.omg.CORBA.REPLY_HERE(interm.create_input_stream());
			}
		    }
		}
	    }

	    org.omg.CORBA.Any result;
	    if(ids==null||ids.length<1) type_id="";
	    else
		type_id=ids[0];
	    org.omg.FT.ParametersHolder call_arguments=new org.omg.FT.ParametersHolder(((org.openorb.net.ClientRequest)ri).arguments());

	    last_retention_id=ret_id;
	    last_client_id=client_id;

	    result=gateway.invoke(theTarget,type_id,ret_id,client_id,req_start_time,oper,call_arguments,((org.openorb.net.ClientRequest)ri).exceptions(),ri.response_expected(),((org.openorb.net.ClientRequest)ri).result());

	    org.omg.CORBA_2_3.portable.OutputStream interm=(org.omg.CORBA_2_3.portable.OutputStream)orb.create_output_stream();
	    Utils.write_output_stream(result,interm);
	    org.omg.Dynamic.Parameter[] return_parameters=call_arguments.value;
	    for(i=0;i<return_parameters.length;i++)
		if(return_parameters[i].mode.value()==org.omg.CORBA.ParameterMode._PARAM_INOUT||
		   return_parameters[i].mode.value()==org.omg.CORBA.ParameterMode._PARAM_OUT){
		    Utils.write_output_stream(return_parameters[i].argument,interm);
		}
	    throw new org.omg.CORBA.REPLY_HERE(interm.create_input_stream());
	}


	ior = ((org.openorb.CORBA.Delegate)((org.omg.CORBA.portable.ObjectImpl)theTarget)._get_delegate()).ior();
	org.omg.IOP.TaggedProfile etp=null;

	/*extract the group version from any of the profiles*/

	gr_version=Utils.extract_group_version(orb,ior.profiles[0]);
	if(gr_version==-1) return;
	Utils.print_profile(orb,ri.effective_profile());

	org.omg.IOP.ServiceContext serviceCtx_gv = new org.omg.IOP.ServiceContext();
	serviceCtx_gv.context_id = org.omg.IOP.FT_GROUP_VERSION.value;

	FTGroupVersionServiceContext ctx_gv=new FTGroupVersionServiceContext(gr_version);

	try
	    {
		any = orb.create_any();
		org.omg.FT.FTGroupVersionServiceContextHelper.insert(any,ctx_gv);
		serviceCtx_gv.context_data = codec.encode_value(any);
	    }
	catch (java.lang.Exception e){System.out.println("Exception in clientInterceptor line 121");e.printStackTrace(new java.io.PrintStream(System.out));}
	ri.add_request_service_context(serviceCtx_gv,true);
	ri.add_request_service_context(serviceCtx_req,true);
	ri.add_request_service_context(serviceCtx_gi,true);
    }
}
