/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of one of the parties involved in the TRANSORG project. For
 * 	  written permission, please contact diasz@ida.liu.se.
 *
 * 4. Products derived from this Software may not be called "TRANSORG"
 *    nor may "TRANSORG" appear in their names without prior written
 *    permission.
 *
 * 5. Due credit should be given to the TRANSORG Project
 *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
 *
 * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
 *
 * $Id: FTARServerRequestInterceptor.java,v 1.0  diasz Exp $
 *
 * Date         Author  Changes
 */

package org.omg.FT;
/** Class instantiated when active replication is used and the corresponding profile is given
 *  It intercepts incoming requests to, an actively replicated group member
 * @author	<a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
 */
public class FTARServerRequestInterceptor extends org.omg.CORBA.LocalObject implements org.omg.PortableInterceptor.ServerRequestInterceptor{

    org.omg.PortableInterceptor.ORBInitInfo info;
    int ft_slot;
    String location;


    org.omg.CORBA.ORB orb;
    java.lang.Object upd_r_oper_synch,set_state_synch,wait_to_be_first,wait_tot_ord;
    boolean wait_for_set_state=false,get_state_arrived=false,object_deactivated=false;
    String ft_dom_name=null;

    int expected_id,nr_to_wait,nr_replies;
    java.util.Vector thread_vector_s_s; /* this is a vector which will enforce the order of thread
							     execution when waking up threads with notify*/
    MethSignStr[] method_table=new MethSignStr[0];
    LoggingRecoveryController log_rec_ctrl;
    org.omg.IOP.CodecFactory factory=null;
    org.omg.IOP.Codec codec=null;



    private class SynchObj{
	int nr_wait;

	public SynchObj(){
	    nr_wait=0;
	}
    }

    private class Record{
	int retention_id;
	String client_id;
	char meth_type;
	String oper;

	public Record(String c_i,int r_i,char m_type,String op){
	    retention_id=r_i;
	    client_id=c_i;
	    meth_type=m_type;
	    oper=op;
	}
    }


    private void wait_update_g_s_oper(){
	synchronized(upd_r_oper_synch){
	    int replies_in_bw,replies_in_aw;
	    int crt_to_wait=nr_to_wait;
	    nr_to_wait++;
	    while(crt_to_wait>0){
		try{
		    replies_in_bw=nr_replies;
		    upd_r_oper_synch.wait();
		    replies_in_aw=nr_replies;
		    crt_to_wait-=replies_in_aw-replies_in_bw;
		}catch(InterruptedException e){}

	    }
	}
    }


    private void notify_all(){
	synchronized(upd_r_oper_synch){
	    if(nr_to_wait>0)
		nr_to_wait--;
	    nr_replies++;
	    upd_r_oper_synch.notifyAll();
	}
    }



    private void wait_set_state(){
	synchronized(set_state_synch){
	    if(((org.openorb.CORBA.ORB)orb).getFeature("wait_set_state")!=null){
		try{
		    set_state_synch.wait();
		}catch(InterruptedException e){}
	    }
	}
    }

    private  void set_state_end(){
	synchronized(set_state_synch){
	    ((org.openorb.CORBA.ORB)orb).setFeature("wait_set_state",null);
	    set_state_synch.notifyAll();
	}
    }

    private class ThreadStructure{
	int retention_id,request_id;
	String client_id,oper;

	public ThreadStructure(int retention_id,String client_id,int request_id,String oper){
	    this.retention_id=retention_id;
	    this.client_id=client_id;
	    this.oper=oper;
	    this.request_id=request_id;
	}
    }


    private void set_meth_table(){
	org.omg.CORBA.Any any;
	if(method_table.length==0){
	    any=(org.omg.CORBA.Any)((org.openorb.PI.OpenORBInitInfo)info).orb().getFeature("methsign");
	    if(any!=null)
		method_table=MethSignStrsHelper.extract(any);
	}
    }

    private void unblock_all_oper_send_exceptions(){
	synchronized(upd_r_oper_synch){
	    upd_r_oper_synch.notifyAll();
	}
    }

    private void atomic_delivery(int tot_ord_req_id){
	synchronized(wait_tot_ord){
	    if(expected_id<tot_ord_req_id)
		while(expected_id<tot_ord_req_id)
		    try{
			wait_tot_ord.wait();
		    }catch(InterruptedException e){}
	    expected_id++;
	    wait_update_g_s_oper();
	    wait_tot_ord.notifyAll();
	}
    }

	/** Constructor
	  * @param	inf the ORB info used e.g. to obtain the instance of the ORB on which the CORBA object is running
	  *			ft_s the slot of this interceptor
	  */

    public FTARServerRequestInterceptor(org.omg.PortableInterceptor.ORBInitInfo inf, int ft_s)
    {
	info = inf;
	ft_slot = ft_s;
	try{
	    location = java.net.InetAddress.getLocalHost().getHostName();
	}catch(java.net.UnknownHostException e){System.out.println("Error in req interc. constr.");e.printStackTrace();location=null;}

	orb=((org.openorb.PI.OpenORBInitInfo)info).orb();
	upd_r_oper_synch=new java.lang.Object();
	set_state_synch=new java.lang.Object();
	wait_to_be_first=new java.lang.Object();
	wait_tot_ord=new java.lang.Object();
	thread_vector_s_s=new java.util.Vector();
	expected_id=0;
	nr_to_wait=0;
	nr_replies=0;
    }

    /** Method called from a specific interception point, when the active replica group member receives a request from the client (via the gateway)
     * Operations are performed if the called method is get_state, wait_set_state, activate/deactivate_object,etc. (mainly operations that are not allowed
     * to reach the application object, because it does not implement those methods
     * @param	ri the server request information
	 */
    public void receive_request_service_contexts(org.omg.PortableInterceptor.ServerRequestInfo ri) throws org.omg.PortableInterceptor.ForwardRequest
    {

	String oper;

	org.omg.CORBA.Any any=null;
	String client_id="",stringIOR;
	int retention_id=-1;
	org.omg.IOP.ServiceContext serviceCtx_req=null,serviceCtx_rtob=null;
	boolean ex_req=true,ex_rtob=true;
	org.omg.PortableServer.POA poa=null;
	byte[] objI=null;

	oper=ri.operation();
	if(oper.equals("set_state")||oper.equals("is_alive")) return;
	if(oper.equals("set_method_table")){
	    org.omg.CORBA.Any methods_wr;
	    methods_wr=((org.openorb.net.ServerRequest)ri).argument_stream().read_any();
	    method_table=org.omg.FT.MethSignStrsHelper.extract(methods_wr);
	    throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
	}
	if(oper.equals("install_gateway")){
	    stringIOR=(((org.openorb.net.ServerRequest)ri).argument_stream()).read_string();
	    ((org.openorb.CORBA.ORB)orb).setFeature("gateway",stringIOR);
	    ((org.openorb.net.ServerRequest)ri).createNoInvokeReply();
	    throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
	}

	if(oper.equals("deactivate_object"))
	    {
		if(!object_deactivated){
		    poa=(org.omg.PortableServer.POA)((org.openorb.CORBA.ORB)orb).getFeature("thePOA");
		    objI=(byte[])((org.openorb.CORBA.ORB)orb).getFeature("theOID");
		    if(poa!=null&&objI!=null){
			try{
			    poa.deactivate_object(objI);

			}catch(java.lang.Exception e){}
		    }
		    else
			throw new org.omg.CORBA.BAD_CONTEXT();
		}
		((org.openorb.net.ServerRequest)ri).createNoInvokeReply();
		throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
	    }

	if(oper.equals("get_state")&&((org.openorb.CORBA.ORB)orb).getFeature("wait_set_state")!=null){
	    synchronized(set_state_synch){
		get_state_arrived=true;
	    }
	    throw new org.omg.CORBA.BAD_CONTEXT();
	}

	if(oper.equals("FT_HB"))
	    if(info._get_policy(HEARTBEAT_ENABLED_POLICY.value)!=null){
		/*send reply*/
		return;
	    }

	synchronized(set_state_synch){
	    if(!get_state_arrived&&((org.openorb.CORBA.ORB)orb).getFeature("wait_set_state")!=null){
		throw new org.omg.CORBA.BAD_CONTEXT();
	    }
	}

	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	try{
	    serviceCtx_req=ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){System.out.println("No req service context");ex_req=false;}

	if(codec==null){System.out.println("Codec not properly initialized");return;}

	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(java.lang.Exception e){System.out.println(name()+" Error in req interc. cont.");e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	}
	wait_set_state();
    }

     /** Method called from a specific interception point, when the active replica group member receives a request from the client (via the gateway)
	     * Operations are performed mainly on incoming requests that have to be called on the application object as methods; for active replication
	     * these are dispatched in a total order
	     * @param	ri the server request information
	 */

    public void receive_request(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {
	org.omg.CORBA.Object or,ogr;
	org.omg.PortableServer.POA poa;
	int group_ref_v,k;
	String oper,m_der_i;
	String client_id="";
	long expiration_time=-1;
	int retention_id=-1;
	int repl_style;
	long c_kn_group_ref_v;
	boolean ex_req=true,ex_rtob=true;
	org.omg.CORBA.Any any=null;

	byte[] objI;

	oper=ri.operation();
	objI=ri.object_id();

	if(oper.equals("set_state")||oper.equals("is_alive")) return;

	org.omg.IOP.ServiceContext serviceCtx_ar = null,serviceCtx_req=null,serviceCtx_rtob=null;
	repl_style=-1;
	org.omg.FT.FTActiveReplAccessServiceContext arasc=null;

	if(oper.equals("get_state")){
	    wait_update_g_s_oper();
	    return;
	}
	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	try{
	    serviceCtx_ar = ri.get_request_service_context(org.omg.IOP.FT_GATEWAY_SOURCE.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){repl_style=-2;}
	try{
	    serviceCtx_req=ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){System.out.println("No req service context");ex_req=false;}

	if(codec==null){System.out.println("Codec not properly initialized");return;}
	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){System.out.println(name()+" Error in req interc. cont.");e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	    expiration_time=frctx.expiration_time;
	    System.out.println(name()+" "+client_id+" "+retention_id+" "+expiration_time);
	    if(System.currentTimeMillis()>expiration_time){
			throw new org.omg.CORBA.REQUEST_EXPIRED();
	    }
	}


	log_rec_ctrl=(LoggingRecoveryController)((org.openorb.CORBA.ORB)orb).getFeature("logging_recovery_ctrl");

	int reply_pos,i;

	org.omg.CORBA.Any reply=null,reply_args[]=new org.omg.CORBA.Any[0];
	int tkvr=-1,tkvrc=Integer.MIN_VALUE;
	ResultResultArgs result_result_args;
	boolean error=false;

	if(log_rec_ctrl!=null)
	   if(client_id!=null&&!client_id.equals("")&&retention_id!=-1){
	       System.out.println("Looking for an answer ....");
	       long objI_l=Utils.makeByteId(objI);
	       result_result_args=log_rec_ctrl.call_reply(objI_l,client_id,retention_id);
	       try{
		    result_result_args.result.type().id();
	       }catch(org.omg.CORBA.TypeCodePackage.BadKind bk){error=true;}
	       if(!error){
		   if(ri.response_expected()){
		       reply=result_result_args.result;
		       reply_args=result_result_args.result_arguments;
		       if(reply!=null){
			   org.omg.CORBA.portable.OutputStream os;
			    os=((org.openorb.net.ServerRequest)ri).createNoInvokeReply();
			    tkvr=reply.type().kind().value();
			    if(tkvr==org.omg.CORBA.TCKind._tk_TypeCode)
				tkvrc=reply.extract_TypeCode().kind().value();
			    if(tkvr!=org.omg.CORBA.TCKind._tk_TypeCode||tkvrc!=org.omg.CORBA.TCKind._tk_void)
				Utils.write_output_stream(reply,os);
			    for(i=0;i<reply_args.length;i++)
				Utils.write_output_stream(reply_args[i],os);
		       }
		   }
		  	   throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
	       }
	   }

	try{
	    serviceCtx_rtob=ri.get_request_service_context(org.omg.IOP.FT_RTOBR.value);
	}catch(org.omg.CORBA.BAD_PARAM bp){ex_rtob=false;}

	org.omg.FT.FTRelTotOrdBrServiceContext frtobctx=null;

	if(serviceCtx_rtob!=null&&ex_rtob){
	    try{
		any = codec.decode_value(serviceCtx_rtob.context_data, org.omg.FT.FTRelTotOrdBrServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    frtobctx = org.omg.FT.FTRelTotOrdBrServiceContextHelper.extract(any);
	}

	if(method_table==null)
	    set_meth_table();

	MethSignStr the_method=null;
	the_method=Utils.method_structure(oper,method_table);
	if(the_method.type!=org.omg.FT.Utils.READ)
	    if(frtobctx!=null){
		atomic_delivery(frtobctx.tot_ord_req_id);
	    }
	    }

    /** Method called from a specific interception point, when the activelly replicated object group member sends a reply to the client
     * @param	ri the server request information
     */
    public void send_reply(org.omg.PortableInterceptor.ServerRequestInfo ri)
    {


	org.omg.CORBA.Object or;
	org.omg.PortableServer.POA poa;
	org.omg.CORBA.Any any=null;
	org.omg.IOP.Encoding encoding;
	org.omg.IOP.ServiceContext serviceCtx_gv = null, serviceCtx_req=null;
	boolean ex_req=true;
	String oper,client_id="";
	byte[] objI;
	long expiration_time=0;
	int retention_id=-1;

	oper=ri.operation();
	objI=ri.object_id();

	if(oper.equals("is_alive")) {
	    org.omg.CORBA.Any res=((org.openorb.net.ServerRequest)ri).result();
	    boolean is_alive_r=res.extract_boolean();
	    if(!is_alive_r){
		poa=(org.omg.PortableServer.POA)((org.openorb.CORBA.ORB)orb).getFeature("thePOA");
		objI=(byte[])((org.openorb.CORBA.ORB)orb).getFeature("theOID");
		if(poa!=null&&objI!=null){
		    try{
			poa.deactivate_object(objI);
		    }catch(java.lang.Exception e){}
		    object_deactivated=true;
		}
		else
		    throw new org.omg.CORBA.BAD_CONTEXT();
	    }
	    return;
	}

	if(oper.equals("set_state")) {set_state_end();return;}
	if(oper.equals("get_state")){notify_all();return;}
	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	try{
	    serviceCtx_req=ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}


	if(codec==null){System.out.println("Codec not properly initialized");return;}
	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	    expiration_time=frctx.expiration_time;
	}


	log_rec_ctrl=(LoggingRecoveryController)((org.openorb.CORBA.ORB)orb).getFeature("logging_recovery_ctrl");
	org.omg.CORBA.Any res=((org.openorb.net.ServerRequest)ri).result(),res_args[]=((org.openorb.net.ServerRequest)ri).result_arguments();
	if(log_rec_ctrl!=null){
	    long objI_l=Utils.makeByteId(objI);
	    log_rec_ctrl.record_reply_stateless(objI_l,client_id,retention_id,expiration_time,res,res_args);
	}
	if(Utils.method_type(oper,method_table)!=Utils.READ)
	    notify_all();
	if(res==null) return;

	org.omg.IOP.ServiceContext serviceCtx_repl=new org.omg.IOP.ServiceContext();
	FTReplyServiceContext ctx_repl=new FTReplyServiceContext(res,res_args);

	try
	    {
		any = orb.create_any();
		org.omg.FT.FTReplyServiceContextHelper.insert(any,ctx_repl);
		serviceCtx_repl.context_data = codec.encode_value(any);
	    }
	catch (Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	ri.add_reply_service_context(serviceCtx_repl,true);
    }

    /** Method called from a specific interception point when the replica sends an exception to the client
     * @param	ri the server request information
     */
    public void send_exception(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {


	org.omg.CORBA.Any any=null;
	org.omg.IOP.ServiceContext  serviceCtx_req=null;
	boolean ex_req=true;
	String client_id="";
	int retention_id=-1;
	String oper=ri.operation();


	if(ri==null){
	    unblock_all_oper_send_exceptions();
	    return;
	}

	if(ri.operation().equals("get_state")){notify_all();return;}

	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	try{
	    serviceCtx_req=ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}

	if(codec==null){System.out.println("Codec not properly initialized");return;}

	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	    if(Utils.method_type(ri.operation(),method_table)!=Utils.READ)
		notify_all();
	}
    }

    /** Method called from a specific interception point when the replica sends something similar to an exception to the client
        * @param	ri the server request information
     */
    public void send_other(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {

	org.omg.CORBA.Any any=null;
	org.omg.IOP.ServiceContext  serviceCtx_req=null;
	boolean ex_req=true;
	String client_id="";
	int retention_id=-1;
	String oper=ri.operation();

	if(ri.operation().equals("get_state")){notify_all();return;}
	try{
	    serviceCtx_req=ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}
	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	if(codec==null){System.out.println("Codec not properly initialized");return;}

	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	    if(Utils.method_type(ri.operation(),method_table)!=Utils.READ)
		notify_all();
	}
    }


    public java.lang.String name()
    {
	return "FTARServerRequestInterceptor";
    }

}
