/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of one of the parties involved in the TRANSORG project. For
 * 	  written permission, please contact diasz@ida.liu.se.
 *
 * 4. Products derived from this Software may not be called "TRANSORG"
 *    nor may "TRANSORG" appear in their names without prior written
 *    permission.
 *
 * 5. Due credit should be given to the TRANSORG Project
 *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
 *
 * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
 *
 * $Id: FTCPRClientRequestInterceptor.java,v 1.0  diasz Exp $
 *
 * Date         Author  Changes
 */

package org.omg.FT;
/** Class instantiated when cold passive replication is used and the corresponding profile is given
 *  It intercepts outgoing requests from, as well as replies to requests of, a cold passively replicated group member
 * @author	<a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
 */
public class FTCPRClientRequestInterceptor extends org.omg.CORBA.LocalObject implements org.omg.PortableInterceptor.ClientRequestInterceptor{


    org.omg.PortableInterceptor.ORBInitInfo info;
    int ft_slot;
    org.omg.CORBA.ORB orb;
    String location;
    LoggingRecoveryController log_rec_ctrl=null;

/** Constructor
  * @param	inf the ORB info used e.g. to obtain the instance of the ORB on which the CORBA object is running
  *			ft_s the slot of this interceptor
  */


    public FTCPRClientRequestInterceptor(org.omg.PortableInterceptor.ORBInitInfo inf, int ft_s)
	{
	    info = inf;
	    ft_slot = ft_s;

	    java.util.Properties pr=System.getProperties();
	    orb=((org.openorb.PI.OpenORBInitInfo)info).orb();



	    try{
		location=java.net.InetAddress.getLocalHost().getHostName();
	    }catch(Exception e){location="localhost";}
	}



    public java.lang.String name()
    {
	return "FTPRClientRequestInterceptor";
    }

    private void set_logg_rec_c(){
	log_rec_ctrl=(LoggingRecoveryController)((org.openorb.CORBA.ORB)orb).getFeature("logging_recovery_ctrl");
    }


    public void receive_exception(org.omg.PortableInterceptor.ClientRequestInfo ri) {
	System.out.println("CPR Client: Receive exception "+ri.request_id()+" "+ri.received_exception_id());
    }

    public void receive_other(org.omg.PortableInterceptor.ClientRequestInfo ri) {
	System.out.println("CPR Client: Receive other "+ri.request_id());
    }

    /** Method called from a specific interception point when the cold passively replicated object group member receives a
    * normal reply to a request; an operation that is performed in the interceptor is loging of the outgoing request reply
    * @param	ri the client request information
    */

    public void receive_reply(org.omg.PortableInterceptor.ClientRequestInfo ri) {
	int ret_id;
	String cl_id;


	String oper=ri.operation();

	if(oper.equals("_is_a")) return;

	ret_id=((org.openorb.net.ClientRequest)ri).retention_id();
	cl_id=((org.openorb.net.ClientRequest)ri).client_id();



	if(log_rec_ctrl==null)
	    set_logg_rec_c();

	org.omg.IOP.ServiceContext serviceCtx_repl=null;
	boolean ex_repl=true;
	org.omg.IOP.CodecFactory factory;
	org.omg.IOP.Codec codec;
	org.omg.CORBA.Any any=null;

	if(log_rec_ctrl!=null){
	    try{
		serviceCtx_repl=ri.get_reply_service_context(org.omg.IOP.FT_REPLY.value);
	    }catch(java.lang.Exception bio){ex_repl=false;}

	    if(serviceCtx_repl!=null&&ex_repl){
		try{
		    factory = info.codec_factory();
		    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
		    any = codec.decode_value(serviceCtx_repl.context_data, org.omg.FT.FTReplyServiceContextHelper.type());
		}catch(Exception e){e.printStackTrace();return;}

		org.omg.FT.FTReplyServiceContext frctx = org.omg.FT.FTReplyServiceContextHelper.extract(any);
		log_rec_ctrl.record_request_repl(((org.openorb.net.ClientRequest)ri).client_id(),ri.target(),ri.operation(),frctx.result_arguments,ret_id,frctx.result);
	    }
	}

    }

    public void send_poll(org.omg.PortableInterceptor.ClientRequestInfo ri) {}


/** Method called when the cold passively replicated group member sends an outgoing request; different service contexts are added to the request
  * @param ri	the client request information
  */
    public void send_request(org.omg.PortableInterceptor.ClientRequestInfo ri){

	org.omg.IOP.IOR ior;
	int i,gr_version;
	long gr_id,req_start_time;
	int ret_id;
	String client_id;
	org.omg.IOP.CodecFactory codecFactory;
	org.omg.IOP.Encoding encoding;
	org.omg.IOP.Codec codec;
	org.omg.CORBA.Any any;
	String oper;

	org.omg.CORBA.Object theTarget=ri.target();

	String ids[]=((org.omg.CORBA.portable.ObjectImpl)theTarget)._ids();

	oper=ri.operation();



	if(oper.equals("_is_a")) {return;}

	for(i=0;i<ids.length;i++)
	    if(ids[i].equals("IDL:omg.org/FT/LoggingRecoveryController:1.0")||ids[i].equals("IDL:omg.org/FT/GroupReferenceManager:1.0"))
		return;



	ret_id=((org.openorb.net.ClientRequest)ri).retention_id();
	client_id=((org.openorb.net.ClientRequest)ri).client_id();
	req_start_time=((org.openorb.net.ClientRequest)ri).request_start_time();


	/* Check for reply*/
	if(log_rec_ctrl==null)
	    set_logg_rec_c();


	ResultResultArgs result_result_args;
	org.omg.CORBA.Any reply=null,reply_args[]=new org.omg.CORBA.Any[0];
	boolean error=false;
	int tkvr=-1,tkvrc=Integer.MIN_VALUE;

	if(log_rec_ctrl!=null){
	    result_result_args=log_rec_ctrl.request_reply(client_id,ret_id,theTarget);
	    try{
		result_result_args.result.type().id();
	    }catch(org.omg.CORBA.TypeCodePackage.BadKind bk){error=true;}
	    if(!error){
		org.omg.CORBA_2_3.portable.OutputStream interm=(org.omg.CORBA_2_3.portable.OutputStream)orb.create_output_stream();
		if(ri.response_expected()){
		    reply=result_result_args.result;
		    reply_args=result_result_args.result_arguments;
		    if(reply!=null){

			tkvr=reply.type().kind().value();
			if(tkvr==org.omg.CORBA.TCKind._tk_TypeCode)
			    tkvrc=reply.extract_TypeCode().kind().value();
			if(tkvr!=org.omg.CORBA.TCKind._tk_TypeCode||tkvrc!=org.omg.CORBA.TCKind._tk_void)
			    Utils.write_output_stream(reply,interm);
			for(i=0;i<reply_args.length;i++)
			    Utils.write_output_stream(reply_args[i],interm);
		    }
		}
		throw new org.omg.CORBA.REPLY_HERE(interm.create_input_stream());
	    }
	}

	ior = ((org.openorb.CORBA.Delegate)((org.omg.CORBA.portable.ObjectImpl)theTarget)._get_delegate()).ior();
	org.omg.IOP.TaggedProfile etp=null;

	/*extract the group version from any of the profiles*/

	long[] gr_v_id;

	gr_v_id=Utils.extract_group_vers_and_id(orb,ior.profiles[0]);

	if(gr_v_id.length<2) {return;}

	gr_version=(int)gr_v_id[0];
	gr_id=gr_v_id[1];

	org.omg.IOP.ServiceContext serviceCtx_gv = new org.omg.IOP.ServiceContext(), serviceCtx_req = new org.omg.IOP.ServiceContext(), serviceCtx_gi=new org.omg.IOP.ServiceContext();

	serviceCtx_gv.context_id = org.omg.IOP.FT_GROUP_VERSION.value;
	serviceCtx_req.context_id =  org.omg.IOP.FT_REQUEST.value;
	serviceCtx_gi.context_id = org.omg.IOP.FT_GROUP_ID.value;

	FTGroupVersionServiceContext ctx_gv=new FTGroupVersionServiceContext(gr_version);
	FTGroupIdServiceContext ctx_gi=new FTGroupIdServiceContext(gr_id);
	long req_dur=DEFAULT_REQUEST_DURATION.value;
	/* Calculate here the expiration time*/
	Integer the_dur=(Integer)((org.openorb.CORBA.ORB)orb).getFeature("request_duration");

	if(the_dur!=null)
	    req_dur=the_dur.intValue();
	FTRequestServiceContext ctx_req=new FTRequestServiceContext(client_id,ret_id,req_start_time+req_dur,location);

	try
	    {
		codecFactory = info.codec_factory();
		encoding = new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value,(byte)1,(byte)2);
		codec = codecFactory.create_codec(encoding);
		any = orb.create_any();
		org.omg.FT.FTGroupVersionServiceContextHelper.insert(any,ctx_gv);
		serviceCtx_gv.context_data = codec.encode_value(any);
		any = orb.create_any();
		org.omg.FT.FTRequestServiceContextHelper.insert(any,ctx_req);
		serviceCtx_req.context_data = codec.encode_value(any);
		any = orb.create_any();
		org.omg.FT.FTGroupIdServiceContextHelper.insert(any,ctx_gi);
		serviceCtx_gi.context_data=codec.encode_value(any);
	    }
	catch (Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	ri.add_request_service_context(serviceCtx_gv,true);
	ri.add_request_service_context(serviceCtx_req,true);
	ri.add_request_service_context(serviceCtx_gi,true);
    }
}
