/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of one of the parties involved in the TRANSORG project. For
 * 	  written permission, please contact diasz@ida.liu.se.
 *
 * 4. Products derived from this Software may not be called "TRANSORG"
 *    nor may "TRANSORG" appear in their names without prior written
 *    permission.
 *
 * 5. Due credit should be given to the TRANSORG Project
 *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
 *
 * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
 *
 * $Id: FTGatewayServerRequestInterceptor.java,v 1.0  diasz Exp $
 *
 * Date         Author  Changes
 */

package org.omg.FT;
/** Class instantiated when active replication is used and a gateway is needed; this class has the role to intercept incoming requests from clients
* to be broadcast to active replica group members; also, it intercepts requests received from actively replicated server group members, and for these
* it has the role to eliminate the duplicates, and to send the answer to all calling replicas
* @author	<a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
* @version $Revision: 1.0 $ $Date:2003/06/23$
*/
public class FTGatewayServerRequestInterceptor extends org.omg.CORBA.LocalObject implements org.omg.PortableInterceptor.ServerRequestInterceptor{

    org.omg.PortableInterceptor.ORBInitInfo info;
    int ft_slot;
    String location;
    org.omg.CORBA.ORB orb;
    String ft_dom_name=null;
    SynchObj send_th_synch,stop_new_calls,ann_come,waiting_to_send,method_wait,change_wait;
    boolean memb_failed,in_change_membership=false;
    java.util.Vector request_evid,result_evid,call_log,group_length_evid,m_ch_list;
    GatewayForActiveRepl_impl its_gateway=null;
    java.lang.Object membership_changed=new java.lang.Object(),tot_ord_r_i_guard=new java.lang.Object();
    int tot_ord_req_id=-1;
    MethSignStr[] method_table;
    org.omg.IOP.CodecFactory factory=null;
    org.omg.IOP.Codec codec=null;


    private class GroupLStr{
	int length;
	public GroupLStr(int l){
	    length=l;
	}
    }

    private class SynchObj{
	int nr_wait;
	boolean ready;
	int notifier_index;

	public SynchObj(){
	    nr_wait=0;
	    ready=false;
	    notifier_index=-1;
	}
    }

    private class RITargetStr{
	org.omg.CORBA.Object target;
	int retention_id;
	SynchObj wait_res;

	public RITargetStr(org.omg.CORBA.Object target,int retention_id){
	    this.target=target;
	    this.retention_id=retention_id;
	    wait_res=new SynchObj();
	}
    }

    private class RIRTargetStr{
	org.omg.CORBA.Object target;
	int retention_id;
	org.omg.CORBA.Any result;
	org.omg.Dynamic.Parameter[] call_arguments;
	int nr_after_to_del;


	public RIRTargetStr(org.omg.CORBA.Object target,int retention_id,org.omg.CORBA.Any result,org.omg.Dynamic.Parameter[] call_arguments,int nr_after_to_del){
	    this.target=target;
	    this.retention_id=retention_id;
	    this.result=result;
	    this.call_arguments=call_arguments;
	    this.nr_after_to_del=nr_after_to_del;
	}
    }

    private class DynCallStr{
	String oper;
	org.omg.CORBA.NVList arguments;
	org.omg.CORBA.NamedValue result;
	org.omg.CORBA.ExceptionList exceptions;
	int nr_out_inout_r;

	int ret_id;
	String cl_id;
	long exp_time;

	public DynCallStr(String oper,int ret_id,String cl_id,long exp_time,org.omg.CORBA.NVList arguments,org.omg.CORBA.NamedValue result,org.omg.CORBA.ExceptionList exceptions,int nr_out_inout_r)
	{
	    this.oper=oper;
	    this.arguments=arguments;
	    this.result=result;
	    this.exceptions=exceptions;
	    this.nr_out_inout_r=nr_out_inout_r;
	    this.ret_id=ret_id;
	    this.cl_id=cl_id;
	    this.exp_time=exp_time;
	}
    }


/** Class instantiated when a request has to be broadcast to all members of the active group that the gateway intermediates for*/
    private class SendRequestThread extends Thread{
	SynchObj one_done;
	org.openorb.CORBA.dii.Request dyn_request;
	org.omg.CORBA.AnyHolder result_holder,exception_holder;
	org.omg.CORBA.AnyHolder[] out_inout_result_holders;
	boolean[] error_table;
	int index;
	char meth_type;




	public SendRequestThread(SynchObj one_done,org.openorb.CORBA.dii.Request dyn_request,org.omg.CORBA.AnyHolder result_holder,org.omg.CORBA.AnyHolder[] out_inout_result_holders,org.omg.CORBA.AnyHolder exception_holder,boolean[] error_table,int index,char meth_type){
	    this.one_done=one_done;
	    this.dyn_request=dyn_request;
	    this.result_holder=result_holder;
	    this.exception_holder=exception_holder;
	    this.out_inout_result_holders=out_inout_result_holders;
	    this.error_table=error_table;
	    this.index=index;
	    this.meth_type=meth_type;
	}


	public void run(){

	    int i,l,flags,j;
	    org.omg.CORBA.NVList all_arguments;
	    org.omg.CORBA.NamedValue item;
	    java.lang.Exception exception;
	    boolean error=false;
	    try{
		if(meth_type==Utils.UPDATE_ONEWAY)
		    dyn_request.send_oneway();
		else
		    dyn_request.invoke();
		exception=dyn_request.env().exception();
		if(exception!=null){
		    if(exception instanceof org.omg.CORBA.UnknownUserException)
			exception_holder.value=((org.omg.CORBA.UnknownUserException)exception).except;
		    else
			if(exception instanceof org.omg.CORBA.UNKNOWN)
			    org.omg.CORBA.UNKNOWNHelper.insert(exception_holder.value,(org.omg.CORBA.UNKNOWN)exception);
			else
			    if(exception instanceof org.omg.CORBA.SystemException)
				org.openorb.CORBA.SystemExceptionHelper.insert(exception_holder.value,(org.omg.CORBA.SystemException)exception);
			    else
				exception_holder.value.type(orb.get_primitive_tc(org.omg.CORBA.TCKind.tk_string));
		    error_table[index]=true;
		}
		else{
		    result_holder.value=dyn_request.result().value();
		    all_arguments=dyn_request.arguments();
		    if(all_arguments!=null&&out_inout_result_holders.length>0){
			l=all_arguments.count();
			j=0;
			for(i=0;i<l;i++){
			    item=all_arguments.item(i);
			    flags=item.flags();
			    if(flags==org.omg.CORBA.ARG_OUT.value||flags==org.omg.CORBA.ARG_INOUT.value){
				out_inout_result_holders[j]=new org.omg.CORBA.AnyHolder(item.value());
				j++;
			    }
			}
		    }
		}
	    }catch(java.lang.Exception e){
		if(e instanceof org.omg.CORBA.SystemException){
		    if(exception_holder.value==null){ exception_holder.value=orb.create_any();}
		    org.openorb.CORBA.SystemExceptionHelper.insert(exception_holder.value,(org.omg.CORBA.SystemException)e);
		}

		e.printStackTrace(new java.io.PrintStream(System.out));
		synchronized(error_table){
		    error_table[index]=true;
		}
	    }
	    if(one_done!=null&&(!error_table[index]||all_true(error_table))){
		notify_end(one_done,index);
	    }
	}
    }


/** Class instantiated when the state has to be read from one of the replicas of the active group in order to be transfered to a new
* joining replica*/

    private class GetStateThread extends Thread{
	SynchObj state_here;
	org.omg.CORBA.AbstractBaseHolder state_res_holder;
	boolean[] error_table;
	int index;
	java.lang.Object member;


	    public GetStateThread(java.lang.Object member,SynchObj state_here,org.omg.CORBA.AbstractBaseHolder state_res_holder,boolean[] error_table,int index){
	    this.state_here=state_here;
	    this.member=member;
	    this.state_res_holder=state_res_holder;
	    this.error_table=error_table;
	    this.index=index;
	}


	public void run(){

	   java.lang.Object obj_res;

	    boolean error=false;
	    try{
		obj_res=Class.forName(member.getClass().getName()).getMethod("get_state",new Class[0]).invoke(member,new java.lang.Object[0]);
		state_res_holder.value=obj_res;
		if(obj_res==null||((byte[])obj_res).length==0)
		    error_table[index]=true;
	    }catch(java.lang.Exception e){
		e.printStackTrace(new java.io.PrintStream(System.out));
		synchronized(error_table){
		    error_table[index]=true;
		}
	    }
	    if(state_here!=null&&(!error_table[index]||all_true(error_table))){
		notify_end(state_here,index);
	    }
	}
    }




    private int next_tot_ord_id(){
	synchronized(tot_ord_r_i_guard){
	    tot_ord_req_id++;
	    return tot_ord_req_id;
	}
    }

    private boolean all_true(boolean table[]){
	int i;
	for(i=0;i<table.length;i++)
	    if(!table[i]) return false;
	return true;
    }

    private void set_all_false(boolean table[]){
	int i;
	for(i=0;i<table.length;i++)
	    table[i]=false;
    }

    private void notify_end(SynchObj s_obj,int n_index){
	synchronized(s_obj){
	    if(s_obj.ready) return;
	    s_obj.notifier_index=n_index;
	    s_obj.ready=true;
	    while(s_obj.nr_wait>0){
		s_obj.notify();
		s_obj.nr_wait--;
	    }
	}
    }

    private void block_until_first_ready(SynchObj s_obj){
	synchronized(s_obj){
	    if(!s_obj.ready){
		try{
		    s_obj.nr_wait++;
		    s_obj.wait();
		}catch(InterruptedException e){}
	    }
	}
    }

    private boolean is_first_ready(SynchObj s_obj){
	synchronized(s_obj){
	    return s_obj.ready;
	}
    }

    private void set_not_first_ready(SynchObj s_obj){
	synchronized(s_obj){
	    s_obj.nr_wait=0;
	    s_obj.ready=false;
	}
    }



    private void block_knows_failed(){
	synchronized(ann_come){
	    while(!memb_failed){
		ann_come.nr_wait++;
		try{
		    ann_come.wait();
		}catch(InterruptedException e){}
	    }
	    if(ann_come.nr_wait<=0) memb_failed=false;
	}
    }


    private void knows_failed(){
	synchronized(ann_come){
	    memb_failed=true;
	    while(ann_come.nr_wait>0){
		ann_come.notify();
		ann_come.nr_wait--;
	    }
	}
    }

    private void block_until_change_membership(){
	synchronized(waiting_to_send){
	    if(in_change_membership){
		waiting_to_send.nr_wait++;
		try{
		    waiting_to_send.wait();
		}catch(InterruptedException e){}
	    }
	}
    }

    private void notify_membership_changed(){
	synchronized(waiting_to_send){
	    in_change_membership=false;
	    if(waiting_to_send.nr_wait>0){
		while(waiting_to_send.nr_wait>0){
		    waiting_to_send.notify();
		    waiting_to_send.nr_wait--;
		}
	    }
	}
    }


    private org.omg.CORBA.Object[] give_dest(int retention_id,String client_id,String oper){
	org.omg.CORBA.Object[] group_members;

	block_until_change_membership();
	group_members=its_gateway.group_members();
	if(group_members.length==0) {throw new org.omg.CORBA.TRANSIENT(org.omg.CORBA.OMGVMCID.value | 3, org.omg.CORBA.CompletionStatus.COMPLETED_NO);}

	return group_members;
    }

    private int meth_pos(String m_name,MethSignStr[] methods){
	int i;

	for(i=0;i<methods.length;i++)
	    if(methods[i].name.equals(m_name))
		return i;
	return -1;
    }


    private SynchObj sent_one(int retention_id,org.omg.CORBA.Object target){
	int i,l;
	RITargetStr crt;
	synchronized(request_evid){
	    l=request_evid.size();
	    for(i=0;i<l;i++){
		crt=(RITargetStr)request_evid.elementAt(i);
		if(Utils.refs_equal(orb,crt.target,target,1)&&retention_id==crt.retention_id)
		    return crt.wait_res;
	    }
	    request_evid.addElement(new RITargetStr(target,retention_id));
	    return null;
	}
    }

    private void all_arrived(int retention_id,org.omg.CORBA.Object target){
	int i,l;
	RITargetStr crt;
	boolean found=false;
	synchronized(request_evid){
	    l=request_evid.size();
	    for(i=0;i<l;i++){
		crt=(RITargetStr)request_evid.elementAt(i);
		if(Utils.refs_equal(orb,crt.target,target,1)&&retention_id==crt.retention_id){
		    found=true;
		    break;
		}
	    }
	    if(found)
		request_evid.remove(i);
	}
    }

    private RIRTargetStr result_str_here(int ret_id,org.omg.CORBA.Object target){
	RIRTargetStr crt;
	int i,l;
	synchronized(result_evid){
	    l=result_evid.size();
	    for(i=0;i<l;i++){
		crt=(RIRTargetStr)result_evid.elementAt(i);
		if(crt.retention_id==ret_id&&Utils.refs_equal(orb,crt.target,target,1)){
		    crt.nr_after_to_del--;
		    if(crt.nr_after_to_del==0) {
			result_evid.remove(i);
			all_arrived(ret_id,target);
		    }
		    return crt;
		}
	    }
	    return null;
	}
    }

    private void insert_result_str(int ret_id,org.omg.CORBA.Object target,org.omg.CORBA.Any result,org.omg.Dynamic.Parameter[] call_args,int nr_after_to_del){
	SynchObj to_notify;
	synchronized(result_evid){
	    to_notify=sent_one(ret_id,target);
	    result_evid.addElement(new RIRTargetStr(target,ret_id,result,call_args,nr_after_to_del));
	    if(to_notify==null) return;
	    synchronized(to_notify){
		to_notify.notifyAll();
	    }
	}
    }

    private void simulate_stub(org.omg.CORBA.portable.ObjectImpl target,int ret_id,String client_id,long expiration_time,String oper,org.omg.Dynamic.Parameter[] arguments,org.omg.CORBA.TypeCode[] exceptions,boolean resp_exp,org.omg.CORBA.Any result) throws org.omg.CORBA.portable.ApplicationException{
	int i;

	while( true )
	    {
		org.omg.CORBA.portable.InputStream _input = null;
		try
		    {
			org.omg.CORBA.portable.OutputStream _output=null;
			_output = target._request(ret_id,client_id,expiration_time,oper,resp_exp);

			for(i=0;i<arguments.length;i++)
			    if(arguments[i].mode.value()!=org.omg.CORBA.ParameterMode._PARAM_OUT){

				Utils.write_output_stream(arguments[i].argument,_output);
			    }

			_input = target._invoke(_output);
			Utils.read_input_stream(_input,result,true);
			for(i=0;i<arguments.length;i++)
			    switch(arguments[i].mode.value()){
			    case org.omg.CORBA.ParameterMode._PARAM_INOUT:
				Utils.read_input_stream(_input,arguments[i].argument,false);
				break;
			    case org.omg.CORBA.ParameterMode._PARAM_OUT:
				Utils.read_input_stream(_input,arguments[i].argument,true);
				break;
			    case org.omg.CORBA.ParameterMode._PARAM_IN:
				arguments[i].argument.insert_string("_PARAM_IN");
				break;
			    }
			return;
		    }
		catch( org.omg.CORBA.portable.RemarshalException _exception )
		    {
			continue;
		    }
		catch( org.omg.CORBA.portable.ApplicationException _exception )
		    {
			String _exception_id = _exception.getId();
			java.lang.Object params[];
			String type_id=null;
			for(i=0;i<exceptions.length;i++){
			    try{
				type_id=exceptions[i].id();
			    }catch(Exception e){type_id=null;}
			    if(_exception_id.equals(Utils.get_class_name(type_id))){
				    params=new java.lang.Object[1];
				    params[0]=_exception.getInputStream();
				    Utils.invoke_helper_method("read",_exception_id,params);
				    throw _exception;
			    }
			}
			throw new org.omg.CORBA.UNKNOWN("Unexcepected User Exception: "+ _exception_id);
		    }
		finally
		    {
			target._releaseReply(_input);
		    }
	    }
    }



    public FTGatewayServerRequestInterceptor(org.omg.PortableInterceptor.ORBInitInfo inf, int ft_s)
    {
	info = inf;
	ft_slot = ft_s;

	try{
	    location = java.net.InetAddress.getLocalHost().getHostName();
	}catch(java.net.UnknownHostException e){e.printStackTrace(new java.io.PrintStream(System.out));location=null;}
	send_th_synch=new SynchObj();
	stop_new_calls=new SynchObj();
	ann_come=new SynchObj();
	waiting_to_send=new SynchObj();
	memb_failed=false;
	request_evid=new java.util.Vector();
	result_evid=new java.util.Vector();
	call_log=new java.util.Vector();
	group_length_evid=new java.util.Vector();
	orb=((org.openorb.PI.OpenORBInitInfo)info).orb();

    }


/** Method called from a specific interception point; it has the role to stop the method call "set_method_table" to reach the server object (the gateway)
* @param	ri server request information
*/

    public void receive_request_service_contexts(org.omg.PortableInterceptor.ServerRequestInfo ri) throws org.omg.PortableInterceptor.ForwardRequest
    {

	String oper;

	oper=ri.operation();


	if(oper.equals("set_method_table")){
	    org.omg.CORBA.Any methods_wr;
	    methods_wr=((org.openorb.net.ServerRequest)ri).argument_stream().read_any();
	    method_table=org.omg.FT.MethSignStrsHelper.extract(methods_wr);
	    throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
	}

	if(oper.equals("FT_HB"))
	    if(info._get_policy(HEARTBEAT_ENABLED_POLICY.value)!=null){
		/*send reply*/
		return;
	    }
    }

/** Method called from a specific interception point; it has the role to stop calls from clients to active group memebers to reach the gateway
* object; instead these are either broadcast to active server group replicas - if they come from the "true" clients, or after duplicate supression they
* are sent to the sercer for which the active group is a client
* @param ri the server request information
*/
    public void receive_request(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {

	java.lang.Object[] real_params;
	org.omg.CORBA.Any result=orb.create_any(),argument;
	SynchObj one_done=new SynchObj(),state_here=new SynchObj();
	int start_member=0,retention_id=-1,i;
	org.omg.IOP.ServiceContext serviceCtx_rg=null,serviceCtx_req=null;
	boolean ex_req;
	long expiration_time=0;

	org.omg.CORBA.Any any=null;
	byte[] ret_id;
	boolean[] error_table;
	RIRTargetStr result_str;

	String type_id,client_id="";
	Class par_cl_rm[],par_cl[];
	java.lang.Object params[],member,obj_res;
	org.omg.CORBA.AbstractBaseHolder state_res_holder_t[];


	String oper=ri.operation();
	if(oper.equals("push_structured_event")){
	    return;
	}
	if(oper.equals("init_membership"))
	    return;

	if(oper.equals("change_membership"))
	    {
		block_knows_failed();
		synchronized(waiting_to_send){
		    in_change_membership=true;
		}
		return;
	    }

	its_gateway=(GatewayForActiveRepl_impl)((org.openorb.CORBA.ORB)orb).getFeature("ItsGateway");
	int flags=-1,l;
	org.omg.CORBA.Object[] group_members,old_group_members,new_group_members;

	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	ex_req=true;
	try{
	    serviceCtx_req=ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}

	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	    expiration_time=frctx.expiration_time;
	    if(System.currentTimeMillis()>expiration_time){
		throw new org.omg.CORBA.REQUEST_EXPIRED();
	    }
	}

	/* Operation get_state is called on the gateway, USUALLY, when the replication manager creates a new replica
	   for an actively replicated group; at this moment the state of a working replica is read and set with "set_state"
	   on the new replica*/

	if(oper.equals("get_state")){
	    boolean ex_rg=true;
	    try{
		serviceCtx_rg = ri.get_request_service_context(org.omg.IOP.FT_RM_TO_GATEWAY.value);
	    }
	    catch (org.omg.CORBA.BAD_PARAM bp){ex_rg=false;}
	    if(serviceCtx_rg!=null&&ex_rg){
		try{
		    if(its_gateway==null){System.out.println("No gateway available"); throw new org.omg.CORBA.TRANSIENT(org.omg.CORBA.OMGVMCID.value | 3, org.omg.CORBA.CompletionStatus.COMPLETED_NO);}
		    old_group_members=its_gateway.old_group_members();
		    new_group_members=its_gateway.new_group_members();
		    if(old_group_members==null||old_group_members.length==0) {throw new org.omg.CORBA.TRANSIENT(org.omg.CORBA.OMGVMCID.value | 3, org.omg.CORBA.CompletionStatus.COMPLETED_NO);}
		    type_id=Utils.type_id(old_group_members[0]);
		    state_res_holder_t=new org.omg.CORBA.AbstractBaseHolder[1+new_group_members.length];
		    group_members=new org.omg.CORBA.Object[1+new_group_members.length];
		    group_members[0]=old_group_members[0];
		    for(i=1;i<1+new_group_members.length;i++)
			group_members[i]=new_group_members[i-1];

		    set_not_first_ready(state_here);
		    error_table=new boolean[group_members.length];
		    set_all_false(error_table);
		    for(i=0;i<group_members.length;i++){
			state_res_holder_t[i]=new org.omg.CORBA.AbstractBaseHolder();
			state_res_holder_t[i].value=null;
			member=Utils.narrow(group_members[i],type_id);
			new GetStateThread(member,state_here,state_res_holder_t[i],error_table,i).start();
		    }
		    block_until_first_ready(state_here);

		    byte[] fake_state;
		    obj_res=state_res_holder_t[state_here.notifier_index].value;
		    if(obj_res!=null){
			fake_state=new byte[4+((byte[])obj_res).length];
			ret_id=Utils.makeIntByte(((org.openorb.CORBA.ORB)orb).current_retention_id());
			for(i=0;i<4;i++)
			    fake_state[i]=ret_id[i];
			for(;i<4+((byte[])obj_res).length;i++)
			    fake_state[i]=((byte[])obj_res)[i-4];
		    }
		    else
			fake_state=new byte[0];

		    org.omg.CORBA.portable.OutputStream os;
		    os=((org.openorb.net.ServerRequest)ri).createNoInvokeReply();
		    org.omg.FT.StateHelper.write(os,fake_state);
		}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	    }
	    else{
		org.omg.CORBA.portable.OutputStream os;
		os=((org.openorb.net.ServerRequest)ri).createNoInvokeReply();
		org.omg.FT.StateHelper.write(os,new byte[0]);
	    }
	    throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
	}
	else
	    /* Gateway letting CALLS OUT from the ACTIVELY replicated group*/
	    if(oper.equals("invoke")){
		try{
		    int nr_to_send; /* how many invoke method invocations do I expect for a certain outgoing request, i.e. how many members are there
				       in the group*/
		    org.omg.Dynamic.Parameter[] call_arguments;
		    org.omg.CORBA.portable.InputStream argument_stream=((org.openorb.net.ServerRequest)ri).argument_stream();
		    org.omg.CORBA.Object target=argument_stream.read_Object();
		    type_id=argument_stream.read_string();
		    retention_id=argument_stream.read_long();
		    client_id=argument_stream.read_string();
		    expiration_time=argument_stream.read_longlong();
		    String oper_to_call=argument_stream.read_string();
		    call_arguments=ParametersHelper.read(argument_stream);
		    SynchObj wait_res=sent_one(retention_id,target);
		    ((org.openorb.CORBA.ORB)orb).retention_id(retention_id);
		    if(wait_res==null){

			org.omg.CORBA.TypeCode[] exceptions=TypeCodeSeqHelper.read(argument_stream);
			boolean resp_exp=argument_stream.read_boolean();
			result=argument_stream.read_any();
			try{
			    java.lang.Object real_target=Utils.narrow(target,type_id);
			    simulate_stub((org.omg.CORBA.portable.ObjectImpl)real_target,retention_id,client_id,expiration_time,oper_to_call,call_arguments,exceptions,resp_exp,result);
			}catch(org.omg.CORBA.portable.ApplicationException ae){ae.printStackTrace(new java.io.PrintStream(System.out));}
			insert_result_str(retention_id,target,result,call_arguments,its_gateway.group_members().length-1);
			/*if(nr_to_send==0)
			  remove_result_str(retention_id,target);*/
		    }
		    else{

			while(true){
			    result_str=result_str_here(retention_id,target);
			    if(result_str==null){
				wait_res.nr_wait++;
				synchronized(wait_res){
				    try{
					wait_res.wait();
				    }catch(InterruptedException e){}
				}
			    }
			    else
				break;
			}
			if(result_str!=null){
			    result=result_str.result;
			    call_arguments=result_str.call_arguments;
			}
		    }
		    org.omg.CORBA.portable.OutputStream os;
		    os=((org.openorb.net.ServerRequest)ri).createNoInvokeReply();
		    os.write_any(result);
		    ParametersHelper.write(os,call_arguments);
		}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
		throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
	    }

	    else{ /* Gateway letting CALLS IN to the ACTIVELY replicated group*/
		boolean success=true;
		int local_tot_ord_r_i=-1;
		if(method_table==null){
		    org.omg.CORBA.Any m_sign_wr=(org.omg.CORBA.Any)((org.openorb.PI.OpenORBInitInfo)info).orb().getFeature("methsign");
		    if(m_sign_wr==null) {return;}
		    method_table=MethSignStrsHelper.extract(m_sign_wr);
		}
		int meth_pos=meth_pos(oper,method_table);
		if(meth_pos==-1) {return;}
		org.omg.CORBA.TypeCode[] param_type_codes=method_table[meth_pos].param_types,except_type_codes=method_table[meth_pos].except_types;
		int[] modes=method_table[meth_pos].param_modes;
		char meth_type=method_table[meth_pos].type;
		org.omg.CORBA.AnyHolder result_holder_t[],exception_holder_t[],out_inout_result_holders_t[][];
		org.openorb.CORBA.dii.Request dyn_request;
		org.omg.CORBA.NVList dyn_arguments=orb.create_list(0);
		org.omg.CORBA.ExceptionList dyn_exceptions=orb.create_exception_list();
		org.omg.CORBA.NamedValue dyn_result;
		int out_inout_results=0;
		org.omg.CORBA.portable.InputStream argument_stream=((org.openorb.net.ServerRequest)ri).argument_stream();
		result.type(param_type_codes[param_type_codes.length-1]);
		dyn_result=orb.create_named_value("result",result,-1);
		exception_holder_t=new org.omg.CORBA.AnyHolder[0];
		result_holder_t=new org.omg.CORBA.AnyHolder[0];
		out_inout_result_holders_t=new org.omg.CORBA.AnyHolder[0][];
		/* modes (arguments modes) are given for the parameters of the method, not the result)
		   - the number of mdoes is one less the number of parameters;
		   - e.g., if there are no call parameters then there are "no modes"*/
		for(i=0;i<modes.length;i++){
		    argument=orb.create_any();
		    argument.type(param_type_codes[i]);
		    if(modes[i]==org.omg.CORBA.ParameterMode._PARAM_IN||modes[i]==org.omg.CORBA.ParameterMode._PARAM_INOUT){
			Utils.read_input_stream(argument_stream,argument,false);
		    }
		    switch(modes[i]){
		    case org.omg.CORBA.ParameterMode._PARAM_IN:  flags=org.omg.CORBA.ARG_IN.value;break;
		    case org.omg.CORBA.ParameterMode._PARAM_INOUT:  flags=org.omg.CORBA.ARG_INOUT.value;out_inout_results++;break;
		    case org.omg.CORBA.ParameterMode._PARAM_OUT:  flags=org.omg.CORBA.ARG_OUT.value;out_inout_results++;break;
		    }
		    dyn_arguments.add_value("param"+i,argument,flags);
		}

		for(i=0;i<except_type_codes.length;i++)
		    dyn_exceptions.add(except_type_codes[i]);

		try{

		    org.omg.IOP.ServiceContext serviceCtx_rtob=null;
		    FTRelTotOrdBrServiceContext ctx_rtob=null;
		    try{
			group_members=give_dest(retention_id,client_id,oper);
		    }catch(org.omg.CORBA.TRANSIENT tr){
			throw tr;
		    }

		    if(meth_type!=org.omg.FT.Utils.READ){
			/* only the update operations have to be executed in the same total order at every replica*/
			local_tot_ord_r_i=next_tot_ord_id();

			/* only the update operations have to be executed in the same total order at every replica*/
			serviceCtx_rtob=new org.omg.IOP.ServiceContext();
			serviceCtx_rtob.context_id=org.omg.IOP.FT_RTOBR.value;
			ctx_rtob=new FTRelTotOrdBrServiceContext(local_tot_ord_r_i,new org.omg.IOP.IOR("",new org.omg.IOP.TaggedProfile[0]),/*Utils.get_ior(group_members)*/new org.omg.IOP.IOR("",new org.omg.IOP.TaggedProfile[0]));
			any = orb.create_any();
			org.omg.FT.FTRelTotOrdBrServiceContextHelper.insert(any,ctx_rtob);
			serviceCtx_rtob.context_data = codec.encode_value(any);
		    }
		    type_id=Utils.type_id(group_members[0]);
		    set_not_first_ready(one_done);
		    l=group_members.length;
		    exception_holder_t=new org.omg.CORBA.AnyHolder[l];
		    result_holder_t=new org.omg.CORBA.AnyHolder[l];
		    out_inout_result_holders_t=new org.omg.CORBA.AnyHolder[l][];

		    one_done.nr_wait++;
		    error_table=new boolean[l];
		    set_all_false(error_table);
		    for(i=0;i<l;i++){
			member=Utils.narrow(group_members[i],type_id);
			dyn_request=new org.openorb.CORBA.dii.Request((org.omg.CORBA.Object)member,retention_id,client_id,expiration_time,serviceCtx_rtob,oper,dyn_arguments,dyn_result,orb.create_environment(),dyn_exceptions,orb.create_context_list(),orb);
			result_holder_t[i]=new org.omg.CORBA.AnyHolder();
			result_holder_t[i].value=orb.create_any();
			exception_holder_t[i]=new org.omg.CORBA.AnyHolder();
			exception_holder_t[i].value=orb.create_any();
			exception_holder_t[i].value.type(orb.get_primitive_tc(org.omg.CORBA.TCKind.tk_void));
			out_inout_result_holders_t[i]=new org.omg.CORBA.AnyHolder[out_inout_results];
			new SendRequestThread(one_done,dyn_request,result_holder_t[i],out_inout_result_holders_t[i],exception_holder_t[i],error_table,i,meth_type).start();
		    }
		}catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
		block_until_first_ready(one_done);

		org.omg.CORBA.portable.OutputStream os;
		/* send the first result to the client*/
		switch(exception_holder_t[one_done.notifier_index].value.type().kind().value()){
		case org.omg.CORBA.TCKind._tk_string:
		    throw new org.omg.CORBA.TRANSIENT(org.omg.CORBA.OMGVMCID.value | 3, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
		case org.omg.CORBA.TCKind._tk_except:
		    os=((org.openorb.net.ServerRequest)ri).createExceptionReply();
		    Utils.write_output_stream(exception_holder_t[one_done.notifier_index].value,os);
		    success=false;
		    break;
		default:
		    if(result_holder_t[one_done.notifier_index].value==null)
			throw new org.omg.CORBA.TRANSIENT(org.omg.CORBA.OMGVMCID.value | 3, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
		    os=null;
		    if(out_inout_result_holders_t[one_done.notifier_index].length>0||meth_type!=Utils.UPDATE_ONEWAY)
			os=((org.openorb.net.ServerRequest)ri).createNoInvokeReply();
		    if(meth_type!=Utils.UPDATE_ONEWAY)
			Utils.write_output_stream(result_holder_t[one_done.notifier_index].value,os);
		    for(i=0;i<out_inout_result_holders_t[one_done.notifier_index].length;i++)
			Utils.write_output_stream(out_inout_result_holders_t[one_done.notifier_index][i].value,os);
		}
		throw new org.omg.CORBA.REPLY_ALREADY_THERE(success);
	    }
    }

    /** Method called from a specific interception point and has the role to perform different actions after some of the
     * method calls performed on the gateway CORBA object (such as push_structured_event, change_membership) return
     * @param	ri the server request information
     */
    public void send_reply(org.omg.PortableInterceptor.ServerRequestInfo ri)
    {

	org.omg.CORBA.Object or;
	org.omg.PortableServer.POA poa;
	org.omg.CORBA.Any any=null;
	org.omg.IOP.ServiceContext serviceCtx_gv = null, serviceCtx_req=null;
	boolean ex_req=true;
	String oper,m_der_i,client_id="";
	byte[] objI;
	long expiration_time;
	int retention_id=-1,i;
	org.omg.IOP.ServiceContext serviceCtx_rg=null;
	org.omg.IOP.Encoding encoding;


	oper=ri.operation();
	objI=ri.object_id();

	its_gateway=(GatewayForActiveRepl_impl)((org.openorb.CORBA.ORB)orb).getFeature("ItsGateway");
	if(oper.equals("push_structured_event")) {
	    knows_failed();
	    return;
	}
	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	if(oper.equals("change_membership")){
	    try{
		notify_membership_changed();
		return;
	    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	}

	if(oper.equals("is_alive")){
	    return;
	}
	try{
	    java.lang.reflect.Method[] g_meths=its_gateway.getClass().getMethods();
	    for(i=0;i<g_meths.length;i++)
		if(g_meths[i].getName().equals(oper))
		    return;
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	org.omg.CORBA.Any res=((org.openorb.net.ServerRequest)ri).result(),res_args[]=((org.openorb.net.ServerRequest)ri).result_arguments();

	if(res==null) return;

	org.omg.IOP.ServiceContext serviceCtx_repl=new org.omg.IOP.ServiceContext();
	FTReplyServiceContext ctx_repl=new FTReplyServiceContext(res,res_args);

	try
	    {
		any = orb.create_any();
		org.omg.FT.FTReplyServiceContextHelper.insert(any,ctx_repl);
		serviceCtx_repl.context_data = codec.encode_value(any);
	    }
	catch (Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}

	ex_req=true;
	try{
	    serviceCtx_req=ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}

	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	}

	ri.add_reply_service_context(serviceCtx_repl,true);
}

    /**
     * send exception operation
     */
    public void send_exception(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {

    }

    /**
     * send other operation
     */
    public void send_other(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {}


    public java.lang.String name()
    {
	return "FTGatewayServerInterceptor";
    }

}
