/**
 * Redistribution and use of this software and associated
 * documentation ("Software"), with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright statements
 *    and notices.  Redistributions must also contain a copy of this
 *    document.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of one of the parties involved in the TRANSORG project. For
 * 	  written permission, please contact diasz@ida.liu.se.
 *
 * 4. Products derived from this Software may not be called "TRANSORG"
 *    nor may "TRANSORG" appear in their names without prior written
 *    permission.
 *
 * 5. Due credit should be given to the TRANSORG Project
 *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
 *
 * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
 *
 * $Id: FTSLServerRequestInterceptor.java,v 1.0  diasz Exp $
 *
 * Date         Author  Changes
 */

package org.omg.FT;

/* Server request interceptor for STATELESS Replicated Objects*/

public class FTSLServerRequestInterceptor extends org.omg.CORBA.LocalObject implements org.omg.PortableInterceptor.ServerRequestInterceptor{

    org.omg.PortableInterceptor.ORBInitInfo info;
    int ft_slot;
    String location;
    org.omg.CORBA.ORB orb;
    LoggingRecoveryController log_rec_ctrl=null;
    int group_reference_version;
    org.omg.CORBA.Object group_reference;
    java.lang.Object gr_ref_v_guard=new java.lang.Object(),backup_synch=new java.lang.Object();
    MethSignStr[] method_table=new MethSignStr[0];

    public FTSLServerRequestInterceptor(org.omg.PortableInterceptor.ORBInitInfo inf, int ft_s)
    {
	info = inf;
	ft_slot = ft_s;
	try{
	    location = java.net.InetAddress.getLocalHost().getHostName();
	}catch(java.net.UnknownHostException e){e.printStackTrace(new java.io.PrintStream(System.out));location=null;}

	java.util.Properties pr=System.getProperties();
	orb=((org.openorb.PI.OpenORBInitInfo)info).orb();

    }

    private void set_logg_rec_c(){
	log_rec_ctrl=(LoggingRecoveryController)((org.openorb.CORBA.ORB)orb).getFeature("logging_recovery_ctrl");
    }

   private void set_meth_table(){
	org.omg.CORBA.Any any;
	if(method_table.length==0){
	    any=(org.omg.CORBA.Any)((org.openorb.PI.OpenORBInitInfo)info).orb().getFeature("methsign");
	    if(any!=null)
		method_table=MethSignStrsHelper.extract(any);
	}
    }

    public void receive_request_service_contexts(org.omg.PortableInterceptor.ServerRequestInfo ri) throws org.omg.PortableInterceptor.ForwardRequest
    {

	String oper;
	org.omg.PortableServer.POA poa;
	org.omg.PortableServer.Servant theServ=null;
	byte[] objI=null;
	org.omg.CORBA.Object ogr;
	oper=ri.operation();


	if(oper.equals("activate_object"))
	    {

		poa=(org.omg.PortableServer.POA)((org.openorb.CORBA.ORB)orb).getFeature("thePOA");
		theServ=(org.omg.PortableServer.Servant)((org.openorb.CORBA.ORB)orb).getFeature("theServant");
		objI=(byte[])((org.openorb.CORBA.ORB)orb).getFeature("theOID");
		if(poa!=null&&theServ!=null&&objI!=null){
		    try{
			poa.activate_object_with_id(objI,theServ);
		    }catch(java.lang.Exception e){}
		    ((org.openorb.net.ServerRequest)ri).createNoInvokeReply();
		    throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
		}
		else
		    throw new org.omg.CORBA.BAD_CONTEXT();
	    }

	if(oper.equals("deactivate_object"))
	    {
		poa=(org.omg.PortableServer.POA)((org.openorb.CORBA.ORB)orb).getFeature("thePOA");
		objI=(byte[])((org.openorb.CORBA.ORB)orb).getFeature("theOID");
		if(poa!=null&&objI!=null){
		    try{
			poa.deactivate_object(objI);

		    }catch(java.lang.Exception e){}
		    ((org.openorb.net.ServerRequest)ri).createNoInvokeReply();
		    throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
		}
		else
		    throw new org.omg.CORBA.BAD_CONTEXT();
	    }

	if(oper.equals("set_latest_group_reference")){
	    synchronized(gr_ref_v_guard){
		group_reference_version=((org.openorb.net.ServerRequest)ri).argument_stream().read_long();
		group_reference=((org.openorb.net.ServerRequest)ri).argument_stream().read_Object();
	    }
	    throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
	}

	if(oper.equals("is_alive"))
	    return;

	if(oper.equals("replay_last_requests")){
	    LogRecord[] last_r_r,last_r_nr;
	    LoggingRecoveryController callback;
	    int curr_r_i;
	    theServ=(org.omg.PortableServer.Servant)((org.openorb.CORBA.ORB)orb).getFeature("theServant");
	    objI=(byte[])((org.openorb.CORBA.ORB)orb).getFeature("theOID");
	    long objI_l=Utils.makeByteId(objI);
	    last_r_r=LogRecordsHelper.read(((org.openorb.net.ServerRequest)ri).argument_stream());
	    last_r_nr=LogRecordsHelper.read(((org.openorb.net.ServerRequest)ri).argument_stream());
	    callback=LoggingRecoveryControllerHelper.read(((org.openorb.net.ServerRequest)ri).argument_stream());
	    curr_r_i=((org.openorb.net.ServerRequest)ri).argument_stream().read_long();
	    set_meth_table();
	    Utils.replay_last_requests((java.lang.Object)theServ,objI_l,last_r_r,last_r_nr,callback,curr_r_i,(org.openorb.CORBA.ORB)orb,method_table);
	    throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
	}

	if(oper.equals("FT_HB"))
	    if(info._get_policy(HEARTBEAT_ENABLED_POLICY.value)!=null){
		/*send reply*/
		return;
	    }

	synchronized(backup_synch){
	    if(((org.openorb.CORBA.ORB)orb).getFeature("backup")!=null){
		synchronized(gr_ref_v_guard){
		    ogr=group_reference;
		}
		throw new org.omg.PortableInterceptor.ForwardRequest(ogr,true);
	    }
	}
    }
    /**
     * receive request operation
     */

    public void receive_request(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {
	org.omg.CORBA.Object or,ogr;
	org.omg.PortableServer.POA poa;
	int group_ref_v,k;
	String oper,m_der_i;
	String client_id="";
	long expiration_time=-1;
	int retention_id=-1;
	long c_kn_group_ref_v;

	org.omg.IOP.CodecFactory factory=null;
	org.omg.IOP.Codec codec=null;
	org.omg.CORBA.Any any=null;

	byte[] objI;

	oper=ri.operation();
	objI=ri.object_id();

	if(oper.equals("is_alive")) {return;}

	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	/*a normal request to the replicated server*/
	org.omg.IOP.ServiceContext serviceCtx_gv = null, serviceCtx_req=null;
	boolean ex_gv=true,ex_req=true;

	try{
	    serviceCtx_req = ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}

	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	    expiration_time=frctx.expiration_time;
	}
	try
	    {
		serviceCtx_gv = ri.get_request_service_context(org.omg.IOP.FT_GROUP_VERSION.value);
	    }
	catch (org.omg.CORBA.BAD_PARAM bp){ex_gv=false;}

	if(log_rec_ctrl==null)
	    set_logg_rec_c();

	/*this is the case when the object is not part of a group
	  (the object is a normal unreplicated one); or when the group
	  manager does not have a reference to that group*/


	/* first of all check if the request arrived at the right place */
	synchronized(gr_ref_v_guard){
	    group_ref_v=group_reference_version;
	    ogr=group_reference;
	}
	if(serviceCtx_gv!=null&&ex_gv) {

	    try{
		any = codec.decode_value(serviceCtx_gv.context_data, org.omg.FT.FTGroupVersionServiceContextHelper.type());
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	    org.omg.FT.FTGroupVersionServiceContext ftctx = org.omg.FT.FTGroupVersionServiceContextHelper.extract(any);
	    c_kn_group_ref_v=ftctx.object_group_ref_version;



	    /* here we extract the group reference version which is known to the Server ORB*/

	    /* we do the necessary comparisions between the two versions - the one known by the client,
	       and the one known by the server ORB, and we follow the indications in the specification*/

	    if(c_kn_group_ref_v<group_ref_v){
		throw new org.omg.PortableInterceptor.ForwardRequest(ogr,true);
	    }
	    if(c_kn_group_ref_v>group_ref_v)
		synchronized(gr_ref_v_guard){
		    try{
			gr_ref_v_guard.wait();
		    }catch(InterruptedException e){}
		}

	    /* the versions are the same - the request arrived in the right place*/

	    if(serviceCtx_req!=null&&ex_req)
		if(System.currentTimeMillis()>expiration_time){
		    throw new org.omg.CORBA.REQUEST_EXPIRED();
		}

	    int reply_pos,i;
	    org.omg.CORBA.Any reply=null,reply_args[]=new org.omg.CORBA.Any[0];
	    int tkvr=-1,tkvrc=Integer.MIN_VALUE;
	    ResultResultArgs result_result_args;
	    boolean error=false;
	    long objI_l=Utils.makeByteId(objI);

	    if(log_rec_ctrl!=null)
		if(client_id!=null&&!client_id.equals("")&&retention_id!=-1){
		    result_result_args=log_rec_ctrl.call_reply(objI_l,client_id,retention_id);
		    try{
			result_result_args.result.type().id();
		    }catch(org.omg.CORBA.TypeCodePackage.BadKind bk){error=true;}
		    if(!error){
			if(ri.response_expected()){
			    reply=result_result_args.result;
			    reply_args=result_result_args.result_arguments;
			    if(reply!=null){
				org.omg.CORBA.portable.OutputStream os;
				os=((org.openorb.net.ServerRequest)ri).createNoInvokeReply();
				tkvr=reply.type().kind().value();
				if(tkvr==org.omg.CORBA.TCKind._tk_TypeCode)
				    tkvrc=reply.extract_TypeCode().kind().value();
				if(tkvr!=org.omg.CORBA.TCKind._tk_TypeCode||tkvrc!=org.omg.CORBA.TCKind._tk_void)
				    Utils.write_output_stream(reply,os);
				for(i=0;i<reply_args.length;i++)
				    Utils.write_output_stream(reply_args[i],os);
			    }
			}
			throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
		    }
		}
	}
	else /* The client does not have a FT ORB; it did not include any of the required service contexts*/
	    if(!Utils.is_primary(orb,ogr,location)) {
		throw new org.omg.CORBA.TRANSIENT(org.omg.CORBA.OMGVMCID.value | 3, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
	    }
    }

    /**
     * send reply operation
     */
    public void send_reply(org.omg.PortableInterceptor.ServerRequestInfo ri)
    {
	String oper,client_id;
	int retention_id;
	long expiration_time;
	org.omg.IOP.ServiceContext  serviceCtx_req=null;
	org.omg.IOP.CodecFactory factory=null;
	org.omg.IOP.Codec codec=null;
	org.omg.CORBA.Any any;
	boolean ex_req=true;
	org.omg.IOP.Encoding encoding;

	oper=ri.operation();

	if(oper.equals("is_alive"))
	    return;

	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	if(log_rec_ctrl==null)
	    set_logg_rec_c();

	client_id="";
	retention_id=-1;
	expiration_time=0;
	try
	    {
		serviceCtx_req = ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	    }
	catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}

	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	    expiration_time=frctx.expiration_time;
	}

	org.omg.CORBA.Any res=((org.openorb.net.ServerRequest)ri).result(),res_args[]=((org.openorb.net.ServerRequest)ri).result_arguments();


	if(log_rec_ctrl!=null)
	    {

		try{
		    long objI_l=Utils.makeByteId(ri.object_id());
		    log_rec_ctrl.record_reply_stateless(objI_l,client_id,retention_id,expiration_time,res,res_args);
		}catch(Exception rt){rt.printStackTrace(new java.io.PrintStream(System.out));}
	    }

	if(res==null) return;

	org.omg.IOP.ServiceContext serviceCtx_repl=new org.omg.IOP.ServiceContext();
	FTReplyServiceContext ctx_repl=new FTReplyServiceContext(res,res_args);

	try
	    {
		encoding = new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value,(byte)1,(byte)2);
		codec = factory.create_codec(encoding);
		any = orb.create_any();
		org.omg.FT.FTReplyServiceContextHelper.insert(any,ctx_repl);
		serviceCtx_repl.context_data = codec.encode_value(any);
	    }
	catch (Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}

	ri.add_reply_service_context(serviceCtx_repl,true);
    }

    /**
     * send exception operation
     */
    public void send_exception(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {


    }

    /**
     * send other operation
     */
    public void send_other(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {

    }


    public java.lang.String name()
    {
	return "FTSLServerRequestInterceptor";
    }
}
