/**
  * Redistribution and use of this software and associated
  * documentation ("Software"), with or without modification, are
  * permitted provided that the following conditions are met:
  *
  * 1. Redistributions of source code must retain copyright statements
  *    and notices.  Redistributions must also contain a copy of this
  *    document.
  *
  * 2. Redistributions in binary form must reproduce the above
  *    copyright notice, this list of conditions and the following
  *    disclaimer in the documentation and/or other materials provided
  *    with the distribution.
  *
  * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
  *    products derived from this Software without prior written
  *    permission of one of the parties involved in the TRANSORG project. For
  * 	  written permission, please contact diasz@ida.liu.se.
  *
  * 4. Products derived from this Software may not be called "TRANSORG"
  *    nor may "TRANSORG" appear in their names without prior written
  *    permission.
  *
  * 5. Due credit should be given to the TRANSORG Project
  *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
  *
  * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  * PARTICULAR PURPOSE ARE DISCLAIMED.
  *
  * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
  * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
  *
  * $Id: FTWPRServerRequestInterceptor.java,v 1.0  diasz Exp $
  *
  * Date         Author  Changes
  */

package org.omg.FT;
/** Class instantiated when warm passive replication is used and the corresponding profile is given
 *  It intercepts incoming requests sent to the warm passively replicated group member
 * @author	<a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
 */
public class FTWPRServerRequestInterceptor extends org.omg.CORBA.LocalObject implements org.omg.PortableInterceptor.ServerRequestInterceptor{

    org.omg.PortableInterceptor.ORBInitInfo info;
    int ft_slot;
    String location;
    org.omg.CORBA.Object logger_o;
    LoggingRecoveryController log_rec_ctrl;
    org.omg.CORBA.ORB orb;
    org.omg.CosNaming.NamingContext rootC,ftDomC,hostC;

    SynchObject backup_synch=new SynchObject(0);
    String ft_dom_name=null;
    boolean send_exceptions=false,object_deactivated=false;
    MethSignStr[] method_table=new MethSignStr[0];

    java.util.Vector thread_vector; /* this is a vector which will enforce the order of thread
				       execution when waking up threads with notify*/
    int group_reference_version;

    java.lang.Object failed_synch=new java.lang.Object(),gr_ref_v_guard=new java.lang.Object();

    org.omg.CORBA.Object group_reference;
    org.omg.IOP.CodecFactory factory=null;
    org.omg.IOP.Codec codec=null;
    /* some stupid, "no other solution available right now in my head" semaphor solution*/



    private void set_meth_table(){
	org.omg.CORBA.Any any;
	if(method_table.length==0){
	    any=(org.omg.CORBA.Any)((org.openorb.PI.OpenORBInitInfo)info).orb().getFeature("methsign");
	    if(any!=null)
		method_table=MethSignStrsHelper.extract(any);
	}
    }

   private void wait_for_activation(){
	synchronized(backup_synch){
	    while(((org.openorb.CORBA.ORB)orb).getFeature("backup")!=null){
		try{
		    backup_synch.wait();
		}catch(InterruptedException e){}
	    }
	}
   }

    private void atomic_notify_wait_act(){
	synchronized(backup_synch){
	    backup_synch.notifyAll();
	}
    }

    public FTWPRServerRequestInterceptor(org.omg.PortableInterceptor.ORBInitInfo inf, int ft_s)
    {
	info = inf;
	ft_slot = ft_s;
	try{
	    location = java.net.InetAddress.getLocalHost().getHostName();
	}catch(java.net.UnknownHostException e){e.printStackTrace(new java.io.PrintStream(System.out));location=null;}
	orb=((org.openorb.PI.OpenORBInitInfo)info).orb();
	thread_vector=new java.util.Vector();

    }


    private void set_logg_rec_c(){
	log_rec_ctrl=(LoggingRecoveryController)((org.openorb.CORBA.ORB)orb).getFeature("logging_recovery_ctrl");
    }

    /** Method called from a specific interception point, and mainly used to perform operations on method calls which are not allowed to
	     * reach the application object (activate_object, deactivate_object, etc.)
	     * @param	ri the server request information
     */
    public void receive_request_service_contexts(org.omg.PortableInterceptor.ServerRequestInfo ri) throws org.omg.PortableInterceptor.ForwardRequest
    {

	String oper;
	org.omg.CORBA.Object ogr;
	byte[] objI=null;
    	org.omg.PortableServer.POA poa;
	oper=ri.operation();
	org.omg.PortableServer.Servant theServ=null;
	org.omg.CORBA.Any any;


	if(oper.equals("deactivate_object"))
	    {
		if(!object_deactivated){
		    poa=(org.omg.PortableServer.POA)((org.openorb.CORBA.ORB)orb).getFeature("thePOA");
		    objI=(byte[])((org.openorb.CORBA.ORB)orb).getFeature("theOID");
		    if(poa!=null&&objI!=null){
			try{
			    poa.deactivate_object(objI);
			}catch(java.lang.Exception e){}
		    }
		    else
			throw new org.omg.CORBA.BAD_CONTEXT();
		}
		((org.openorb.net.ServerRequest)ri).createNoInvokeReply();
		throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
	    }

	if(oper.equals("set_latest_group_reference")){
	    synchronized(gr_ref_v_guard){
		group_reference_version=((org.openorb.net.ServerRequest)ri).argument_stream().read_long();
		group_reference=((org.openorb.net.ServerRequest)ri).argument_stream().read_Object();
	    }
	    throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
	}

	if(oper.equals("replay_last_requests")){
	    LogRecord[] last_r_r,last_r_nr;
	    LoggingRecoveryController callback;
	    int curr_r_i;
	    objI=(byte[])((org.openorb.CORBA.ORB)orb).getFeature("theOID");
	    long objI_l=Utils.makeByteId(objI);
	    theServ=(org.omg.PortableServer.Servant)((org.openorb.CORBA.ORB)orb).getFeature("theServant");
	    last_r_r=LogRecordsHelper.read(((org.openorb.net.ServerRequest)ri).argument_stream());
	    last_r_nr=LogRecordsHelper.read(((org.openorb.net.ServerRequest)ri).argument_stream());
	    callback=LoggingRecoveryControllerHelper.read(((org.openorb.net.ServerRequest)ri).argument_stream());
	    curr_r_i=((org.openorb.net.ServerRequest)ri).argument_stream().read_long();
	    Utils.replay_last_requests((java.lang.Object)theServ,objI_l,last_r_r,last_r_nr,callback,curr_r_i,(org.openorb.CORBA.ORB)orb,method_table);
	    ((org.openorb.net.ServerRequest)ri).createNoInvokeReply();
	    throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
	}

	if(oper.equals("set_method_table")){
	    org.omg.CORBA.Any methods_wr;
	    methods_wr=((org.openorb.net.ServerRequest)ri).argument_stream().read_any();
	    method_table=org.omg.FT.MethSignStrsHelper.extract(methods_wr);
	    throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
	}

	if(oper.equals("set_state")){
	    org.omg.IOP.ServiceContext serviceCtx_cri=null;
	    boolean ex_ctxri=true;

	    try{
		serviceCtx_cri=ri.get_request_service_context(org.omg.IOP.FT_CURRORI.value);
	    }catch(java.lang.Exception bio){bio.printStackTrace(new java.io.PrintStream(System.out));ex_ctxri=false;}

	    if(serviceCtx_cri!=null&&ex_ctxri){
		try{
		    factory = info.codec_factory();
		    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
		    any = codec.decode_value(serviceCtx_cri.context_data, org.omg.FT.FTCurrORIServiceContextHelper.type());
		    org.omg.FT.FTCurrORIServiceContext frctx = org.omg.FT.FTCurrORIServiceContextHelper.extract(any);
		    ((org.openorb.CORBA.ORB)orb).retention_id(frctx.current_retention_id);
		}catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    }
	    return;
	}

	if(oper.equals("is_alive")||oper.equals("get_state")||oper.equals("get_update"))
	    return;
	if(oper.equals("FT_HB"))
	    if(info._get_policy(HEARTBEAT_ENABLED_POLICY.value)!=null){
		/*send reply*/
		return;
	    }


	synchronized(backup_synch){
	    if(((org.openorb.CORBA.ORB)orb).getFeature("backup")!=null){
		//throw new org.omg.CORBA.TRANSIENT(org.omg.CORBA.OMGVMCID.value | 3, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
		synchronized(gr_ref_v_guard){
		    ogr=group_reference;
		}
		throw new org.omg.PortableInterceptor.ForwardRequest(ogr,true);
	    }
	}
    }

     /** Method called from a specific interception point when receiving a request, and used to perform operations such as logging on the incoming call
        * @param	ri the server request information
     */

    public void receive_request(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {
	org.omg.CORBA.Object or,ogr;
	org.omg.PortableServer.POA poa;
	int group_ref_v,k,c_kn_group_ref_v;
	String oper,m_der_i;

	String client_id="";
	long expiration_time=-1;
	int retention_id=-1,ret_id;
	org.omg.CORBA.Any any=null;

	byte[] objI;

	oper=ri.operation();
	objI=ri.object_id();

	long objI_l=Utils.makeByteId(objI);

	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	if(oper.equals("set_state")||oper.equals("set_update"))
	    return;


	if(oper.equals("is_alive")) {

	    return;
	}

	if(log_rec_ctrl==null)
	    set_logg_rec_c();

	if(method_table==null)
	    set_meth_table();



	if(oper.equals("get_state")||oper.equals("get_update")){
	    synchronized(failed_synch){
		if(send_exceptions)
		    throw new org.omg.CORBA.OBJECT_NOT_EXIST(0,org.omg.CORBA.CompletionStatus.COMPLETED_NO);
	    }

	}
	else{ /*a normal request to the replicated server*/
	    org.omg.IOP.ServiceContext  serviceCtx_req=null,serviceCtx_gv = null;
	    boolean ex_req=true,ex_gv=true;



	    try{
		serviceCtx_req = ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	    }catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}

	    if(serviceCtx_req!=null&&ex_req){
		try{
		    any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		    ri.set_slot(ft_slot,any);
		}catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
		org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
		client_id=frctx.client_id;
		retention_id=frctx.retention_id;
		expiration_time=frctx.expiration_time;
	    }
	    try
		{
		    serviceCtx_gv = ri.get_request_service_context(org.omg.IOP.FT_GROUP_VERSION.value);
		}
	    catch (org.omg.CORBA.BAD_PARAM bp){ex_gv=false;}

	    /*this is the case when the object is not part of a group
	      (the object is a normal unreplicated one); or when the group
	      manager does not have a reference to that group*/

	    /* first of all check if the request arrived at the right place */
	    synchronized(gr_ref_v_guard){
		group_ref_v=group_reference_version;
		ogr=group_reference;
	    }
	    if(serviceCtx_gv!=null&&ex_gv) {

		try{
		    any = codec.decode_value(serviceCtx_gv.context_data, org.omg.FT.FTGroupVersionServiceContextHelper.type());
		}catch(Exception e){
		    e.printStackTrace(new java.io.PrintStream(System.out));
		    return;
		}

		org.omg.FT.FTGroupVersionServiceContext ftctx = org.omg.FT.FTGroupVersionServiceContextHelper.extract(any);
		c_kn_group_ref_v=ftctx.object_group_ref_version;
		/* here we extract the group reference version which is known to the Server ORB*/


		/* we do the necessary comparisions between the two versions - the one known by
		   the client,and the one known by the server ORB, and we follow the indications
		   in the specification*/

		if(c_kn_group_ref_v<group_ref_v){
		    throw new org.omg.PortableInterceptor.ForwardRequest(ogr,true);
		}

		if(c_kn_group_ref_v>group_ref_v) /* This case will not appear*/
		    synchronized(gr_ref_v_guard){
			try{
			    gr_ref_v_guard.wait();
			}catch(InterruptedException e){}
		    }
		/* the versions are the same - the request arrived in the right place*/

		if(serviceCtx_req!=null&&ex_req)
		    if(System.currentTimeMillis()>expiration_time){
			throw new org.omg.CORBA.REQUEST_EXPIRED();
		    }
		/* Check if there is a reply to this request (a repeated one) and if there is send it to
		   the client, without reexecuting the method call*/
		int reply_pos,i;

		org.omg.CORBA.Any reply=null,reply_args[]=new org.omg.CORBA.Any[0];
		int tkvr=-1,tkvrc=Integer.MIN_VALUE;
		ResultResultArgs result_result_args=null;
		boolean error=false;
		char m_t=Utils.method_type(oper,method_table);
		if(log_rec_ctrl!=null&&m_t!=org.omg.FT.Utils.READ){
		    org.openorb.iiop.CDRInputStream is=(org.openorb.iiop.CDRInputStream)((org.openorb.net.ServerRequest)ri).argument_stream();
		    byte[] buf=is.read_arg_buffer();
		    boolean bigEndian=is.bigEndian();
		    org.omg.GIOP.Version version=is.version();
		    try{
			result_result_args=log_rec_ctrl.record_method_call_as(objI_l,oper,buf,bigEndian,version,client_id,retention_id,expiration_time,m_t,(m_t!=Utils.READ));
		    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
		    if(result_result_args==null)
			error=true;
		    try{
			result_result_args.result.type().id();
		    }catch(org.omg.CORBA.TypeCodePackage.BadKind bk){error=true;}
		    if(!error){
			if(ri.response_expected()){
			    reply=result_result_args.result;
			    reply_args=result_result_args.result_arguments;
			    if(reply!=null){
				org.omg.CORBA.portable.OutputStream os;
				os=((org.openorb.net.ServerRequest)ri).createNoInvokeReply();
				tkvr=reply.type().kind().value();
				if(tkvr==org.omg.CORBA.TCKind._tk_TypeCode)
				    tkvrc=reply.extract_TypeCode().kind().value();
				if(tkvr!=org.omg.CORBA.TCKind._tk_TypeCode||tkvrc!=org.omg.CORBA.TCKind._tk_void)
				    Utils.write_output_stream(reply,os);
				for(i=0;i<reply_args.length;i++)
				    Utils.write_output_stream(reply_args[i],os);
			    }
			}

			throw new org.omg.CORBA.REPLY_ALREADY_THERE(true);
		    }
		}

		synchronized(failed_synch){
		    if(send_exceptions)
			throw new org.omg.CORBA.OBJECT_NOT_EXIST(0,org.omg.CORBA.CompletionStatus.COMPLETED_NO);
		}
		/* if there is a get_state request in the line, stop all others (coming after it)
		   from being executed,until the get_state gets its answer*/
	    }
	    else /* The client does not have a FT ORB; it did not include any of the required service contexts*/
		if(!Utils.is_primary(orb,ogr,location)) {
		    throw new org.omg.CORBA.TRANSIENT(org.omg.CORBA.OMGVMCID.value | 3, org.omg.CORBA.CompletionStatus.COMPLETED_NO);
		}
	}
    }

  /** Method called from a specific reply sending interception point; an operation performed here is mainly the logging of request reply
       * @param	ri the server request information
     */
    public void send_reply(org.omg.PortableInterceptor.ServerRequestInfo ri)
    {
	org.omg.CORBA.Object or;
	org.omg.PortableServer.POA poa;
	org.omg.CORBA.Any any=null,result;
	org.omg.IOP.ServiceContext serviceCtx_gv = null, serviceCtx_req=null,serviceCtx_ori=new org.omg.IOP.ServiceContext();
	boolean ex_req=true;
	String oper,client_id;
	byte[] objI,in_res;
	long expiration_time;
	int retention_id;
	org.omg.IOP.Encoding encoding;

	oper=ri.operation();
	objI=ri.object_id();
	if(oper.equals("is_alive")) {
	  org.omg.CORBA.Any res=((org.openorb.net.ServerRequest)ri).result();
	  boolean is_alive_r=res.extract_boolean();
	  if(!is_alive_r){
	    poa=(org.omg.PortableServer.POA)((org.openorb.CORBA.ORB)orb).getFeature("thePOA");
	    objI=(byte[])((org.openorb.CORBA.ORB)orb).getFeature("theOID");
	    if(poa!=null&&objI!=null){
		    try{
		      poa.deactivate_object(objI);
		    }catch(java.lang.Exception e){}
		    object_deactivated=true;
	    }
	    else
	      throw new org.omg.CORBA.BAD_CONTEXT();
	  }
	  return;
	}

	if(log_rec_ctrl==null)
	    set_logg_rec_c();

	long objI_l=Utils.makeByteId(objI);

	if(oper.equals("get_state")||oper.equals("get_update")){
	    int current_retention_id=((org.openorb.CORBA.ORB)orb).current_retention_id(); /*for the moment*/
	    FTCurrORIServiceContext ctx_ori=new FTCurrORIServiceContext(objI_l,current_retention_id);
	    try
		{
		    any = orb.create_any();
		    org.omg.FT.FTCurrORIServiceContextHelper.insert(any,ctx_ori);
		    serviceCtx_ori=new org.omg.IOP.ServiceContext();
		    serviceCtx_ori.context_data = codec.encode_value(any);
		    serviceCtx_ori.context_id=org.omg.IOP.FT_CURRORI.value;
		}
	    catch (Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}
	    ri.add_reply_service_context(serviceCtx_ori,true);
	    return;
	}


	if(oper.equals("set_state")||oper.equals("set_update")){
	    if(log_rec_ctrl!=null){
		log_rec_ctrl.react_set_st_upd_reply(objI_l);
	    }
	    return;
	}

	client_id="";
	retention_id=-1;

	try{
	    serviceCtx_req = ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch(org.omg.CORBA.BAD_PARAM bp){ex_req=false;}

	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	}
	org.omg.CORBA.Any res=((org.openorb.net.ServerRequest)ri).result(),res_args[]=((org.openorb.net.ServerRequest)ri).result_arguments();
	char meth_type=Utils.method_type(oper,method_table);

	if(log_rec_ctrl!=null&&meth_type!=org.omg.FT.Utils.READ)
	    {
		try{
		    log_rec_ctrl.record_reply(objI_l,client_id,retention_id,res,res_args,meth_type,(meth_type==Utils.READ));
		}catch(Exception rt){rt.printStackTrace(new java.io.PrintStream(System.out));}
	    }
	if(res==null) {
	    return;
	}

	org.omg.IOP.ServiceContext serviceCtx_repl=new org.omg.IOP.ServiceContext();
	FTReplyServiceContext ctx_repl=new FTReplyServiceContext(res,res_args);

	try
	    {
		any = orb.create_any();
		org.omg.FT.FTReplyServiceContextHelper.insert(any,ctx_repl);
		serviceCtx_repl.context_data = codec.encode_value(any);
	    }
	catch (Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}

	ri.add_reply_service_context(serviceCtx_repl,true);
    }

/** Method called from an exception sending interception point on the way of the reply from the server to client
     * @param	ri the server request information
     */
    public void send_exception(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {
	if(ri==null){
	    synchronized(failed_synch){
		send_exceptions=true;
	    }
	    return;
	}

	org.omg.CORBA.Any any=null;
	org.omg.IOP.ServiceContext  serviceCtx_req=null;
	boolean ex_req=true;
	String client_id="";
	int retention_id=-1;
	byte[] objI=null;
	String oper=ri.operation();
	try{
	    serviceCtx_req=ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}

	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	    try{
		objI=ri.object_id();
	    }catch(org.omg.CORBA.BAD_INV_ORDER bio){
		return;
	    }
	    long objI_l=Utils.makeByteId(objI);
	    log_rec_ctrl.record_reply_exc(objI_l,client_id,retention_id,Utils.method_type(oper,method_table));
	}
    }

    /** Method called from a specific interception point (the send other interception point) on the way of the reply from the server to client
           * @param	ri the server request information
     */
    public void send_other(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {

	org.omg.CORBA.Any any=null;
	org.omg.IOP.ServiceContext  serviceCtx_req=null;
	boolean ex_req=true;
	String client_id="";
	int retention_id=-1;
	byte[] objI=null;
	if(ri==null){
	    atomic_notify_wait_act();
	    return;
	}
	String oper=ri.operation();
	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	try{
	    serviceCtx_req=ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}


	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	    try{
		objI=ri.object_id();
	    }catch(org.omg.CORBA.BAD_INV_ORDER bio){
		return;
	    }
	    long objI_l=Utils.makeByteId(objI);
	    log_rec_ctrl.record_reply_exc(objI_l,client_id,retention_id,Utils.method_type(oper,method_table));
 }
    }


    public java.lang.String name()
    {
	return "FTWPRServerRequestInterceptor";
    }

}
