/**
  * Redistribution and use of this software and associated
  * documentation ("Software"), with or without modification, are
  * permitted provided that the following conditions are met:
  *
  * 1. Redistributions of source code must retain copyright statements
  *    and notices.  Redistributions must also contain a copy of this
  *    document.
  *
  * 2. Redistributions in binary form must reproduce the above
  *    copyright notice, this list of conditions and the following
  *    disclaimer in the documentation and/or other materials provided
  *    with the distribution.
  *
  * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
  *    products derived from this Software without prior written
  *    permission of one of the parties involved in the TRANSORG project. For
  * 	  written permission, please contact diasz@ida.liu.se.
  *
  * 4. Products derived from this Software may not be called "TRANSORG"
  *    nor may "TRANSORG" appear in their names without prior written
  *    permission.
  *
  * 5. Due credit should be given to the TRANSORG Project
  *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
  *
  * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  * PARTICULAR PURPOSE ARE DISCLAIMED.
  *
  * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
  * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
  *
  * $Id: JustMeasureTimeInterceptorCPR.java,v 1.0  diasz Exp $
  *
  * Date         Author  Changes
  */

package org.omg.FT;
/** Class instantiated when roundtrip time measurements have to be made for a non-replicated, non-FT-CORBA-supported application to compare
*results with a cold passive replication scenario; the interceptor is inserted in order to avoid the probe effect
* @author <a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
 */

public class JustMeasureTimeInterceptorCPR extends org.omg.CORBA.LocalObject implements org.omg.PortableInterceptor.ServerRequestInterceptor{

    org.omg.PortableInterceptor.ORBInitInfo info;
    int ft_slot;
    String location;
    org.omg.CORBA.ORB orb;
    java.io.RandomAccessFile trace_file=null;

    public JustMeasureTimeInterceptorCPR(org.omg.PortableInterceptor.ORBInitInfo inf, int ft_s)
    {
	info = inf;
	ft_slot = ft_s;
	try{
	    location = java.net.InetAddress.getLocalHost().getHostName();
	}catch(java.net.UnknownHostException e){System.out.println("Error in req interc. constr.");e.printStackTrace(new java.io.PrintStream(System.out));location=null;}
	orb=((org.openorb.PI.OpenORBInitInfo)info).orb();
    }



    public void receive_request_service_contexts(org.omg.PortableInterceptor.ServerRequestInfo ri) throws org.omg.PortableInterceptor.ForwardRequest
    {


    }

    /**
     * receive request operation
     */

    public void receive_request(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {
	org.omg.CORBA.Object or,ogr;
	org.omg.PortableServer.POA poa;
	int group_ref_v,k;
	String oper,m_der_i;
	String client_id="";
	long expiration_time=-1;
	int retention_id=-1;
	long c_kn_group_ref_v;

	org.omg.IOP.CodecFactory factory=null;
	org.omg.IOP.Codec codec=null;
	org.omg.CORBA.Any any=null;

	byte[] objI;

	oper=ri.operation();
	objI=ri.object_id();


 /*a normal request to the replicated server*/
	org.omg.IOP.ServiceContext serviceCtx_req=null;
	boolean ex_req=true;

	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(Exception e){System.out.println(location+" "+name()+" Error in req interc. cont.");e.printStackTrace(new java.io.PrintStream(System.out));return;}

	trace_file=(java.io.RandomAccessFile)((org.openorb.CORBA.ORB)orb).getFeature("trace_file");

	try{
	    serviceCtx_req = ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}

	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){System.out.println(location+" "+name()+" Error in req interc. cont.");e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	    expiration_time=frctx.expiration_time;
	}



	long time=System.currentTimeMillis();
	Utils.trace(Utils.ENTER_RECEIVE_REQUEST,retention_id,client_id,time,oper,trace_file);
	Utils.trace(Utils.YES_REQUEST_RIGHT_PLACE,retention_id,client_id,time,oper,trace_file);
	Utils.trace(Utils.START_REC_METHOD_CALL,retention_id,client_id,time,oper,trace_file);
	Utils.trace(Utils.STOP_REC_METHOD_CALL,retention_id,client_id,time,oper,trace_file);
	Utils.trace(Utils.EXIT_RECEIVE_REQUEST,retention_id,client_id,time,oper,trace_file);
    }

    /**
     * send reply operation
     */
    public void send_reply(org.omg.PortableInterceptor.ServerRequestInfo ri)
    {

	org.omg.IOP.CodecFactory factory;
	org.omg.IOP.Codec codec;
	org.omg.CORBA.Object or;
	org.omg.PortableServer.POA poa;
	org.omg.CORBA.Any any=null;
	org.omg.IOP.ServiceContext serviceCtx_gv = null, serviceCtx_req=null;
	boolean ex_req=true;
	String oper,client_id;
	byte[] objI;
	long expiration_time;
	int retention_id;

	oper=ri.operation();
	objI=ri.object_id();

	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(Exception e){System.out.println(location+" "+name()+" Error in req interc. cont.");e.printStackTrace(new java.io.PrintStream(System.out));return;}
	/* access is not via a gateway (or not active repl.)*/

	trace_file=(java.io.RandomAccessFile)((org.openorb.CORBA.ORB)orb).getFeature("trace_file");


	client_id="";
	retention_id=-1;

	try{
	    serviceCtx_req = ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch(org.omg.CORBA.BAD_PARAM bp){ex_req=false;}

	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	}
	long time=System.currentTimeMillis();
	Utils.trace(Utils.ENTER_SEND_REPLY,retention_id,client_id,time,oper,trace_file);
	Utils.trace(Utils.START_REC_REPLY,retention_id,client_id,time,oper,trace_file);
	Utils.trace(Utils.STOP_REC_REPLY,retention_id,client_id,time,oper,trace_file);
	Utils.trace(Utils.EXIT_SEND_REPLY,retention_id,client_id,time,oper,trace_file);

    }

    /**
     * send exception operation
     */
    public void send_exception(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {


	org.omg.IOP.CodecFactory factory;
	org.omg.IOP.Codec codec;
	org.omg.CORBA.Any any=null;
	org.omg.IOP.ServiceContext  serviceCtx_req=null;
	boolean ex_req=true;
	String client_id="";
	int retention_id=-1;
	String oper=ri.operation();


	trace_file=(java.io.RandomAccessFile)((org.openorb.CORBA.ORB)orb).getFeature("trace_file");

	try{
	    serviceCtx_req=ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}

	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(Exception e){e.printStackTrace();return;}

	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	}
	Utils.trace(Utils.SEND_EXCEPTION,retention_id,client_id,System.currentTimeMillis(),oper,trace_file);
    }

    /**
     * send other operation
     */
    public void send_other(org.omg.PortableInterceptor.ServerRequestInfo ri)
	throws org.omg.PortableInterceptor.ForwardRequest
    {
	org.omg.IOP.CodecFactory factory;
	org.omg.IOP.Codec codec;
	org.omg.CORBA.Any any=null;
	org.omg.IOP.ServiceContext  serviceCtx_req=null;
	boolean ex_req=true;
	String client_id="";
	int retention_id=-1;

	trace_file=(java.io.RandomAccessFile)((org.openorb.CORBA.ORB)orb).getFeature("trace_file");
	String oper=ri.operation();
	try{
	    serviceCtx_req=ri.get_request_service_context(org.omg.IOP.FT_REQUEST.value);
	}catch (org.omg.CORBA.BAD_PARAM bp){ex_req=false;}

	try{
	    factory = info.codec_factory();
	    codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
	}catch(Exception e){e.printStackTrace();return;}

	if(serviceCtx_req!=null&&ex_req){
	    try{
		any = codec.decode_value(serviceCtx_req.context_data, org.omg.FT.FTRequestServiceContextHelper.type());
		ri.set_slot(ft_slot,any);
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}
	    org.omg.FT.FTRequestServiceContext frctx = org.omg.FT.FTRequestServiceContextHelper.extract(any);
	    client_id=frctx.client_id;
	    retention_id=frctx.retention_id;
	}
	Utils.trace(Utils.SEND_OTHER,retention_id,client_id,System.currentTimeMillis(),oper,trace_file);
    }


    public java.lang.String name()
    {
	return "FTCPRJServerRequestInterceptor";
    }

}
