/**
  * Redistribution and use of this software and associated
  * documentation ("Software"), with or without modification, are
  * permitted provided that the following conditions are met:
  *
  * 1. Redistributions of source code must retain copyright statements
  *    and notices.  Redistributions must also contain a copy of this
  *    document.
  *
  * 2. Redistributions in binary form must reproduce the above
  *    copyright notice, this list of conditions and the following
  *    disclaimer in the documentation and/or other materials provided
  *    with the distribution.
  *
  * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
  *    products derived from this Software without prior written
  *    permission of one of the parties involved in the TRANSORG project. For
  * 	  written permission, please contact diasz@ida.liu.se.
  *
  * 4. Products derived from this Software may not be called "TRANSORG"
  *    nor may "TRANSORG" appear in their names without prior written
  *    permission.
  *
  * 5. Due credit should be given to the TRANSORG Project
  *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
  *
  * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  * PARTICULAR PURPOSE ARE DISCLAIMED.
  *
  * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
  * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
  *
  * $Id: LoggRClInterceptor.java,v 1.0  diasz Exp $
  *
  * Date         Author  Changes
  */

package org.omg.FT;
/** Class used to intercept requests going out from the Logging Recovery Server Object: e.g. the get_state or set_state method call
* on the server replicas is intercepted and some service cotext is added
* @author <a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
 */
public class LoggRClInterceptor extends org.omg.CORBA.LocalObject implements org.omg.PortableInterceptor.ClientRequestInterceptor{

    org.omg.PortableInterceptor.ORBInitInfo info;
    int ft_slot;
    org.omg.CORBA.ORB orb;

    public LoggRClInterceptor(org.omg.PortableInterceptor.ORBInitInfo inf, int ft_s)
	{
	    info = inf;
	    ft_slot = ft_s;
	    orb=((org.openorb.PI.OpenORBInitInfo)info).orb();
	}


    public java.lang.String name()
    {
	return "LoggRClInterceptor";
    }


    public void receive_exception(org.omg.PortableInterceptor.ClientRequestInfo ri) {
    }

    public void receive_other(org.omg.PortableInterceptor.ClientRequestInfo ri) {
    }

    public void receive_reply(org.omg.PortableInterceptor.ClientRequestInfo ri) {

	org.omg.IOP.CodecFactory factory;
	org.omg.IOP.Codec codec;
	org.omg.CORBA.Any any;
	String oper=ri.operation();

	if(!oper.equals("get_state")&&!oper.equals("get_update")) {
	    return;
	}


	LoggingRecoveryControllerOperations logg_rec_c=(LoggingRecoveryControllerOperations)((org.openorb.CORBA.ORB)orb).getFeature("ItsLoggRecC");
	if(logg_rec_c==null) return;

	org.omg.IOP.ServiceContext serviceCtx_cri=null;
	boolean ex_ctxri=true;

	try{
	    serviceCtx_cri=ri.get_reply_service_context(org.omg.IOP.FT_CURRORI.value);
	}catch(java.lang.Exception bio){bio.printStackTrace(new java.io.PrintStream(System.out));ex_ctxri=false;}

	if(serviceCtx_cri!=null&&ex_ctxri){

	    try{
		factory = info.codec_factory();
		codec = factory.create_codec(new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value, (byte)1, (byte)2));
		any = codec.decode_value(serviceCtx_cri.context_data, org.omg.FT.FTCurrORIServiceContextHelper.type());
	    }catch(Exception e){e.printStackTrace(new java.io.PrintStream(System.out));return;}

	    org.omg.FT.FTCurrORIServiceContext frctx = org.omg.FT.FTCurrORIServiceContextHelper.extract(any);
	    logg_rec_c.record_part_get_st_upd_reply(frctx.object_id,frctx.current_retention_id,oper);
	}
    }

    public void send_poll(org.omg.PortableInterceptor.ClientRequestInfo ri) {}

    public void send_request(org.omg.PortableInterceptor.ClientRequestInfo ri){
    }
}
