/**
  * Redistribution and use of this software and associated
  * documentation ("Software"), with or without modification, are
  * permitted provided that the following conditions are met:
  *
  * 1. Redistributions of source code must retain copyright statements
  *    and notices.  Redistributions must also contain a copy of this
  *    document.
  *
  * 2. Redistributions in binary form must reproduce the above
  *    copyright notice, this list of conditions and the following
  *    disclaimer in the documentation and/or other materials provided
  *    with the distribution.
  *
  * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
  *    products derived from this Software without prior written
  *    permission of one of the parties involved in the TRANSORG project. For
  * 	  written permission, please contact diasz@ida.liu.se.
  *
  * 4. Products derived from this Software may not be called "TRANSORG"
  *    nor may "TRANSORG" appear in their names without prior written
  *    permission.
  *
  * 5. Due credit should be given to the TRANSORG Project
  *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
  *
  * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  * PARTICULAR PURPOSE ARE DISCLAIMED.
  *
  * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
  * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
  *
  * $Id: NonCORBAObjClientRequestInterceptor.java,v 1.0  diasz Exp $
  *
  * Date         Author  Changes
  */

package org.omg.FT;

/** Class instantiated when the corresponding profile (a non replicated client calling a non-replicated server) is chosen
 *  It intercepts requests sent from the client to add request identifying information, only for tracing purposes
 * @author	<a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
 * @version $Revision: 1.0 $ $Date:2003/06/23$
 */

public class NonCORBAObjClientRequestInterceptor extends org.omg.CORBA.LocalObject implements org.omg.PortableInterceptor.ClientRequestInterceptor{

    org.omg.PortableInterceptor.ORBInitInfo info;
    int ft_slot;
    org.omg.CORBA.ORB orb;
    String location;


    public NonCORBAObjClientRequestInterceptor(org.omg.PortableInterceptor.ORBInitInfo inf, int ft_s)
	{
	    info = inf;
	    ft_slot = ft_s;

	    java.util.Properties pr=System.getProperties();
	    orb=((org.openorb.PI.OpenORBInitInfo)info).orb();


	    try{
		location=java.net.InetAddress.getLocalHost().getHostName();
	    }catch(Exception e){location="localhost";}
	}

    /* Extract the TAG_FT_GROUP component of the profile which contains the
       TAG_FT_PRIMARY component*/



    public java.lang.String name()
    {
	return "NonCORBAClientRequestInterceptor";
    }


    public void receive_exception(org.omg.PortableInterceptor.ClientRequestInfo ri) {

    }

    public void receive_other(org.omg.PortableInterceptor.ClientRequestInfo ri) {

    }

    public void receive_reply(org.omg.PortableInterceptor.ClientRequestInfo ri) {

    }

    public void send_poll(org.omg.PortableInterceptor.ClientRequestInfo ri) {}

    public void send_request(org.omg.PortableInterceptor.ClientRequestInfo ri){

	org.omg.IOP.IOR ior;
	int i;
	long req_start_time;
	int ret_id;
	String client_id;
	org.omg.IOP.CodecFactory codecFactory;
	org.omg.IOP.Encoding encoding;
	org.omg.IOP.Codec codec;
	org.omg.CORBA.Any any;
	String oper=ri.operation();

	ret_id=((org.openorb.net.ClientRequest)ri).retention_id();
	client_id=((org.openorb.net.ClientRequest)ri).client_id();
	req_start_time=((org.openorb.net.ClientRequest)ri).request_start_time();

	org.omg.IOP.ServiceContext  serviceCtx_req = new org.omg.IOP.ServiceContext();
	serviceCtx_req.context_id =  org.omg.IOP.FT_REQUEST.value;
	long req_dur=org.omg.FT.DEFAULT_REQUEST_DURATION.value;
	/* Calculate here the expiration time*/

	Integer the_dur=(Integer)((org.openorb.CORBA.ORB)orb).getFeature("request_duration");
	if(the_dur!=null)
	    req_dur=the_dur.intValue();

	FTRequestServiceContext ctx_req=new FTRequestServiceContext(client_id,ret_id,req_start_time+req_dur,location);

	try
	    {
		codecFactory = info.codec_factory();
		encoding = new org.omg.IOP.Encoding(org.omg.IOP.ENCODING_CDR_ENCAPS.value,(byte)1,(byte)2);
		codec = codecFactory.create_codec(encoding);
		any = orb.create_any();
		org.omg.FT.FTRequestServiceContextHelper.insert(any,ctx_req);
		serviceCtx_req.context_data = codec.encode_value(any);
	    }
	catch (Exception e){System.out.println("Exception in clientInterceptor line 121");e.printStackTrace(new java.io.PrintStream(System.out));}

	ri.add_request_service_context(serviceCtx_req,true);
    }
}
